Package: sphunif
Type: Package
Title: Uniformity Tests on the Circle, Sphere, and Hypersphere
Version: 1.4.4
Date: 2026-06-02
Authors@R: c(
     person(given = "Eduardo", family = "García-Portugués",
            role = c("aut", "cre"), email = "edgarcia@est-econ.uc3m.es",
            comment = c(ORCID = "0000-0002-9224-4111")),
     person(given = "Thomas", family = "Verdebout", role = "aut",
            comment = c(ORCID = "0000-0002-3277-9587")),
     person(given = "Alberto", family = "Fernández-de-Marcos", role = "ctb",
            email = "albertfe@est-econ.uc3m.es"),
     person(given = "Paula", family = "Navarro", role = "ctb",
            email = "paula.navarro@unican.es")
     )
Description: Implementation of uniformity tests on the circle and
    (hyper)sphere. The main function of the package is unif_test(), which
    conveniently collects more than 35 tests for assessing uniformity on
    S^{p-1} = {x in R^p : ||x|| = 1}, p >= 2. The test statistics are 
    implemented in the unif_stat() function, which allows computing several
    statistics for different samples within a single call, thus facilitating
    Monte Carlo experiments. Furthermore, the unif_stat_MC() function allows
    parallelizing them in a simple way. The asymptotic null distributions of
    the statistics are available through the function unif_stat_distr(). The
    core of 'sphunif' is coded in C++ by relying on the 'Rcpp' package.
    The package also provides several novel datasets and gives the replicability
    for the data applications/simulations in García-Portugués et al. (2021)
    <doi:10.1007/978-3-030-69944-4_12>, García-Portugués et al. (2023)
    <doi:10.3150/21-BEJ1454>, Fernández-de-Marcos and
    García-Portugués (2024) <doi:10.1016/j.spl.2024.110218>, and
    García-Portugués et al. (2025) <doi:10.1080/01621459.2025.2566414>.
License: GPL-3
LazyData: true
Depends: R (>= 3.5.0), Rcpp
Imports: doFuture, doRNG, foreach, future, gsl, rotasym
Suggests: CompQuadForm, goftest, knitr, markdown, mvtnorm, numDeriv,
        progress, progressr, rmarkdown, scatterplot3d, testthat,
        viridisLite
LinkingTo: Rcpp, RcppArmadillo
URL: https://github.com/egarpor/sphunif
BugReports: https://github.com/egarpor/sphunif/issues
Encoding: UTF-8
RoxygenNote: 7.3.3
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-06-02 20:59:28 UTC; Eduardo
Author: Eduardo García-Portugués [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9224-4111>),
  Thomas Verdebout [aut] (ORCID: <https://orcid.org/0000-0002-3277-9587>),
  Alberto Fernández-de-Marcos [ctb],
  Paula Navarro [ctb]
Maintainer: Eduardo García-Portugués <edgarcia@est-econ.uc3m.es>
Repository: CRAN
Date/Publication: 2026-06-03 08:50:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-06-03 12:28:35 UTC; unix
Archs: sphunif.so.dSYM

Package: sphunif
Type: Package
Title: Uniformity Tests on the Circle, Sphere, and Hypersphere
Version: 1.4.4
Date: 2026-06-02
Authors@R: c(
     person(given = "Eduardo", family = "García-Portugués",
            role = c("aut", "cre"), email = "edgarcia@est-econ.uc3m.es",
            comment = c(ORCID = "0000-0002-9224-4111")),
     person(given = "Thomas", family = "Verdebout", role = "aut",
            comment = c(ORCID = "0000-0002-3277-9587")),
     person(given = "Alberto", family = "Fernández-de-Marcos", role = "ctb",
            email = "albertfe@est-econ.uc3m.es"),
     person(given = "Paula", family = "Navarro", role = "ctb",
            email = "paula.navarro@unican.es")
     )
Description: Implementation of uniformity tests on the circle and
    (hyper)sphere. The main function of the package is unif_test(), which
    conveniently collects more than 35 tests for assessing uniformity on
    S^{p-1} = {x in R^p : ||x|| = 1}, p >= 2. The test statistics are 
    implemented in the unif_stat() function, which allows computing several
    statistics for different samples within a single call, thus facilitating
    Monte Carlo experiments. Furthermore, the unif_stat_MC() function allows
    parallelizing them in a simple way. The asymptotic null distributions of
    the statistics are available through the function unif_stat_distr(). The
    core of 'sphunif' is coded in C++ by relying on the 'Rcpp' package.
    The package also provides several novel datasets and gives the replicability
    for the data applications/simulations in García-Portugués et al. (2021)
    <doi:10.1007/978-3-030-69944-4_12>, García-Portugués et al. (2023)
    <doi:10.3150/21-BEJ1454>, Fernández-de-Marcos and
    García-Portugués (2024) <doi:10.1016/j.spl.2024.110218>, and
    García-Portugués et al. (2025) <doi:10.1080/01621459.2025.2566414>.
License: GPL-3
LazyData: true
Depends: R (>= 3.5.0), Rcpp
Imports: doFuture, doRNG, foreach, future, gsl, rotasym
Suggests: CompQuadForm, goftest, knitr, markdown, mvtnorm, numDeriv,
        progress, progressr, rmarkdown, scatterplot3d, testthat,
        viridisLite
LinkingTo: Rcpp, RcppArmadillo
URL: https://github.com/egarpor/sphunif
BugReports: https://github.com/egarpor/sphunif/issues
Encoding: UTF-8
RoxygenNote: 7.3.3
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-06-02 20:59:28 UTC; Eduardo
Author: Eduardo García-Portugués [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9224-4111>),
  Thomas Verdebout [aut] (ORCID: <https://orcid.org/0000-0002-3277-9587>),
  Alberto Fernández-de-Marcos [ctb],
  Paula Navarro [ctb]
Maintainer: Eduardo García-Portugués <edgarcia@est-econ.uc3m.es>
Repository: CRAN
Date/Publication: 2026-06-03 08:50:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-06-03 12:06:56 UTC; unix
Archs: sphunif.so.dSYM

# Generated by roxygen2: do not edit by hand

export(A_theta_x)
export(F_from_f)
export(F_inv_from_f)
export(Gauss_Legen_nodes)
export(Gauss_Legen_weights)
export(Gegen_coefs)
export(Gegen_coefs_2d)
export(Gegen_coefs_Pn)
export(Gegen_norm)
export(Gegen_norm_2d)
export(Gegen_polyn)
export(Gegen_polyn_2d)
export(Gegen_series)
export(Gegen_series_2d)
export(H_ab)
export(Psi_mat)
export(Theta_to_X)
export(X_to_Theta)
export(akx)
export(angles_to_sphere)
export(avail_cir_tests)
export(avail_sph_tests)
export(bk_to_uk)
export(bk_to_vk2)
export(c_unif_cap)
export(cir_gaps)
export(cir_stat_Ajne)
export(cir_stat_Bakshaev)
export(cir_stat_Bingham)
export(cir_stat_CCF09)
export(cir_stat_Cressie)
export(cir_stat_FG01)
export(cir_stat_Gine_Fn)
export(cir_stat_Gine_Gn)
export(cir_stat_Gini)
export(cir_stat_Gini_squared)
export(cir_stat_Greenwood)
export(cir_stat_Hermans_Rasson)
export(cir_stat_Hodges_Ajne)
export(cir_stat_Kuiper)
export(cir_stat_Log_gaps)
export(cir_stat_Max_uncover)
export(cir_stat_Num_uncover)
export(cir_stat_PAD)
export(cir_stat_PCvM)
export(cir_stat_PRt)
export(cir_stat_Poisson)
export(cir_stat_Pycke)
export(cir_stat_Pycke_q)
export(cir_stat_Range)
export(cir_stat_Rao)
export(cir_stat_Rayleigh)
export(cir_stat_Riesz)
export(cir_stat_Rothman)
export(cir_stat_Sobolev)
export(cir_stat_Softmax)
export(cir_stat_Vacancy)
export(cir_stat_Watson)
export(cir_stat_Watson_1976)
export(con_f)
export(cutoff_locdev)
export(cutoff_wschisq)
export(d_Kolmogorov)
export(d_Sobolev)
export(d_cir_stat_Ajne)
export(d_cir_stat_Bakshaev)
export(d_cir_stat_Bingham)
export(d_cir_stat_Gine_Fn)
export(d_cir_stat_Gine_Gn)
export(d_cir_stat_Gini)
export(d_cir_stat_Gini_squared)
export(d_cir_stat_Greenwood)
export(d_cir_stat_Hermans_Rasson)
export(d_cir_stat_Hodges_Ajne)
export(d_cir_stat_Kuiper)
export(d_cir_stat_Log_gaps)
export(d_cir_stat_Max_uncover)
export(d_cir_stat_Num_uncover)
export(d_cir_stat_PAD)
export(d_cir_stat_PCvM)
export(d_cir_stat_PRt)
export(d_cir_stat_Poisson)
export(d_cir_stat_Pycke)
export(d_cir_stat_Pycke_q)
export(d_cir_stat_Range)
export(d_cir_stat_Rao)
export(d_cir_stat_Rayleigh)
export(d_cir_stat_Riesz)
export(d_cir_stat_Rothman)
export(d_cir_stat_Sobolev)
export(d_cir_stat_Softmax)
export(d_cir_stat_Stein)
export(d_cir_stat_Vacancy)
export(d_cir_stat_Watson)
export(d_cir_stat_Watson_1976)
export(d_locdev)
export(d_p_k)
export(d_proj_unif)
export(d_proj_unif_cap)
export(d_sph_stat_Ajne)
export(d_sph_stat_Bakshaev)
export(d_sph_stat_Bingham)
export(d_sph_stat_CJ12)
export(d_sph_stat_Gine_Fn)
export(d_sph_stat_Gine_Gn)
export(d_sph_stat_PAD)
export(d_sph_stat_PCvM)
export(d_sph_stat_PRt)
export(d_sph_stat_Poisson)
export(d_sph_stat_Rayleigh)
export(d_sph_stat_Rayleigh_HD)
export(d_sph_stat_Riesz)
export(d_sph_stat_Sobolev)
export(d_sph_stat_Softmax)
export(d_sph_stat_Stein)
export(d_sph_stat_Stereo)
export(d_unif_cap)
export(d_wschisq)
export(f_locdev)
export(f_locdev_Pn)
export(g_i_k)
export(int_sph_MC)
export(p_Kolmogorov)
export(p_Sobolev)
export(p_cir_stat_Ajne)
export(p_cir_stat_Bakshaev)
export(p_cir_stat_Bingham)
export(p_cir_stat_Gine_Fn)
export(p_cir_stat_Gine_Gn)
export(p_cir_stat_Gini)
export(p_cir_stat_Gini_squared)
export(p_cir_stat_Greenwood)
export(p_cir_stat_Hermans_Rasson)
export(p_cir_stat_Hodges_Ajne)
export(p_cir_stat_Hodges_Ajne2)
export(p_cir_stat_Kuiper)
export(p_cir_stat_Log_gaps)
export(p_cir_stat_Max_uncover)
export(p_cir_stat_Num_uncover)
export(p_cir_stat_PAD)
export(p_cir_stat_PCvM)
export(p_cir_stat_PRt)
export(p_cir_stat_Poisson)
export(p_cir_stat_Pycke)
export(p_cir_stat_Pycke_q)
export(p_cir_stat_Range)
export(p_cir_stat_Rao)
export(p_cir_stat_Rayleigh)
export(p_cir_stat_Riesz)
export(p_cir_stat_Rothman)
export(p_cir_stat_Sobolev)
export(p_cir_stat_Softmax)
export(p_cir_stat_Stein)
export(p_cir_stat_Vacancy)
export(p_cir_stat_Watson)
export(p_cir_stat_Watson_1976)
export(p_proj_unif)
export(p_proj_unif_cap)
export(p_sph_stat_Ajne)
export(p_sph_stat_Bakshaev)
export(p_sph_stat_Bingham)
export(p_sph_stat_CJ12)
export(p_sph_stat_Gine_Fn)
export(p_sph_stat_Gine_Gn)
export(p_sph_stat_PAD)
export(p_sph_stat_PCvM)
export(p_sph_stat_PRt)
export(p_sph_stat_Poisson)
export(p_sph_stat_Rayleigh)
export(p_sph_stat_Rayleigh_HD)
export(p_sph_stat_Riesz)
export(p_sph_stat_Sobolev)
export(p_sph_stat_Softmax)
export(p_sph_stat_Stein)
export(p_sph_stat_Stereo)
export(p_wschisq)
export(psi_Pn)
export(q_Sobolev)
export(q_proj_unif)
export(q_proj_unif_cap)
export(q_wschisq)
export(r_alt)
export(r_locdev)
export(r_proj_unif)
export(r_proj_unif_cap)
export(r_unif_cap)
export(r_unif_cir)
export(r_unif_sph)
export(r_wschisq)
export(rot_ab)
export(sph_stat_Ajne)
export(sph_stat_Bakshaev)
export(sph_stat_Bingham)
export(sph_stat_CCF09)
export(sph_stat_CJ12)
export(sph_stat_Gine_Fn)
export(sph_stat_Gine_Gn)
export(sph_stat_PAD)
export(sph_stat_PCvM)
export(sph_stat_PRt)
export(sph_stat_Poisson)
export(sph_stat_Pycke)
export(sph_stat_Rayleigh)
export(sph_stat_Rayleigh_HD)
export(sph_stat_Riesz)
export(sph_stat_Sobolev)
export(sph_stat_Softmax)
export(sph_stat_Stereo)
export(sphere_to_angles)
export(uk_to_bk)
export(uk_to_vk2)
export(unif_stat)
export(unif_stat_MC)
export(unif_stat_distr)
export(unif_test)
export(upper_tri_ind)
export(vk2_to_bk)
export(vk2_to_uk)
export(weights_dfs_Sobolev)
import(Rcpp)
importFrom(graphics,abline)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,par)
importFrom(graphics,segments)
importFrom(stats,approxfun)
importFrom(stats,dchisq)
importFrom(stats,dgamma)
importFrom(stats,integrate)
importFrom(stats,na.omit)
importFrom(stats,nlm)
importFrom(stats,pchisq)
importFrom(stats,pgamma)
importFrom(stats,qchisq)
importFrom(stats,qgamma)
importFrom(stats,quantile)
importFrom(stats,rbeta)
importFrom(stats,rchisq)
importFrom(stats,runif)
importFrom(stats,splinefun)
importFrom(stats,uniroot)
importFrom(utils,capture.output)
useDynLib(sphunif)

# sphunif

[![License:
GPLv3](https://img.shields.io/badge/license-GPLv3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)
[![R build
status](https://github.com/egarpor/sphunif/workflows/R-CMD-check/badge.svg)](https://github.com/egarpor/sphunif/actions)
[![R build
status](https://github.com/egarpor/sphunif/workflows/test-coverage/badge.svg)](https://github.com/egarpor/sphunif/actions)
[![](https://codecov.io/gh/egarpor/sphunif/branch/master/graph/badge.svg)](https://codecov.io/gh/egarpor/sphunif)
[![](https://www.r-pkg.org/badges/version/sphunif?color=green)](https://cran.r-project.org/package=sphunif)
[![](http://cranlogs.r-pkg.org/badges/grand-total/sphunif?color=green)](https://cran.r-project.org/package=sphunif)
[![](http://cranlogs.r-pkg.org/badges/last-month/sphunif?color=green)](https://cran.r-project.org/package=sphunif)

<!-- <img src="" alt="sphunif hexlogo" align="right" width="200" style="padding: 0 15px; float: right;"/> -->

## Overview

Implementation of more than 30 tests of uniformity on the circle,
sphere, and hypersphere. Software companion for the (evolving) review
“*An overview of uniformity tests on the hypersphere*” (García-Portugués
and Verdebout, 2018) and the paper “*On a projection-based class of
uniformity tests on the hypersphere*” (García-Portugués, Navarro-Esteban
and Cuesta-Albertos, 2020).

## Installation

Get the latest version from GitHub:

``` r
# Install the package and the vignettes
library(devtools)
install_github("egarpor/sphunif", build_vignettes = TRUE)

# Load package
library(sphunif)

# See main vignette
vignette("sphunif")
```

## Usage

### Circular data

You want to test if a sample of *circular* data is uniformly
distributed. For example, the following circular uniform sample in
*radians*:

``` r
set.seed(987202226)
cir_data <- runif(n = 10, min = 0, max = 2 * pi)
```

Call the main function in the `sphunif` package, `unif_test`, specifying
the `type` of test to be performed. For example, the `"Watson"` test:

``` r
library(sphunif)
unif_test(data = cir_data, type = "Watson") # An htest object
#> 
#>  Watson test of circular uniformity
#> 
#> data:  cir_data
#> statistic = 0.036003, p-value = 0.8694
#> alternative hypothesis: any alternative to circular uniformity
```

By default, the calibration of the test statistic is done by Monte
Carlo. This can be changed with `p_value = "asymp"` to employ asymptotic
distributions (faster, but not available for all tests):

``` r
unif_test(data = cir_data, type = "Watson", p_value = "MC") # Monte Carlo
#> 
#>  Watson test of circular uniformity
#> 
#> data:  cir_data
#> statistic = 0.036003, p-value = 0.8844
#> alternative hypothesis: any alternative to circular uniformity
unif_test(data = cir_data, type = "Watson", p_value = "asymp") # Asymp. distr.
#> 
#>  Watson test of circular uniformity
#> 
#> data:  cir_data
#> statistic = 0.036003, p-value = 0.8694
#> alternative hypothesis: any alternative to circular uniformity
```

You can perform *several* tests with a *single* call to `unif_test`.
Choose the available circular uniformity tests from

``` r
avail_cir_tests
#>  [1] "Ajne"           "Bakshaev"       "Bingham"        "Cressie"       
#>  [5] "CCF09"          "FG01"           "Gine_Fn"        "Gine_Gn"       
#>  [9] "Gini"           "Gini_squared"   "Greenwood"      "Hermans_Rasson"
#> [13] "Hodges_Ajne"    "Kuiper"         "Log_gaps"       "Max_uncover"   
#> [17] "Num_uncover"    "PAD"            "PCvM"           "PRt"           
#> [21] "Pycke"          "Pycke_q"        "Range"          "Rao"           
#> [25] "Rayleigh"       "Riesz"          "Rothman"        "Vacancy"       
#> [29] "Watson"         "Watson_1976"
```

For example:

``` r
# A *list* of htest objects
unif_test(data = cir_data, type = c("Watson", "PAD", "Ajne"))
#> $Watson
#> 
#>  Watson test of circular uniformity
#> 
#> data:  cir_data
#> statistic = 0.036003, p-value = 0.8694
#> alternative hypothesis: any alternative to circular uniformity
#> 
#> 
#> $PAD
#> 
#>  Projected Anderson-Darling test of circular uniformity
#> 
#> data:  cir_data
#> statistic = 0.47247, p-value = 0.9051
#> alternative hypothesis: any alternative to circular uniformity
#> 
#> 
#> $Ajne
#> 
#>  Ajne test of circular uniformity
#> 
#> data:  cir_data
#> statistic = 0.11016, p-value = 0.7361
#> alternative hypothesis: any non-axial alternative to circular uniformity
```

### Spherical data

You want to test if a sample of *spherical* data is uniformly
distributed. Consider a *non*-uniformly-generated sample of directions
in **Cartesian coordinates**:

``` r
# Sample data on S^2
set.seed(987202226)
theta <- runif(n = 50, min = 0, max = 2 * pi)
phi <- runif(n = 50, min = 0, max = pi)
sph_data <- cbind(cos(theta) * sin(phi), sin(theta) * sin(phi), cos(phi))
```

The available spherical uniformity tests:

``` r
avail_sph_tests
#>  [1] "Ajne"        "Bakshaev"    "Bingham"     "CJ12"        "CCF09"      
#>  [6] "Gine_Fn"     "Gine_Gn"     "PAD"         "PCvM"        "PRt"        
#> [11] "Pycke"       "Rayleigh"    "Rayleigh_HD" "Riesz"
```

The default `type = "all"` equals `type = avail_sph_tests`:

``` r
head(unif_test(data = sph_data, type = "all", p_value = "MC"))
#> $Ajne
#> 
#>  Ajne test of spherical uniformity
#> 
#> data:  sph_data
#> statistic = 0.079876, p-value = 0.9562
#> alternative hypothesis: any non-axial alternative to spherical uniformity
#> 
#> 
#> $Bakshaev
#> 
#>  Bakshaev (2010) test of spherical uniformity
#> 
#> data:  sph_data
#> statistic = 1.2727, p-value = 0.4389
#> alternative hypothesis: any alternative to spherical uniformity
#> 
#> 
#> $Bingham
#> 
#>  Bingham test of spherical uniformity
#> 
#> data:  sph_data
#> statistic = 22.455, p-value = 5e-04
#> alternative hypothesis: scatter matrix different from constant
#> 
#> 
#> $CJ12
#> 
#>  Cai and Jiang (2012) test of spherical uniformity
#> 
#> data:  sph_data
#> statistic = 27.401, p-value = 0.2691
#> alternative hypothesis: unclear consistency
#> 
#> 
#> $CCF09
#> 
#>  Cuesta-Albertos et al. (2009) test of spherical uniformity with k = 50
#> 
#> data:  sph_data
#> statistic = 1.4619, p-value = 0.278
#> alternative hypothesis: any alternative to spherical uniformity
#> 
#> 
#> $Gine_Fn
#> 
#>  Gine's Fn test of spherical uniformity
#> 
#> data:  sph_data
#> statistic = 1.8889, p-value = 0.2267
#> alternative hypothesis: any alternative to spherical uniformity
unif_test(data = sph_data, type = "Rayleigh", p_value = "asymp")
#> 
#>  Rayleigh test of spherical uniformity
#> 
#> data:  sph_data
#> statistic = 0.52692, p-value = 0.9129
#> alternative hypothesis: mean direction different from zero
```

### Higher dimensions

The *hyperspherical* setting is treated analogously to the spherical
setting, and the available tests are exactly the same
(`avail_sph_tests`). An example of testing uniformity with a sample of
size `100` on the 9-sphere:

``` r
# Sample data on S^9
set.seed(987202226)
hyp_data <- r_unif_sph(n = 50, p = 10)

# Test
unif_test(data = hyp_data, type = "Rayleigh", p_value = "asymp")
#> 
#>  Rayleigh test of spherical uniformity
#> 
#> data:  hyp_data
#> statistic = 11.784, p-value = 0.2997
#> alternative hypothesis: mean direction different from zero
```

## Data application in astronomy

The data application in García-Portugués, Navarro-Esteban and
Cuesta-Albertos (2020) can be reproduced through the script
[data-application-ecdf.R](https://github.com/egarpor/sphunif/blob/master/application/data-application-ecdf.R).
The code snippet below illustrates the testing of the uniformity of
craters in Rhea.

``` r
# Load data
data(rhea)

# Add Cartesian coordinates
rhea$X <- cbind(cos(rhea$theta) * cos(rhea$phi),
                sin(rhea$theta) * cos(rhea$phi),
                sin(rhea$phi))

# Distribution of diameter
quantile(rhea$diameter)
#>       0%      25%      50%      75%     100% 
#>  10.0000  13.2475  17.0500  24.5600 449.8200

# Subsets of craters, according to diameter
ind_15_20 <- rhea$diameter > 15 & rhea$diameter < 20
ind_20 <- rhea$diameter > 20

# Sample sizes
nrow(rhea)
#> [1] 3596
sum(ind_15_20)
#> [1] 867
sum(ind_20)
#> [1] 1373

# Tests to be performed
type_tests <- c("PCvM", "PAD", "PRt")

# Tests
tests_rhea_15_20 <- unif_test(data = rhea$X[ind_15_20, ], type = type_tests,
                              p_value = "asymp", K_max = 5e4)
tests_rhea_20 <- unif_test(data = rhea$X[ind_20, ], type = type_tests,
                           p_value = "asymp", K_max = 5e4)
tests_rhea_15_20
#> $PCvM
#> 
#>  Projected Cramer-von Mises test of spherical uniformity
#> 
#> data:  rhea$X[ind_15_20, ]
#> statistic = 0.26452, p-value = 0.1176
#> alternative hypothesis: any alternative to spherical uniformity
#> 
#> 
#> $PAD
#> 
#>  Projected Anderson-Darling test of spherical uniformity
#> 
#> data:  rhea$X[ind_15_20, ]
#> statistic = 1.6854, p-value = 0.07209
#> alternative hypothesis: any alternative to spherical uniformity
#> 
#> 
#> $PRt
#> 
#>  Projected Rothman test of spherical uniformity with t = 0.333
#> 
#> data:  rhea$X[ind_15_20, ]
#> statistic = 0.31352, p-value = 0.1856
#> alternative hypothesis: any alternative to spherical uniformity if t is irrational
tests_rhea_20
#> $PCvM
#> 
#>  Projected Cramer-von Mises test of spherical uniformity
#> 
#> data:  rhea$X[ind_20, ]
#> statistic = 3.6494, p-value = 2.202e-09
#> alternative hypothesis: any alternative to spherical uniformity
#> 
#> 
#> $PAD
#> 
#>  Projected Anderson-Darling test of spherical uniformity
#> 
#> data:  rhea$X[ind_20, ]
#> statistic = 18.482, p-value < 2.2e-16
#> alternative hypothesis: any alternative to spherical uniformity
#> 
#> 
#> $PRt
#> 
#>  Projected Rothman test of spherical uniformity with t = 0.333
#> 
#> data:  rhea$X[ind_20, ]
#> statistic = 5.3485, p-value = 3.481e-09
#> alternative hypothesis: any alternative to spherical uniformity if t is irrational
```

## References

García-Portugués, E., Navarro-Esteban, P., and Cuesta-Albertos, J. A.
(2021). A Cramér–von Mises test of uniformity on the hypersphere. In
Balzano, S., Porzio, G. C., Salvatore, R., Vistocco, D., and Vichi, M.
(Eds.), *Statistical Learning and Modeling in Data Analysis*, Studies in
Classification, Data Analysis and Knowledge Organization, pp. 107–-116.
Springer, Cham.
[doi:10.1007/978-3-030-69944-4\_12](https://doi.org/10.1007/978-3-030-69944-4_12).

García-Portugués, E., Navarro-Esteban, P., and Cuesta-Albertos, J. A.
(2020). On a projection-based class of uniformity tests on the
hypersphere. *arXiv:2008.09897*. <https://arxiv.org/abs/2008.09897>.

García-Portugués, E. and Verdebout, T. (2018). An overview of uniformity
tests on the hypersphere. *arXiv:1804.00286*.
<https://arxiv.org/abs/1804.00286>.

García-Portugués, E., Paindaveine, D., and Verdebout, T. (2021). On the
power of Sobolev tests for isotropy under local rotationally symmetric
alternatives. *arXiv:2108.09874*. <https://arxiv.org/abs/2108.09874>.

---
output:
  md_document:
    variant: gfm
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE, comment = "#>", fig.path = "README/README-",
  message = FALSE, warning = FALSE, fig.asp = 1, fig.align = 'center'
)
```

sphunif
=======

```{r, echo = FALSE, results = 'asis'}
cat(
  badger::badge_license(license = "GPLv3", color = "blue",
                        url = "https://www.gnu.org/licenses/gpl-3.0"),
  badger::badge_github_actions(action = "R-CMD-check"),
  badger::badge_github_actions(action = "test-coverage"),
  badger::badge_codecov(),
  badger::badge_cran_release(color = "green"),
  badger::badge_cran_download(pkg = NULL, type = "grand-total"),
  badger::badge_cran_download(pkg = NULL, type = "last-month")
)
```

<!-- <img src="" alt="sphunif hexlogo" align="right" width="200" style="padding: 0 15px; float: right;"/> -->

## Overview

Implementation of more than 30 tests of uniformity on the circle, sphere, and hypersphere. Software companion for the (evolving) review "*An overview of uniformity tests on the hypersphere*" (García-Portugués and Verdebout, 2018) and the paper "*On a projection-based class of uniformity tests on the hypersphere*" (García-Portugués, Navarro-Esteban and Cuesta-Albertos, 2020).

## Installation

Get the latest version from GitHub:

```{r, install-devtools, eval = FALSE}
# Install the package and the vignettes
library(devtools)
install_github("egarpor/sphunif", build_vignettes = TRUE)

# Load package
library(sphunif)

# See main vignette
vignette("sphunif")
```

```{r, load, echo = FALSE}
# Load package
library(sphunif)
```

## Usage

### Circular data

You want to test if a sample of *circular* data is uniformly distributed. For example, the following circular uniform sample in *radians*:

```{r, cir_data}
set.seed(987202226)
cir_data <- runif(n = 10, min = 0, max = 2 * pi)
```

Call the main function in the `sphunif` package, `unif_test`, specifying the `type` of test to be performed. For example, the `"Watson"` test:

```{r, unif_test_cir}
library(sphunif)
unif_test(data = cir_data, type = "Watson") # An htest object
```

By default, the calibration of the test statistic is done by Monte Carlo. This can be changed with `p_value = "asymp"` to employ asymptotic distributions (faster, but not available for all tests):

```{r, asymp}
unif_test(data = cir_data, type = "Watson", p_value = "MC") # Monte Carlo
unif_test(data = cir_data, type = "Watson", p_value = "asymp") # Asymp. distr.
```

You can perform *several* tests with a *single* call to `unif_test`. Choose the available circular uniformity tests from

```{r, avail_cir}
avail_cir_tests
```

For example:

```{r, unif_test_avail_cir}
# A *list* of htest objects
unif_test(data = cir_data, type = c("Watson", "PAD", "Ajne"))
```

### Spherical data

You want to test if a sample of *spherical* data is uniformly distributed. Consider a *non*-uniformly-generated sample of directions in **Cartesian coordinates**:

```{r, sph_data}
# Sample data on S^2
set.seed(987202226)
theta <- runif(n = 50, min = 0, max = 2 * pi)
phi <- runif(n = 50, min = 0, max = pi)
sph_data <- cbind(cos(theta) * sin(phi), sin(theta) * sin(phi), cos(phi))
```

The available spherical uniformity tests:

```{r, avail_sph}
avail_sph_tests
```

The default `type = "all"` equals `type = avail_sph_tests`:

```{r, unif_test_avail_sph}
head(unif_test(data = sph_data, type = "all", p_value = "MC"))
unif_test(data = sph_data, type = "Rayleigh", p_value = "asymp")
```

### Higher dimensions

The *hyperspherical* setting is treated analogously to the spherical setting, and the available tests are exactly the same (`avail_sph_tests`). An example of testing uniformity with a sample of size `100` on the $9$-sphere:

```{r, hyp_data}
# Sample data on S^9
set.seed(987202226)
hyp_data <- r_unif_sph(n = 50, p = 10)

# Test
unif_test(data = hyp_data, type = "Rayleigh", p_value = "asymp")
```

## Data application in astronomy

The data application in García-Portugués, Navarro-Esteban and Cuesta-Albertos (2020) can be reproduced through the script [data-application-ecdf.R](https://github.com/egarpor/sphunif/blob/master/application/data-application-ecdf.R). The code snippet below illustrates the testing of the uniformity of craters in Rhea.

```{r, rhea}
# Load data
data(rhea)

# Add Cartesian coordinates
rhea$X <- cbind(cos(rhea$theta) * cos(rhea$phi),
                sin(rhea$theta) * cos(rhea$phi),
                sin(rhea$phi))

# Distribution of diameter
quantile(rhea$diameter)

# Subsets of craters, according to diameter
ind_15_20 <- rhea$diameter > 15 & rhea$diameter < 20
ind_20 <- rhea$diameter > 20

# Sample sizes
nrow(rhea)
sum(ind_15_20)
sum(ind_20)

# Tests to be performed
type_tests <- c("PCvM", "PAD", "PRt")

# Tests
tests_rhea_15_20 <- unif_test(data = rhea$X[ind_15_20, ], type = type_tests,
                              p_value = "asymp", K_max = 5e4)
tests_rhea_20 <- unif_test(data = rhea$X[ind_20, ], type = type_tests,
                           p_value = "asymp", K_max = 5e4)
tests_rhea_15_20
tests_rhea_20
```

## References

García-Portugués, E., Navarro-Esteban, P., and Cuesta-Albertos, J. A. (2021). A Cramér–von Mises test of uniformity on the hypersphere. In Balzano, S., Porzio, G. C., Salvatore, R., Vistocco, D., and Vichi, M. (Eds.), *Statistical Learning and Modeling in Data Analysis*, Studies in Classification, Data Analysis and Knowledge Organization, pp. 107–-116. Springer, Cham. [doi:10.1007/978-3-030-69944-4_12](https://doi.org/10.1007/978-3-030-69944-4_12).

García-Portugués, E., Navarro-Esteban, P., and Cuesta-Albertos, J. A. (2020). On a projection-based class of uniformity tests on the hypersphere. *arXiv:2008.09897*. <https://arxiv.org/abs/2008.09897>.

García-Portugués, E. and Verdebout, T. (2018). An overview of uniformity tests on the hypersphere. *arXiv:1804.00286*. <https://arxiv.org/abs/1804.00286>.

García-Portugués, E., Paindaveine, D., and Verdebout, T. (2021). On the power of Sobolev tests for isotropy under local rotationally symmetric alternatives. *arXiv:2108.09874*. <https://arxiv.org/abs/2108.09874>.

## ----setup, include = FALSE---------------------------------------------------
options(rmarkdown.html_vignette.check_title = FALSE)
knitr::opts_chunk$set(
  collapse = TRUE,
  message = FALSE,
  comment = "#>"
)

## ----cir_data-----------------------------------------------------------------
set.seed(987202226)
cir_data <- runif(n = 30, min = 0, max = 2 * pi)

## ----unif_test_cir------------------------------------------------------------
library(sphunif)
unif_test(data = cir_data, type = "Watson") # An htest object

## ----asymp--------------------------------------------------------------------
unif_test(data = cir_data, type = "Watson", p_value = "MC") # Monte Carlo
unif_test(data = cir_data, type = "Watson", p_value = "asymp") # Asymp. distr.

## ----avail_cir----------------------------------------------------------------
avail_cir_tests

## ----unif_test_avail_cir------------------------------------------------------
# A *list* of htest objects
unif_test(data = cir_data, type = c("PAD", "Watson"))

## ----sph_data-----------------------------------------------------------------
# Sample data on S^2
set.seed(987202226)
theta <- runif(n = 30, min = 0, max = 2 * pi)
phi <- runif(n = 30, min = 0, max = pi)
sph_data <- cbind(cos(theta) * sin(phi), sin(theta) * sin(phi), cos(phi))

## ----avail_sph----------------------------------------------------------------
avail_sph_tests

## ----unif_test_avail_sph------------------------------------------------------
unif_test(data = sph_data, type = "all", p_value = "MC", M = 1e2)
unif_test(data = sph_data, type = "Rayleigh", p_value = "asymp")

## ----hyp_data-----------------------------------------------------------------
# Sample data on S^9
hyp_data <- r_unif_sph(n = 30, p = 10)

# Test
unif_test(data = hyp_data, type = "Rayleigh", p_value = "asymp")

## ----venus--------------------------------------------------------------------
# Load spherical data
data(venus)
head(venus)
nrow(venus)

# Compute Cartesian coordinates from polar coordinates
venus$X <- cbind(cos(venus$theta) * cos(venus$phi),
                 sin(venus$theta) * cos(venus$phi),
                 sin(venus$phi))

# Test uniformity using the Projected Cramér-von Mises and Projected
# Anderson-Darling tests on S^2, both using asymptotic distributions
unif_test(data = venus$X, type = c("PCvM", "PAD"), p_value = "asymp")

# Define a handler for progressr
require(progress)
require(progressr)
handlers(handler_progress(
  format = paste("(:spin) [:bar] :percent Iter: :current/:total Rate:",
                 ":tick_rate iter/sec ETA: :eta Elapsed: :elapsedfull"),
  clear = FALSE))

# Test uniformity using Monte-Carlo approximated null distributions
with_progress(
  unif_test(data = venus$X, type = c("PCvM", "PAD"),
            p_value = "MC", chunks = 1e2, M = 5e2, cores = 2)
)

## ----unif_stat_vec------------------------------------------------------------
# M samples of size n on S^2
samps_sph <- r_unif_sph(n = 30, p = 3, M = 10)

# Apply all statistics to the M samples
unif_stat(data = samps_sph, type = "all")

## ----MC-----------------------------------------------------------------------
# Break the simulation in 10 chunks of tasks to be divided between 2 cores
sim <- unif_stat_MC(n = 30, type = "all", p = 3, M = 1e2, cores = 2,
                    chunks = 10)

# Critical values for test statistics
sim$crit_val_MC

# Test statistics
head(sim$stats_MC)

# Power computation using a pre-built sampler for the alternative
pow <- unif_stat_MC(n = 30, type = "all", p = 3, M = 1e2, cores = 2,
                    chunks = 10, r_H1 = r_alt, crit_val = sim$crit_val_MC,
                    alt = "vMF", kappa = 1)
pow$power_MC

# How to use a custom sampler according to ?unif_stat_MC
r_H1 <- function(n, p, M, l = 1) {

  samp <- array(dim = c(n, p, M))
  for (j in 1:M) {

    samp[, , j] <- mvtnorm::rmvnorm(n = n, mean = c(l, rep(0, p - 1)),
                                    sigma = diag(rep(1, p)))
    samp[, , j] <- samp[, , j] / sqrt(rowSums(samp[, , j]^2))

  }
  return(samp)

}
pow <- unif_stat_MC(n = 30, type = "all", p = 3, M = 1e2, cores = 2,
                    chunks = 5, r_H1 = r_H1, crit_val = sim$crit_val_MC)
pow$power_MC

## ----crit_val-----------------------------------------------------------------
# Using precomputed critical values
ht1 <- unif_test(data = samps_sph[, , 1], type = "Rayleigh",
                 p_value = "crit_val", crit_val = sim$crit_val_MC)
ht1
ht1$reject

# Using precomputed Monte Carlo statistics
ht2 <- unif_test(data = samps_sph[, , 1], type = "Rayleigh",
                 p_value = "MC", stats_MC = sim$stats_MC)
ht2
ht2$reject

# Faster than
unif_test(data = samps_sph[, , 1], type = "Rayleigh", p_value = "MC")

## -----------------------------------------------------------------------------
citation("sphunif")

# Clean workspace
rm(list = ls())

# Load packages
library(rotasym)
library(sphunif)

## Sunspots

# Load data used in García-Portugués, Paindaveine, and Verdebout (2020)
data("sunspots_births")

# Add Cartesian coordinates
sunspots_births$X <-
  cbind(cos(sunspots_births$phi) * cos(sunspots_births$theta),
        cos(sunspots_births$phi) * sin(sunspots_births$theta),
        sin(sunspots_births$phi))

# Data associated to the 23rd and 22nd solar cycles
sunspots_23 <- subset(sunspots_births, cycle == 23)
sunspots_22 <- subset(sunspots_births, cycle == 22)

# Obtain the signs with respect to the north pole c(0, 0, 1) -- these are the
# positions in S^1 of the longitudes associated to the data. The latitudes are
# disregarded to investigate if the longitudes are uniformly distributed, which
# is implied if rotational symmetry about (0, 0, 1) holds
signs_23 <- signs(sunspots_23$X, theta = c(0, 0, 1))
signs_22 <- signs(sunspots_22$X, theta = c(0, 0, 1))

# Sample sizes
nrow(signs_23)
nrow(signs_22)

# Tests to be performed
type_tests <- c("PCvM", "PAD", "PRt")

# Tests
tests_sunspots_23 <- unif_test(data = signs_23, type = type_tests,
                               p_value = "asymp", K_max = 5e4)
tests_sunspots_22 <- unif_test(data = signs_22, type = type_tests,
                               p_value = "asymp", K_max = 5e4)
tests_sunspots_23
tests_sunspots_22

## Planets

# Load data
data("planets")

# Add normal vectors
planets$normal <- cbind(sin(planets$i) * sin(planets$om),
                       -sin(planets$i) * cos(planets$om),
                       cos(planets$i))

# Tests to be performed
type_tests <- c("PCvM", "PAD", "PRt")

# Tests with Pluto
unif_test(data = planets$normal, type = type_tests, M = 1e6,
          p_value = "MC", cores = 4, chunks = 100, seed = 1:100)

# Tests without Pluto
unif_test(data = planets$normal[-9, ], type = type_tests, M = 1e6,
          p_value = "MC", cores = 4, chunks = 100, seed = 1:100)

## Comets

# Load data
data("comets")

# Add normal vectors
comets$normal <- cbind(sin(comets$i) * sin(comets$om),
                       -sin(comets$i) * cos(comets$om),
                       cos(comets$i))

# Tests to be performed
type_tests <- c("PCvM", "PAD", "PRt")

# Excluding the C/1882 R1-X (Great September comet) records with X = B, C, D
comets_ccf2009 <- comets[comets$ccf2009, ][-c(13:15), ]

# Sample size
nrow(comets_ccf2009)

# Tests for the data in Cuesta-Albertos et al. (2009)
tests_ccf2009 <- c(unif_test(data = comets_ccf2009$normal, type = type_tests,
                             p_value = "asymp", K_max = 5e4),
                   list("Cuesta_Albertos" =
                          unif_test(data = comets_ccf2009$normal,
                                    type = "Cuesta_Albertos", p_value = "MC",
                                    M = 1e4, cores = 4, chunks = 100,
                                    seed = 1:100))
                   )
tests_ccf2009

# Numbered comets
comets$numbered <- substr(comets$id, 1, 1) == "c"

# Filter comets as in Cuesta-Albertos et al. (2009)
comets_ellip <- subset(x = comets, subset = !(class %in% c("HYP", "PAR")) &
                         per_y >= 200 & !numbered)

# Remove "duplicated" records (as done in Cuesta-Albertos et al. (2009))
dup <- duplicated(round(comets_ellip$normal, 2))
comets_ellip <- comets_ellip[!dup, ]

# Sample size
nrow(comets_ellip)

# Tests for the extended dataset
tests_comets <- c(unif_test(data = comets_ellip$normal, type = type_tests,
                            p_value = "asymp", K_max = 5e4),
                   list("Cuesta_Albertos" =
                          unif_test(data = comets_ellip$normal,
                                    type = "Cuesta_Albertos", p_value = "MC",
                                    M = 1e4, cores = 4, chunks = 100,
                                    seed = 1:100))
)
tests_comets

## Craters named by the IUA

# Read data of named craters
data("craters")

# Add Cartesian coordinates
craters$X <- cbind(cos(craters$theta) * cos(craters$phi),
                   sin(craters$theta) * cos(craters$phi),
                   sin(craters$phi))

# Most crater-populated bodies and its type
num_craters <- sort(table(craters$target), decreasing = TRUE)
body_type <- craters$target_type[match(names(num_craters), craters$target)]
tab_craters <- data.frame("body" = names(num_craters), "type" = body_type,
                          "craters" = c(num_craters))
rownames(tab_craters) <- NULL
tab_craters

# Tests to be performed
type_tests <- c("PCvM", "PAD", "PRt")

# Select the bodies with more than 30 craters and that are either planets
# or moons
sel_bodies <- which(tab_craters$craters >= 30 & tab_craters$type != "Asteroid")
names_sel_bodies <- tab_craters$body[sel_bodies]
tab_craters[sel_bodies, ]

# How many filtered bodies?
length(sel_bodies)

# How many craters?
sum(tab_craters$craters[sel_bodies])

# Run the tests
tests_craters <- lapply(names_sel_bodies, function(body) {
  unif_test(data = craters$X[craters$target == body, , drop = FALSE],
            type = type_tests, p_value = "asymp", K_max = 5e4)
})
names(tests_craters) <- names_sel_bodies

# Matrix with the p-values for the bodies
bodies_pvalues <- t(sapply(tests_craters, function(body) {
  sapply(body, function(test) test$p.value)
}))
bodies_pvalues <- as.data.frame(cbind("n" = num_craters[sel_bodies],
                                      bodies_pvalues))
bodies_pvalues

## Rhea craters

# Load data
data(rhea)

# Add Cartesian coordinates
rhea$X <- cbind(cos(rhea$theta) * cos(rhea$phi),
                sin(rhea$theta) * cos(rhea$phi),
                sin(rhea$phi))

# Distribution of diameter
quantile(rhea$diameter)

# Subsets of craters, according to diameter
ind_15_20 <- rhea$diameter > 15 & rhea$diameter < 20
ind_20 <- rhea$diameter > 20
ind_15 <- rhea$diameter > 15

# Sample sizes
nrow(rhea)
sum(ind_15_20)
sum(ind_20)
sum(ind_15)

# Tests to be performed
type_tests <- c("PCvM", "PAD", "PRt")

# Tests
tests_rhea_15_20 <- unif_test(data = rhea$X[ind_15_20, ], type = type_tests,
                              p_value = "asymp", K_max = 5e4)
tests_rhea_20 <- unif_test(data = rhea$X[ind_20, ], type = type_tests,
                           p_value = "asymp", K_max = 5e4)
tests_rhea_15 <- unif_test(data = rhea$X[ind_15, ], type = type_tests,
                           p_value = "asymp", K_max = 5e4)
tests_rhea_15_20
tests_rhea_20
tests_rhea_15

## Venus craters

# Load data
data(venus)

# Add Cartesian coordinates
venus$X <- cbind(cos(venus$theta) * cos(venus$phi),
                 sin(venus$theta) * cos(venus$phi),
                 sin(venus$phi))

# Sample size
nrow(venus)

# Tests
type_tests <- c("PCvM", "PAD", "PRt")
tests_venus <- unif_test(data = venus$X, type = type_tests,
                         p_value = "asymp", K_max = 5e4)
tests_venus

#' @title Circular and (hyper)spherical uniformity statistics
#'
#' @description Implementation of several statistics for assessing uniformity
#' on the (hyper)sphere
#' \eqn{S^{p-1} := \{{\bf x} \in R^p : ||{\bf x}|| = 1\}}{
#' S^{p-1} := \{x \in R^p : ||x|| = 1\}}, \eqn{p\ge 2}, for a sample
#' \eqn{{\bf X}_1,\ldots,{\bf X}_n\in S^{p-1}}{X_1,\ldots,X_n\in S^{p-1}}.
#'
#' \code{unif_stat} receives a (several) sample(s) of directions in
#' \emph{Cartesian coordinates}, except for the circular case (\eqn{p=2}) in
#' which the sample(s) can be \emph{angles}
#' \eqn{\Theta_1,\ldots,\Theta_n\in [0, 2\pi)}.
#'
#' \code{unif_stat} allows to compute several statistics to several samples
#' within a single call, facilitating thus Monte Carlo experiments.
#'
#' @param data sample to compute the test statistic. An \bold{array} of size
#' \code{c(n, p, M)} containing \code{M} samples of size \code{n} of directions
#' (in Cartesian coordinates) on \eqn{S^{p-1}}. Alternatively, a
#' \bold{matrix} of size \code{c(n, M)} with the angles on \eqn{[0, 2\pi)} of
#' the \code{M} circular samples of size \code{n} on \eqn{S^{1}}. Other objects
#' accepted are an array of size \code{c(n, 1, M)} or a vector of size
#' \code{n} with angular data. Must not contain \code{NA}'s.
#' @inheritParams unif_test
#' @param data_sorted is the circular data sorted? If \code{TRUE}, certain
#' statistics are faster to compute. Defaults to \code{FALSE}.
#' @param Rayleigh_m integer \eqn{m} for the \eqn{m}-modal Rayleigh test.
#' Defaults to \code{m = 1} (the standard Rayleigh test).
#' @param cov_a \eqn{a_n = a / n} parameter used in the length of the arcs
#' of the coverage-based tests. Must be positive. Defaults to \code{2 * pi}.
#' @param Rothman_t \eqn{t} parameter for the Rothman test, a real in
#' \eqn{(0, 1)}. Defaults to \code{1 / 3}.
#' @param Cressie_t \eqn{t} parameter for the Cressie test, a real in
#' \eqn{(0, 1)}. Defaults to \code{1 / 3}.
#' @param Pycke_q \eqn{q} parameter for the Pycke "\eqn{q}-test", a real in
#' \eqn{(0, 1)}. Defaults to \code{1 / 2}.
#' @param Riesz_s \eqn{s} parameter for the \eqn{s}-Riesz test, a real in
#' \eqn{(0, 2)}. Defaults to \code{1}.
#' @param CCF09_dirs a matrix of size \code{c(n_proj, p)} containing
#' \code{n_proj} random directions (in Cartesian coordinates) on \eqn{S^{p-1}}
#' to perform the CCF09 test. If \code{NULL} (default), a sample of size
#' \code{n_proj = 50} directions is computed internally.
#' @param K_CCF09 integer giving the truncation of the series present in the
#' asymptotic distribution of the Kolmogorov-Smirnov statistic. Defaults to
#' \code{5e2}.
#' @param CJ12_reg type of asymptotic regime for CJ12 test, either \code{1}
#' (sub-exponential regime), \code{2} (exponential), or \code{3}
#' (super-exponential; default).
#' @return A data frame of size \code{c(M, length(type))}, with column names
#' given by \code{type}, that contains the values of the test statistics.
#' @details
#' Descriptions and references for most of the statistics are available
#' in García-Portugués and Verdebout (2018).
#' @references
#' García-Portugués, E. and Verdebout, T. (2018) An overview of uniformity
#' tests on the hypersphere. \emph{arXiv:1804.00286}.
#' \url{https://arxiv.org/abs/1804.00286}.
#' @examples
#' ## Circular data
#'
#' # Sample
#' n <- 10
#' M <- 2
#' Theta <- r_unif_cir(n = n, M = M)
#'
#' # Matrix
#' unif_stat(data = Theta, type = "all")
#'
#' # Array
#' unif_stat(data = array(Theta, dim = c(n, 1, M)), type = "all")
#'
#' # Vector
#' unif_stat(data = Theta[, 1], type = "all")
#'
#' ## Spherical data
#'
#' # Circular sample in Cartesian coordinates
#' n <- 10
#' M <- 2
#' X <- array(dim = c(n, 2, M))
#' for (i in 1:M) X[, , i] <- cbind(cos(Theta[, i]), sin(Theta[, i]))
#'
#' # Array
#' unif_stat(data = X, type = "all")
#'
#' # High-dimensional data
#' X <- r_unif_sph(n = n, p = 3, M = M)
#' unif_stat(data = X, type = "all")
#'
#' ## Specific arguments
#'
#' # Rothman
#' unif_stat(data = Theta, type = "Rothman", Rothman_t = 0.5)
#'
#' # CCF09
#' unif_stat(data = X, type = "CCF09", CCF09_dirs = X[, , 1])
#' unif_stat(data = X, type = "CCF09", CCF09_dirs = X[, , 1], K_CCF09 = 1)
#'
#' # CJ12
#' unif_stat(data = X, type = "CJ12", CJ12_reg = 3)
#' unif_stat(data = X, type = "CJ12", CJ12_reg = 1)
#' @export
unif_stat <- function(data, type = "all", data_sorted = FALSE,
                      Rayleigh_m = 1, cov_a = 2 * pi, Rothman_t = 1 / 3,
                      Cressie_t = 1 / 3, Pycke_q = 0.5, Riesz_s = 1,
                      CCF09_dirs = NULL, K_CCF09 = 25, CJ12_reg = 3) {

  # Stop if NA's
  if (anyNA(data)) {

    stop("NAs present in data, please remove them.")

  }

  # If data is a vector, transform it to matrix
  if (is.vector(data)) {

    data <- matrix(data, ncol = 1)

  }

  # Decode the type of data: vector, matrix, or array
  d <- dim(data)

  # Check sample size
  n <- d[1]
  if (n < 2) {

    stop("Sample size is one!")

  }

  # Array or matrix?
  if (length(d) == 3) {

    # Dimension and number of samples
    p <- d[2]
    M <- d[3]

    # Convert to polar coordinates?
    if (p == 2) {

      data <- X_to_Theta(X = data)

    } else if (p == 1) {

      data <- data[, 1, ]
      p <- 2

    }

  } else {

    # Dimension and number of samples
    p <- 2
    M <- d[2]

  }

  # Available statistics
  if (p == 2) {

    avail_stats <- avail_cir_tests

  } else {

    avail_stats <- avail_sph_tests

  }

  # Get the type of statistics
  if (is.character(type)) {

    type <- gsub("-", "_", type)
    type <- unique(type)
    if ("all" %in% type) {

      stats_type <- avail_stats

    } else {

      # Allow KS, CvM, and AD in unif_stat() (not in unif_test())
      if (p == 2) {

        avail_stats <- c(avail_stats, "KS", "CvM", "AD")

      }
      stats_type <- match.arg(arg = type, choices = avail_stats,
                              several.ok = TRUE)

    }

  } else if (is.numeric(type)) {

    type <- unique(type)
    stats_type <- avail_stats[type]

  } else {

    stop("type must be a character or a numeric vector.")

  }

  # Create a data frame for the statistics
  n_stats <- length(stats_type)
  stats <- matrix(0, nrow = M, ncol = n_stats)
  colnames(stats) <- stats_type
  stats <- as.data.frame(stats)

  # Compute efficiently several statistics
  if (p == 2) {

    # Statistics using gaps and sorted data
    stats_using_gaps <- c("Gini", "Gini_squared", "Greenwood", "Log_gaps",
                          "Max_uncover", "Num_uncover", "Range", "Rao",
                          "Vacancy")
    stats_using_sorted_data <- c("Cressie", "FG01", "Hodges_Ajne", "Kuiper",
                                 "Watson", "Watson_1976", "KS", "CvM", "AD",
                                 stats_using_gaps)

    # Statistics using the shortest angles matrix Psi
    stats_using_Psi <- c("Ajne", "Bakshaev", "Gine_Fn", "Gine_Gn",
                         "Hermans_Rasson", "PAD", "PCvM", "PRt", "Pycke",
                         "Pycke_q", "Rothman", "Riesz")

    # Evaluate which statistics to apply
    run_test <- as.list(c(avail_cir_tests, "KS", "CvM", "AD") %in% stats_type)
    names(run_test) <- c(avail_cir_tests, "KS", "CvM", "AD")

    ## Sorted-data statistics

    # Is it worth to sort Theta? Only if there is more than one statistic
    # using sorted Theta and if it has not already been done
    # CAUTION: replacement of data with sorted_data!
    n_stats_type_sorted <- sum(stats_type %in% stats_using_sorted_data)
    if (!data_sorted & (n_stats_type_sorted > 1)) {

      data <- sort_each_col(data)
      data_sorted <- TRUE

    }
    if (run_test$Cressie) {

      stats$Cressie <- cir_stat_Cressie(Theta = data, t = Cressie_t,
                                        sorted = data_sorted)

    }
    if (run_test$FG01) {

      stats$FG01 <- cir_stat_FG01(Theta = data,
                                                  sorted = data_sorted)

    }
    if (run_test$Hodges_Ajne) {

      stats$Hodges_Ajne <- cir_stat_Hodges_Ajne(Theta = data, asymp_std = FALSE,
                                                sorted = data_sorted,
                                                use_Cressie = TRUE)

    }
    if (run_test$Kuiper) {

      stats$Kuiper <- cir_stat_Kuiper(Theta = data, sorted = data_sorted,
                                      KS = FALSE, Stephens = FALSE)

    }
    if (run_test$Watson) {

      stats$Watson <- cir_stat_Watson(Theta = data, sorted = data_sorted,
                                      CvM = FALSE, Stephens = FALSE)

    }
    if (run_test$KS) {

      stats$KS <- cir_stat_Kuiper(Theta = data, sorted = data_sorted,
                                  KS = TRUE, Stephens = FALSE)

    }
    if (run_test$CvM) {

      stats$CvM <- cir_stat_Watson(Theta = data, sorted = data_sorted,
                                   CvM = TRUE, Stephens = FALSE)

    }
    if (run_test$AD) {

      stats$AD <- cir_stat_PAD(Theta = data, sorted = data_sorted,
                               AD = TRUE)

    }
    if (run_test$Watson_1976) {

      stats$Watson_1976 <- cir_stat_Watson_1976(Theta = data,
                                                sorted = data_sorted,
                                                minus = FALSE)

    }

    ## Gaps-based statistics

    # Is it worth to compute the circular gaps? Only if there is more than
    # one statistic using them
    n_stats_type_gaps <- sum(stats_type %in% stats_using_gaps)
    if (n_stats_type_gaps > 1) {

      gaps <- cir_gaps(Theta = data, sorted = data_sorted)
      gaps_in_Theta <- TRUE

    } else {

      gaps <- data
      gaps_in_Theta <- FALSE

    }
    if (run_test$Range) {

      stats$Range <- cir_stat_Range(Theta = gaps, gaps_in_Theta = gaps_in_Theta,
                                    max_gap = TRUE, sorted = data_sorted)

    }
    if (run_test$Rao) {

      stats$Rao <- cir_stat_Rao(Theta = gaps, gaps_in_Theta = gaps_in_Theta,
                                sorted = data_sorted)

    }
    if (run_test$Gini) {

      stats$Gini <- cir_stat_Gini(Theta = gaps, gaps_in_Theta = gaps_in_Theta,
                                  sorted = data_sorted)

    }
    if (run_test$Gini_squared) {

      stats$Gini_squared <- cir_stat_Gini_squared(Theta = gaps,
                                                  gaps_in_Theta = gaps_in_Theta,
                                                  sorted = data_sorted)

    }
    if (run_test$Greenwood) {

      stats$Greenwood <- cir_stat_Greenwood(Theta = gaps,
                                            gaps_in_Theta = gaps_in_Theta,
                                            sorted = data_sorted)

    }
    if (run_test$Log_gaps) {

      stats$Log_gaps <- cir_stat_Log_gaps(Theta = gaps,
                                          gaps_in_Theta = gaps_in_Theta,
                                          sorted = data_sorted)

    }
    if (run_test$Max_uncover) {

      stats$Max_uncover <- cir_stat_Max_uncover(Theta = gaps, a = cov_a,
                                                gaps_in_Theta = gaps_in_Theta,
                                                sorted = data_sorted)

    }
    if (run_test$Num_uncover) {

      stats$Num_uncover <- cir_stat_Num_uncover(Theta = gaps, a = cov_a,
                                                gaps_in_Theta = gaps_in_Theta,
                                                sorted = data_sorted)

    }
    if (run_test$Vacancy) {

      stats$Vacancy <- cir_stat_Vacancy(Theta = gaps, a = cov_a,
                                        gaps_in_Theta = gaps_in_Theta,
                                        sorted = data_sorted)

    }

    ## Other statistics

    if (run_test$Rayleigh) {

      stats$Rayleigh <- cir_stat_Rayleigh(Theta = data, m = Rayleigh_m)

    }
    if (run_test$Bingham) {

      stats$Bingham <- cir_stat_Bingham(Theta = data)

    }
    if (run_test$CCF09) {

      # Sample random directions
      if (is.null(CCF09_dirs)) {

        CCF09_dirs <- r_unif_sph(n = 50, p = 2, M = 1)[, , 1]

      }
      stats$CCF09 <- cir_stat_CCF09(Theta = data, dirs = CCF09_dirs,
                                    K_CCF09 = K_CCF09, original = FALSE)

    }

    ## Psi-based statistics

    # Is it worth to compute Psi? Only if there is more than one statistic
    # using it (after discounting equivalent tests if they are simultaneously
    # present) AND if 0.5 * n * (n - 1) * M is not too large.
    # CAUTION: replacement of data with Psi!
    n_stats_type_Psi <- sum(stats_type %in% stats_using_Psi)
    n_stats_type_Psi <- n_stats_type_Psi - ((run_test$Watson && run_test$PCvM) +
         (run_test$Rothman && run_test$PRt) + (Riesz_s == 2) +
         (Riesz_s == 0 && run_test$Pycke && run_test$Riesz) +
         (Riesz_s == 1 && run_test$Bakshaev && run_test$Riesz) +
         (run_test$Gine_Fn && run_test$Gine_Gn && run_test$Ajne))
    if (n_stats_type_Psi > 1 & 0.5 * n * (n - 1) * M <= 1e8) {

      dim(data) <- c(n, 1, M)
      data <- Psi_mat(data = data)
      Psi_in_Theta <- TRUE

    } else {

      Psi_in_Theta <- FALSE

    }
    if (run_test$Ajne) {

      stats$Ajne <- cir_stat_Ajne(Theta = data, Psi_in_Theta = Psi_in_Theta)

    }
    if (run_test$Bakshaev) {

      stats$Bakshaev <- cir_stat_Riesz(Theta = data,
                                       Psi_in_Theta = Psi_in_Theta, s = 1)

    }
    if (run_test$Riesz) {

      if (Riesz_s == 1 && run_test$Bakshaev) {

        stats$Riesz <- stats$Bakshaev

      } else if (Riesz_s == 2 && (run_test$Rayleigh | !Psi_in_Theta)) {

        if (run_test$Rayleigh) {

          stats$Riesz <- stats$Rayleigh

        } else {

          stats$Riesz <- cir_stat_Rayleigh(Theta = data, m = 1)

        }

      } else {

        stats$Riesz <- cir_stat_Riesz(Theta = data, Psi_in_Theta = Psi_in_Theta,
                                      s = Riesz_s)

      }

    }
    if (run_test$Gine_Gn) {

      stats$Gine_Gn <- cir_stat_Gine_Gn(Theta = data,
                                        Psi_in_Theta = Psi_in_Theta)

    }
    if (run_test$Gine_Fn) {

      if (run_test$Ajne & run_test$Gine_Gn) {

        stats$Gine_Fn <- 4 * stats$Ajne + stats$Gine_Gn

      } else {

        stats$Gine_Fn <- cir_stat_Gine_Fn(Theta = data,
                                          Psi_in_Theta = Psi_in_Theta)

      }

    }
    if (run_test$Hermans_Rasson) {

      stats$Hermans_Rasson <-
        cir_stat_Hermans_Rasson(Theta = data, Psi_in_Theta = Psi_in_Theta)

    }
    if (run_test$PAD) {

      stats$PAD <- cir_stat_PAD(Theta = data, Psi_in_Theta = Psi_in_Theta,
                                AD = FALSE, sorted = FALSE)

    }
    if (run_test$PCvM) {

      if (run_test$Watson) {

        stats$PCvM <- 2 * stats$Watson

      } else {

        stats$PCvM <- cir_stat_PCvM(Theta = data, Psi_in_Theta = Psi_in_Theta)

      }

    }
    if (run_test$Rothman) {

      stats$Rothman <- cir_stat_Rothman(Theta = data, t = Rothman_t,
                                        Psi_in_Theta = Psi_in_Theta)

    }
    if (run_test$PRt) {

      if (run_test$Rothman) {

        stats$PRt <- stats$Rothman

      } else {

        stats$PRt <- cir_stat_PRt(Theta = data, Psi_in_Theta = Psi_in_Theta,
                                  t = Rothman_t)

      }

    }
    if (run_test$Pycke) {

      if (Riesz_s == 0 && run_test$Riesz) {

        stats$Pycke <- (2 * n) / (n - 1) * stats$Riesz

      } else {

        stats$Pycke <- cir_stat_Pycke(Theta = data, Psi_in_Theta = Psi_in_Theta)

      }

    }
    if (run_test$Pycke_q) {

      stats$Pycke_q <- cir_stat_Pycke_q(Theta = data,
                                        Psi_in_Theta = Psi_in_Theta,
                                        q = Pycke_q)

    }

  } else {

    # Statistics using the shortest angles matrix Psi
    stats_using_Psi <- c("Ajne", "Bakshaev", "CJ12", "Gine_Fn", "Gine_Gn",
                         "PAD", "PCvM", "PRt", "Pycke", "Riesz")

    # Evaluate which statistics to apply
    run_test <- as.list(avail_sph_tests %in% stats_type)
    names(run_test) <- avail_sph_tests

    ## Other statistics

    if (run_test$Bingham) {

      stats$Bingham <- sph_stat_Bingham(X = data)

    }
    if (run_test$CCF09) {

      # Sample random directions
      if (is.null(CCF09_dirs)) {

        CCF09_dirs <- r_unif_sph(n = 50, p = p, M = 1)[, , 1]

      }
      stats$CCF09 <- sph_stat_CCF09(X = data, dirs = CCF09_dirs,
                                    K_CCF09 = K_CCF09, original = FALSE)

    }
    if (run_test$Rayleigh) {

      stats$Rayleigh <- sph_stat_Rayleigh(X = data)

    }
    if (run_test$Rayleigh_HD) {

      if (run_test$Rayleigh) {

        stats$Rayleigh_HD <- (stats$Rayleigh - p) / sqrt(2 * p)

      } else {

        stats$Rayleigh_HD <- sph_stat_Rayleigh_HD(X = data)

      }

    }

    ## Psi-based statistics

    # Is it worth to compute Psi? Only if there is more than one statistic
    # using it (after discounting equivalent tests if they are simultaneously
    # present) AND if 0.5 * n * (n - 1) * M is not too large.
    # CAUTION: replacement of data with Psi!
    n_stats_type_Psi <- sum(stats_type %in% stats_using_Psi)
    n_stats_type_Psi <- n_stats_type_Psi - ((Riesz_s == 2) +
         (Riesz_s == 0 && run_test$Pycke && run_test$Riesz) +
         (Riesz_s == 1 && run_test$Bakshaev && run_test$Riesz) +
         (run_test$PCvM && run_test$Bakshaev && p == 3) +
         (run_test$Gine_Fn && run_test$Gine_Gn && run_test$Ajne))
    if (n_stats_type_Psi > 1 & 0.5 * n * (n - 1) * M <= 1e8) {

      data <- Psi_mat(data = data)
      dim(data) <- c(dim(data), 1)
      Psi_in_X <- TRUE

    } else {

      Psi_in_X <- FALSE

    }
    if (run_test$Ajne) {

      stats$Ajne <- sph_stat_Ajne(X = data, Psi_in_X = Psi_in_X)

    }
    if (run_test$Bakshaev) {

      stats$Bakshaev <- sph_stat_Riesz(X = data, Psi_in_X = Psi_in_X, p = p,
                                       s = 1)

    }
    if (run_test$Riesz) {

      if (Riesz_s == 1 && run_test$Bakshaev) {

        stats$Riesz <- stats$Bakshaev

      } else if (Riesz_s == 2 && (run_test$Rayleigh | !Psi_in_X)) {

        if (run_test$Rayleigh) {

          stats$Riesz <- (2 / p) * stats$Rayleigh

        } else {

          stats$Riesz <- (2 / p) * sph_stat_Rayleigh(X = data)

        }

      } else {

        stats$Riesz <- sph_stat_Riesz(X = data, Psi_in_X = Psi_in_X, p = p,
                                      s = Riesz_s)

      }

    }
    if (run_test$CJ12) {

      if (Psi_in_X) {

        stats$CJ12 <- sph_stat_CJ12(X = cos(data), Psi_in_X = TRUE, p = p,
                                  regime = CJ12_reg)

      } else {

        stats$CJ12 <- sph_stat_CJ12(X = data, Psi_in_X = FALSE, p = p,
                                  regime = CJ12_reg)

      }

    }
    if (run_test$Gine_Gn) {

      stats$Gine_Gn <- sph_stat_Gine_Gn(X = data, Psi_in_X = Psi_in_X, p = p)

    }
    if (run_test$Gine_Fn) {

      if (run_test$Ajne & run_test$Gine_Gn) {

        stats$Gine_Fn <- 4 * stats$Ajne + stats$Gine_Gn

      } else {

        stats$Gine_Fn <- sph_stat_Gine_Fn(X = data, Psi_in_X = Psi_in_X, p = p)

      }

    }
    if (run_test$PAD) {

      stats$PAD <- sph_stat_PAD(X = data, Psi_in_X = Psi_in_X, p = p)

    }
    if (run_test$PCvM) {

      if (run_test$Bakshaev & p == 3) {

        stats$PCvM <- 0.125 * stats$Bakshaev

      } else {

        stats$PCvM <- sph_stat_PCvM(X = data, Psi_in_X = Psi_in_X, p = p)

      }

    }
    if (run_test$PRt) {

      stats$PRt <- sph_stat_PRt(X = data, t = Rothman_t, Psi_in_X = Psi_in_X,
                                p = p)

    }
    if (run_test$Pycke) {

      if (Riesz_s == 0 && run_test$Riesz) {

        if (p == 3) {

          stats$Pycke <- n / (2 * pi * (n - 1)) *
            (stats$Riesz - (log(4) - 1) / 2)

        } else {

          warning(paste("Pycke statistic is only defined for p = 2,3.",
                        "Using Riesz statistic with s = 0 instead,",
                        "which behaves consistently across dimensions."))
          stats$Pycke <- stats$Riesz

        }

      } else {

        if (Psi_in_X) {

          stats$Pycke <- sph_stat_Pycke(X = cos(data), Psi_in_X = TRUE, p = p)

        } else {

          stats$Pycke <- sph_stat_Pycke(X = data, Psi_in_X = FALSE, p = p)

        }

      }

    }

  }

  return(stats)

}

## Retrieve data from https://astrogeology.usgs.gov/search/map/Venus/venuscraters

# The source is the USGS Astrogeology Science Center

# Download
download.file(url = paste0("https://astropedia.astrogeology.usgs.gov/",
                           "download/Venus/venuscraters.csv"),
              destfile = "venuscraters.csv")

## Import and preprocess data

# Read data
venus <- read.csv(file = "venuscraters.csv", header = FALSE, sep = ",",
                  dec = ".")[, 1:4]
names(venus)[1:4] <- c("name", "Center_Latitude", "Center_Longitude", "diameter")

# Remove NAs (there are empty rows)
noNA <- complete.cases(cbind(venus$Center_Latitude, venus$Center_Longitude))
venus <- venus[noNA, ]

# Remove initial spaces in names
venus$name <- gsub(pattern = " (.*)", replacement = "\\1", x = venus$name)

# Transform degrees to radians
venus$theta <- (venus$Center_Longitude / 180 * pi) %% (2 * pi)
venus$phi <- venus$Center_Latitude / 180 * pi
venus$Center_Latitude <- NULL
venus$Center_Longitude <- NULL

# Save object
save(list = "venus", file = "venus.rda", compress = "bzip2")

#' @title Statistics for testing (hyper)spherical uniformity
#'
#' @description Low-level implementation of several statistics for assessing
#' uniformity on the (hyper)sphere
#' \eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{S^{p-1}:=
#' \{x\in R^p:||x||=1\}}, \eqn{p\ge 2}.
#'
#' @param X an \bold{array} of size \code{c(n, p, M)} containing the Cartesian
#' coordinates of \code{M} samples of size \code{n} of directions on
#' \eqn{S^{p-1}}. Must not contain \code{NA}'s.
#' @param Psi_in_X does \code{X} contain the shortest angles matrix
#' \eqn{\boldsymbol\Psi}{\Psi} that is obtained with \code{\link{Psi_mat}(X)}?
#' If \code{FALSE} (default), \eqn{\boldsymbol\Psi}{\Psi} is computed
#' internally.
#' @inheritParams unif_stat_distr
#' @inheritParams unif_stat
#' @inheritParams sph_stat_distr
#' @inheritParams cir_stat
#' @param dirs a matrix of size \code{c(n_proj, p)} containing \code{n_proj}
#' random directions (in Cartesian coordinates) on \eqn{S^{p-1}} to perform
#' the CCF09 test.
#' @param N number of points used in the
#' \link[=Gauss_Legen_nodes]{Gauss-Legendre quadrature}. Defaults to
#' \code{160}.
#' @param L number of discretization points to interpolate angular functions
#' that require evaluating an integral. Defaults to \code{1e3}.
#' @return A matrix of size \code{c(M, 1)} containing the statistics for each
#' of the \code{M} samples.
#' @section Warning:
#' Be careful on avoiding the next bad usages of the functions, which will
#' produce spurious results:
#' \itemize{
#'   \item The directions in \code{X} do \emph{not} have unit norm.
#'   \item \code{X} does \emph{not} contain \code{Psi_mat(X)} when
#'   \code{X_in_Theta = TRUE}.
#'   \item The parameter \code{p} does \emph{not} match with the dimension of
#'   \eqn{R^p}.
#'   \item \emph{Not} passing the scalar products matrix to \code{sph_stat_CJ12}
#'   when \code{Psi_in_X = TRUE}.
#'   \item The directions in \code{dirs} do \emph{not} have unit norm.
#' }
#' @details
#' Detailed descriptions and references of the statistics are available
#' in García-Portugués and Verdebout (2018).
#'
#' The Pycke and CJ12 statistics employ the \emph{scalar products} matrix,
#' rather than the shortest angles matrix, when \code{Psi_in_X = TRUE}. This
#' matrix is obtained by setting \code{scalar_prod = TRUE} in
#' \code{\link{Psi_mat}}.
#' @references
#' García-Portugués, E. and Verdebout, T. (2018) An overview of uniformity
#' tests on the hypersphere. \emph{arXiv:1804.00286}.
#' \url{https://arxiv.org/abs/1804.00286}.
#' @examples
#' ## Sample uniform spherical data
#'
#' M <- 2
#' n <- 100
#' p <- 3
#' set.seed(123456789)
#' X <- r_unif_sph(n = n, p = p, M = M)
#'
#' ## Sobolev tests
#'
#' # Rayleigh
#' sph_stat_Rayleigh(X)
#'
#' # Bingham
#' sph_stat_Bingham(X)
#'
#' # Ajne
#' Psi <- Psi_mat(X)
#' dim(Psi) <- c(dim(Psi), 1)
#' sph_stat_Ajne(X)
#' sph_stat_Ajne(Psi, Psi_in_X = TRUE)
#'
#' # Gine Gn
#' sph_stat_Gine_Gn(X)
#' sph_stat_Gine_Gn(Psi, Psi_in_X = TRUE, p = p)
#'
#' # Gine Fn
#' sph_stat_Gine_Fn(X)
#' sph_stat_Gine_Fn(Psi, Psi_in_X = TRUE, p = p)
#'
#' # Pycke
#' sph_stat_Pycke(X)
#' sph_stat_Pycke(Psi, Psi_in_X = TRUE, p = p)
#'
#' # Bakshaev
#' sph_stat_Bakshaev(X)
#' sph_stat_Bakshaev(Psi, Psi_in_X = TRUE, p = p)
#'
#' # Riesz
#' sph_stat_Riesz(X, s = 1)
#' sph_stat_Riesz(Psi, Psi_in_X = TRUE, p = p, s = 1)
#'
#' # Projected Cramér-von Mises
#' sph_stat_PCvM(X)
#' sph_stat_PCvM(Psi, Psi_in_X = TRUE, p = p)
#'
#' # Projected Rothman
#' sph_stat_PRt(X)
#' sph_stat_PRt(Psi, Psi_in_X = TRUE, p = p)
#'
#' # Projected Anderson-Darling
#' sph_stat_PAD(X)
#' sph_stat_PAD(Psi, Psi_in_X = TRUE, p = p)
#'
#' ## Other tests
#'
#' # CCF09
#' dirs <- r_unif_sph(n = 3, p = p, M = 1)[, , 1]
#' sph_stat_CCF09(X, dirs = dirs)
#'
#' ## High-dimensional tests
#'
#' # Rayleigh HD-Standardized
#' sph_stat_Rayleigh_HD(X)
#'
#' # CJ12
#' sph_stat_CJ12(X, regime = 1)
#' sph_stat_CJ12(Psi, regime = 1, Psi_in_X = TRUE, p = p)
#' sph_stat_CJ12(X, regime = 2)
#' sph_stat_CJ12(Psi, regime = 2, Psi_in_X = TRUE, p = p)
#' sph_stat_CJ12(X, regime = 3)
#' sph_stat_CJ12(Psi, regime = 3, Psi_in_X = TRUE, p = p)
#' @name sph_stat
NULL

#' @title Utilities for projected-ecdf statistics of spherical uniformity
#'
#' @description Computation of the kernels
#' \deqn{\psi_p^W(\theta) := \int_{-1}^1 A_x(\theta)\,\mathrm{d}W(F_p(x)),}{
#' \psi_p^W(\theta) := \int_{-1}^1 A_x(\theta) dW(F_p(x)),
#' }
#' where \eqn{A_x(\theta)} is the proportion of area surface of
#' \eqn{S^{p - 1}} covered by the
#' \link[=A_theta_x]{intersection of two hyperspherical caps} with common solid
#' angle \eqn{\pi - \cos^{-1}(x)} and centers separated by
#' an angle \eqn{\theta \in [0, \pi]}, \eqn{F_p} is the distribution function
#' of the \link[=p_proj_unif]{projected spherical uniform distribution},
#' and \eqn{W} is a measure on \eqn{[0, 1]}.
#'
#' Also, computation of the \link[=Gegen_coefs]{Gegenbauer coefficients} of
#' \eqn{\psi_p^W}:
#' \deqn{b_{k, p}^W := \frac{1}{c_{k, p}}\int_0^\pi \psi_p^W(\theta)
#' C_k^{p / 2 - 1}(\cos\theta)\,\mathrm{d}\theta.}{
#' b_{k, p}^W := \frac{1}{c_{k, p}} \int_0^\pi \psi_p^W(\theta)
#' C_k^(p / 2 - 1)(\cos\theta) d\theta.}
#' These coefficients can also be computed via
#' \deqn{b_{k, p}^W = \int_{-1}^1 a_{k, p}^x\,\mathrm{d}W(F_p(x))}{
#' b_{k, p}^W = \int_{-1}^1 a_{k, p}^x dW(F_p(x))}
#' for a certain function \eqn{x\rightarrow a_{k, p}^x}. They serve to define
#' \link[=locdev]{projected alternatives to uniformity}.
#' @inheritParams Gegenbauer
#' @param q integer giving the dimension of the sphere \eqn{S^q}.
#' @param type type of projected-ecdf test statistic. Must be either
#' \code{"PCvM"} (Cramér--von Mises), \code{"PAD"} (Anderson--Darling), or
#' \code{"PRt"} (Rothman).
#' @inheritParams unif_stat
#' @param tilde include the constant and bias term? Defaults to \code{FALSE}.
#' @param psi_Gauss use a \link[=Gauss_Legen_nodes]{Gauss--Legendre quadrature}
#' rule with \code{psi_N} nodes in the computation of the kernel function?
#' Defaults to \code{TRUE}.
#' @param psi_N number of points used in the Gauss--Legendre quadrature for
#' computing the kernel function. Defaults to \code{320}.
#' @param verbose flag to print informative messages. Defaults to \code{FALSE}.
#' @param x evaluation points for \eqn{a_{k, p}^x}, a vector with values in
#' \eqn{[-1, 1]}.
#' @param sqr return the \emph{signed} square root of \eqn{a_{k, p}^x}?
#' Defaults to \code{FALSE}.
#' @param K number of equispaced points on \eqn{[-1, 1]} used for evaluating
#' \eqn{f} and then interpolating. Defaults to \code{1e3}.
#' @inheritParams locdev
#' @return
#' \itemize{
#'   \item \code{psi_Pn}: a vector of size \code{length(theta)} with the
#'   evaluation of \eqn{\psi}.
#'   \item \code{Gegen_coefs_Pn}: a vector of size \code{length(k)} containing
#'   the coefficients \eqn{b_{k, p}^W}.
#'   \item \code{akx}: a matrix of size \code{c(length(x), length(k))}
#'   containing the coefficients \eqn{a_{k, p}^x}.
#'   \item \code{f_locdev_Pn}: the projected alternative \eqn{f} as a function
#'   ready to be evaluated.
#' }
#' @author Eduardo García-Portugués and Paula Navarro-Esteban.
#' @details
#' The evaluation of \eqn{\psi_p^W} and \eqn{b_{k, p}^W} depends on the type of
#' projected-ecdf statistic:
#' \itemize{
#' \item PCvM: closed-form expressions for \eqn{\psi_p^W} and \eqn{b_{k, p}^W}
#' with \eqn{p = 2, 3, 4}, numerical integration required for \eqn{p \ge 5}.
#' \item PAD: closed-form expressions for \eqn{\psi_2^W} and \eqn{b_{k, 3}^W},
#' numerical integration required for \eqn{\psi_p^W} with \eqn{p \ge 3} and
#' \eqn{b_{k, p}^W} with \eqn{p = 2} and \eqn{p \ge 4}.
#' \item PRt: closed-form expressions for \eqn{\psi_p^W} and \eqn{b_{k, p}^W}
#' for any \eqn{p \ge 2}.
#' }
#' See García-Portugués et al. (2020) for more details.
#' @references
#' García-Portugués, E., Navarro-Esteban, P., Cuesta-Albertos, J. A. (2020)
#' On a projection-based class of uniformity tests on the hypersphere.
#' \emph{arXiv:2008.09897}. \url{https://arxiv.org/abs/2008.09897}
#' @examples
#' # Kernels in the projected-ecdf test statistics
#' k <- 0:10
#' coefs <- list()
#' (coefs$PCvM <- t(sapply(2:5, function(p)
#'   Gegen_coefs_Pn(k = k, p = p, type = "PCvM"))))
#' (coefs$PAD <- t(sapply(2:5, function(p)
#'   Gegen_coefs_Pn(k = k, p = p, type = "PAD"))))
#' (coefs$PRt <- t(sapply(2:5, function(p)
#'   Gegen_coefs_Pn(k = k, p = p, type = "PRt"))))
#'
#' # Gegenbauer expansion
#' th <- seq(0, pi, length.out = 501)[-501]
#' old_par <- par(mfrow = c(3, 4))
#' for (type in c("PCvM", "PAD", "PRt")) {
#'
#'   for (p in 2:5) {
#'
#'     plot(th, psi_Pn(theta = th, q = p - 1, type = type), type = "l",
#'          main = paste0(type, ", p = ", p), xlab = expression(theta),
#'          ylab = expression(psi(theta)), axes = FALSE, ylim = c(-1.5, 1))
#'     axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
#'          labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
#'     axis(2); box()
#'     lines(th, Gegen_series(theta = th, coefs = coefs[[type]][p - 1, ],
#'                            k = k, p = p), col = 2)
#'
#'   }
#'
#' }
#' par(old_par)
#'
#' # Analytical coefficients vs. numerical integration
#' test_coef <- function(type, p, k = 0:20) {
#'
#'   plot(k, log1p(abs(Gegen_coefs_Pn(k = k, p = p, type = type))),
#'        ylab = "Coefficients", main = paste0(type, ", p = ", p))
#'   points(k, log1p(abs(Gegen_coefs(k = k, p = p, psi = psi_Pn, type = type,
#'                                   q = p - 1))), col = 2)
#'   legend("topright", legend = c("log(1 + Gegen_coefs_Pn))",
#'                                 "log(1 + Gegen_coefs(psi_Pn))"),
#'          lwd = 2, col = 1:2)
#'
#' }
#'
#' # PCvM statistic
#' old_par <- par(mfrow = c(2, 2))
#' for (p in 2:5) test_coef(type = "PCvM", p = p)
#' par(old_par)
#'
#' # PAD statistic
#' old_par <- par(mfrow = c(2, 2))
#' for (p in 2:5) test_coef(type = "PAD", p = p)
#' par(old_par)
#'
#' # PRt statistic
#' old_par <- par(mfrow = c(2, 2))
#' for (p in 2:5) test_coef(type = "PRt", p = p)
#' par(old_par)
#'
#' # akx
#' akx(x = seq(-1, 1, l = 5), k = 1:4, p = 2)
#' akx(x = 0, k = 1:4, p = 3)
#'
#' # PRt alternative to uniformity
#' z <- seq(-1, 1, l = 1e3)
#' p <- c(2:5, 10, 15, 17)
#' col <- viridisLite::viridis(length(p))
#' plot(z, f_locdev_Pn(p = p[1], type = "PRt")(z), type = "s",
#'      col = col[1], ylim = c(0, 0.6), ylab = expression(f[Rt](z)))
#' for (k in 2:length(p)) {
#'   lines(z, f_locdev_Pn(p = p[k], type = "PRt")(z), type = "s", col = col[k])
#' }
#' legend("topleft", legend = paste("p =", p), col = col, lwd = 2)
#' @name Pn


#' @rdname Pn
#' @export
psi_Pn <- function(theta, q, type, Rothman_t = 1 / 3, tilde = FALSE,
                   psi_Gauss = TRUE, psi_N = 320, tol = 1e-6) {

  # Check dimension
  stopifnot(q >= 1)

  # Check theta
  stopifnot(all(0 <= theta & theta <= pi))

  # Type of statistic
  if (type == "PCvM") {

    # Compute psi_q
    ps <-
      switch(as.character(q),
             "1" = { # psi_1

               1 / 2 + theta * (theta - 2 * pi)  / (4 * pi^2)

             },
             "2" = { # psi_2

               1 / 2 - sin(theta / 2) / 4

             },
             "3" = { # psi_3

               psi_Pn(theta = theta, q = 1, type = "PCvM") +
                 ((pi - theta) * tan(theta / 2) - 2 * sin(theta / 2)^2) /
                 (4 * pi^2)

             },
             { # psi_q, q >= 4

               # Explicit part
               psi_q <- -3 / 4 + theta / (2 * pi) +
                 2 * drop(p_proj_unif(x = cos(theta / 2), p = q + 1))^2

               # Integral
               psi_q <- psi_q -
                 4 * sapply(theta, function(th) {

                   # Gauss--Legendre quadrature? Otherwise, call integrate
                   if (psi_Gauss) {

                     # Nodes and weights
                     t_k <- Gauss_Legen_nodes(a = 0, b = cos(th / 2), N = psi_N)
                     w_k <- Gauss_Legen_weights(a = 0, b = cos(th / 2),
                                                N = psi_N)

                     # Integrand evaluation
                     f_k <- exp(p_proj_unif(x = t_k, p = q + 1, log = TRUE) +
                                  p_proj_unif(x = tan(th / 2) * t_k /
                                                sqrt(1 - t_k^2),
                                              p = q, log = TRUE) +
                                  d_proj_unif(x = t_k, p = q + 1, log = TRUE))

                     # Remove non-finite data
                     f_k[!is.finite(f_k)] <- NA

                     # Integration
                     sum(w_k * f_k, na.rm = TRUE)

                   } else {

                     # Integrand
                     f <- function(t) {

                       f_k <- exp(p_proj_unif(x = t, p = q + 1, log = TRUE) +
                                    p_proj_unif(x = tan(th / 2) * t /
                                                  sqrt(1 - t^2),
                                                p = q, log = TRUE) +
                                    d_proj_unif(x = t, p = q + 1, log = TRUE))
                       f_k[!is.finite(f_k)] <- NA
                       return(f_k)

                     }

                     # Integrate
                     integrate(f = f, lower = 0, upper = cos(th / 2),
                               subdivisions = 1e3, rel.tol = tol, abs.tol = tol,
                               stop.on.error = FALSE)$value

                   }

                 })

             }
      )

    # Add constant?
    if (tilde) {

      ps <- ps - 1 / 3

    }

  } else if (type == "PAD") {

    # Compute psi_q
    ps <-
      switch(as.character(q),
             "1" = { # psi_1

               -2 * log(2 * pi) + (theta * log(theta) + (2 * pi - theta) *
                                     log(2 * pi - theta)) / pi

             },
             "2" = { # psi_2

               -log(4) + 2 / pi * sapply(theta, function(th) {

                 # Gauss--Legendre quadrature? Otherwise, call integrate
                 if (psi_Gauss) {

                   # Nodes and weights
                   t_k <- Gauss_Legen_nodes(a = 0, b = cos(th / 2), N = psi_N)
                   w_k <- Gauss_Legen_weights(a = 0, b = cos(th / 2), N = psi_N)

                   # Integrand evaluation
                   f_k <- log((1 + t_k) / (1 - t_k)) *
                     acos(tan(th / 2) * t_k / sqrt(1 - t_k^2))

                   # Remove non-finite data
                   f_k[!is.finite(f_k)] <- NA

                   # Integration
                   sum(w_k * f_k, na.rm = TRUE)

                 } else {

                   # Integrand
                   f <- function(t) {

                     f_k <- log((1 + t) / (1 - t)) *
                       acos(tan(th / 2) * t / sqrt(1 - t^2))
                     f_k[!is.finite(f_k)] <- NA
                     return(f_k)

                   }

                   # Integrate
                   integrate(f = f, lower = 0, upper = cos(th / 2),
                             subdivisions = 1e3, rel.tol = tol, abs.tol = tol,
                             stop.on.error = FALSE)$value

                 }

               })

             },
             "3" = { # psi_3

               s_theta <- theta - sin(theta)
               -2 * log(2 * pi) + (s_theta * log(s_theta) + (2 * pi - s_theta) *
                                     log(2 * pi - s_theta)) / pi -
                 4 / pi * tan(theta / 2) * sapply(theta, function(th) {

                 # Gauss--Legendre quadrature? Otherwise, call integrate
                 if (psi_Gauss) {

                   # Nodes and weights
                   t_k <- Gauss_Legen_nodes(a = 0, b = cos(th / 2), N = psi_N)
                   w_k <- Gauss_Legen_weights(a = 0, b = cos(th / 2), N = psi_N)

                   # Integrand evaluation
                   f_k <- t_k *
                     log(pi / (acos(t_k) - t_k * sqrt(1 - t_k^2)) - 1)

                   # Remove non-finite data
                   f_k[!is.finite(f_k)] <- NA

                   # Integration
                   sum(w_k * f_k, na.rm = TRUE)

                 } else {

                   # Integrand
                   f <- function(t) {

                     f_k <- t * log(pi / (acos(t) - t * sqrt(1 - t^2)) - 1)
                     f_k[!is.finite(f_k)] <- NA
                     return(f_k)

                   }

                   # Integrate
                   integrate(f = f, lower = 0, upper = cos(th / 2),
                             subdivisions = 1e3, rel.tol = tol, abs.tol = tol,
                             stop.on.error = FALSE)$value

                 }

               })

             },
             { # psi_q, q >= 4

               -log(4) + 4 * sapply(theta, function(th) {

                 # Gauss--Legendre quadrature? Otherwise, call integrate
                 if (psi_Gauss) {

                   # Nodes and weights
                   t_k <- Gauss_Legen_nodes(a = 0, b = cos(th / 2), N = psi_N)
                   w_k <- Gauss_Legen_weights(a = 0, b = cos(th / 2), N = psi_N)

                   # Integrand evaluation
                   log_F_k <- p_proj_unif(x = t_k, p = q + 1, log = TRUE) -
                     p_proj_unif(x = -t_k, p = q + 1, log = TRUE)
                   f_k <- log_F_k *
                     exp(p_proj_unif(x = -tan(th / 2) * t_k / sqrt(1 - t_k^2),
                                     p = q, log = TRUE) +
                           d_proj_unif(x = t_k, p = q + 1, log = TRUE))

                   # Remove non-finite data
                   f_k[!is.finite(f_k)] <- NA

                   # Integration
                   sum(w_k * f_k, na.rm = TRUE)

                 } else {

                   # Integrand
                   f <- function(t) {

                     log_F_k <- p_proj_unif(x = t, p = q + 1, log = TRUE) -
                       p_proj_unif(x = -t, p = q + 1, log = TRUE)
                     f_k <- log_F_k *
                       exp(p_proj_unif(x = -tan(th / 2) * t / sqrt(1 - t^2),
                                       p = q, log = TRUE) +
                             d_proj_unif(x = t, p = q + 1, log = TRUE))
                     f_k[!is.finite(f_k)] <- NA
                     return(f_k)

                   }

                   # Integrate
                   integrate(f = f, lower = 0, upper = cos(th / 2),
                             subdivisions = 1e3, rel.tol = tol, abs.tol = tol,
                             stop.on.error = FALSE)$value

                 }

               })

             }
      )

    # Set NaNs to 0, as they correspond to theta = 0
    ps[!is.finite(ps)] <- 0

    # Add constant?
    if (tilde) {

      ps <- ps + 1

    }

  } else if (type == "PRt") {

    # Separation angle
    t_m <- min(Rothman_t, 1 - Rothman_t)
    theta_t_m <- 2 * acos(drop(q_proj_unif(u = 1 - t_m, p = q + 1)))

    # Two cases
    psi_q <- numeric(length(theta))
    ind <- (theta >= theta_t_m)

    # Explicit part: theta >= theta_t_m
    psi_q[ind] <- 1 / 2 - t_m

    # Compute psi_q
    ps <-
      switch(as.character(q),
             "1" = { # psi_1

               -pmin(theta / (2 * pi) - t_m * (1 - t_m), t_m^2) +
                 1 / 2 - t_m * (1 - t_m)

             },
             "2" = { # psi_2

               # Longer part
               psi_q[!ind] <- -t_m + 1 / 2 - (1 - 2 * t_m) / pi *
                 acos((1 / 2 - t_m) / sqrt(t_m * (1 - t_m)) *
                        tan(theta[!ind] / 2)) +
                 atan(sqrt(cos(theta[!ind] / 2)^2 - (1 - 2 * t_m)^2) /
                        sin(theta[!ind] / 2)) / pi
               psi_q

             },
             "3" = { # psi_3

               # Longer part
               psi_q[!ind] <- 1 / 2 + t_m -
                 (theta[!ind] + theta_t_m) / (2 * pi) +
                 (sin(theta_t_m) / 2 + tan(theta[!ind] / 2) *
                    cos(theta_t_m / 2)^2) / pi
               psi_q

             },
             { # psi_q, q >= 4

               # Integral part
               psi_q[!ind] <-
                 t_m - theta[!ind] / (2 * pi) +
                 2 * sapply(theta[!ind], function(th) {

                   # Gauss--Legendre quadrature? Otherwise, call integrate
                   if (psi_Gauss) {

                     # Nodes and weights
                     u_k <- Gauss_Legen_nodes(a = 0, b = cos(theta_t_m / 2),
                                              N = psi_N)
                     w_k <- Gauss_Legen_weights(a = 0, b = cos(theta_t_m / 2),
                                                N = psi_N)

                     # Integrand
                     f_k <- p_proj_unif(x = (u_k / sqrt(1 - u_k^2)) *
                                          tan(th / 2), p = q) *
                       d_proj_unif(x = u_k, p = q + 1)

                     # Remove non-finite data
                     f_k[!is.finite(f_k)] <- NA

                     # Integration
                     sum(w_k * f_k, na.rm = TRUE)

                   } else {

                     # Integrand
                     f <- function(u) {

                       f_k <- p_proj_unif(x = (u / sqrt(1 - u^2)) *
                                            tan(th / 2), p = q) *
                         d_proj_unif(x = u, p = q + 1)
                       f_k[!is.finite(f_k)] <- NA
                       return(f_k)

                     }

                     # Integrate
                     integrate(f = f, lower = 0, upper = cos(theta_t_m / 2),
                               subdivisions = 1e3, rel.tol = tol, abs.tol = tol,
                               stop.on.error = FALSE)$value

                   }

                 })
               psi_q

             }
      )

    # Add constant?
    if (tilde) {

      ps <- ps + t_m * (1 - t_m) - 1 / 2

    }

  } else {

    stop("type must be either \"PCvM\", \"PAD\", or \"PRt\".")

  }

  return(ps)

}


#' @rdname Pn
#' @export
Gegen_coefs_Pn <- function(k, p, type, Rothman_t = 1 / 3, Gauss = TRUE,
                           N = 320, tol = 1e-6, verbose = FALSE) {

  # Check dimension
  stopifnot(p >= 2)

  # Check orders
  stopifnot(all(k >= 0))

  # Coefficients
  coefs <- numeric(length(k))

  # Positive indexes
  ind1 <- k > 0
  k1 <- k[ind1]
  dok1 <- any(ind1)

  # Type of statistic
  if (type == "PCvM") {

    # k = 0
    coefs[!ind1] <- 1 / 3

    # k > 0
    if (dok1) {

      if (p == 2) {

        coefs[ind1] <- 1 / (k1 * pi)^2

      } else if (p == 3) {

        coefs[ind1] <- 1 / (8 * (k1 + 3 / 2) * (k1 - 1 / 2))

      } else if (p == 4) {

        coefs[ind1] <- (k1 == 1) * (35 / (72 * pi^2)) +
          (k1 > 1) / (2 * pi^2) * exp(log(3 * k1^2 + 6 * k1 + 4) - 2 * log(k1) -
                                        log(k1 + 1) - 2 * log(k1 + 2))

      } else {

        # Gauss--Legendre quadrature?
        if (Gauss) {

          # Nodes and weights
          w_k <- drop(Gauss_Legen_weights(a = -1, b = 1, N = N))
          x_k <- drop(Gauss_Legen_nodes(a = -1, b = 1, N = N))

          # k > 0
          coefs[ind1] <- colSums(w_k * akx(x = x_k, p = p, k = k1,
                                           sqr = FALSE) *
                                   drop(d_proj_unif(x = x_k, p = p)),
                                 na.rm = TRUE)

        } else {

          # k > 0
          coefs[ind1] <- sapply(k1, function(l) {
            f <- function(x) drop(akx(x = x, p = p, k = l, sqr = FALSE) *
                                    d_proj_unif(x = x, p = p))
            integrate(f = f, lower = -1, upper = 1,
                      subdivisions = 1e3, abs.tol = tol, rel.tol = tol,
                      stop.on.error = FALSE)$value
          })

        }
        if (verbose) {

          message("Gegenbauer coefficients of the ", type,
                  " statistic with p = ", p, " unknown in analytical form, ",
                  "using numerical integration of akx.")

        }

      }

    }

  } else if (type == "PAD") {

    # k = 0
    coefs[!ind1] <- -1

    # k > 0
    if (dok1) {

      if (p == 2) {

        # Gauss--Legendre quadrature?
        if (Gauss) {

          # Nodes and weights
          th_k <- drop(Gauss_Legen_nodes(a = 0, b = pi, N = N))
          w_k <- drop(Gauss_Legen_weights(a = 0, b = pi, N = N))

          # Integrand evaluation
          f_k <- (1 - cos(2 * th_k %o% k1)) / ((pi - th_k) * th_k)

          # k > 0
          coefs[ind1] <- colSums(w_k * f_k) / (pi * k1^2)

        } else {

          # k > 0
          coefs[ind1] <- sapply(k1, function(m) {
            f <- function(th) (1 - cos(2 * th * m)) / ((pi - th) * th)
            integrate(f = f, lower = 0 + tol, upper = pi - tol,
                      subdivisions = 1e3, abs.tol = tol, rel.tol = tol,
                      stop.on.error = FALSE)$value
          }) / (pi * k1^2)

        }

      } else if (p == 3) {

        # k > 0
        coefs[ind1] <- 1 / ((k1 + 1) * k1)

      } else {

        # Gauss--Legendre quadrature?
        if (Gauss) {

          # Nodes and weights
          stopifnot(N > 5)
          w_k <- drop(Gauss_Legen_weights(a = -1, b = 1, N = N))
          x_k <- drop(Gauss_Legen_nodes(a = -1, b = 1, N = N))

          # Weight evaluation
          dWF_k <- drop(p_proj_unif(x = x_k, p = p, log = TRUE))
          dWF_k <- rep(dWF_k[1:(N / 2)] + dWF_k[(N / 2 + 1):N], 2)
          dWF_k <- exp(drop(d_proj_unif(x = x_k, p = p, log = TRUE)) - dWF_k)
          dWF_k[!is.finite(dWF_k)] <- NA

          # k > 0
          coefs[ind1] <-
            colSums(w_k * akx(x = x_k, p = p, k = k1, sqr = FALSE) * dWF_k,
                    na.rm = TRUE)

        } else {

          # k > 0
          coefs[ind1] <- sapply(k1, function(l) {
            f <- function(x) {
              dWF_k <- p_proj_unif(x = x, p = p, log = TRUE) +
                p_proj_unif(x = -x, p = p, log = TRUE)
              dWF_k <- drop(exp(d_proj_unif(x = x, p = p, log = TRUE) - dWF_k))
              dWF_k[!is.finite(dWF_k)] <- NA
              drop(akx(x = x, p = p, k = l, sqr = FALSE) * dWF_k)
            }
            integrate(f = f, lower = -1, upper = 1, subdivisions = 1e3,
                      abs.tol = tol, rel.tol = tol, stop.on.error = FALSE)$value
          })

        }
        if (verbose) {

          message("Gegenbauer coefficients of the ", type,
                  " statistic with p = ", p, " unknown in analytical form, ",
                  "using numerical integration of akx.")

        }

      }

    }

  } else if (type == "PRt") {

    # k = 0
    t_m <- min(Rothman_t, 1 - Rothman_t)
    coefs[!ind1] <- 0.5 - t_m * (1 - t_m)

    # k > 0
    if (dok1) {

      # Call akx
      coefs[ind1] <- ifelse(p == 2, 2, 1) *
        drop(akx(x = drop(q_proj_unif(u = t_m, p = p)), p = p, k = k1,
                 sqr = FALSE))

    }

  } else {

    stop("type must be either \"PCvM\", \"PAD\", or \"PRt\".")

  }

  return(coefs)

}


#' @rdname Pn
#' @export
akx <- function(x, p, k, sqr = FALSE) {

  # Check dimension
  stopifnot(p >= 2)

  # Check orders
  stopifnot(all(k >= 0))

  # Coeffiecients
  a <- matrix(nrow = length(x), ncol = length(k))

  # Positive orders
  ind1 <- k > 0
  k1 <- k[ind1]
  dok1 <- any(ind1)

  # k = 0
  a[, !ind1] <- drop(p_proj_unif(x = x, p = p))

  # k > 0
  if (dok1) {

    if (p == 2) {

      a[, ind1] <- t(sin(k1 %o% acos(x)) / (k1 * pi))

    } else {

      q <- p - 1
      a[, ind1] <- t(Gegen_polyn(theta = acos(x), k = k1 - 1, p = p + 2))
      a[, ind1] <- a[, ind1] *
        t(exp(0.5 * log(1 + 2 * k1 / (q - 1)) +
                ((q - 1) * log(2) + 2 * lgamma((q + 1) / 2) +
                   lgamma(k1) - log(pi) - lgamma(k1 + q)) +
                rep(q / 2, length(k1)) %o% log(1 - x^2)))

    }

  }

  # Return sqrt(a) or a
  if (!sqr) {

    a <- a^2

  }
  return(a)

}


#' @rdname Pn
#' @export
f_locdev_Pn <- function(p, type, K = 1e3, N = 320, K_max = 1e4, thre = 1e-3,
                        Rothman_t = 1 / 3, verbose = FALSE) {

  # Exact f for PRt family
  if (type %in% c("PRt", "Rothman", "Ajne")) {

    t <- ifelse(type == "Ajne", 0.5, Rothman_t)
    x_t <- drop(q_proj_unif(u = t, p = p))
    f <- function(x) ((x >= x_t) + t) / rotasym::w_p(p = p)

  # Exact f for PCvM and p = 2
  } else if (type == "PCvM" & p == 2) {

    f <- function(x) (1 - log(2 * (1 - x)) * sqrt(2) / (2 * pi)) / (2 * pi)

  # Series expansion otherwise
  } else {

    uk <- cutoff_locdev(p = p, K_max = K_max, type = type, thre = thre,
                        Rothman_t = Rothman_t, verbose = verbose,
                        Gauss = TRUE, N = N)
    z <- seq(-1, 1, l = 1e4)
    f <- splinefun(x = z, y = pmax(f_locdev(z = z, p = p, uk = uk), 0))

  }
  return(f)

}

#' @title \code{sphunif}: Uniformity Tests on the Circle, Sphere, and
#' Hypersphere
#'
#' @description Implementation of uniformity tests on the circle and
#' (hyper)sphere. The main function of the package is \code{\link{unif_test}},
#' which conveniently collects more than 30 tests for assessing uniformity on
#' \eqn{S^{p-1}=\{{\bf x}\in R^p:||{\bf x}||=1\}}{
#' S^{p-1}=\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}. The test statistics are
#' implemented in the \code{\link{unif_stat}} function, which allows computing
#' several statistics to different samples within a single call, thus
#' facilitating Monte Carlo experiments. Furthermore, the
#' \code{\link{unif_stat_MC}} function allows parallelizing them in
#' a simple way. The asymptotic null distributions of the statistics are
#' available through the function \code{\link{unif_stat_distr}}. The core of
#' \code{\link{sphunif-package}} is coded in C++ by relying on the
#' \code{\link[Rcpp]{Rcpp-package}}. The package also provides several
#' novel datasets and gives the reproducibility for the data application
#' in García-Portugués, Navarro-Esteban and Cuesta-Albertos (2020)
#' <arXiv:2008.09897>.
#'
#' @author Eduardo García-Portugués and Thomas Verdebout.
#' @references
#' García-Portugués, E. and Verdebout, T. (2018) An overview of uniformity
#' tests on the hypersphere. \emph{arXiv:1804.00286}.
#' \url{https://arxiv.org/abs/1804.00286}.
#'
#' García-Portugués, E., Navarro-Esteban, P., Cuesta-Albertos, J. A. (2020)
#' On a projection-based class of uniformity tests on the hypersphere.
#' \emph{arXiv:2008.09897}. \url{https://arxiv.org/abs/2008.09897}
#'
#' García-Portugués, E., Navarro-Esteban, P., and Cuesta-Albertos, J. A. (2021).
#' A Cramér–von Mises test of uniformity on the hypersphere. In Balzano, S.,
#' Porzio, G. C., Salvatore, R., Vistocco, D., and Vichi, M. (Eds.), \emph{
#' Statistical Learning and Modeling in Data Analysis}, Studies in
#' Classification, Data Analysis and Knowledge Organization, pp. 107–-116.
#' Springer, Cham. \doi{10.1007/978-3-030-69944-4_12}.
#'
#' García-Portugués, E., Paindaveine, D., and Verdebout, T. (2021). On the
#' power of Sobolev tests for isotropy under local rotationally symmetric
#' alternatives. \emph{arXiv:2108.09874}. \url{https://arxiv.org/abs/2108.09874}
#' @docType package
#' @name sphunif-package
#' @import graphics Rcpp stats
#' @useDynLib sphunif
#' @aliases sphunif sphunif-package
NULL

#' @rdname proj_unif
#' @export
r_proj_unif <- function(n, p) {

  # Check p
  if (p <= 1) {

    stop("p must be >= 2.")

  }

  # The projection is the square of a B(1 / 2, (p - 1)/2)
  return(sqrt(rbeta(n = n, shape1 = 0.5, shape2 = (p - 1) / 2)) *
           sample(c(-1, 1), size = n, replace = TRUE))

}

## Retrieve data from https://ssd.jpl.nasa.gov/txt/aprx_pos_planets.pdf

# Extract I (inclination in degs) and Omega (longitude of the ascending
# node in degs) from Table 2a
planets <- matrix(c(7.00559432, 48.33961819,
                    3.39777545, 76.67261496,
                    0, 0,
                    1.85181869, 49.71320984,
                    1.29861416, 100.29282654,
                    2.49424102, 113.63998702,
                    0.77298127, 73.96250215,
                    1.77005520, 131.78635853,
                    17.14104260, 110.30167986),
                  ncol = 2, byrow = TRUE)

# As data.frame
names <- c("Mercury","Venus", "Earth", "Mars", "Jupiter",
           "Saturn", "Uranus", "Neptune", "Pluto")
planets <- data.frame("planet" = names, "i" = planets[, 1], "om" = planets[, 2])

# Transform degrees to radians
planets$i <- planets$i / 180 * pi
planets$om <- planets$om / 180 * pi

# Save object
save(list = "planets", file = "planets.rda", compress = "bzip2")

#' @title Asymptotic distributions for spherical uniformity statistics
#'
#' @description Computation of the asymptotic null distributions of
#' spherical uniformity statistics.
#'
#' @inheritParams cir_stat_distr
#' @inheritParams Sobolev
#' @inheritParams r_unif
#' @inheritParams wschisq
#' @inheritParams unif_stat_distr
#' @param regime type of asymptotic regime for the CJ12 test, either \code{1}
#' (sub-exponential regime), \code{2} (exponential), or \code{3}
#' (super-exponential; default).
#' @param beta \eqn{\beta} parameter in the exponential regime of the CJ12
#' test, a nonnegative real. Defaults to \code{0}.
#' @inheritParams cir_stat
#' @param ... further parameters passed to \code{\link{p_Sobolev}} or
#' \code{\link{d_Sobolev}} (such as \code{x_tail}).
#' @return
#' \itemize{
#'   \item \code{r_sph_stat_*}: a matrix of size \code{c(n, 1)} containing
#'   the sample.
#'   \item \code{p_sph_stat_*}, \code{d_sph_stat_*}: a matrix of size
#'   \code{c(nx, 1)} with the evaluation of the distribution or density
#'   functions at \code{x}.
#' }
#' @details
#' Descriptions and references on most of the asymptotic distributions
#' are available in García-Portugués and Verdebout (2018).
#' @examples
#' # Ajne
#' curve(d_sph_stat_Ajne(x, p = 3, method = "HBE"), n = 2e2, ylim = c(0, 4))
#' curve(p_sph_stat_Ajne(x, p = 3, method = "HBE"), n = 2e2, col = 2,
#'       add = TRUE)
#'
#' # Bakshaev
#' curve(d_sph_stat_Bakshaev(x, p = 3, method = "HBE"), to = 5, n = 2e2,
#'       ylim = c(0, 2))
#' curve(p_sph_stat_Bakshaev(x, p = 3, method = "HBE"), n = 2e2, col = 2,
#'       add = TRUE)
#'
#' # Bingham
#' curve(d_sph_stat_Bingham(x, p = 3), to = 20, n = 2e2, ylim = c(0, 1))
#' curve(p_sph_stat_Bingham(x, p = 3), n = 2e2, col = 2, add = TRUE)
#'
#' # CJ12
#' curve(d_sph_stat_CJ12(x, regime = 1), from = -10, to = 10, n = 2e2,
#'       ylim = c(0, 1))
#' curve(d_sph_stat_CJ12(x, regime = 2, beta = 0.1), n = 2e2, col = 2,
#'       add = TRUE)
#' curve(d_sph_stat_CJ12(x, regime = 3), n = 2e2, col = 3, add = TRUE)
#' curve(p_sph_stat_CJ12(x, regime = 1), n = 2e2, col = 1, add = TRUE)
#' curve(p_sph_stat_CJ12(x, regime = 2, beta = 0.1), n = 2e2, col = 2,
#'       add = TRUE)
#' curve(p_sph_stat_CJ12(x, regime = 3), col = 3, add = TRUE)
#'
#' # Gine Fn
#' curve(d_sph_stat_Gine_Fn(x, p = 3, method = "HBE"), to = 2, n = 2e2,
#'       ylim = c(0, 2))
#' curve(p_sph_stat_Gine_Fn(x, p = 3, method = "HBE"), n = 2e2, col = 2,
#'       add = TRUE)
#'
#' # Gine Gn
#' curve(d_sph_stat_Gine_Gn(x, p = 3, method = "HBE"), to = 1.5, n = 2e2,
#'       ylim = c(0, 2.5))
#' curve(p_sph_stat_Gine_Gn(x, p = 3, method = "HBE"), n = 2e2, col = 2,
#'       add = TRUE)
#'
#' # PAD
#' curve(d_sph_stat_PAD(x, p = 3, method = "HBE"), to = 3, n = 2e2,
#'       ylim = c(0, 1.5))
#' curve(p_sph_stat_PAD(x, p = 3, method = "HBE"), n = 2e2, col = 2,
#'       add = TRUE)
#'
#' # PCvM
#' curve(d_sph_stat_PCvM(x, p = 3, method = "HBE"), to = 0.6, n = 2e2,
#'       ylim = c(0, 7))
#' curve(p_sph_stat_PCvM(x, p = 3, method = "HBE"), n = 2e2, col = 2,
#'       add = TRUE)
#'
#' # PRt
#' curve(d_sph_stat_PRt(x, p = 3, method = "HBE"), n = 2e2, ylim = c(0, 5))
#' curve(p_sph_stat_PRt(x, p = 3, method = "HBE"), n = 2e2, col = 2, add = TRUE)
#'
#' # Rayleigh
#' curve(d_sph_stat_Rayleigh(x, p = 3), to = 15, n = 2e2, ylim = c(0, 1))
#' curve(p_sph_stat_Rayleigh(x, p = 3), n = 2e2, col = 2, add = TRUE)
#'
#' # HD-standardized Rayleigh
#' curve(d_sph_stat_Rayleigh_HD(x, p = 3), from = -4, to = 4, n = 2e2,
#'       ylim = c(0, 1))
#' curve(p_sph_stat_Rayleigh_HD(x, p = 3), n = 2e2, col = 2, add = TRUE)
#'
#' # Riesz
#' curve(d_sph_stat_Riesz(x, p = 3, method = "HBE"), n = 2e2, from = 0, to = 5,
#'       ylim = c(0, 2))
#' curve(p_sph_stat_Riesz(x, p = 3, method = "HBE"), n = 2e2, col = 2,
#'       add = TRUE)
#' @name sph_stat_distr
NULL


#' @rdname sph_stat_distr
#' @export
p_sph_stat_Ajne <- function(x, p, K_max = 1e3, thre = 0, ...) {

  cbind(p_Sobolev(x = x, p = p, type = "Ajne", K_max = K_max, thre = thre, ...))

}


#' @rdname sph_stat_distr
#' @export
d_sph_stat_Ajne <- function(x, p, K_max = 1e3, thre = 0, ...) {

  cbind(d_Sobolev(x = x, p = p, type = "Ajne", K_max = K_max, thre = thre, ...))

}


#' @rdname sph_stat_distr
#' @export
p_sph_stat_Bakshaev <- function(x, p, K_max = 1e3, thre = 0, ...) {

  cbind(p_Sobolev(x = x, p = p, type = "Bakshaev", K_max = K_max, thre = thre,
                  ...))

}


#' @rdname sph_stat_distr
#' @export
d_sph_stat_Bakshaev <- function(x, p, K_max = 1e3, thre = 0, ...) {

  cbind(d_Sobolev(x = x, p = p, type = "Bakshaev", K_max = K_max, thre = thre,
                  ...))

}


#' @rdname sph_stat_distr
#' @export
p_sph_stat_Gine_Fn <- function(x, p, K_max = 1e3, thre = 0, ...) {

  cbind(p_Sobolev(x = x, p = p, type = "Gine_Fn", K_max = K_max, thre = thre,
                  ...))

}


#' @rdname sph_stat_distr
#' @export
d_sph_stat_Gine_Fn <- function(x, p, K_max = 1e3, thre = 0, ...) {

  cbind(d_Sobolev(x = x, p = p, type = "Gine_Fn", K_max = K_max, thre = thre,
                  ...))

}


#' @rdname sph_stat_distr
#' @export
p_sph_stat_Gine_Gn <- function(x, p, K_max = 1e3, thre = 0, ...) {

  cbind(p_Sobolev(x = x, p = p, type = "Gine_Gn", K_max = K_max, thre = thre,
                  ...))

}


#' @rdname sph_stat_distr
#' @export
d_sph_stat_Gine_Gn <- function(x, p, K_max = 1e3, thre = 0, ...) {

  cbind(d_Sobolev(x = x, p = p, type = "Gine_Gn", K_max = K_max, thre = thre,
                  ...))

}


#' @rdname sph_stat_distr
#' @export
p_sph_stat_PAD <- function(x, p, K_max = 1e3, thre = 0, ...) {

  cbind(p_Sobolev(x = x, p = p, type = "PAD", K_max = K_max, thre = thre,
                  ...))

}


#' @rdname sph_stat_distr
#' @export
d_sph_stat_PAD <- function(x, p, K_max = 1e3, thre = 0, ...) {

  cbind(d_Sobolev(x = x, p = p, type = "PAD", K_max = K_max, thre = thre,
                  ...))

}


#' @rdname sph_stat_distr
#' @export
p_sph_stat_PCvM <- function(x, p, K_max = 1e3, thre = 0, ...) {

  cbind(p_Sobolev(x = x, p = p, type = "PCvM", K_max = K_max, thre = thre,
                  ...))

}


#' @rdname sph_stat_distr
#' @export
d_sph_stat_PCvM <- function(x, p, K_max = 1e3, thre = 0, ...) {

  cbind(d_Sobolev(x = x, p = p, type = "PCvM", K_max = K_max, thre = thre,
                  ...))

}


#' @rdname sph_stat_distr
#' @export
p_sph_stat_PRt <- function(x, p, t = 1 / 3, K_max = 1e3, thre = 0, ...) {

  cbind(p_Sobolev(x = x, p = p, type = "PRt", Rothman_t = t,
                  K_max = K_max, thre = thre, ...))

}


#' @rdname sph_stat_distr
#' @export
d_sph_stat_PRt <- function(x, p, t = 1 / 3, K_max = 1e3, thre = 0, ...) {

  cbind(d_Sobolev(x = x, p = p, type = "PRt", Rothman_t = t,
                  K_max = K_max, thre = thre, ...))

}


#' @rdname sph_stat_distr
#' @export
p_sph_stat_Riesz <- function(x, p, s = 1, K_max = 1e3, thre = 0, ...) {

  cbind(p_Sobolev(x = x, p = p, type = "Riesz", Riesz_s = s, K_max = K_max,
                  thre = thre, ...))

}


#' @rdname sph_stat_distr
#' @export
d_sph_stat_Riesz <- function(x, p, s = 1, K_max = 1e3, thre = 0, ...) {

  cbind(d_Sobolev(x = x, p = p, type = "Riesz", Riesz_s = s, K_max = K_max,
                  thre = thre, ...))

}

## Retrieve data from https://ssd.jpl.nasa.gov/sbdb_query.cgi

# The source is NASA's Jet Propusion Laboratory Small-Body Database
# Search Engine

# The database was accessed on the 2020-05-07 using the following query:
#
# Step 1.
# - Limit by object type/group:
#   * object group:    [x] All objects [ ] NEOs      [ ] PHAs
#   * object kind:     [ ] All objects [ ] Asteroids [x] Comets
#   * numbered state:  [x] All objects [ ] Numbered  [ ] Unnumbered
# - Limit to selected orbit class(es):
#   * Asteroid Orbit Classes: mark none
#   * Comet Orbit Classes: mark all
# - Limit by object characteristics: add no constraints
#
# Step 2.
# - Object Fields: "object internal database ID", "object full name/designation"
# - Orbital and Model Parameter Fields:
#   "[i] inclination; angle with respect to x-y ecliptic plane (deg)",
#   "longitude of the ascending node (deg)", "sidereal orbital period (years)",
#   "orbit classification"
#
# Step 3, 4, 5. Not relevant.
#
# Step 6. [ ] HTML (default) [x] CSV (download)
#
# The returned dataset was saved as "comets-accessed-2020-05-07.csv". The urls
# with the database query (done in 2020-05-07) are
# CSV: https://ssd.jpl.nasa.gov/sbdb_query.cgi?obj_group=all;obj_kind=com;obj_numbered=all;com_orbit_class=HYP;com_orbit_class=ETc;com_orbit_class=PAR;com_orbit_class=CTc;com_orbit_class=JFC;com_orbit_class=JFc;com_orbit_class=HTC;com_orbit_class=COM;OBJ_field=0;ORB_field=0;table_format=CSV;format_option=full;query=Generate%20Table;c_fields=AaAcBjBkBsCiAp;c_sort=;.cgifields=format_option;.cgifields=ast_orbit_class;.cgifields=table_format;.cgifields=obj_kind;.cgifields=obj_group;.cgifields=obj_numbered;.cgifields=com_orbit_class
# HTML: https://ssd.jpl.nasa.gov/sbdb_query.cgi?obj_group=all;obj_kind=com;obj_numbered=all;com_orbit_class=HYP;com_orbit_class=ETc;com_orbit_class=PAR;com_orbit_class=CTc;com_orbit_class=JFC;com_orbit_class=JFc;com_orbit_class=HTC;com_orbit_class=COM;OBJ_field=0;ORB_field=0;table_format=HTML;max_rows=500;format_option=full;query=Generate%20Table;c_fields=AaAcBjBkBsCiAp;c_sort=;.cgifields=format_option;.cgifields=ast_orbit_class;.cgifields=table_format;.cgifields=obj_kind;.cgifields=obj_group;.cgifields=obj_numbered;.cgifields=com_orbit_class

# Alternatively
download.file(url = paste0("https://ssd.jpl.nasa.gov/sbdb_query.cgi?",
                           "obj_group=all;obj_kind=com;obj_numbered=all;",
                           "com_orbit_class=HYP;com_orbit_class=ETc;",
                           "com_orbit_class=PAR;com_orbit_class=CTc;",
                           "com_orbit_class=JFC;com_orbit_class=JFc;",
                           "com_orbit_class=HTC;com_orbit_class=COM;",
                           "OBJ_field=0;ORB_field=0;table_format=CSV;",
                           "format_option=full;query=Generate%20Table;",
                           "c_fields=AaAcBjBkBsCiAp;c_sort=;",
                           ".cgifields=format_option;",
                           ".cgifields=ast_orbit_class;",
                           ".cgifields=table_format;.cgifields=obj_kind;",
                           ".cgifields=obj_group;.cgifields=obj_numbered;",
                           ".cgifields=com_orbit_class"),
              destfile = paste0("comets-accessed-", Sys.Date(), ".csv"))

# The classes of comet orbits are:
#
# - "COM" = Comet. Comet orbit not matching any defined orbit class.
# - "CTc" = Chiron-type Comet. Chiron-type comet, as defined by Levison and
#           Duncan (T_Jupiter > 3; a > a_Jupiter).
# - "ETc" = Encke-type Comet. Encke-type comet, as defined by Levison and
#           Duncan (T_Jupiter > 3; a < a_Jupiter).
# - "HTC" = Halley-type Comet*. Halley-type comet, classical definition
#           (20y < P < 200y).
# - "HYP" = Hyperbolic Comet. Comets on hyperbolic orbits (e > 1.0).
# - "JFc" = Jupiter-family Comet. Jupiter-family comet, as defined by
#           Levison and Duncan (2 < T_Jupiter < 3).
# - "JFC" = Jupiter-family Comet*. Jupiter-family comet, classical
#           definition (P < 20y).
# - "PAR" = Parabolic Comet. Comets on parabolic orbits (e = 1.0).

## Import and preprocess data

# Read data
comets <- read.csv(file = "comets-accessed-2020-05-07.csv", header = TRUE)

# Transform degrees to radians
comets$i <- comets$i / 180 * pi
comets$om <- comets$om / 180 * pi

# Normal vectors to the ecliptic planes of the comets
comets_normal <- cbind(sin(comets$i) * sin(comets$om),
                       -sin(comets$i) * cos(comets$om),
                       cos(comets$i))

## Compare with the data in Cuesta-Albertos et al. (2009)

# Read data directly retrieved from the authors' file
comets_ccf2009 <- readxl::read_xls("LP-unumb.xls")
comets_ccf2009_normal <- cbind(comets_ccf2009$...5,
                               comets_ccf2009$...6,
                               comets_ccf2009$...7)
ccf2009 <- match(comets_ccf2009$id, comets$id)

# Check equality of observations
max(rowSums((comets_normal[ccf2009, ] - comets_ccf2009_normal)^2))

# Add flag to larger database
comets$ccf2009 <- FALSE
comets$ccf2009[ccf2009] <- TRUE

# Save object
save(list = "comets", file = "comets.rda", compress = "bzip2")

#' @title Local projected alternatives to uniformity
#'
#' @description Density and random generation for local projected alternatives
#' to uniformity with densities
#' \deqn{f_{\kappa, \boldsymbol{\mu}}({\bf x}): =
#' \frac{1 - \kappa}{\omega_p} + \kappa f({\bf x}'\boldsymbol{\mu})}{
#' f_{\kappa, \mu}(x) = (1 - \kappa) / \omega_p + \kappa f(x'\mu)}
#' where
#' \deqn{f(z) = \frac{1}{\omega_p}\left\{1 + \sum_{k = 1}^\infty u_{k, p}
#' C_k^{p / 2 - 1}(z)\right\}}{f(x) = (1 / \omega_p)
#' \{1 + \sum_{k = 1}^\infty u_{k, p} C_k^(p / 2 - 1)(z)\}}
#' is the \emph{angular function} controlling the local alternative in a
#' \link[=Gegenbauer]{Gegenbauer series}, \eqn{0\le \kappa \le 1},
#' \eqn{\boldsymbol{\mu}}{\mu} is a direction on \eqn{S^{p - 1}}, and
#' \eqn{\omega_p} is the surface area of \eqn{S^{p - 1}}. The sequence
#' \eqn{\{u_{k, p}\}} is typically such that
#' \eqn{u_{k, p} = \left(1 + \frac{2k}{p - 2}\right) b_{k, p}}{
#' u_{k, p} = (1 + 2k / (p - 2)) b_{k, p}} for the Gegenbauer coefficients
#' \eqn{\{b_{k, p}\}} of the kernel function of a Sobolev statistic (see the
#' \link[=Sobolev_coefs]{transformation} between the coefficients \eqn{u_{k, p}}
#' and \eqn{b_{k, p}}).
#'
#' Also, automatic truncation of the series \eqn{\sum_{k = 1}^\infty u_{k, p}
#' C_k^{p / 2 - 1}(z)}{\sum_{k = 1}^\infty u_{k, p} C_k^(p / 2 - 1)(z)}
#' according to the proportion of \link[=Gegenbauer]{"Gegenbauer norm"}
#' explained.
#'
#' @param z projected evaluation points for \eqn{f}, a vector with entries on
#' \eqn{[-1, 1]}.
#' @inheritParams Sobolev_coefs
#' @inheritParams rotasym::d_tang_norm
#' @param mu a unit norm vector of size \code{p} giving the axis of rotational
#' symmetry.
#' @param f angular function defined on \eqn{[-1, 1]}. Must be vectorized.
#' @param kappa the strength of the local alternative, between \code{0}
#' and \code{1}.
#' @inheritParams r_unif
#' @param F_inv quantile function associated to \eqn{f}. Computed by
#' \code{\link{F_inv_from_f}} if \code{NULL} (default).
#' @inheritParams Gegenbauer
#' @param ... further parameters passed to \code{\link{F_inv_from_f}}.
#' @param K_max integer giving the truncation of the series. Defaults to
#' \code{1e4}.
#' @param thre proportion of norm \emph{not} explained by the first terms of the
#' truncated series. Defaults to \code{1e-3}.
#' @inheritParams Sobolev
#' @param verbose output information about the truncation (\code{TRUE} or
#' \code{1}) and a diagnostic plot (\code{2})? Defaults to \code{FALSE}.
#' @return
#' \itemize{
#'   \item \code{f_locdev}: angular function evaluated at \code{x}, a vector.
#'   \item \code{con_f}: normalizing constant \eqn{c_f} of \eqn{f}, a scalar.
#'   \item \code{d_locdev}: density function evaluated at \code{x}, a vector.
#'   \item \code{r_locdev}: a matrix of size \code{c(n, p)} containing a random
#'   sample from the density \eqn{f_{\kappa, \boldsymbol{\mu}}}{
#'   f_{\kappa, \mu}}.
#'   \item \code{cutoff_locdev}: vector of coefficients \eqn{\{u_{k, p}\}}
#'   automatically truncated according to \code{K_max} and \code{thre}
#'   (see details).
#' }
#' @details
#' See the definitions of local alternatives in Prentice (1978) and in
#' García-Portugués et al. (2020).
#'
#' The truncation of \eqn{\sum_{k = 1}^\infty u_{k, p} C_k^{p / 2 - 1}(z)}{
#' \sum_{k = 1}^\infty u_{k, p} C_k^(p / 2 - 1)(z)} is done to the first
#' \code{K_max} terms and then up to the index such that the first terms
#' leave unexplained the proportion \code{thre} of the norm of the whole series.
#' Setting \code{thre = 0} truncates to \code{K_max} terms exactly. If the
#' series only contains odd or even non-zero terms, then only \code{K_max / 2}
#' addends are \emph{effectively} taken into account in the first truncation.
#' @references
#' García-Portugués, E., Navarro-Esteban, P., Cuesta-Albertos, J. A. (2020)
#' On a projection-based class of uniformity tests on the hypersphere.
#' \emph{arXiv:2008.09897}. \url{https://arxiv.org/abs/2008.09897}
#'
#' Prentice, M. J. (1978). On invariant tests of uniformity for directions and
#' orientations. \emph{The Annals of Statistics}, 6(1):169--176.
#' \doi{10.1214/aos/1176344075}
#' @examples
#' ## Local alternatives diagnostics
#'
#' loc_alt_diagnostic  <- function(p, type, thre = 1e-3, K_max = 1e3) {
#'
#'   # Coefficients of the alternative
#'   uk <- cutoff_locdev(K_max = K_max, p = p, type = type, thre = thre,
#'                       N = 640)
#'
#'   old_par <- par(mfrow = c(2, 2))
#'
#'   # Construction of f
#'   z <- seq(-1, 1, l = 1e3)
#'   f <- function(z) f_locdev(z = z, p = p, uk = uk)
#'   plot(z, f(z), type = "l", xlab = expression(z), ylab = expression(f(z)),
#'        main = paste0("Local alternative f, ", type, ", p = ", p), log = "y")
#'
#'   # Projected density on [-1, 1]
#'   f_proj <- function(z) rotasym::w_p(p = p - 1) * f(z) *
#'     (1 - z^2)^((p - 3) / 2)
#'   plot(z, f_proj(z), type = "l", xlab = expression(z),
#'        ylab = expression(omega[p - 1] * f(z) * (1 - z^2)^{(p - 3) / 2}),
#'        main = paste0("Projected density, ", type, ", p = ", p), log = "y",
#'        sub = paste("Integral:", round(con_f(f = f, p = p), 4)))
#'
#'   # Quantile function for projected density
#'   mu <- c(rep(0, p - 1), 1)
#'   F_inv <- F_inv_from_f(f = f, p = p, K = 5e2)
#'   plot(F_inv, xlab = expression(x), ylab = expression(F^{-1}*(x)),
#'        main = paste0("Quantile function, ", type, ", p = ", p))
#'
#'   # Sample from the alternative and plot the projected sample
#'   n <- 5e4
#'   samp <- r_locdev(n = n, mu = mu, f = f, kappa = 1, F_inv = F_inv)
#'   plot(z, f_proj(z), col = 2, type = "l",
#'        main = paste0("Simulated projected data, ", type, ", p = ", p),
#'        ylim = c(0, 1.75))
#'   hist(samp %*% mu, freq = FALSE, breaks = seq(-1, 1, l = 50), add = TRUE)
#'
#'   par(old_par)
#'
#' }
#' \donttest{
#' ## Local alternatives for the PCvM test
#'
#' loc_alt_diagnostic(p = 2, type = "PCvM")
#' loc_alt_diagnostic(p = 3, type = "PCvM")
#' loc_alt_diagnostic(p = 4, type = "PCvM")
#' loc_alt_diagnostic(p = 5, type = "PCvM")
#' loc_alt_diagnostic(p = 11, type = "PCvM")
#'
#' ## Local alternatives for the PAD test
#'
#' loc_alt_diagnostic(p = 2, type = "PAD")
#' loc_alt_diagnostic(p = 3, type = "PAD")
#' loc_alt_diagnostic(p = 4, type = "PAD")
#' loc_alt_diagnostic(p = 5, type = "PAD")
#' loc_alt_diagnostic(p = 11, type = "PAD")
#'
#' ## Local alternatives for the PRt test
#'
#' loc_alt_diagnostic(p = 2, type = "PRt")
#' loc_alt_diagnostic(p = 3, type = "PRt")
#' loc_alt_diagnostic(p = 4, type = "PRt")
#' loc_alt_diagnostic(p = 5, type = "PRt")
#' loc_alt_diagnostic(p = 11, type = "PRt")
#' }
#' @name locdev


#' @rdname locdev
#' @export
f_locdev <- function(z, p, uk) {

  # Check dimension
  stopifnot(p >= 2)

  # Unnormalized local alternative
  f <- 1 + Gegen_series(theta = acos(z), coefs = uk, p = p, k = seq_along(uk))

  # Normalize by \omega_p such that
  # \omega_{p - 1} * f(z) * (1 - z^2)^((p - 3) / 2) integrates one in [-1, 1]
  f <- f / rotasym::w_p(p = p)
  return(f)

}


#' @rdname locdev
#' @export
con_f <- function(f, p, N = 320) {

  # Gauss--Legendre nodes and weights
  th_k <- drop(Gauss_Legen_nodes(a = 0, b = pi, N = N))
  w_k <- drop(Gauss_Legen_weights(a = 0, b = pi, N = N))

  # Integral of w_{p - 1} * f(z) * (1 - z^2)^((p - 3) / 2) in [-1, 1],
  # using z = cos(theta) for theta in [0, pi]
  int <- rotasym::w_p(p = p - 1) *
    sum(w_k * f(cos(th_k)) * sin(th_k)^(p - 2), na.rm = TRUE)
  return(1 / int)

}


#' @rdname locdev
#' @export
d_locdev <- function(x, mu, f, kappa) {

  # Check dimension
  if (is.null(dim(x))) {

    x <- rbind(x)

  }
  stopifnot(ncol(x) == length(mu))

  # Check kappa
  stopifnot(0 <= kappa & kappa <= 1)

  # Alternative density
  if (kappa > 0) {

    f1 <- rotasym::d_tang_norm(x = x, theta = mu,
                               d_U = rotasym::d_unif_sphere,
                               g_scaled = function(z, log = TRUE) log(f(z)))

  } else {

    f1 <- 0

  }

  # Uniform density
  f0 <- rotasym::d_unif_sphere(x = x)

  # Merge densities
  return((1 - kappa) * f0 + kappa * f1)

}


#' @rdname locdev
#' @export
r_locdev <- function(n, mu, f, kappa, F_inv = NULL, ...) {

  # Dimension
  p <- length(mu)

  # Check kappa
  stopifnot(0 <= kappa & kappa <= 1)
  if (kappa == 0) {

    return(r_unif_sph(n = n, p = p, M = 1)[, , 1])

  }

  # Compute the inverse of the distribution function F?
  if (is.null(F_inv)) {

    F_inv <- F_inv_from_f(f = f, p = p, ...)

  }

  # Sample object
  samp <- matrix(0, nrow = n, ncol = p)
  ind_1 <- runif(n = n) <= kappa
  n_1 <- sum(ind_1)

  # Sample under the alternative
  if (n_1 > 0) {

    r_V <- function(n) F_inv(runif(n = n))
    r_U <- function(n) r_unif_sph(n = n, p = p - 1, M = 1)[, , 1]
    samp[ind_1, ] <- rotasym::r_tang_norm(n = n_1, theta = mu,
                                          r_V = r_V, r_U = r_U)

  }

  # Sample under the null
  if (n_1 < n) {

    samp[!ind_1, ] <- r_unif_sph(n = n - n_1, p = p, M = 1)[, , 1]

  }

  # Sample
  return(samp)

}


#' @rdname locdev
#' @export
cutoff_locdev <- function(p, K_max = 1e4, thre = 1e-3, type, Rothman_t = 1 / 3,
                          Pycke_q = 0.5, verbose = FALSE, Gauss = TRUE, N = 320,
                          tol = 1e-6) {

  # vk2
  vk2 <- weights_dfs_Sobolev(p, K_max = K_max, thre = 0, type = type,
                             Rothman_t = Rothman_t, Pycke_q = Pycke_q,
                             log = FALSE, verbose = FALSE, Gauss = Gauss,
                             N = N, tol = tol)$weights
  K_max_new <- length(vk2)

  # Signs
  if (type %in% c("PRt", "Rothman", "Ajne")) {

    x_t <- drop(q_proj_unif(u = ifelse(type == "Ajne", 0.5, Rothman_t), p = p))
    signs <- akx(x = x_t, p = p, k = seq_len(K_max_new), sqr = TRUE)

  } else {

    if (verbose > 1) {

      message("Signs unknown for the ", type,
              " statistic, using positive signs experimentally.")

    }
    signs <- 1

  }

  # uk
  uk <- vk2_to_uk(vk2 = vk2, p = p, signs = signs)

  # Cutoff based on the explained squared norm
  cum_norm <- Gegen_norm(coefs = uk, k = seq_len(K_max_new), p = p,
                         cumulative = TRUE)^2
  cum_norm <- cum_norm / cum_norm[K_max_new]
  cutoff <- which(cum_norm >= 1 - thre)[1]

  # Truncate displaying optional information
  uk_cutoff <- uk[1:cutoff]
  if (verbose) {

    message("Series truncated from ", K_max_new, " to ", cutoff,
            " terms (", 100 * (1 - thre),
            "% of cumulated norm; last coefficient = ",
            sprintf("%.3e", uk_cutoff[cutoff]), ").")

    # Diagnostic plots
    if (verbose > 1) {

      old_par <- par(mfrow = c(1, 2), mar = c(5, 5.5, 4, 2) + 0.1)

      # Cumulated norm
      plot(seq_len(K_max), 100 * c(cum_norm, rep(1, K_max - K_max_new)),
           xlab = "k", ylab = "Percentage of cumulated squared norm",
           type = "s", log = "x")
      segments(x0 = cutoff, y0 = par()$usr[3],
               x1 = cutoff, y1 = 100 * cum_norm[cutoff], col = 3)
      segments(x0 = 1, y0 = 100 * (1 - thre),
               x1 = cutoff, y1 = 100 * (1 - thre), col = 2)
      abline(v = K_max_new, col = "gray", lty = 2)

      # Function and truncation
      z <- seq(-1, 1, l = 1e3)[-c(1, 1e3)]
      th <- acos(z)
      G1 <- Gegen_series(theta = th, coefs = c(1, uk),
                         k = c(0, seq_along(uk)), p = p)
      G2 <- Gegen_series(theta = th, coefs = c(1, uk_cutoff),
                         k = c(0, seq_along(uk_cutoff)), p = p)
      e <- expression(f(z) == 1 + sum(u[k] * C[k]^(p / 2 - 1) * (z), k == 1, K))
      plot(z, G1, ylim = c(1e-3, max(c(G1, G2))), xlab = expression(z),
           ylab = e, type = "l", log = "y")
      lines(z, G2, col = 2)
      legend("top", legend = paste("K =", c(K_max_new, cutoff)),
             col = 1:2, lwd = 2)
      par(old_par)

    }

  }
  return(uk_cutoff)

}


#' @title Distribution and quantile functions from angular function
#'
#' @description Numerical computation of the distribution function \eqn{F} and
#' the quantile function \eqn{F^{-1}} for an \link[=locdev]{angular function}
#' \eqn{f} in a \link[=tang-norm-decomp]{tangent-normal decomposition}.
#' \eqn{F^{-1}(x)} results from the inversion of
#' \deqn{F(x) = \int_{-1}^x \omega_{p - 1}c_f f(z) (1 - z^2)^{(p - 3) / 2}
#' \,\mathrm{d}z}{F(x) = \int_{-1}^x \omega_{p - 1}c_f f(z)
#' (1 - z^2)^{(p - 3) / 2} dz}
#' for \eqn{x\in [-1, 1]}, where \eqn{c_f} is a normalizing constant and
#' \eqn{\omega_{p - 1}} is the surface area of \eqn{S^{p - 2}}.
#'
#' @inheritParams locdev
#' @inheritParams r_unif
#' @param Gauss use a \link[=Gauss_Legen_nodes]{Gauss--Legendre quadrature}
#' rule to integrate \eqn{f} with \code{N} nodes? Otherwise, rely on
#' \code{\link{integrate}} Defaults to \code{TRUE}.
#' @param N number of points used in the Gauss--Legendre quadrature. Defaults
#' to \code{320}.
#' @param K number of equispaced points on \eqn{[-1, 1]} used for evaluating
#' \eqn{F^{-1}} and then interpolating. Defaults to \code{1e3}.
#' @param tol tolerance passed to \code{\link{uniroot}} for the inversion of
#' \eqn{F}. Also, passed to \code{\link{integrate}}'s \code{rel.tol} and
#' \code{abs.tol} if \code{Gauss = FALSE}. Defaults to \code{1e-6}.
#' @param ... further parameters passed to \code{f}.
#' @details
#' The normalizing constant \eqn{c_f} is such that \eqn{F(1) = 1}. It does not
#' need to be part of \code{f} as it is computed internally.
#'
#' Interpolation is performed by a monotone cubic spline. \code{Gauss = TRUE}
#' yields more accurate results, at expenses of a heavier computation.
#'
#' If \code{f} yields negative values, these are silently truncated to zero.
#' @return A \code{\link{splinefun}} object ready to evaluate \eqn{F} or
#' \eqn{F^{-1}}, as specified.
#' @examples
#' f <- function(x) rep(1, length(x))
#' plot(F_from_f(f = f, p = 4, Gauss = TRUE), ylab = "F(x)", xlim = c(-1, 1))
#' plot(F_from_f(f = f, p = 4, Gauss = FALSE), col = 2, add = TRUE,
#'      xlim = c(-1, 1))
#' curve(p_proj_unif(x = x, p = 4), col = 3, add = TRUE, n = 300)
#' plot(F_inv_from_f(f = f, p = 4, Gauss = TRUE), ylab = "F^{-1}(x)")
#' plot(F_inv_from_f(f = f, p = 4, Gauss = FALSE), col = 2, add = TRUE)
#' curve(q_proj_unif(u = x, p = 4), col = 3, add = TRUE, n = 300)
#' @name F_from_f


#' @rdname F_from_f
#' @export
F_from_f <- function(f, p, Gauss = TRUE, N = 320, K = 1e3, tol = 1e-6, ...) {

  # Integration grid
  z <- seq(-1, 1, length.out = K)

  # Integrated \omega_{p - 1} * f(z) * (1 - z^2)^{(p - 3) / 2}
  if (Gauss) {

    # Using Gauss--Legendre quadrature but ensuring monotonicity?
    F_grid <- rotasym::w_p(p = p - 1) * sapply(z, function(t) {
      z_k <- drop(Gauss_Legen_nodes(a = -1, b = t, N = N))
      w_k <- drop(Gauss_Legen_weights(a = -1, b = t, N = N))
      sum(w_k * pmax(f(z_k, ...), 0) * (1 - z_k^2)^((p - 3) / 2), na.rm = TRUE)
    })

    # Normalize f (the normalizing constant may not be included in f)
    c_f <- F_grid[length(F_grid)]
    F_grid <- F_grid / c_f

  } else {

    # Using integrate
    g <- function(t) pmax(f(t, ...), 0) * (1 - t^2)^((p - 3) / 2)
    F_grid <- sapply(z[-1], function(u) rotasym::w_p(p = p - 1) *
                       integrate(f = g, lower = -1, upper = u,
                                 subdivisions = 1e3, rel.tol = tol,
                                 abs.tol = tol, stop.on.error = FALSE)$value)

    # Normalize f (the normalizing constant may not be included in f)
    c_f <- F_grid[length(F_grid)]
    F_grid <- F_grid / c_f

    # Add 0
    F_grid <- c(0, F_grid)

  }

  # Use method = "hyman" for monotone interpolations if possible
  if (anyNA(F_grid)) stop("Numerical error (NAs) in F_grid")
  F_appf <- switch(is.unsorted(F_grid) + 1,
                   splinefun(x = z, y = F_grid, method = "hyman"),
                   approxfun(x = z, y = F_grid, method = "linear", rule = 2))
  return(F_appf)

}


#' @rdname F_from_f
#' @export
F_inv_from_f <- function(f, p, Gauss = TRUE, N = 320, K = 1e3, tol = 1e-6,
                         ...) {

  # Approximate F
  F_appf <- F_from_f(f = f, p = p, Gauss = Gauss, N = N, K = K, tol = tol, ...)

  # Inversion of F
  u <- seq(0, 1, length.out = K)
  F_inv_grid <- sapply(u[-c(1, K)], function(v) {
    uniroot(f = function(x) F_appf(x) - v, lower = -1, upper = 1,
            tol = tol)$root
  })
  F_inv_grid <- c(-1, F_inv_grid, 1)

  # Use method = "hyman" for monotone interpolations if possible
  stopifnot(!anyNA(F_inv_grid))
  F_inv <- switch(is.unsorted(F_inv_grid) + 1,
                  splinefun(x = u, y = F_inv_grid, method = "hyman"),
                  approxfun(x = u, y = F_inv_grid, method = "linear",
                            rule = 2))
  return(F_inv)

}


#' @title Transformation between different coefficients in Sobolev statistics
#'
#' @description Given a Sobolev statistic
#' \deqn{S_{n, p} = \sum_{i, j = 1}^n \psi(\cos^{-1}({\bf X}_i'{\bf X}_j)),}{
#' S_{n, p} = \sum_{i, j = 1}^n \psi(\cos^{-1}(X_i'X_j)),}
#' for a sample \eqn{{\bf X}_1, \ldots, {\bf X}_n \in S^{p - 1} := \{{\bf x}
#' \in R^p : ||{\bf x}|| = 1\}}{X_1, \ldots, X_n \in S^{p - 1} :=
#' \{x \in R^p : ||x|| = 1\}}, \eqn{p\ge 2}, three important sequences
#' are related to \eqn{S_{n, p}}.
#' \itemize{
#' \item \link[=Gegen_coefs]{Gegenbauer coefficients} \eqn{\{b_{k, p}\}} of
#' \eqn{\psi_p} (see, e.g., the \link[=Pn]{projected-ecdf statistics}), given
#' by
#' \deqn{b_{k, p} := \frac{1}{c_{k, p}}\int_0^\pi \psi_p(\theta)
#' C_k^{p / 2 - 1}(\cos\theta)\,\mathrm{d}\theta.}{
#' b_{k, p} := \frac{1}{c_{k, p}} \int_0^\pi \psi_p(\theta)
#' C_k^(p / 2 - 1)(\cos\theta) d\theta.}
#' \item Weights \eqn{\{v_{k, p}^2\}} of the
#' \link[=Sobolev]{asymptotic distribution} of the Sobolev statistic,
#' \eqn{\sum_{k = 1}^\infty v_k^2 \chi^2_{d_{p, k}}}, given by
#' \deqn{v_{k, p}^2 = \left(1 + \frac{2k}{p - 2}\right)^{-1} b_{k, p},
#' \quad p \ge 3.}{v_{k, p}^2 = (1 + 2k / (p - 2))^{-1} b_{k, p}, p \ge 3.}
#' \item Gegenbauer coefficients \eqn{\{u_{k, p}\}} of the
#' \link[=locdev]{local projected alternative} associated to \eqn{S_{n, p}},
#' given by
#' \deqn{u_{k, p} = \left(1 + \frac{2k}{p - 2}\right) v_{k, p},
#' \quad p \ge 3.}{u_{k, p} = (1 + 2k / (p - 2)) b_{k, p}, p \ge 3.}
#' }
#' For \eqn{p = 2}, the factor \eqn{(1 + 2k / (p - 2))} is replaced by \eqn{2}.
#'
#' @param bk coefficients \eqn{b_{k, p}} associated to the indexes
#' \code{1:length(bk)}, a vector.
#' @param vk2 \bold{squared} coefficients \eqn{v_{k, p}^2} associated to the
#' indexes \code{1:length(vk2)}, a vector.
#' @param uk coefficients \eqn{u_{k, p}} associated to the indexes
#' \code{1:length(uk)}, a vector.
#' @inheritParams r_unif_sph
#' @param signs signs of the coefficients \eqn{u_{k, p}}, a vector of the
#' same size as \code{vk2} or \code{bk}, or a scalar. Defaults to \code{1}.
#' @return
#' The corresponding vectors of coefficients \code{vk2}, \code{bk}, or
#' \code{uk}, depending on the call.
#' @details
#' See more details in Prentice (1978) and García-Portugués et al. (2020). The
#' adequate signs of \code{uk} for the \code{"PRt"} \link[=Pn]{Rothman test}
#' can be retrieved with \code{\link{akx}} and \code{sqr = TRUE}, see the
#' examples.
#' @references
#' García-Portugués, E., Navarro-Esteban, P., Cuesta-Albertos, J. A. (2020)
#' On a projection-based class of uniformity tests on the hypersphere.
#' \emph{arXiv:2008.09897}. \url{https://arxiv.org/abs/2008.09897}
#'
#' Prentice, M. J. (1978). On invariant tests of uniformity for directions and
#' orientations. \emph{The Annals of Statistics}, 6(1):169--176.
#' \doi{10.1214/aos/1176344075}
#' @examples
#' # bk, vk2, and uk for the PCvM test in p = 3
#' (bk <- Gegen_coefs_Pn(k = 1:5, type = "PCvM", p = 3))
#' (vk2 <- bk_to_vk2(bk = bk, p = 3))
#' (uk <- bk_to_uk(bk = bk, p = 3))
#'
#' # vk2 is the same as
#' weights_dfs_Sobolev(K_max = 10, thre = 0, p = 3, type = "PCvM")$weights
#'
#' # bk and uk for the Rothman test in p = 3, with adequate signs
#' t <- 1 / 3
#' (bk <- Gegen_coefs_Pn(k = 1:5, type = "PRt", p = 3, Rothman_t = t))
#' (ak <- akx(x = drop(q_proj_unif(t, p = 3)), p = 3, k = 1:5, sqr = TRUE))
#' (uk <- bk_to_uk(bk = bk, p = 3, signs = ak))
#' @name Sobolev_coefs


#' @rdname Sobolev_coefs
#' @export
bk_to_vk2 <- function(bk, p) {

  # Check dimension
  p <- as.integer(p)
  stopifnot(p >= 2)

  # Add factor
  if (p == 2) {

    return(bk / 2)

  } else {

    return(bk / (1 + 2 * seq_along(bk) / (p - 2)))

  }

}


#' @rdname Sobolev_coefs
#' @export
bk_to_uk <- function(bk, p, signs = 1) {

  # Check dimension
  p <- as.integer(p)
  stopifnot(p >= 2)

  # Check signs
  stopifnot(length(signs) %in% c(1, length(bk)))

  # Add factor
  if (p == 2) {

    return(sign(signs) * sqrt(2 * bk))

  } else {

    return(sign(signs) * sqrt((1 + 2 * seq_along(bk) / (p - 2)) * bk))

  }

}


#' @rdname Sobolev_coefs
#' @export
vk2_to_bk <- function(vk2, p) {

  # Check dimension
  p <- as.integer(p)
  stopifnot(p >= 2)

  # Add factor
  if (p == 2) {

    return(2 * vk2)

  } else {

    return((1 + 2 * seq_along(vk2) / (p - 2)) * vk2)

  }

}


#' @rdname Sobolev_coefs
#' @export
vk2_to_uk <- function(vk2, p, signs = 1) {

  # Check dimension
  p <- as.integer(p)
  stopifnot(p >= 2)

  # Check signs
  stopifnot(length(signs) %in% c(1, length(vk2)))

  # Add factor
  if (p == 2) {

    return(2 * sign(signs) * sqrt(vk2))

  } else {

    return((1 + 2 * seq_along(vk2) / (p - 2)) * sign(signs) * sqrt(vk2))

  }

}


#' @rdname Sobolev_coefs
#' @export
uk_to_vk2 <- function(uk, p) {

  # Check dimension
  p <- as.integer(p)
  stopifnot(p >= 2)

  # Add factor
  if (p == 2) {

    return((uk / 2)^2)

  } else {

    return((uk / (1 + 2 * seq_along(uk) / (p - 2)))^2)

  }

}


#' @rdname Sobolev_coefs
#' @export
uk_to_bk <- function(uk, p) {

  # Check dimension
  p <- as.integer(p)
  stopifnot(p >= 2)

  # Add factor
  if (p == 2) {

    return(uk^2 / 2)

  } else {

    return(uk^2 / (1 + 2 * seq_along(uk) / (p - 2)))

  }

}


#' @title Sample non-uniformly distributed spherical data
#'
#' @description Simple simulation of prespecified non-uniform spherical
#' distributions: von Mises--Fisher (vMF), Mixture of vMF (MvMF),
#' Angular Central Gaussian (ACG), Small Circle (SC), Watson (W), or
#' Cauchy-like (C).
#'
#' @inheritParams r_unif
#' @param alt alternative, must be \code{"vMF"}, \code{"MvMF"},
#' \code{"ACG"}, \code{"SC"}, \code{"W"}, or \code{"C"}. See details below.
#' @param kappa non-negative parameter measuring the strength of the deviation
#' with respect to uniformity (obtained with \eqn{\kappa = 0}).
#' @param nu projection along \eqn{{\bf e}_p}{e_p} controlling the modal
#' strip of the small circle distribution. Must be in (-1, 1). Defaults to
#' \code{0.5}.
#' @param F_inv quantile function returned by \code{\link{F_inv_from_f}}. Used
#' for \code{"SC"}, \code{"W"}, and \code{"C"}. Computed by internally if
#' \code{NULL} (default).
#' @inheritParams F_inv_from_f
#' @param axial_MvMF use a mixture of vMF that is axial (i.e., symmetrically
#' distributed about the origin)? Defaults to \code{TRUE}.
#' @details
#' The parameter \code{kappa} is used as \eqn{\kappa} in the following
#' distributions:
#' \itemize{
#'   \item \code{"vMF"}: von Mises--Fisher distribution with concentration
#'   \eqn{\kappa} and directional mean \eqn{{\bf e}_p = (0, 0, \ldots, 1)}{
#'   e_p = (0, 0, \ldots, 1)}.
#'   \item \code{"MvMF"}: equally-weighted mixture of \eqn{p} von Mises--Fisher
#'   distributions with common concentration \eqn{\kappa} and directional means
#'   \eqn{\pm{\bf e}_1, \ldots, \pm{\bf e}_p}{±e_1, \ldots, ±e_p} if
#'   \code{axial_MvMF = TRUE}. If \code{axial_MvMF = FALSE}, then only means
#'   with positive signs are considered.
#'   \item \code{"ACG"}: Angular Central Gaussian distribution with diagonal
#'   shape matrix with diagonal given by
#'   \deqn{(1, \ldots, 1, 1 + \kappa) / (p + \kappa).}
#'   \item \code{"SC"}: Small Circle distribution with axis mean
#'   \eqn{{\bf e}_p = (0, 0, \ldots, 1)}{e_p = (0, 0, \ldots, 1)} and
#'   concentration \eqn{\kappa} about the projection along the mean, \eqn{\nu}.
#'   \item \code{"W"}: Watson distribution with axis mean
#'   \eqn{{\bf e}_p = (0, 0, \ldots, 1)}{e_p = (0, 0, \ldots, 1)} and
#'   concentration \eqn{\kappa}. The Watson distribution is a particular case
#'   of the Bingham distribution.
#'   \item \code{"C"}: Cauchy-like distribution with directional mode
#'   \eqn{{\bf e}_p = (0, 0, \ldots, 1)}{e_p = (0, 0, \ldots, 1)} and
#'   concentration \eqn{\kappa = \rho / (1 - \rho^2)}. The circular Wrapped
#'   Cauchy distribution is a particular case of this Cauchy-like distribution.
#' }
#' @return An \bold{array} of size \code{c(n, p, M)} with \code{M} random
#' samples of size \code{n} of non-uniformly-generated directions on
#' \eqn{S^{p-1}}.
#' @details
#' Much faster sampling for \code{"SC"}, \code{"W"}, and \code{"C"} is achieved
#' providing \code{F_inv}, see examples.
#' @examples
#' ## Simulation with p = 2
#'
#' p <- 2
#' n <- 200
#' kappa <- 20
#' nu <- 0.5
#' rho <- ((2 * kappa + 1) - sqrt(4 * kappa + 1)) / (2 * kappa)
#' F_inv_SC_2 <- F_inv_from_f(f = function(z) exp(-kappa * (z - nu)^2), p = 2)
#' F_inv_W_2 <- F_inv_from_f(f = function(z) exp(kappa * z^2), p = 2)
#' F_inv_C_2 <- F_inv_from_f(f = function(z) (1 - rho^2) /
#'                             (1 + rho^2 - 2 * rho * z)^(p / 2), p = 2)
#' x1 <- r_alt(n = n, p = p, alt = "vMF", kappa = kappa)[, , 1]
#' x2 <- r_alt(n = n, p = p, alt = "MvMF", kappa = kappa)[, , 1]
#' x3 <- r_alt(n = n, p = p, alt = "ACG", kappa = kappa)[, , 1]
#' x4 <- r_alt(n = n, p = p, alt = "SC", F_inv = F_inv_SC_2)[, , 1]
#' x5 <- r_alt(n = n, p = p, alt = "W", F_inv = F_inv_W_2)[, , 1]
#' x6 <- r_alt(n = n, p = p, alt = "C", F_inv = F_inv_C_2)[, , 1]
#' r <- runif(n, 0.95, 1.05) # Radius perturbation to improve visualization
#' plot(r * x1, pch = 16, xlim = c(-1.1, 1.1), ylim = c(-1.1, 1.1), col = 1)
#' points(r * x2, pch = 16, col = 2)
#' points(r * x3, pch = 16, col = 3)
#' points(r * x4, pch = 16, col = 4)
#' points(r * x5, pch = 16, col = 5)
#' points(r * x6, pch = 16, col = 6)
#'
#' ## Simulation with p = 3
#'
#' n <- 200
#' p <- 3
#' kappa <- 20
#' nu <- 0.5
#' rho <- ((2 * kappa + 1) - sqrt(4 * kappa + 1)) / (2 * kappa)
#' F_inv_SC_3 <- F_inv_from_f(f = function(z) exp(-kappa * (z - nu)^2), p = 3)
#' F_inv_W_3 <- F_inv_from_f(f = function(z) exp(kappa * z^2), p = 3)
#' F_inv_C_3 <- F_inv_from_f(f = function(z) (1 - rho^2) /
#'                             (1 + rho^2 - 2 * rho * z)^(p / 2), p = 3)
#' x1 <- r_alt(n = n, p = p, alt = "vMF", kappa = kappa)[, , 1]
#' x2 <- r_alt(n = n, p = p, alt = "MvMF", kappa = kappa)[, , 1]
#' x3 <- r_alt(n = n, p = p, alt = "ACG", kappa = kappa)[, , 1]
#' x4 <- r_alt(n = n, p = p, alt = "SC", F_inv = F_inv_SC_3)[, , 1]
#' x5 <- r_alt(n = n, p = p, alt = "W", F_inv = F_inv_W_3)[, , 1]
#' x6 <- r_alt(n = n, p = p, alt = "C", F_inv = F_inv_C_3)[, , 1]
#' s3d <- scatterplot3d::scatterplot3d(x1, pch = 16, xlim = c(-1.1, 1.1),
#'                                     ylim = c(-1.1, 1.1), zlim = c(-1.1, 1.1))
#' s3d$points3d(x2, pch = 16, col = 2)
#' s3d$points3d(x3, pch = 16, col = 3)
#' s3d$points3d(x4, pch = 16, col = 4)
#' s3d$points3d(x5, pch = 16, col = 5)
#' s3d$points3d(x6, pch = 16, col = 6)
#' @export
r_alt <- function(n, p, M = 1, alt = "vMF", kappa = 1, nu = 0.5, F_inv = NULL,
                  K = 1e3, axial_MvMF = TRUE) {

  # Common mean (North pole)
  mu <- c(rep(0, p - 1), 1)

  # Check concentration parameter and sample size
  stopifnot(kappa >= 0)
  stopifnot(n >= 1)

  # Sampling from uniform
  if (kappa == 0) {

    return(r_unif_sph(n = n, p = p, M = M))

  }

  # Choose alternative
  if (alt == "vMF") {

    long_samp <- rotasym::r_vMF(n = n * M, mu = mu, kappa = kappa)

  } else if (alt == "MvMF") {

    # Mixture components
    j <- sample(x = 1:p, size = n * M, replace = TRUE)
    nM_j <- tabulate(bin = j, nbins = p)
    mu_j <- diag(1, nrow = p, ncol = p)

    # Sample components
    long_samp <- matrix(nrow = n * M, ncol = p)
    for (k in which(nM_j > 0)) {

      long_samp[j == k, ] <- rotasym::r_vMF(n = nM_j[k], mu = mu_j[k, ],
                                            kappa = kappa)

    }

    # Add plus and minus means
    if (axial_MvMF) {

      long_samp <- sample(x = c(-1, 1), size = n * M, replace = TRUE) *
        long_samp

    }

    # Shuffle data
    long_samp <- long_samp[sample(x = n * M), , drop = FALSE]

  } else if (alt == "ACG") {

    Lambda <- diag(c(rep(1 / (p + kappa), p - 1),
                     (1 + kappa) / (p + kappa)), nrow = p, ncol = p)
    long_samp <- rotasym::r_ACG(n = n * M, Lambda = Lambda)

  } else if (alt == "SC") {

    # Compute the inverse of the distribution function F?
    if (is.null(F_inv)) {

      stopifnot(-1 < nu & nu < 1)
      f <- function(z) exp(-kappa * (z - nu)^2)
      F_inv <- F_inv_from_f(f = f, p = p, K = K)

    }

    # Sample the small circle distribution
    r_U <- function(n) r_unif_sph(n = n, p = p - 1, M = 1)[, , 1]
    r_V <- function(n) F_inv(runif(n = n))
    long_samp <- rotasym::r_tang_norm(n = n * M, theta = mu,
                                      r_U = r_U, r_V = r_V)

  } else if (alt == "W") {

    # Compute the inverse of the distribution function F?
    if (is.null(F_inv)) {

      f <- function(z) exp(kappa * z^2)
      F_inv <- F_inv_from_f(f = f, p = p, K = K)

    }

    # Sample the small circle distribution
    r_U <- function(n) r_unif_sph(n = n, p = p - 1, M = 1)[, , 1]
    r_V <- function(n) F_inv(runif(n = n))
    long_samp <- rotasym::r_tang_norm(n = n * M, theta = mu,
                                      r_U = r_U, r_V = r_V)

  } else if (alt == "C") {

    # Compute the inverse of the distribution function F?
    if (is.null(F_inv)) {

      rho <- ifelse(kappa == 0, 0,
                    ((2 * kappa + 1) - sqrt(4 * kappa + 1)) / (2 * kappa))
      f <- function(z) (1 - rho^2) / (1 + rho^2 - 2 * rho * z)^(p / 2)
      F_inv <- F_inv_from_f(f = f, p = p, K = K)

    }

    # Sample the small circle distribution
    r_U <- function(n) r_unif_sph(n = n, p = p - 1, M = 1)[, , 1]
    r_V <- function(n) F_inv(runif(n = n))
    long_samp <- rotasym::r_tang_norm(n = n * M, theta = mu,
                                      r_U = r_U, r_V = r_V)

  } else {

    stop(paste("Wrong alt; must be \"vMF\", \"MvMF\", \"Bing\"",
               "\"ACG\", \"SC\", \"W\", or \"C\"."))

  }

  # As an array
  samp <- array(dim = c(n, p, M))
  for (j in 1:M) {

    samp[, , j] <- long_samp[(1 + (j - 1) * n):(j * n), , drop = FALSE]

  }
  return(samp)

}

Sys.unsetenv("R_TESTS")

set.seed(21332)
u <- seq(0, 1, l = 13)
x <- seq(-1, 1, l = 13)
v <- runif(1e3)
f0 <- function(x) rep(1, length(x))
f1 <- function(x, kappa) exp(kappa * x)
f2 <- function(x, kappa) exp(kappa * x^2)
f3 <- function(x, kappa, nu) exp(-kappa * (x - nu)^2)
f4 <- function(x, kappa, q) {
  rho <- ((2 * kappa + 1) - sqrt(4 * kappa + 1)) / (2 * kappa)
  (1 - rho^2) / (1 + rho^2 - 2 * rho * x)^((q + 1) / 2) /
    rotasym::w_p(p = q + 1)
}

test_that("Correct integration of con_f", {

  skip_on_cran()
  for (p in c(2:4, 11)) {
    expect_equal(con_f(f = function(x)
      rotasym::g_vMF(t = x, p = p, kappa = 3, scaled = TRUE),
      p = p, N = 320), 1)
    expect_equal(con_f(f = function(x) f4(x, kappa = 1, q = p - 1),
                       p = p, N = 320), 1)
  }

})

test_that("d_locdev", {

  skip_on_cran()
  for (p in 2:4) {
    xp <- r_unif_sph(n = 5, p = p)[, , 1]
    mu <- c(rep(0, p - 1), 1)
    expect_equal(d_locdev(x = xp[1, , drop = FALSE], mu = mu, kappa = 0.25,
                          f = function(z)
                            rotasym::g_vMF(t = z, kappa = 3, p = p)),
                 unname(d_locdev(x = xp[1, ], mu = mu, kappa = 0.25,
                                 f = function(z)
                                   rotasym::g_vMF(t = z, kappa = 3, p = p))))
    expect_equal(d_locdev(x = xp, mu = mu, kappa = 0, f = NULL),
                 rep(1 / rotasym::w_p(p = p), nrow(xp)))
    expect_equal(d_locdev(x = xp, mu = mu, kappa = 0.25,
                          f = function(z)
                            rotasym::g_vMF(t = z, kappa = 3, p = p)),
                 0.25 * rotasym::g_vMF(t = xp[, p], kappa = 3, p = p) +
                 0.75 / rotasym::w_p(p = p))
  }

})

test_that("r_locdev coherence with d_locdev", {

  skip_on_cran()
  for (p in 2:4) {
    mu <- c(rep(0, p - 1), 1)
    samp_1 <- r_locdev(n = 1e3, mu = mu, kappa = 0.25,
                       f = function(z) f4(x = z, kappa = 3, q = p - 1))[, p]
    samp_2 <- F_inv_from_f(f = function(z)
      0.25 * f4(x = z, kappa = 3, q = p - 1) + 0.75 / rotasym::w_p(p = p),
      p = p)(runif(n = 1e3))
    expect_gt(ks.test(samp_1, samp_2)$p.value, 0.01)
    samp_1 <- r_locdev(n = 1e3, mu = mu, kappa = 0, f = NULL)[, 1]
    samp_2 <- r_unif_sph(n = 1e3, p = p)[, 1, 1]
    expect_gt(ks.test(samp_1, samp_2)$p.value, 0.01)
  }

})

test_that("Edge cases d_locdev and r_locdev", {

  skip_on_cran()
  expect_error(d_locdev(x = 1, mu = 1, kappa = -1, f = NULL))
  expect_error(d_locdev(x = 1:2, mu = 1:3, kappa = -1, f = NULL))
  expect_error(r_locdev(n = 1, mu = 1, kappa = -1))

})

test_that("F_from_f via Gauss--Legendre", {

  skip_on_cran()
  for (p in c(2:4, 11)) {
    expect_equal(F_from_f(f = f0, p = p, Gauss = TRUE, K = 1e2)(x),
                 drop(p_proj_unif(x = x, p = p)), tolerance = 1e-3)
  }

})

test_that("F_from_f via integrate()", {

  skip_on_cran()
  for (p in c(2:4, 11)) {
    expect_equal(F_from_f(f = f0, p = p, Gauss = FALSE, K = 1e2)(x),
                 drop(p_proj_unif(x = x, p = p)), tolerance = 1e-3)
  }

})

test_that("F_from_f for vMF", {

  skip_on_cran()
  for (p in c(2:4, 11)) {
    samp_g <- drop(rotasym::r_g_vMF(n = 100, p = p, kappa = 3))
    expect_gt(ks.test(x = F_from_f(f = f1, p = p, Gauss = TRUE,
                                   K = 1e2, kappa = 3)(samp_g),
                      y = "punif")$p.value, 0.01)
  }
  expect_error(F_from_f(f = f1, p = 2, kappa = 1e5))

})

test_that("F_inv_from_f via Gauss--Legendre", {

  skip_on_cran()
  for (p in c(2:4, 11)) {
    expect_equal(F_inv_from_f(f = f0, p = p, Gauss = TRUE, K = 1e2)(u),
                 drop(q_proj_unif(u = u, p = p)), tolerance = 5e-3)
  }

})

test_that("F_inv_from_f via integrate()", {

  skip_on_cran()
  for (p in c(2:4, 11)) {
    expect_equal(F_inv_from_f(f = f0, p = p, Gauss = FALSE, K = 1e2)(u),
                 drop(q_proj_unif(u = u, p = p)), tolerance = 5e-3)
  }

})

test_that("F_inv_from_f for vMF", {

  skip_on_cran()
  expect_gt(ks.test(x = F_inv_from_f(f = f1, p = 2, Gauss = TRUE,
                                     K = 1e2, kappa = 3)(v),
                    y = rotasym::r_g_vMF(n = 100, p = 2,
                                         kappa = 3))$p.value, 0.01)
  expect_gt(ks.test(x = F_inv_from_f(f = f1, p = 3, Gauss = TRUE,
                                      K = 1e2, kappa = 5)(v),
                    y = rotasym::r_g_vMF(n = 100, p = 3,
                                          kappa = 5))$p.value, 0.01)
  expect_gt(ks.test(x = F_inv_from_f(f = f1, p = 4, Gauss = TRUE,
                                     K = 1e2, kappa = 5)(v),
                    y = rotasym::r_g_vMF(n = 100, p = 4,
                                         kappa = 5))$p.value, 0.01)
  expect_gt(ks.test(x = F_inv_from_f(f = f1, p = 5, Gauss = TRUE,
                                      K = 1e2, kappa = 10)(v),
                    y = rotasym::r_g_vMF(n = 100, p = 5,
                                         kappa = 10))$p.value, 0.01)
  expect_gt(ks.test(x = F_inv_from_f(f = f1, p = 11, Gauss = TRUE,
                                     K = 1e2, kappa = 20)(v),
                    y = rotasym::r_g_vMF(n = 100, p = 11,
                                         kappa = 20))$p.value, 0.01)
  expect_error(F_inv_from_f(f = f1, p = 2, kappa = 1e5))

})

test_that("r_alt rotationally symmetric", {

  skip_on_cran()
  for (p in 2:4) {
    samp_g <- r_alt(n = 100, p = p, M = 1, kappa = 2, alt = "vMF")[, p, 1]
    expect_gt(ks.test(x = F_from_f(f = f1, p = p, kappa = 2)(samp_g),
                      y = "punif")$p.value, 0.01)
    samp_g <- r_alt(n = 100, p = p, M = 1, kappa = 2, alt = "W")[, p, 1]
    expect_gt(ks.test(x = F_from_f(f = f2, p = p, kappa = 2)(samp_g),
                      y = "punif")$p.value, 0.01)
    samp_g <- r_alt(n = 100, p = p, M = 1, kappa = 2, nu = 0.5,
                    alt = "SC")[, p, 1]
    expect_gt(ks.test(x = F_from_f(f = f3, p = p, kappa = 2, nu = 0.5)(samp_g),
                      y = "punif")$p.value, 0.01)
    samp_g <- r_alt(n = 100, p = p, M = 1, kappa = 2, alt = "C")[, p, 1]
    expect_gt(ks.test(x = F_from_f(f = f4, p = p, kappa = 2, q = p - 1)(samp_g),
                      y = "punif")$p.value, 0.01)
    samp_1 <- r_alt(n = 1e3, p = p, M = 1, kappa = 0, alt = "MvMF")[, p, 1]
    samp_2 <- r_unif_sph(n = 1e3, p = p, M = 1)[, p, 1]
    expect_gt(ks.test(x = samp_1, y = samp_2)$p.value, 0.01)
  }

})

test_that("r_alt non-rotationally symmetric", {

  skip_on_cran()
  for (p in c(2:4, 11)) {
    samp_1a <- r_alt(n = 1e3, p = p, M = 1, kappa = 2, alt = "MvMF",
                     axial_MvMF = TRUE)[, p, 1]
    samp_1b <- r_alt(n = 1e3, p = p, M = 1, kappa = 2, alt = "MvMF",
                     axial_MvMF = FALSE)[, p, 1]
    samp_2b <- c(apply(diag(rep(1, p)), 1, function(mu)
      t(rotasym::r_vMF(n = round(1e3 / p), mu = mu, kappa = 2))))
    samp_2b <- matrix(samp_2b, ncol = p, byrow = TRUE)[, p]
    samp_2a <- samp_2b * sample(c(-1, 1), size = length(samp_2b),
                                replace = TRUE)
    expect_gt(ks.test(x = samp_1a, y = samp_2a)$p.value, 0.01)
    expect_gt(ks.test(x = samp_1b, y = samp_2b)$p.value, 0.01)
    samp_1 <- r_alt(n = 1e3, p = p, M = 1, kappa = 1, alt = "ACG")[, p, 1]
    samp_2 <- mvtnorm::rmvnorm(n = 1e3, mean = rep(0, p),
                               sigma = diag(c(rep(1, p - 1), 1 + 1)))
    samp_2 <- samp_2 / sqrt(rowSums(samp_2^2))
    samp_2 <- samp_2[, p]
    expect_gt(ks.test(x = samp_1, y = samp_2)$p.value, 0.01)
  }

})

test_that("Edge cases in r_alt", {

  skip_on_cran()
  for (p in c(2:4, 11)) {

    expect_length(r_alt(n = 5, p = p, M = 1, alt = "MvMF", kappa = 1),
                  5 * p)
    expect_equal(dim(r_alt(n = 1, p = p, M = 1, alt = "vMF", kappa = 1)),
                 c(1, p, 1))
    expect_equal(dim(r_alt(n = 1, p = p, M = 1, alt = "MvMF", kappa = 1)),
                 c(1, p, 1))
    expect_equal(dim(r_alt(n = 1, p = p, M = 1, alt = "SC", kappa = 1)),
                 c(1, p, 1))
    expect_equal(dim(r_alt(n = 1, p = p, M = 1, alt = "C", kappa = 1)),
                 c(1, p, 1))
    expect_equal(dim(r_alt(n = 1, p = p, M = 1, alt = "W", kappa = 1)),
                 c(1, p, 1))
    expect_equal(dim(r_alt(n = 1, p = p, M = 1, alt = "ACG", kappa = 1)),
                 c(1, p, 1))
    expect_error(r_alt(n = 100, p = p, M = 1, kappa = 1, alt = "WC"))
    expect_error(r_alt(n = 100, p = p, M = 1, kappa = -1, alt = "C"))
    expect_error(r_alt(n = 0, p = p, M = 1, kappa = 1, alt = "C"))

  }

})

set.seed(12311)
n <- 20
X_2 <- r_unif_sph(n = n, p = 2, M = 1)
X_3 <- r_unif_sph(n = n, p = 3, M = 1)
X_4 <- r_unif_sph(n = n, p = 4, M = 1)
X_11 <- r_unif_sph(n = n, p = 11, M = 1)
th_k <- Gauss_Legen_nodes(a = 0, b = 2 * pi, N = 320)
w_k <- Gauss_Legen_weights(a = 0, b = 2 * pi, N = 320)
z <- seq(-1, 1, l = 1e3)
z8 <- seq(-0.8, 1, l = 1e3)

# PCvM
k_PCvM <- 1:1e3
uk_PCvM_2 <- bk_to_uk(Gegen_coefs_Pn(k = k_PCvM, p = 2, type = "PCvM",
                                     N = 0), p = 2)
f_locdev_PCvM_2 <- function(z) f_locdev(z = z, p = 2, uk = uk_PCvM_2)
integrand_vec_PCvM_2 <- function(x) {
  f_gamma <- matrix(f_locdev_PCvM_2(c(X_2[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 2) * f_gamma - 1)^2 / n / rotasym::w_p(p = 2)
}
f_locdev_PCvM_2_anal <- f_locdev_Pn(p = 2, type = "PCvM")
integrand_vec_PCvM_2_anal <- function(x) {
  f_gamma <- matrix(f_locdev_PCvM_2_anal(c(X_2[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 2) * f_gamma - 1)^2 / n / rotasym::w_p(p = 2)
}
uk_PCvM_3 <- bk_to_uk(Gegen_coefs_Pn(k = k_PCvM, p = 3, type = "PCvM",
                                     N = 0), p = 3)
f_locdev_PCvM_3 <- function(z) f_locdev(z = z, p = 3, uk = uk_PCvM_3)
integrand_vec_PCvM_3 <- function(x) {
  f_gamma <- matrix(f_locdev_PCvM_3(c(X_3[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 3) * f_gamma - 1)^2 / n / rotasym::w_p(p = 3)
}
uk_PCvM_4 <- bk_to_uk(Gegen_coefs_Pn(k = k_PCvM, p = 4, type = "PCvM",
                                     N = 0), p = 4)
f_locdev_PCvM_4 <- function(z) f_locdev(z = z, p = 4, uk = uk_PCvM_4)
integrand_vec_PCvM_4 <- function(x) {
  f_gamma <- matrix(f_locdev_PCvM_4(c(X_4[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 4) * f_gamma - 1)^2 / n / rotasym::w_p(p = 4)
}
uk_PCvM_11 <- bk_to_uk(pmax(Gegen_coefs_Pn(k = k_PCvM, p = 11, type = "PCvM",
                                           N = 5120), 0), p = 11)
f_locdev_PCvM_11 <- function(z) f_locdev(z = z, p = 11, uk = uk_PCvM_11)
integrand_vec_PCvM_11 <- function(x) {
  f_gamma <- matrix(f_locdev_PCvM_11(c(X_11[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 11) * f_gamma - 1)^2 / n / rotasym::w_p(p = 11)
}

# PAD
k_PAD <- 1:1e3
uk_PAD_2 <- bk_to_uk(pmax(Gegen_coefs_Pn(k = k_PAD, p = 2, type = "PAD",
                                         N = 5120), 0), p = 2)
f_locdev_PAD_2 <- function(z) f_locdev(z = z, p = 2, uk = uk_PAD_2)
integrand_vec_PAD_2 <- function(x) {
  f_gamma <- matrix(f_locdev_PAD_2(c(X_2[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 2) * f_gamma - 1)^2 / n / rotasym::w_p(p = 2)
}
uk_PAD_3 <- bk_to_uk(pmax(Gegen_coefs_Pn(k = k_PAD, p = 3, type = "PAD",
                                         N = 5120), 0), p = 3)
f_locdev_PAD_3 <- function(z) f_locdev(z = z, p = 3, uk = uk_PAD_3)
integrand_vec_PAD_3 <- function(x) {
  f_gamma <- matrix(f_locdev_PAD_3(c(X_3[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 3) * f_gamma - 1)^2 / n / rotasym::w_p(p = 3)
}
uk_PAD_4 <- bk_to_uk(pmax(Gegen_coefs_Pn(k = k_PAD, p = 4, type = "PAD",
                                         N = 5120), 0), p = 4)
f_locdev_PAD_4 <- function(z) f_locdev(z = z, p = 4, uk = uk_PAD_4)
integrand_vec_PAD_4 <- function(x) {
  f_gamma <- matrix(f_locdev_PAD_4(c(X_4[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 4) * f_gamma - 1)^2 / n / rotasym::w_p(p = 4)
}
uk_PAD_11 <- bk_to_uk(pmax(Gegen_coefs_Pn(k = k_PAD, p = 11, type = "PAD",
                                          N = 5120), 0), p = 11)
f_locdev_PAD_11 <- function(z) f_locdev(z = z, p = 11, uk = uk_PAD_11)
integrand_vec_PAD_11 <- function(x) {
  f_gamma <- matrix(f_locdev_PAD_11(c(X_11[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 11) * f_gamma - 1)^2 / n / rotasym::w_p(p = 11)
}

# PRt 1 / 3
k_PRt <- 1:1e3
uk_PRt_2 <- bk_to_uk(Gegen_coefs_Pn(k = k_PRt, p = 2, type = "PRt",
                                    N = 0, Rothman_t = 1 / 3),
                     p = 2)
f_locdev_PRt_2 <- function(z) f_locdev(z = z, p = 2, uk = uk_PRt_2)
integrand_vec_PRt_2 <- function(x) {
  f_gamma <- matrix(f_locdev_PRt_2(c(X_2[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 2) * f_gamma - 1)^2 / n / rotasym::w_p(p = 2)
}
f_locdev_PRt_2_anal <- f_locdev_Pn(p = 2, type = "PRt", Rothman_t = 1 / 3)
integrand_vec_PRt_2_anal <- function(x) {
  f_gamma <- matrix(f_locdev_PRt_2_anal(c(X_2[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 2) * f_gamma - 1)^2 / n / rotasym::w_p(p = 2)
}
uk_PRt_3 <- bk_to_uk(Gegen_coefs_Pn(k = k_PRt, p = 3, type = "PRt",
                                    N = 0, Rothman_t = 1 / 3),
                     p = 3)
f_locdev_PRt_3 <- function(z) f_locdev(z = z, p = 3, uk = uk_PRt_3)
integrand_vec_PRt_3 <- function(x) {
  f_gamma <- matrix(f_locdev_PRt_3(c(X_3[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 3) * f_gamma - 1)^2 / n / rotasym::w_p(p = 3)
}
f_locdev_PRt_3_anal <- f_locdev_Pn(p = 3, type = "PRt", Rothman_t = 1 / 3)
integrand_vec_PRt_3_anal <- function(x) {
  f_gamma <- matrix(f_locdev_PRt_3_anal(c(X_3[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 3) * f_gamma - 1)^2 / n / rotasym::w_p(p = 3)
}
uk_PRt_4 <- bk_to_uk(Gegen_coefs_Pn(k = k_PRt, p = 4, type = "PRt",
                                    N = 0, Rothman_t = 1 / 3),
                     p = 4)
f_locdev_PRt_4 <- function(z) f_locdev(z = z, p = 4, uk = uk_PRt_4)
integrand_vec_PRt_4 <- function(x) {
  f_gamma <- matrix(f_locdev_PRt_4(c(X_4[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 4) * f_gamma - 1)^2 / n / rotasym::w_p(p = 4)
}
f_locdev_PRt_4_anal <- f_locdev_Pn(p = 4, type = "PRt", Rothman_t = 1 / 3)
integrand_vec_PRt_4_anal <- function(x) {
  f_gamma <- matrix(f_locdev_PRt_4_anal(c(X_4[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 4) * f_gamma - 1)^2 / n / rotasym::w_p(p = 4)
}
uk_PRt_11 <- bk_to_uk(Gegen_coefs_Pn(k = k_PRt, p = 11, type = "PRt",
                                     N = 0, Rothman_t = 1 / 3),
                      p = 11)
f_locdev_PRt_11 <- function(z) f_locdev(z = z, p = 11, uk = uk_PRt_11)
integrand_vec_PRt_11 <- function(x) {
  f_gamma <- matrix(f_locdev_PRt_11(c(X_11[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 11) * f_gamma - 1)^2 / n / rotasym::w_p(p = 11)
}
f_locdev_PRt_11_anal <- f_locdev_Pn(p = 11, type = "PRt", Rothman_t = 1 / 3)
integrand_vec_PRt_11_anal <- function(x) {
  f_gamma <- matrix(f_locdev_PRt_11_anal(c(X_11[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 11) * f_gamma - 1)^2 / n / rotasym::w_p(p = 11)
}

# Ajne
k_Ajne <- 1:1e3
vk2_Ajne_2 <- weights_dfs_Sobolev(p = 2, K_max = max(k_Ajne),
                                  type = "Ajne", thre = 0,
                                  verbose = FALSE)$weights
uk_Ajne_2 <- vk2_to_uk(vk2 = vk2_Ajne_2, p = 2,
                       signs = (-1)^((seq_along(vk2_Ajne_2) - 1) %/% 2))
f_locdev_Ajne_2 <- function(z) f_locdev(z = z, p = 2, uk = uk_Ajne_2)
integrand_vec_Ajne_2 <- function(x) {
  f_gamma <- matrix(f_locdev_Ajne_2(c(X_2[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 2) * f_gamma - 1)^2 / n / rotasym::w_p(p = 2)
}
vk2_Ajne_3 <- weights_dfs_Sobolev(p = 3, K_max = max(k_Ajne),
                                  type = "Ajne", thre = 0,
                                  verbose = FALSE)$weights
uk_Ajne_3 <- vk2_to_uk(vk2 = vk2_Ajne_3, p = 3)
f_locdev_Ajne_3 <- function(z) f_locdev(z = z, p = 3, uk = uk_Ajne_3)
integrand_vec_Ajne_3 <- function(x) {
  f_gamma <- matrix(f_locdev_Ajne_3(c(X_3[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 3) * f_gamma - 1)^2 / n / rotasym::w_p(p = 3)
}
vk2_Ajne_4 <- weights_dfs_Sobolev(p = 4, K_max = max(k_Ajne),
                                  type = "Ajne", thre = 0,
                                  verbose = FALSE)$weights
uk_Ajne_4 <- vk2_to_uk(vk2 = vk2_Ajne_4, p = 4)
f_locdev_Ajne_4 <- function(z) f_locdev(z = z, p = 4, uk = uk_Ajne_4)
integrand_vec_Ajne_4 <- function(x) {
  f_gamma <- matrix(f_locdev_Ajne_4(c(X_4[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 4) * f_gamma - 1)^2 / n / rotasym::w_p(p = 4)
}
vk2_Ajne_11 <- weights_dfs_Sobolev(p = 11, K_max = max(k_Ajne),
                                   type = "Ajne", thre = 0,
                                   verbose = FALSE)$weights
uk_Ajne_11 <- vk2_to_uk(vk2 = vk2_Ajne_11, p = 11)
f_locdev_Ajne_11 <- function(z) f_locdev(z = z, p = 11, uk = uk_Ajne_11)
integrand_vec_Ajne_11 <- function(x) {
  f_gamma <- matrix(f_locdev_Ajne_11(c(X_11[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 11) * f_gamma - 1)^2 / n / rotasym::w_p(p = 11)
}

# Ajne analytical
k_Ajne <- 1:1e3
f_Ajne_2 <- function(z) (1 * (z >= 0) + 0.5) / rotasym::w_p(p = 2)
integrand_vec_f_Ajne_2  <- function(x) {
  f_gamma <- matrix(f_Ajne_2(c(X_2[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 2) * f_gamma - 1)^2 / n / rotasym::w_p(p = 2)
}
f_Ajne_3 <- function(z) (1 * (z >= 0) + 0.5) / rotasym::w_p(p = 3)
integrand_vec_f_Ajne_3 <- function(x) {
  f_gamma <- matrix(f_Ajne_3(c(X_3[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 3) * f_gamma - 1)^2 / n / rotasym::w_p(p = 3)
}
f_Ajne_4 <- function(z) (1 * (z >= 0) + 0.5) / rotasym::w_p(p = 4)
integrand_vec_f_Ajne_4 <- function(x) {
  f_gamma <- matrix(f_Ajne_4(c(X_4[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 4) * f_gamma - 1)^2 / n / rotasym::w_p(p = 4)
}
f_Ajne_11 <- function(z) (1 * (z >= 0) + 0.5) / rotasym::w_p(p = 11)
integrand_vec_f_Ajne_11 <- function(x) {
  f_gamma <- matrix(f_Ajne_11(c(X_11[, , 1] %*% t(x))),
                    nrow = n, ncol = nrow(x))
  colSums(rotasym::w_p(p = 11) * f_gamma - 1)^2 / n / rotasym::w_p(p = 11)
}

test_that("PCvM as the integral of its local alternative", {

  skip_on_cran()
  expect_equal(drop(sph_stat_PCvM(X = X_2)),
               sum(w_k * integrand_vec_PCvM_2(cbind(cos(th_k), sin(th_k)))),
               tolerance = 5e-2)
  expect_equal(drop(sph_stat_PCvM(X = X_2)),
               integrate(function(th)
                 integrand_vec_PCvM_2_anal(x = cbind(cos(th), sin(th))),
                 lower = 0, upper = 2 * pi, abs.tol = 1e-5,
                 subdivisions = 1e3)$value,
               tolerance = 5e-2)
  expect_equal(drop(sph_stat_PCvM(X = X_3)), {
    int_sph_MC(f = integrand_vec_PCvM_3, p = 3, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16, M = 1e4)
  }, tolerance = 5e-2)
  expect_equal(drop(sph_stat_PCvM(X = X_4)), {
    int_sph_MC(f = integrand_vec_PCvM_4, p = 4, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16, M = 1e4)
  }, tolerance = 5e-2)
  expect_equal(drop(sph_stat_PCvM(X = X_11)), {
    int_sph_MC(f = integrand_vec_PCvM_11, p = 11, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16 + 10, M = 1e4)
  }, tolerance = 1e-1)

})

test_that("PAD as the integral of its local alternative", {

  skip_on_cran()
  expect_equal(drop(sph_stat_PAD(X = X_2)),
               sum(w_k * integrand_vec_PAD_2(cbind(cos(th_k), sin(th_k)))),
               tolerance = 1e-1)
  expect_equal(drop(sph_stat_PAD(X = X_3)), {
    int_sph_MC(f = integrand_vec_PAD_3, p = 3, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16, M = 1e4)
  }, tolerance = 5e-2)
  expect_equal(drop(sph_stat_PAD(X = X_4)), {
    int_sph_MC(f = integrand_vec_PAD_4, p = 4, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16, M = 1e4)
  }, tolerance = 2e-2)
  expect_equal(drop(sph_stat_PAD(X = X_11)), {
    int_sph_MC(f = integrand_vec_PAD_11, p = 11, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16 + 20, M = 1e4)
  }, tolerance = 3e-2)

})

test_that("PRt t = 1 / 3 as the integral of its local alternative", {

  skip_on_cran()
  expect_equal(drop(sph_stat_PRt(X = X_2, t = 1 / 3)),
               sum(w_k * integrand_vec_PRt_2(cbind(cos(th_k), sin(th_k)))),
               tolerance = 5e-2)
  expect_equal(drop(sph_stat_PRt(X = X_2, t = 1 / 3)),
               integrate(function(th)
                 integrand_vec_PRt_2_anal(x = cbind(cos(th), sin(th))),
                 lower = 0, upper = 2 * pi, abs.tol = 1e-5,
                 subdivisions = 1e3)$value,
               tolerance = 1e-4)
  expect_equal(drop(sph_stat_PRt(X = X_3, t = 1 / 3)), {
    int_sph_MC(f = integrand_vec_PRt_3, p = 3, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16, M = 1e4)
  }, tolerance = 5e-2)
  expect_equal(drop(sph_stat_PRt(X = X_3, t = 1 / 3)), {
    int_sph_MC(f = integrand_vec_PRt_3_anal, p = 3, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16, M = 1e4)
  }, tolerance = 3e-2)
  expect_equal(drop(sph_stat_PRt(X = X_4, t = 1 / 3)), {
    int_sph_MC(f = integrand_vec_PRt_4, p = 4, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16, M = 1e4)
  }, tolerance = 5e-2)
  expect_equal(drop(sph_stat_PRt(X = X_4, t = 1 / 3)), {
    int_sph_MC(f = integrand_vec_PRt_4_anal, p = 4, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16, M = 1e4)
  }, tolerance = 3e-2)
  expect_equal(drop(sph_stat_PRt(X = X_11, t = 1 / 3)), {
    int_sph_MC(f = integrand_vec_PRt_11, p = 11, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16 + 10, M = 1e4)
  }, tolerance = 5e-2)
  expect_equal(drop(sph_stat_PRt(X = X_11, t = 1 / 3)), {
    int_sph_MC(f = integrand_vec_PRt_11_anal, p = 11, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16, M = 1e4)
  }, tolerance = 3e-2)

})

test_that("Ajne as the integral of its local alternative (series expansion)", {

  expect_equal(drop(sph_stat_Ajne(X = X_2)),
               sum(w_k * integrand_vec_Ajne_2(cbind(cos(th_k), sin(th_k)))),
               tolerance = 5e-2)
  expect_equal(drop(sph_stat_Ajne(X = X_3)), {
    int_sph_MC(f = integrand_vec_Ajne_3, p = 3, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16, M = 1e4)
  }, tolerance = 5e-2)
  expect_equal(drop(sph_stat_Ajne(X = X_4)), {
    int_sph_MC(f = integrand_vec_Ajne_4, p = 4, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16, M = 1e4)
  }, tolerance = 5e-2)
  expect_equal(drop(sph_stat_Ajne(X = X_11)), {
    int_sph_MC(f = integrand_vec_Ajne_11, p = 11, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16 + 10, M = 1e4)
  }, tolerance = 1e-1)

})

test_that("Ajne as the integral of its local alternative (analytical)", {

  skip_on_cran()
  expect_equal(drop(sph_stat_Ajne(X = X_2)),
               sum(w_k * integrand_vec_f_Ajne_2(cbind(cos(th_k), sin(th_k)))),
               tolerance = 5e-2)
  expect_equal(drop(sph_stat_Ajne(X = X_3)), {
    int_sph_MC(f = integrand_vec_f_Ajne_3, p = 3, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16, M = 1e4)
  }, tolerance = 5e-2)
  expect_equal(drop(sph_stat_Ajne(X = X_4)), {
    int_sph_MC(f = integrand_vec_f_Ajne_4, p = 4, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16, M = 1e4)
  }, tolerance = 5e-2)
  expect_equal(drop(sph_stat_Ajne(X = X_11)), {
    int_sph_MC(f = integrand_vec_f_Ajne_11, p = 11, cores = 2, chunks = 16,
               .export = ls(), seeds = 1:16 + 10, M = 1e4)
  }, tolerance = 3e-2)

})

test_that("Check positivity of f_PCvM and f_PAD", {

  skip_on_cran()
  expect_true(all(f_locdev_PCvM_2(z = z) > 0))
  expect_true(all(f_locdev_PCvM_3(z = z) > 0))
  expect_true(all(f_locdev_PCvM_4(z = z) > 0))
  expect_true(all(f_locdev_PCvM_11(z = z8) > 0))
  expect_true(all(f_locdev_PAD_2(z = z) > 0))
  expect_true(all(f_locdev_PAD_3(z = z) > 0))
  expect_true(all(f_locdev_PAD_4(z = z) > 0))
  expect_true(all(f_locdev_PAD_11(z = z8) > 0))

})

vk2 <- 1:5
bk <- 5:1
uk <- seq(-1, 1, l = 5)

test_that("Conversion between coefficients", {

  skip_on_cran()
  expect_equal(bk_to_vk2(vk2_to_bk(vk2, p = 2), p = 2), vk2)
  expect_equal(bk_to_vk2(vk2_to_bk(vk2, p = 3), p = 3), vk2)
  expect_equal(bk_to_uk(uk_to_bk(uk, p = 2), p = 2, signs = sign(uk)), uk)
  expect_equal(bk_to_uk(uk_to_bk(uk, p = 3), p = 3, signs = sign(uk)), uk)
  expect_equal(vk2_to_bk(bk_to_vk2(bk, p = 2), p = 2), bk)
  expect_equal(vk2_to_bk(bk_to_vk2(bk, p = 3), p = 3), bk)
  expect_equal(vk2_to_uk(uk_to_vk2(uk, p = 2), p = 2, signs = sign(uk)), uk)
  expect_equal(vk2_to_uk(uk_to_vk2(uk, p = 3), p = 3, signs = sign(uk)), uk)
  expect_equal(uk_to_vk2(vk2_to_uk(vk2, p = 2), p = 2), vk2)
  expect_equal(uk_to_vk2(vk2_to_uk(vk2, p = 3), p = 3), vk2)
  expect_equal(uk_to_bk(bk_to_uk(bk, p = 2), p = 2), bk)
  expect_equal(uk_to_bk(bk_to_uk(bk, p = 3), p = 3), bk)

})

test_that("Conversion of bk to vk2 in projected-ecdf statistics", {

  skip_on_cran()
  for (type in c("PCvM", "PAD", "PRt")) {
    for (p in c(2, 3, 4, 9)) {
      expect_equal(bk_to_vk2(Gegen_coefs_Pn(k = 1:9, p = p, type = type),
                             p = p),
                   weights_dfs_Sobolev(K_max = 9, thre = 0, p = p, type = type,
                                       verbose = FALSE)$weights)
    }
  }

})

test_that("Conversion of bk to uk in projected-ecdf statistics", {

  skip_on_cran()
  for (type in c("PCvM", "PAD", "PRt")) {
    for (p in c(2, 3, 4, 9)) {
      expect_equal(bk_to_uk(Gegen_coefs_Pn(k = 1:9, p = p, type = type,
                                           Rothman_t = 0.1), p = p),
                   abs(cutoff_locdev(K_max = 9, thre = 0, p = p, type = type,
                                     Rothman_t = 0.1)))
    }
  }

})

test_that("cutoff_locdev verbose", {

  skip_on_cran()
  expect_message(cutoff_locdev(K_max = 9, thre = 0, p = 9, type = "PAD",
                               verbose = 1))
  suppressMessages(
    expect_message(cutoff_locdev(K_max = 1e1, thre = 1e-4, p = 4, type = "PCvM",
                                 verbose = 2)))

})

# MJ (2000) page 114 and applying modulus
f1_orig <- function(theta) {
  return(theta^2 / (2 * pi^2))
}
f1_mod <- function(theta) {
  theta <- (theta + pi) %% (2 * pi) - pi
  return(theta^2 / (2 * pi^2))
}

# f^CvM local deviation
f2 <- function(theta) {
  res <- (1 - log(2 * (1 - cos(theta))) * sqrt(2) / (2 * pi)) / (2 * pi)
  res[is.infinite(res)] <- 0
  return(res)
}

# Simulation
M <- 50
n <- 20
int1_orig <- int1_mod <- int2 <- wat <- numeric(M)
set.seed(1323131)
for (i in 1:M) {

  # Sample
  theta_i <- rnorm(n = n) %% (2 * pi)

  # MJ (2000) (6.3.70) + (6.3.60)
  int1_orig[i] <- integrate(f = function(th) {
    sapply(th, function(theta) (sum(f1_mod(theta + theta_i)) - n / (2 * pi))^2)
  }, lower = 0, upper = 2 * pi, abs.tol = 1e-6, subdivisions = 1e4,
  stop.on.error = TRUE)$value / (2 * pi) / (4 * n)

  # MJ (2000) (6.3.70) + (6.3.60) applying modulus
  int1_mod[i] <- integrate(f = function(th) {
    sapply(th, function(theta) (sum(f1_orig(theta + theta_i)) - n / (2 * pi))^2)
  }, lower = 0, upper = 2 * pi, abs.tol = 1e-6, subdivisions = 1e4,
  stop.on.error = TRUE)$value / (2 * pi) / (4 * n)

  # f^CvM local deviation
  int2[i] <- 0.5 * integrate(f = function(th) {
    sapply(th, function(theta) sum((2 * pi) * f2(theta + theta_i) - 1)^2)
  }, lower = 0, upper = 2 * pi, abs.tol = 1e-6, subdivisions = 1e4,
  stop.on.error = TRUE)$value / (2 * pi * n)

  # Watson statistic
  wat[i] <- sphunif::cir_stat_Watson(Theta = cbind(theta_i))

}

test_that("Watson vs. Sobolev statistic using f^PCvM", {

  expect_equal(wat, int2, tolerance = 1e-4)

})

test_that("Watson vs. Sobolev statistic using f in MJ (2000) page 114", {

  skip_on_cran()
  expect_false(isTRUE(all.equal(wat, int1_orig, tolerance = 1e-3)))
  expect_false(isTRUE(all.equal(diff(wat / int1_orig), rep(0, M - 1),
                                tolerance = 1e-3)))
  expect_false(isTRUE(all.equal(wat, int1_mod, tolerance = 1e-3)))
  expect_false(isTRUE(all.equal(diff(wat / int1_mod), rep(0, M - 1),
                                tolerance = 1e-3)))

})

# par(mfrow = c(2, 2))
# plot(wat, int1_orig, main = "MJ (2000) page 114 + (6.3.70) + (6.3.60)")
# plot(wat, int1_mod, main = "With modulus")
# plot(wat, int2, main = "f^CvM local alternative")
# curve(f1_orig, from = 0, to = 2 * pi, n = 1e4, ylim = c(0, 1),
#        main = "f functions")
# curve(f1_mod, add = TRUE, col = 2, n = 1e4)
# curve(f2, add = TRUE, col = 3, n = 1e4)

#' @title Monte Carlo simulation of circular and (hyper)spherical uniformity
#' statistics
#'
#' @description Utility for performing Monte Carlo simulation of several
#' statistics for assessing uniformity on the (hyper)sphere
#' \eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{
#' S^{p-1}:=\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}.
#'
#' \code{unif_stat_MC} provides a convenient wrapper for parallel
#' evaluation of \code{unif_stat}, the estimation of critical values under the
#' null distribution, and the computation of empirical powers under the
#' alternative.
#'
#' @inheritParams unif_test
#' @inheritParams r_unif
#' @param M number of Monte Carlo replications. Defaults to \code{1e4}.
#' @param r_H1 if provided, the computation of empirical powers is
#' carried out for the alternative hypothesis sampled with \code{r_H1}.
#' This must be a function with the same arguments and value as
#' \code{\link{r_unif_sph}} (see examples). Defaults to \code{NULL}, indicating
#' that the critical values are estimated from samples of \code{r_unif_sph}.
#' @param crit_val if provided, must be the critical values as returned by
#' \code{$stats_MC} in a call to \code{unif_stat_MC}. They are used for
#' computing the empirical powers of the tests present in \code{type}.
#' Defaults to \code{NULL}, which means that no power computation is done.
#' @param return_stats return the Monte Carlo statistics? If only the critical
#' values or powers are desired, \code{FALSE} saves memory in the returned
#' object. Defaults to \code{TRUE}.
#' @param stats_sorted sort the returned Monte Carlo statistics? If
#' \code{TRUE}, this is useful for evaluating faster the empirical cumulative
#' distribution function when approximating the distribution in
#' \code{\link{unif_stat_distr}}. Defaults to \code{FALSE}.
#' @param chunks number of chunks to split the \code{M} Monte Carlo
#' replications. Useful for parallelizing the simulation study in \code{chunks}
#' tasks containing \code{ceiling(M / chunks)} replications. Useful also for
#' avoiding memory bottlenecks when \code{M} is large. Defaults to
#' \cr\code{ceiling((n * M) / 1e5)}.
#' @param cores number of cores to perform the simulation. Defaults to \code{1}.
#' @param seeds if provided, a vector of size \code{chunks} for fixing the
#' seeds on each of the simulation chunks (useful for reproducing parallel
#' simulations). Specifically, for \code{k in 1:chunks}, seeds are
#' set as \code{set.seed(seeds[k], kind = "Mersenne-Twister")} in each chunk.
#' Defaults to \code{NULL} (no seed setting is done).
#' @inheritParams unif_stat
#' @param ... optional arguments to be passed to the \code{r_H1} sampler or to
#' \code{\link[foreach]{foreach}} (for example, \code{.export} to export global
#' variables or other functions to the \code{foreach} environment).
#' @return A list with the following entries:
#' \itemize{
#'   \item \code{crit_val_MC}: a data frame of size
#'   \code{c(length(alpha), length(type))}, with column names given by
#'   \code{type} and rows corresponding to the significance levels \code{alpha},
#'   that contains the estimated critical values of the tests.
#'   \item \code{power_MC}: a data frame of size
#'   \code{c(nrow(crit_val), length(type))}, with column names given by
#'   \code{type} and rows corresponding to the significance levels of
#'   \code{crit_val}, that contains the empirical powers of the tests. \code{NA}
#'   if \code{crit_val = NULL}.
#'   \item \code{stats_MC}: a data frame of size \code{c(M, length(type))}, with
#'   column names given by \code{type}, that contains the Monte Carlo
#'   statistics.
#' }
#' @details
#' It is possible to have a progress bar if \code{unif_stat_MC} is wrapped with
#' \code{\link[progressr:with_progress]{progressr::with_progress}} or if
#' \code{progressr::handlers(global = TRUE)} is invoked (once) by the user.
#' See the examples below. The progress bar is updated with the number of
#' finished chunks.
#'
#' All the tests reject for large values of the test statistic
#' (\code{max_gap = TRUE} is assumed for the Range test), so the critical
#' values for the significance levels \code{alpha} correspond to the
#' \code{alpha}-upper quantiles of the null distribution of the test statistic.
#'
#' The Monte Carlo simulation for the CCF09 test is made conditionally
#' on the choice of \code{CCF09_dirs}. That is, all the Monte
#' Carlo statistics share the same random directions.
#' @examples
#' ## Critical values
#'
#' # Single statistic, specific alpha
#' cir <- unif_stat_MC(n = 10, M = 1e2, type = "Ajne", p = 2, alpha = 0.15)
#' summary(cir$stats_MC)
#' cir$crit_val_MC
#'
#' # All circular statistics
#' cir <- unif_stat_MC(n = 10, M = 1e2, p = 2)
#' head(cir$stats_MC)
#' cir$crit_val_MC
#'
#' # All spherical statistics
#' sph <- unif_stat_MC(n = 10, M = 1e2, p = 3)
#' head(sph$stats_MC)
#' sph$crit_val_MC
#'
#' ## Using a progress bar
#'
#' # Define a progress bar
#' require(progress)
#' require(progressr)
#' handlers(handler_progress(
#'   format = ":spin [:bar] :percent Total: :elapsedfull End \u2248 :eta",
#'   clear = FALSE))
#'
#' # Call unif_stat_MC() within with_progress()
#' with_progress(unif_stat_MC(n = 10, M = 1e2, p = 3, chunks = 10))
#'
#' # With several cores
#' with_progress(unif_stat_MC(n = 10, M = 1e2, p = 3, chunks = 10, cores = 2))
#'
#' # Instead of using with_progress() each time, it is more practical to run
#' # handlers(global = TRUE)
#' # once to activate progress bars in your R session
#'
#' ## Power computation
#'
#' # Single statistic
#' cir_pow <- unif_stat_MC(n = 10, M = 1e2, type = "Ajne", p = 2,
#'                         crit_val = cir$crit_val_MC)
#' cir_pow$crit_val_MC
#' cir_pow$power_MC
#'
#' # All circular statistics
#' cir_pow <- unif_stat_MC(n = 10, M = 1e2, p = 2, crit_val = cir$crit_val_MC)
#' cir_pow$crit_val_MC
#' cir_pow$power_MC
#'
#' # All spherical statistics
#' sph_pow <- unif_stat_MC(n = 10, M = 1e2, p = 3, crit_val = sph$crit_val_MC)
#' sph_pow$crit_val_MC
#' sph_pow$power_MC
#' \donttest{
#' ## Custom r_H1
#'
#' # Circular
#' r_H1 <- function(n, p, M, l = 0.05) {
#'
#'   stopifnot(p == 2)
#'   Theta_to_X(matrix(runif(n * M, 0, (2 - l) * pi), n, M))
#'
#' }
#' dirs <- r_unif_sph(n = 5, p = 2, M = 1)[, , 1]
#' cir <- unif_stat_MC(n = 50, M = 1e2, p = 2, CCF09_dirs = dirs)
#' cir_pow <- unif_stat_MC(n = 50, M = 1e2, p = 2, r_H1 = r_H1, l = 0.10,
#'                         crit_val = cir$crit_val_MC, CCF09_dirs = dirs)
#' cir_pow$crit_val_MC
#' cir_pow$power_MC
#'
#' # Spherical
#' r_H1 <- function(n, p, M, l = 0.5) {
#'
#'   samp <- array(dim = c(n, p, M))
#'   for (j in 1:M) {
#'
#'     samp[, , j] <- mvtnorm::rmvnorm(n = n, mean = c(l, rep(0, p - 1)),
#'                                     sigma = diag(rep(1, p)))
#'     samp[, , j] <- samp[, , j] / sqrt(rowSums(samp[, , j]^2))
#'
#'   }
#'   return(samp)
#'
#' }
#' dirs <- r_unif_sph(n = 5, p = 3, M = 1)[, , 1]
#' sph <- unif_stat_MC(n = 50, M = 1e2, p = 3, CCF09_dirs = dirs)
#' sph_pow <- unif_stat_MC(n = 50, M = 1e2, p = 3, r_H1 = r_H1, l = 0.5,
#'                        crit_val = sph$crit_val_MC, CCF09_dirs = dirs)
#' sph_pow$power_MC
#'
#' ## Pre-built r_H1
#'
#' # Circular
#' dirs <- r_unif_sph(n = 5, p = 2, M = 1)[, , 1]
#' cir_pow <- unif_stat_MC(n = 50, M = 1e2, p = 2, r_H1 = r_alt, alt = "vMF",
#'                         kappa = 1, crit_val = cir$crit_val_MC,
#'                         CCF09_dirs = dirs)
#' cir_pow$power_MC
#'
#' # Spherical
#' dirs <- r_unif_sph(n = 5, p = 3, M = 1)[, , 1]
#' sph_pow <- unif_stat_MC(n = 50, M = 1e2, p = 3, r_H1 = r_alt, alt = "vMF",
#'                         kappa = 1, crit_val = sph$crit_val_MC,
#'                         CCF09_dirs = dirs)
#' sph_pow$power_MC
#' }
#' @export
unif_stat_MC <- function(n, type = "all", p, M = 1e4, r_H1 = NULL,
                         crit_val = NULL, alpha = c(0.10, 0.05, 0.01),
                         return_stats = TRUE, stats_sorted = FALSE,
                         chunks = ceiling((n * M) / 1e5), cores = 1,
                         seeds = NULL, Rayleigh_m = 1, cov_a = 2 * pi,
                         Rothman_t = 1 / 3, Cressie_t = 1 / 3, Pycke_q = 0.5,
                         Riesz_s = 1, CCF09_dirs = NULL, K_CCF09 = 25,
                         CJ12_reg = 3, ...) {

  # Check dimension
  if (p < 2) {

    stop("Dimension p must be p >= 2.")

  }

  # Chunk large n * M to avoid memory issues
  small_M <- M %/% chunks

  # Fix projections for the CCF09 test
  if ("CCF09" %in% type & is.null(CCF09_dirs)) {

    CCF09_dirs <- r_unif_sph(n = 50, p = p, M = 1)[, , 1]

  }

  # If the uniformity is simulated
  if (is.null(r_H1)) {

    r_H1 <- r_unif_sph

  }

  # Check if it is a data frame with colnames in avail_cir or avail_sph
  if (!is.null(crit_val)) {

    avail <- switch((p > 2) + 1, avail_cir_tests, avail_sph_tests)
    tests <- colnames(crit_val) %in% avail
    if (!all(tests)) {

      stop(paste0(
        "crit_val must be a data.frame with valid tests names as colnames, ",
        "check: ", paste(paste0("\"", colnames(crit_val)[!tests], "\""),
                         collapse = ", "), "."))

    }

  }

  # Extra arguments for foreach::foreach and r_H1
  dots <- list(...)
  foreach_args <- dots[names(dots) %in% names(formals(foreach::foreach))]
  r_H1_args <- dots[names(dots) %in% names(formals(r_H1))]

  # Check seeds
  if (!is.null(seeds) & length(seeds) != chunks) {

    warning(paste("seeds and chunks do not have the same length:",
                  "seeds are ignored."))
    seeds <- NULL

  }

  # Parallel backend
  old_dopar <- doFuture::registerDoFuture()
  old_plan <- future::plan(future::multisession(), workers = cores)
  options(future.rng.onMisuse = "ignore")
  on.exit({

    with(old_dopar, foreach::setDoPar(fun = fun, data = data, info = info))
    future::plan(old_plan)
    options(future.rng.onMisuse = NULL)

  })
  `%op%` <- foreach::`%dopar%`

  # Measure progress?
  if (requireNamespace("progressr", quietly = TRUE)) {

    prog <- progressr::progressor(along = 1:chunks)

  }

  # Monte Carlo
  k <- 0
  stats <- do.call(what = foreach::foreach,
                   args = c(list(k = 1:chunks, .combine = rbind,
                                 .inorder = TRUE, .multicombine = TRUE,
                                 .maxcombine = 100, .packages = "sphunif"),
                            foreach_args)) %op% {

    # Samples
    if (!is.null(seeds)) {

      set.seed(seeds[k], kind = "Mersenne-Twister")

    }
    X <- do.call(what = r_H1, args = c(list(n = n, p = p, M = small_M),
                                            r_H1_args))

    # Statistics
    stats <- unif_stat(data = X, type = type, Rayleigh_m = Rayleigh_m,
                       cov_a = cov_a, Rothman_t = Rothman_t,
                       Cressie_t = Cressie_t, Pycke_q = Pycke_q,
                       Riesz_s = Riesz_s, CCF09_dirs = CCF09_dirs,
                       K_CCF09 = K_CCF09, CJ12_reg = CJ12_reg)

    # Remove X and clean memory
    rm(X)
    gc()

    # Signal progress
    if (requireNamespace("progressr", quietly = TRUE)) {

      prog()

    }

    # Return stats
    stats

  }

  # Sort statistics
  if (stats_sorted & return_stats) {

    names <- names(stats)
    stats <- as.data.frame(sort_each_col(as.matrix(stats)))
    names(stats) <- names

  }

  # Build tables
  if (is.null(crit_val)) {

    # Critical values
    crit_val <- rbind(apply(stats, 2, quantile, probs = 1 - alpha,
                            na.rm = TRUE))
    crit_val <- as.data.frame(crit_val)
    rownames(crit_val) <- alpha

    # Power
    power <- NA

  } else {

    # In case crit_val contains more tests than the ones considered in type
    # and in a potentially different order, match them accordingly with stats
    names_stats <- colnames(stats)
    ind_tests <- match(x = names_stats, table = colnames(crit_val))
    if (any(is.na(ind_tests))) {

      stop(paste0(
        "No critical values in \"crit_val\" associated with the test(s) ",
        paste(paste0("\"", names_stats[is.na(ind_tests)], "\""),
              collapse = ", "), "."))

    }
    crit_val <- crit_val[, ind_tests, drop = FALSE]

    # Empirical rejection level
    power <- t(apply(crit_val, 1,
                     function(c) rowMeans(t(stats) > c, na.rm = TRUE)))
    if (ncol(crit_val) == 1) {

      power <- t(power)
      colnames(power) <- names_stats

    }
    power <- as.data.frame(power)

  }

  # Empty stats if not desired
  if (!return_stats) {

    stats <- NA

  }

  return(list("crit_val_MC" = crit_val, "power_MC" = power, "stats_MC" = stats))

}

#' @title Asymptotic distributions of Sobolev statistics of spherical uniformity
#'
#' @description Approximated density, distribution, and quantile functions for
#' the asymptotic null distributions of Sobolev statistics of uniformity
#' on \eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{S^{p-1}:=
#' \{x\in R^p:||x||=1\}}. These asymptotic distributions are infinite
#' weighted sums of (central) chi squared random variables:
#' \deqn{\sum_{k = 1}^\infty v_k^2 \chi^2_{d_{p, k}},}
#' where
#' \deqn{d_{p, k} := {{p + k - 3}\choose{p - 2}} + {{p + k - 2}\choose{p - 2}}}
#' is the dimension of the space of eigenfunctions of the Laplacian on
#' \eqn{S^{p-1}}, \eqn{p\ge 2}, associated to the \eqn{k}-th
#' eigenvalue, \eqn{k\ge 1}.
#'
#' @inheritParams r_unif
#' @param k sequence of integer indexes.
#' @param K_max integer giving the truncation of the series that compute the
#' asymptotic p-value of a Sobolev test. Defaults to \code{1e3}.
#' @param thre error threshold for the tail probability given by the
#' the first terms of the truncated series of a Sobolev test. Defaults to
#' \code{1e-3}.
#' @param type Sobolev statistic. For \eqn{p = 2}, either \code{"Watson"},
#' \code{"Rothman"}, \code{"Pycke_q"}, or \code{"Hermans_Rasson"}.
#' For \eqn{p \ge 2}, \code{"Ajne"}, \code{"Gine_Gn"}, \code{"Gine_Fn"},
#' \code{"Bakshaev"}, \code{"Riesz"}, \code{"PCvM"}, \code{"PAD"}, or
#' \code{"PRt"}.
#' @param log compute the logarithm of \eqn{d_{p,k}}? Defaults to
#' \code{FALSE}.
#' @param verbose output information about the truncation? Defaults to
#' \code{TRUE}.
#' @inheritParams unif_stat
#' @inheritParams wschisq
#' @inheritParams Gegenbauer
#' @inheritParams Pn
#' @param force_positive set negative
#' @param ... further parameters passed to \code{*_\link{wschisq}}.
#' @return
#' \itemize{
#'   \item \code{d_p_k}: a vector of size \code{length(k)} with the
#'   evaluation of \eqn{d_{p,k}}.
#'   \item \code{weights_dfs_Sobolev}: a list with entries \code{weights} and
#'   \code{dfs}, automatically truncated according to \code{K_max} and
#'   \code{thre} (see details).
#'   \item \code{d_Sobolev}: density function evaluated at \code{x}, a vector.
#'   \item \code{p_Sobolev}: distribution function evaluated at \code{x},
#'   a vector.
#'   \item \code{q_Sobolev}: quantile function evaluated at \code{u}, a vector.
#' }
#' @author Eduardo García-Portugués and Paula Navarro-Esteban.
#' @details
#' The truncation of \eqn{\sum_{k = 1}^\infty v_k^2 \chi^2_{d_{p, k}}} is
#' done to the first \code{K_max} terms and then up to the index such that
#' the first terms explain the tail probability at the \code{x_tail} with
#' an absolute error smaller than \code{thre} (see details in
#' \code{\link{cutoff_wschisq}}). This automatic truncation takes place when
#' calling \code{*_Sobolev}. Setting \code{thre = 0} truncates to \code{K_max}
#' terms exactly. If the series only contains odd or even non-zero terms, then
#' only \code{K_max / 2} addends are \emph{effectively} taken into account
#' in the first truncation.
#' @examples
#' # Circular-specific statistics
#' curve(p_Sobolev(x = x, p = 2, type = "Watson", method = "HBE"),
#'       n = 2e2, ylab = "Distribution", main = "Watson")
#' curve(p_Sobolev(x = x, p = 2, type = "Rothman", method = "HBE"),
#'       n = 2e2, ylab = "Distribution", main = "Rothman")
#' curve(p_Sobolev(x = x, p = 2, type = "Pycke_q", method = "HBE"), to = 10,
#'       n = 2e2, ylab = "Distribution", main = "Pycke_q")
#' curve(p_Sobolev(x = x, p = 2, type = "Hermans_Rasson", method = "HBE"),
#'       to = 10, n = 2e2, ylab = "Distribution", main = "Hermans_Rasson")
#'
#' # Statistics for arbitrary dimensions
#' test_statistic <- function(type, to = 1, pmax = 5, M = 1e3, ...) {
#'
#'   col <- viridisLite::viridis(pmax - 1)
#'   curve(p_Sobolev(x = x, p = 2, type = type, method = "MC", M = M,
#'                   ...), to = to, n = 2e2, col = col[pmax - 1],
#'                   ylab = "Distribution", main = type, ylim = c(0, 1))
#'   for (p in 3:pmax) {
#'     curve(p_Sobolev(x = x, p = p, type = type, method = "MC", M = M,
#'                     ...), add = TRUE, n = 2e2, col = col[pmax - p + 1])
#'   }
#'   legend("bottomright", legend = paste("p =", 2:pmax), col = rev(col),
#'          lwd = 2)
#'
#' }
#'
#' # Ajne
#' test_statistic(type = "Ajne")
#' \donttest{
#' # Gine_Gn
#' test_statistic(type = "Gine_Gn", to = 1.5)
#'
#' # Gine_Fn
#' test_statistic(type = "Gine_Fn", to = 2)
#'
#' # Bakshaev
#' test_statistic(type = "Bakshaev", to = 3)
#'
#' # Riesz
#' test_statistic(type = "Riesz", Riesz_s = 0.5, to = 3)
#'
#' # PCvM
#' test_statistic(type = "PCvM", to = 0.6)
#'
#' # PAD
#' test_statistic(type = "PAD", to = 3)
#'
#' # PRt
#' test_statistic(type = "PRt", Rothman_t = 0.5)
#'
#' # Quantiles
#' p <- c(2, 3, 4, 11)
#' t(sapply(p, function(p) q_Sobolev(u = c(0.10, 0.05, 0.01), p = p,
#'                                   type = "PCvM")))
#' t(sapply(p, function(p) q_Sobolev(u = c(0.10, 0.05, 0.01), p = p,
#'                                   type = "PAD")))
#' t(sapply(p, function(p) q_Sobolev(u = c(0.10, 0.05, 0.01), p = p,
#'                                   type = "PRt")))
#'
#' # Series truncation for thre = 1e-5
#' sapply(p, function(p) length(weights_dfs_Sobolev(p = p, type = "PCvM")$dfs))
#' sapply(p, function(p) length(weights_dfs_Sobolev(p = p, type = "PRt")$dfs))
#' sapply(p, function(p) length(weights_dfs_Sobolev(p = p, type = "PAD")$dfs))
#' }
#' @name Sobolev


#' @rdname Sobolev
#' @export
d_p_k <- function(p, k, log = FALSE) {

  # log(nu_{p, k})
  p <- p - 2
  log_dfs <- lchoose(n = p + k - 1, k = p) + log(2 + p / k)
  log_dfs[k == 0] <- 0

  # nu_{p, k}
  if (!log) {

    log_dfs <- round(exp(log_dfs))

  }
  return(log_dfs)

}


#' @rdname Sobolev
#' @export
weights_dfs_Sobolev <- function(p, K_max = 1e3, thre = 1e-3, type,
                                Rothman_t = 1 / 3, Pycke_q = 0.5, Riesz_s = 1,
                                log = FALSE, verbose = TRUE, Gauss = TRUE,
                                N = 320, tol = 1e-6, force_positive = TRUE,
                                x_tail = NULL) {

  # alpha
  alpha <- 0.5 * p - 1

  # Sobolev weights and dfs
  if (p == 2 & type %in% c("Watson", "Rothman", "Hermans_Rasson", "Pycke_q")) {

    if (type == "Watson") {

      # Sequence of indexes
      k <- 1:K_max

      # log(v_k^2)
      log_vk2 <- -2 * log(k * pi)

      # log(d_{2, k})
      log_dk <- d_p_k(p = 2, k = k, log = TRUE)

      # Log weights and dfs
      log_weights <- log_vk2
      log_dfs <- log_dk

    } else if (type == "Rothman") {

      # Sequence of indexes
      k <- 1:K_max

      # log(v_k^2)
      log_vk2 <- log((sin(k * pi * Rothman_t) / (pi * k))^2)

      # log(d_{2, k})
      log_dk <- d_p_k(p = 2, k = k, log = TRUE)

      # Log weights and dfs
      log_weights <- log_vk2
      log_dfs <- log_dk

    } else if (type == "Hermans_Rasson") {

      # Sequence of indexes
      k <- 1:K_max

      # log(v_{2 * k - 1}^2)
      log_v2km12 <- log(2 / pi) - log((2 * k - 1)^2)

      # log(v_{2 * k}^2)
      beta2 <- (pi^2 / 36) / (0.5 - 4 / pi^2)
      log_v2k2 <- log(2 * beta2 / pi) - log((2 * k)^2 - 1)

      # log(d_{p, 2 * k - 1})
      log_d2km1 <- d_p_k(p = 2, k = 2 * k - 1, log = TRUE)

      # log(d_{p, 2 * k})
      log_dk <- d_p_k(p = 2, k = 2 * k, log = TRUE)

      # Log weights and dfs
      log_weights <- c(rbind(log_v2km12, log_v2k2))
      log_dfs <- c(rbind(log_d2km1, log_dk))

    } else if (type == "Pycke_q") {

      # Sequence of indexes
      k <- 1:K_max

      # log(v_k^2)
      log_vk2 <- (k - 1) * log(Pycke_q)

      # log(d_{2, k})
      log_dk <- d_p_k(p = 2, k = k, log = TRUE)

      # Log weights and dfs
      log_weights <- log_vk2
      log_dfs <- log_dk

    }

  } else {

    if (type == "Ajne") {

      # Halve K_max since the even coefficients are zero
      K_max <- K_max %/% 2

      # Sequence of indexes
      k <- 1:K_max

      # log(v_{2 * k - 1}^2)
      log_v2km12 <- (p - 2) * log(2) + lgamma(alpha + 1) + lgamma(k + alpha) +
        lgamma(2 * k - 1) - (log(pi) + lgamma(k) + lgamma(2 * k + p - 2))
      log_v2km12 <- 2 * log_v2km12

      # Add the zero coefficients for the even terms
      log_v2km12 <- c(rbind(log_v2km12, -Inf))

      # log(d_{p, k})
      log_dk <- d_p_k(p = p, k = 1:(2 * K_max), log = TRUE)

      # Log weights and dfs
      log_weights <- log_v2km12
      log_dfs <- log_dk

    } else if (type == "Gine_Gn") {

      # Halve K_max since the odd coefficients are zero
      K_max <- K_max %/% 2

      # Sequence of indexes
      k <- 1:K_max

      # log(v_{2 * k}^2)
      log_v2k2 <- log((p - 1) * (2 * k - 1) / (8 * pi * (2 * k + p - 1))) +
        2 * (lgamma(alpha + 0.5) + lgamma(k - 0.5) - lgamma(k + alpha + 0.5))

      # Add the zero coefficients for the odd terms
      log_v2k2 <- c(rbind(-Inf, log_v2k2))

      # log(d_{p, k})
      log_dk <- d_p_k(p = p, k = 1:(2 * K_max), log = TRUE)

      # Log weights and dfs
      log_weights <- log_v2k2
      log_dfs <- log_dk

    } else if (type == "Gine_Fn") {

      # Halve K_max since we sum 4 * Ajne and Gine_Gn
      K_max <- K_max %/% 2

      # Sequence of indexes
      k <- 1:K_max

      # log(v_{2 * k - 1}^2)
      log_v2km12 <- (p - 2) * log(2) + lgamma(alpha + 1) + lgamma(k + alpha) +
        lgamma(2 * k - 1) - (log(pi) + lgamma(k) + lgamma(2 * k + p - 2))
      log_v2km12 <- 2 * log_v2km12

      # log(v_{2 * k}^2)
      log_v2k2 <- log((p - 1) * (2 * k - 1) / (8 * pi * (2 * k + p - 1))) +
        2 * (lgamma(alpha + 0.5) + lgamma(k - 0.5) - lgamma(k + alpha + 0.5))

      # log(d_{p, k})
      log_dk <- d_p_k(p = p, k = 1:(2 * K_max), log = TRUE)

      # Log weights and dfs
      log_weights <- c(rbind(log(4) + log_v2km12, log_v2k2))
      log_dfs <- log_dk

    } else if (type == "Bakshaev") {

      # Sequence of indexes
      k <- 1:K_max

      # log(v_k^2)
      log_vk2 <- switch((p > 2) + 1, log(8) - log(pi) - log(4 * k^2 - 1),
                        (p - 3) * log(2) + log(2 * k + p - 2) +
                          lgamma(k - 1 / 2) + lgamma(p / 2 - 1) +
                          lgamma(p / 2) - log(pi) - lgamma(k + p - 1 / 2))

      # Divide by 2 if p = 2 and (1 + k / alpha) if p > 2
      if (p == 2) {

        log_vk2 <- log_vk2 - log(2)

      } else {

        log_vk2 <- log_vk2 - log(1 + k / alpha)

      }

      # log(d_{p, k})
      log_dk <- d_p_k(p = p, k = k, log = TRUE)

      # Log weights and dfs
      log_weights <- log_vk2
      log_dfs <- log_dk

    } else if (type == "Riesz") {

      # Sequence of indexes
      k <- 1:K_max

      if (Riesz_s == 0) {

        if (p == 2) {

          log_vk2 <- -log(k)

        } else if (p == 3) {

          log_vk2 <- log(1 + 2 * k) - log(2 * k) - log(k + 1)

        } else {

          log_vk2 <- log(2 * k + p - 2) + lgamma(k) + lgamma(p - 2) -
            log(2) - lgamma(k + p - 1)

        }

      } else if (Riesz_s == 2) {

        if (p == 2) {

          log_vk2 <- ifelse(k == 1, log(2), -Inf)

        } else {

          log_vk2 <- ifelse(k == 1, log(2) - log(p - 2), -Inf)

        }

      } else {

        # tau_{s, k, p}
        if (p == 2) {

          tau <- 2^(Riesz_s + 1) / (1 + (k == 0))

        } else {

          tau <- 2^(p + Riesz_s - 3) * (p + 2 * k - 2) * gamma((p - 2) / 2)

        }

        # log(v_k^2) = log(b_{s, k, p})
        log_vk2 <- log(tau / sqrt(pi)) + lgamma((p - 1 + Riesz_s) / 2) +
          lgamma(-Riesz_s / 2 + k) - lgamma(p - 1 + Riesz_s / 2 + k) -
          lgamma(-Riesz_s / 2)

      }

      # Divide by 2 if p = 2 and (1 + k / alpha) if p > 2
      if (p == 2) {

        log_vk2 <- log_vk2 - log(2)

      } else {

        log_vk2 <- log_vk2 - log(1 + k / alpha)

      }

      # log(d_{p, k})
      log_dk <- d_p_k(p = p, k = k, log = TRUE)

      # Log weights and dfs
      log_weights <- log_vk2
      log_dfs <- log_dk

    } else if (type == "PCvM") {

      # Sequence of indexes
      k <- 1:K_max

      # log(v_k^2)
      log_vk2 <- log(Gegen_coefs_Pn(k = k, p = p, type = "PCvM",
                                    Gauss = Gauss, N = N, tol = tol))

      # Divide by 2 if p = 2 and (1 + k / alpha) if p > 2
      if (p == 2) {

        log_vk2 <- log_vk2 - log(2)

      } else {

        log_vk2 <- log_vk2 - log(1 + k / alpha)

      }

      # log(d_{p, k})
      log_dk <- d_p_k(p = p, k = k, log = TRUE)

      # Log weights and dfs
      log_weights <- log_vk2
      log_dfs <- log_dk

    } else if (type == "PAD") {

      # Sequence of indexes
      k <- 1:K_max

      # log(v_k^2)
      log_vk2 <- log(Gegen_coefs_Pn(k = k, p = p, type = "PAD",
                                    Gauss = Gauss, N = N, tol = tol))

      # Divide by 2 if p = 2 and (1 + k / alpha) if p > 2
      if (p == 2) {

        log_vk2 <- log_vk2 - log(2)

      } else {

        log_vk2 <- log_vk2 - log(1 + k / alpha)

      }

      # log(d_{p, k})
      log_dk <- d_p_k(p = p, k = k, log = TRUE)

      # Log weights and dfs
      log_weights <- log_vk2
      log_dfs <- log_dk

    } else if (type == "PRt") {

      # Sequence of indexes
      k <- 1:K_max

      # log(v_k^2)
      log_vk2 <- log(Gegen_coefs_Pn(k = k, p = p, type = "PRt",
                                    Rothman_t = Rothman_t, Gauss = Gauss,
                                    N = N, tol = tol))

      # Divide by 2 if p = 2 and (1 + k / alpha) if p > 2
      if (p == 2) {

        log_vk2 <- log_vk2 - log(2)

      } else {

        log_vk2 <- log_vk2 - log(1 + k / alpha)

      }

      # log(d_{p, k})
      log_dk <- d_p_k(p = p, k = k, log = TRUE)

      # Log weights and dfs
      log_weights <- log_vk2
      log_dfs <- log_dk

    } else {

      stop("Incompatible choice of p and type.")

    }

  }

  # Set NaNs (coming from negative weights) to 0?
  if (force_positive) {

    log_weights[is.nan(log_weights)] <- -Inf

  }

  # Automatic truncation based on the tail probability mean of the sum of
  # weighted chi squared
  K_max <- length(log_weights)
  cutoff <- cutoff_wschisq(thre = thre, weights = log_weights,
                           dfs = log_dfs, log = TRUE, x_tail = x_tail)$prob[1]

  # Truncate displaying optional information
  log_weights <- log_weights[1:cutoff]
  log_dfs <- log_dfs[1:cutoff]
  if (verbose) {

    message("Series truncated from ", K_max, " to ", cutoff,
            " terms (difference <= ", thre, " with the HBE tail probability;",
            " last weight = ", sprintf("%.3e", exp(log_weights[cutoff])), ").")

  }

  # Return weights and dfs?
  if (!log) {

    log_weights <- exp(log_weights)
    log_dfs <- exp(log_dfs)

  }

  return(list("weights" = log_weights, "dfs" = log_dfs))

}


#' @rdname Sobolev
#' @export
d_Sobolev <- function(x, p, type, method = c("I", "SW", "HBE")[1], K_max = 1e3,
                      thre = 1e-3, Rothman_t = 1 / 3, Pycke_q = 0.5,
                      Riesz_s = 1, ncps = 0, verbose = TRUE, N = 320,
                      x_tail = NULL, ...) {

  weights_dfs <- weights_dfs_Sobolev(p = p, K_max = K_max, thre = thre,
                                     type = type, Rothman_t = Rothman_t,
                                     Pycke_q = Pycke_q, Riesz_s = Riesz_s,
                                     verbose = verbose, Gauss = TRUE, N = N,
                                     x_tail = x_tail)
  d_wschisq(x = x, weights = weights_dfs$weights, dfs = weights_dfs$dfs,
            ncps = ncps, method = method, ...)

}


#' @rdname Sobolev
#' @export
p_Sobolev <- function(x, p, type, method = c("I", "SW", "HBE", "MC")[1],
                      K_max = 1e3, thre = 1e-3, Rothman_t = 1 / 3,
                      Pycke_q = 0.5, Riesz_s = 1, ncps = 0, verbose = TRUE,
                      N = 320, x_tail = NULL, ...) {

  weights_dfs <- weights_dfs_Sobolev(p = p, K_max = K_max, thre = thre,
                                     type = type, Rothman_t = Rothman_t,
                                     Pycke_q = Pycke_q, Riesz_s = Riesz_s,
                                     verbose = verbose, Gauss = TRUE,
                                     N = N, x_tail = x_tail)
  p_wschisq(x = x, weights = weights_dfs$weights, dfs = weights_dfs$dfs,
            ncps = ncps, method = method, ...)

}


#' @rdname Sobolev
#' @export
q_Sobolev <- function(u, p, type, method = c("I", "SW", "HBE", "MC")[1],
                      K_max = 1e3, thre = 1e-3, Rothman_t = 1 / 3,
                      Pycke_q = 0.5, Riesz_s = 1, ncps = 0, verbose = TRUE,
                      N = 320, x_tail = NULL, ...) {

  weights_dfs <- weights_dfs_Sobolev(p = p, K_max = K_max, thre = thre,
                                     type = type, Rothman_t = Rothman_t,
                                     Pycke_q = Pycke_q, Riesz_s = Riesz_s,
                                     verbose = verbose, Gauss = TRUE, N = N,
                                     x_tail = x_tail)
  q_wschisq(u = u, weights = weights_dfs$weights, dfs = weights_dfs$dfs,
            ncps = ncps, method = method, ...)

}

K <- 8
k <- 1:K
psi_Ajne <- function(th) 1 / 2 - th / (2 * pi)
psi_PCvM <- function(th, q) psi_Pn(theta = th, q = q, type = "PCvM")
psi_PAD <- function(th, q) psi_Pn(theta = th, q = q, type = "PAD")
psi_PRt <- function(th, q) psi_Pn(theta = th, q = q, type = "PRt")
psi_Gine_Gn <- function(th, q) {
  -q / 4 * (gamma(q / 2) / gamma((q + 1) / 2))^2 * sin(th)
}
psi_Gine_Fn <- function(th, q) {
  psi_Gine_Gn(th = th, q = q) + 4 * psi_Ajne(th = th)
}
psi_Bakshaev <- function(th) -2 * sin(th / 2)
psi_Riesz <- function(th, s) {
  if (s == 0) {
    return(-0.5 * log(2 * (1 - cos(th))))
  } else {
    return(-sign(s) * 2^s * sin(th / 2)^s)
  }
}
psi_Watson <- function(th) 2 * (th^2 / (4 * pi^2) - th / (2 * pi))
psi_Rothman <- function(th, t = 1 / 3) {
  tm <- min(t, 1 - t)
  pmax(0, tm - th / (2 * pi)) - tm^2
}
psi_Hermans_Rasson <- function(th) {
  beta2 <- (pi^2 / 36) / (0.5 - 4 / pi^2)
  -(th + beta2 * sin(th))
}
psi_Pycke_q <- function(th, q = 0.5) {
  2 * (cos(th) - q) / (1 + q^2 - 2 * q * cos(th))
}
alpha <- c(0.10, 0.05, 0.01)
x <- c(0.1, 0.15, 0.2)
eps <- 1e-9
x_eps1 <- x + eps
x_eps2 <- x - eps

test_that("d_p_k", {

  expect_equal(d_p_k(p = 2, k = 0:3), c(1, rep(2, 3)))
  expect_equal(d_p_k(p = 2, k = 0:3, log = TRUE), c(0, rep(log(2), 3)))
  expect_equal(d_p_k(p = 30, k = 0:1), c(1, 30))

})

test_that("Gegen_coefs vs. weights_dfs_Sobolev for Ajne", {

  for (p in c(2, 3, 4, 11)) {
    expect_equal(Gegen_coefs(psi = psi_Ajne, k = k, p = p),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "Ajne", verbose = FALSE)$weights,
                 tolerance = 1e-6)
  }

})

test_that("Gegen_coefs vs. weights_dfs_Sobolev for PCvM", {

  for (p in c(2:4, 11)) {
    expect_equal(Gegen_coefs(psi = psi_PCvM, k = k, p = p, q = p - 1),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "PCvM", verbose = FALSE)$weights,
                 tolerance = 1e-6)
  }

})

test_that("Gegen_coefs vs. weights_dfs_Sobolev for PAD", {

  for (p in c(2:4, 11)) {
    expect_equal(Gegen_coefs(psi = psi_PAD, k = k, p = p, q = p - 1),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "PAD", verbose = FALSE)$weights,
                 tolerance = 1e-6)
  }

})

test_that("Gegen_coefs vs. weights_dfs_Sobolev for PRt", {

  for (p in c(2:4, 11)) {
    expect_equal(Gegen_coefs(psi = psi_PRt, k = k, p = p, q = p - 1),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "PRt", verbose = FALSE)$weights,
                 tolerance = 1e-5)
  }

})

test_that("Gegen_coefs vs. weights_dfs_Sobolev for Gine_Gn", {

  for (p in c(2:4, 11)) {
    expect_equal(Gegen_coefs(psi = psi_Gine_Gn, k = k, p = p, q = p - 1),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "Gine_Gn",
                                       verbose = FALSE)$weights,
                 tolerance = 1e-5)
  }

})

test_that("Gegen_coefs vs. weights_dfs_Sobolev for Gine_Fn", {

  for (p in c(2:4, 11)) {
    expect_equal(Gegen_coefs(psi = psi_Gine_Fn, k = k, p = p, q = p - 1),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "Gine_Fn",
                                       verbose = FALSE)$weights,
                 tolerance = 1e-5)
  }

})

test_that("Gegen_coefs vs. weights_dfs_Sobolev for Bakshaev", {

  for (p in c(2:4, 11)) {
    expect_equal(Gegen_coefs(psi = psi_Bakshaev, k = k, p = p),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "Bakshaev",
                                       verbose = FALSE)$weights,
                 tolerance = 1e-5)
  }

})

test_that("Gegen_coefs vs. weights_dfs_Sobolev for Riesz", {

  for (p in c(2:4, 11)) {
    expect_equal(Gegen_coefs(psi = psi_Riesz, k = k, p = p, s = 0,
                             N = 5120),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "Riesz", Riesz_s = 0,
                                       verbose = FALSE)$weights,
                 tolerance = 1e-5)
    expect_equal(Gegen_coefs(psi = psi_Riesz, k = k, p = p, s = 1),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "Riesz", Riesz_s = 1,
                                       verbose = FALSE)$weights,
                 tolerance = 1e-5)
    expect_equal(Gegen_coefs(psi = psi_Riesz, k = k, p = p, s = 2),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "Riesz", Riesz_s = 2,
                                       verbose = FALSE)$weights,
                 tolerance = 1e-5)
    expect_equal(Gegen_coefs(psi = psi_Riesz, k = k, p = p, s = 1.5),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "Riesz", Riesz_s = 1.5,
                                       verbose = FALSE)$weights,
                 tolerance = 1e-5)
    expect_equal(Gegen_coefs(psi = psi_Riesz, k = k, p = p, s = 0.5),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "Riesz", Riesz_s = 0.5,
                                       verbose = FALSE)$weights,
                 tolerance = 1e-5)
    expect_equal(Gegen_coefs(psi = psi_Riesz, k = k, p = p, s = 0.1,
                             N = 5120),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "Riesz", Riesz_s = 0.1,
                                       verbose = FALSE)$weights,
                 tolerance = 1e-5)
    expect_equal(Gegen_coefs(psi = psi_Riesz, k = k, p = p, s = -0.5,
                             N = 5120),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "Riesz", Riesz_s = -0.5,
                                       verbose = FALSE)$weights,
                 tolerance = 1e-3)
  }

})

test_that(paste("Gegen_coefs vs. weights_dfs_Sobolev for Watson, Rothman,",
                "Hermans_Rasson, Pycke_q"), {

  expect_equal(Gegen_coefs(psi = psi_Watson, k = k, p = 2),
               2 * weights_dfs_Sobolev(p = 2, K_max = K, thre = 0,
                                       type = "Watson",
                                       verbose = FALSE)$weights,
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_Rothman, k = k, p = 2, N = 640),
               2 * weights_dfs_Sobolev(p = 2, K_max = K, thre = 0,
                                       type = "Rothman",
                                       verbose = FALSE)$weights,
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_Hermans_Rasson, k = k, p = 2),
               2 * weights_dfs_Sobolev(p = 2, K_max = K %/% 2, thre = 0,
                                       type = "Hermans_Rasson",
                                       verbose = FALSE)$weights,
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_Pycke_q, k = k, p = 2),
               2 * weights_dfs_Sobolev(p = 2, K_max = K, thre = 0,
                                       type = "Pycke_q",
                                       verbose = FALSE)$weights,
               tolerance = 1e-6)
  expect_error(weights_dfs_Sobolev(p = 3, K_max = K, type = "Pycke_q"))

})

test_that("weights_dfs_Sobolev for Watson vs. PCvM with p = 2", {

  expect_equal(0.5 * weights_dfs_Sobolev(p = 2, K_max = K, thre = 0,
                                         type = "Watson",
                                         verbose = FALSE)$weights,
               weights_dfs_Sobolev(p = 2, K_max = K, thre = 0, type = "PCvM",
                                   verbose = FALSE)$weights,
               tolerance = 1e-6)

})

test_that("weights_dfs_Sobolev for Ajne vs. PRt with t = 1 / 2", {

  for (p in c(2, 3, 4, 11)) {
    expect_equal(weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                     type = "Ajne", verbose = FALSE)$weights,
                 weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                     type = "PRt", Rothman_t = 1 / 2,
                                     verbose = FALSE)$weights,
                 tolerance = 1e-6)
  }

})

test_that("Gegen_coefs_Pn vs. weights_dfs_Sobolev for PCvM", {

  for (p in c(2, 3, 4, 11)) {
    expect_equal(Gegen_coefs_Pn(k = k, p = p, type = "PCvM"),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "PCvM", verbose = FALSE)$weights)
  }

})

test_that("Gegen_coefs_Pn vs. weights_dfs_Sobolev for PAD", {

  for (p in c(2, 3, 4, 11)) {
    expect_equal(Gegen_coefs_Pn(k = k, p = p, type = "PAD"),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "PAD", verbose = FALSE)$weights)
  }

})

test_that("Gegen_coefs_Pn vs. weights_dfs_Sobolev for PRt", {

  for (p in c(2, 3, 4, 11)) {
    expect_equal(Gegen_coefs_Pn(k = k, p = p, type = "PRt"),
                 switch((p == 2) + 1, (1 + 2 * k / (p - 2)), 2) *
                   weights_dfs_Sobolev(p = p, K_max = K, thre = 0,
                                       type = "PRt", verbose = FALSE)$weights)
  }

})

test_that("p_Sobolev vs. q_Sobolev", {

  for (p in c(2:4, 11)) {
    expect_message(
      expect_equal(p_Sobolev(x = q_Sobolev(u = alpha, p = p, type = "Gine_Fn",
                                           thre = 0, method = "I"),
                             p = p, type = "Gine_Fn", verbose = FALSE,
                             thre = 0, method = "I"),
                   alpha, tolerance = 1e-4)
    )
    expect_message(
      expect_equal(q_Sobolev(u = p_Sobolev(x = x + 1, p = p, type = "Gine_Fn",
                                           thre = 0, method = "I"),
                             p = p, type = "Gine_Fn", verbose = FALSE,
                             thre = 0, method = "I"),
                   x + 1, tolerance = 1e-6)
    )
    expect_message(
      expect_equal(p_Sobolev(x = q_Sobolev(u = alpha, p = p, type = "PCvM",
                                           thre = 0, method = "SW"),
                             p = p, type = "PCvM", verbose = FALSE,
                             thre = 0, method = "SW"),
                   alpha, tolerance = 1e-6)
    )
    expect_message(
      expect_equal(q_Sobolev(u = p_Sobolev(x = x, p = p, type = "PCvM",
                                           thre = 0, method = "SW"),
                             p = p, type = "PCvM", verbose = FALSE,
                             thre = 0, method = "SW"),
                   x, tolerance = 1e-6)
    )
    expect_message(
      expect_equal(p_Sobolev(x = q_Sobolev(u = alpha, p = p, type = "PRt",
                                           thre = 0, method = "HBE"),
                             p = p, type = "PRt", verbose = FALSE,
                             thre = 0, method = "HBE"),
                   alpha, tolerance = 1e-6)
    )
    expect_message(
      expect_equal(q_Sobolev(u = p_Sobolev(x = x, p = p, type = "PRt",
                                           thre = 0, method = "HBE"),
                             p = p, type = "PRt", verbose = FALSE,
                             thre = 0, method = "HBE"),
                   x, tolerance = 1e-6)
    )
  }

})

test_that("p_Sobolev vs. d_Sobolev", {

  for (p in c(2:4, 11)) {
    expect_equal((p_Sobolev(x = x_eps1 + 1, p = p, type = "Gine_Fn",
                            verbose = FALSE, thre = 0, method = "I") -
                    p_Sobolev(x = x_eps2 + 1, p = p, type = "Gine_Fn",
                              verbose = FALSE, thre = 0, method = "I")) /
                   (2 * eps),
                   d_Sobolev(x = x + 1, p = p, type = "Gine_Fn",
                             verbose = FALSE, thre = 0, method = "I"),
                   tolerance = 1e-4)
    expect_equal((p_Sobolev(x = x_eps1, p = p, type = "PCvM", verbose = FALSE,
                            thre = 0, method = "SW") -
                    p_Sobolev(x = x_eps2, p = p, type = "PCvM", verbose = FALSE,
                              thre = 0, method = "SW")) / (2 * eps),
                 d_Sobolev(x = x, p = p, type = "PCvM", verbose = FALSE,
                           thre = 0, method = "SW"),
                 tolerance = 1e-4)
    expect_equal((p_Sobolev(x = x_eps1, p = p, type = "PRt", verbose = FALSE,
                            thre = 0, method = "HBE") -
                    p_Sobolev(x = x_eps2, p = p, type = "PRt", verbose = FALSE,
                              thre = 0, method = "HBE")) / (2 * eps),
                 d_Sobolev(x = x, p = p, type = "PRt", verbose = FALSE,
                           thre = 0, method = "HBE"),
                 tolerance = 1e-4)
  }

})

#' @title Circular and (hyper)spherical uniformity tests
#'
#' @description Implementation of several uniformity tests on the (hyper)sphere
#' \eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{
#' S^{p-1}:=\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}, with calibration either in
#' terms of their asymptotic/exact distributions, if available, or Monte Carlo.
#'
#' \code{unif_test} receives a sample of directions
#' \eqn{{\bf X}_1,\ldots,{\bf X}_n\in S^{p-1}}{X_1,\ldots,X_n\in S^{p-1}} in
#' \emph{Cartesian coordinates}, except for the circular case (\eqn{p=2}) in
#' which the sample can be represented in terms of \emph{angles}
#' \eqn{\Theta_1,\ldots,\Theta_n\in [0, 2\pi)}.
#'
#' \code{unif_test} allows to perform several tests within a single call,
#' facilitating thus the exploration of a dataset by applying several tests.
#'
#' @param data sample to perform the test. A matrix of size \code{c(n, p)}
#' containing a sample of size \code{n} of directions (in Cartesian
#' coordinates) on \eqn{S^{p-1}}. Alternatively if \code{p = 2}, a matrix of
#' size \code{c(n, 1)} containing the \code{n} angles on \eqn{[0, 2\pi)} of the
#' circular sample on \eqn{S^{1}}. Other objects accepted are an array of size
#' \code{c(n, p, 1)} with directions (in Cartesian coordinates), or a vector of
#' size \code{n} or an array of size \code{c(n, 1, 1)} with angular data.
#' Must not contain \code{NA}'s.
#' @param type type of test to be applied. A character vector containing any of
#' the following types of tests, depending on the dimension \eqn{p}:
#' \itemize{
#'   \item Circular data: any of the names available at object
#'   \code{\link{avail_cir_tests}}.
#'   \item (Hyper)spherical data: any of the names available at object
#'   \code{\link{avail_sph_tests}}.
#' }
#' If \code{type = "all"} (default), then \code{type} is set as
#' \code{avail_cir_tests} or \code{avail_sph_tests}, depending on the value of
#' \eqn{p}.
#' @param p_value type of \eqn{p}-value computation. Either \code{"MC"} for
#' employing the approximation by Monte Carlo of the exact null distribution,
#' \code{"asymp"} (default) for the use of the asymptotic/exact null
#' distribution (if available), or \code{"crit_val"} for approximation by means
#' of the table of critical values \code{crit_val}.
#' @param alpha vector with significance levels. Defaults to
#' \code{c(0.10, 0.05, 0.01)}.
#' @inheritParams unif_stat_distr
#' @inheritParams unif_stat_MC
#' @param crit_val table with critical values for the tests, to be used if
#' \code{p_value = "crit_val"}. A data frame, with column names containing the
#' character vector \code{type} and rows corresponding to the significance
#' levels \code{alpha}, that results from extracting \code{$crit_val_MC} from
#' a call to \code{\link{unif_stat_MC}}. Internally computed if
#' \code{NULL} (default).
#' @inheritParams unif_stat
#' @param ... If \code{p_value = "MC"} or \code{p_value = "crit_val"}, optional
#' performance parameters to be passed to \code{\link{unif_stat_MC}}:
#' \code{chunks}, \code{cores}, and \code{seed}.
#' @return If only a \bold{single test} is performed, a list with class
#' \code{htest} containing the following components:
#' \itemize{
#'   \item \code{statistic}: the value of the test statistic.
#'   \item \code{p.value}: the p-value of the test. If
#'   \code{p_value = "crit_val"}, an \code{NA}.
#'   \item \code{alternative}: a character string describing the alternative
#'   hypothesis.
#'   \item \code{method}: a character string indicating what type of test was
#'   performed.
#'   \item \code{data.name}: a character string giving the name of the data.
#'   \item \code{reject}: the rejection decision for the levels of significance
#'   \code{alpha}.
#'   \item \code{crit_val}: a vector with the critical values for the
#'   significance levels \code{alpha} used with \code{p_value = "MC"} or
#'   \code{p_value = "asymp"}.
#' }
#' If \bold{several tests} are performed, a \code{type}-named list with
#' entries for each test given by the above list.
#' @details
#' All the tests reject for large values of the test statistic, so the critical
#' values for the significance levels \code{alpha} correspond to the
#' \code{alpha}-upper quantiles of the null distribution of the test statistic.
#'
#' When \code{p_value = "asymp"}, tests that do not have an implemented or
#' known asymptotic are omitted, and a warning is generated.
#'
#' When \code{p_value = "MC"}, it is possible to have a progress bar indicating
#' the Monte Carlo simulation progress if \code{unif_test} is wrapped with
#' \code{\link[progressr:with_progress]{progressr::with_progress}} or if
#' \code{progressr::handlers(global = TRUE)} is invoked (once) by the user.
#' See the examples below. The progress bar is updated with the number of
#' finished chunks.
#'
#' All the statistics are continuous random variables except the
#' Hodges--Ajne statistic (\code{"Hodges_Ajne"}), the Cressie statistic
#' (\code{"Cressie"}), and the number of (different) uncovered spacings
#' (\code{"Num_uncover"}). These three statistics are discrete random variables.
#'
#' The Monte Carlo calibration for the CCF09 test is made conditionally
#' on the choice of \cr\code{CCF09_dirs}. That is, all the Monte
#' Carlo statistics share the same random directions.
#'
#' Descriptions and references for most of the tests are available
#' in García-Portugués and Verdebout (2018).
#' @references
#' García-Portugués, E. and Verdebout, T. (2018) An overview of uniformity
#' tests on the hypersphere. \emph{arXiv:1804.00286}.
#' \url{https://arxiv.org/abs/1804.00286}.
#' @examples
#' ## Asymptotic distribution
#'
#' # Circular data
#' n <- 10
#' samp_cir <- r_unif_cir(n = n)
#'
#' # Matrix
#' unif_test(data = samp_cir, type = "Ajne", p_value = "asymp")
#'
#' # Vector
#' unif_test(data = samp_cir[, 1], type = "Ajne", p_value = "asymp")
#'
#' # Array
#' unif_test(data = array(samp_cir, dim = c(n, 1, 1)), type = "Ajne",
#'           p_value = "asymp")
#' \donttest{
#' # Several tests
#' unif_test(data = samp_cir, type = avail_cir_tests, p_value = "asymp")
#' }
#' # Spherical data
#' n <- 10
#' samp_sph <- r_unif_sph(n = n, p = 3)
#'
#' # Array
#' unif_test(data = samp_sph, type = "Bingham", p_value = "asymp")
#'
#' # Matrix
#' unif_test(data = samp_sph[, , 1], type = "Bingham", p_value = "asymp")
#' \donttest{
#' # Several tests
#' unif_test(data = samp_sph, type = avail_sph_tests, p_value = "asymp")
#'
#' ## Monte Carlo
#'
#' # Circular data
#' unif_test(data = samp_cir, type = "Ajne", p_value = "MC")
#' unif_test(data = samp_cir, type = avail_cir_tests, p_value = "MC")
#'
#' # Spherical data
#' unif_test(data = samp_sph, type = "Bingham", p_value = "MC")
#' unif_test(data = samp_sph, type = avail_sph_tests, p_value = "MC")
#'
#' # Caching stats_MC
#' stats_MC_cir <- unif_stat_MC(n = nrow(samp_cir), p = 2)$stats_MC
#' stats_MC_sph <- unif_stat_MC(n = nrow(samp_sph), p = 3)$stats_MC
#' unif_test(data = samp_cir, type = avail_cir_tests,
#'           p_value = "MC", stats_MC = stats_MC_cir)
#' unif_test(data = samp_sph, type = avail_sph_tests, p_value = "MC",
#'           stats_MC = stats_MC_sph)
#'
#' ## Critical values
#'
#' # Circular data
#' unif_test(data = samp_cir, type = avail_cir_tests, p_value = "crit_val")
#'
#' # Spherical data
#' unif_test(data = samp_sph, type = avail_sph_tests, p_value = "crit_val")
#'
#' # Caching crit_val
#' crit_val_cir <- unif_stat_MC(n = n, p = 2)$crit_val_MC
#' crit_val_sph <- unif_stat_MC(n = n, p = 3)$crit_val_MC
#' unif_test(data = samp_cir, type = avail_cir_tests,
#'           p_value = "crit_val", crit_val = crit_val_cir)
#' unif_test(data = samp_sph, type = avail_sph_tests, p_value = "crit_val",
#'           crit_val = crit_val_sph)
#'
#' ## Specific arguments
#'
#' # Rothman
#' unif_test(data = samp_cir, type = "Rothman", Rothman_t = 0.5)
#'
#' # CCF09
#' unif_test(data = samp_sph, type = "CCF09", p_value = "MC",
#'           CCF09_dirs = samp_sph[1:2, , 1])
#' unif_test(data = samp_sph, type = "CCF09", p_value = "MC",
#'           CCF09_dirs = samp_sph[3:4, , 1])
#'
#' ## Using a progress bar when p_value = "MC"
#'
#' # Define a progress bar
#' require(progress)
#' require(progressr)
#' handlers(handler_progress(
#'   format = ":spin [:bar] :percent Total: :elapsedfull End \u2248 :eta",
#'   clear = FALSE))
#'
#' # Call unif_test() within with_progress()
#' with_progress(
#'   unif_test(data = samp_sph, type = avail_sph_tests, p_value = "MC",
#'             chunks = 10, M = 1e3)
#' )
#'
#' # With several cores
#' with_progress(
#'   unif_test(data = samp_sph, type = avail_sph_tests, p_value = "MC",
#'             cores = 2, chunks = 10, M = 1e3)
#' )
#'
#' # Instead of using with_progress() each time, it is more practical to run
#' # handlers(global = TRUE)
#' # once to activate progress bars in your R session
#' }
#' @export
unif_test <- function(data, type = "all", p_value = "asymp",
                      alpha = c(0.10, 0.05, 0.01), M = 1e4, stats_MC = NULL,
                      crit_val = NULL, data_sorted = FALSE, Rayleigh_m = 1,
                      cov_a = 2 * pi, Rothman_t = 1 / 3, Cressie_t = 1 / 3,
                      Pycke_q = 0.5, Riesz_s = 1, CCF09_dirs = NULL,
                      K_CCF09 = 25, CJ12_reg = 3, CJ12_beta = 0, K_max = 1e4,
                      ...) {

  # Read data's name
  data_name <- deparse(substitute(data))

  # Stop if NA's
  if (anyNA(data)) {

    stop("NAs present in data, please remove them.")

  }

  # If data is a vector, transform it to matrix
  if (is.vector(data)) {

    data <- matrix(data, ncol = 1)

  }

  # If data is an array, transform it to matrix
  d <- dim(data)
  l <- length(d)
  if (l == 3) {

    # As matrix
    data <- matrix(data[, , 1], nrow = d[1], ncol = d[2])

    # First slice only
    if (d[3] != 1) {

      message(paste("data is an array with more than one slice,",
                    "only the first one is employed."))

    }

  } else if (l > 3) {

    stop("data must be a vector, matrix, or a 3-dimensional array.")

  }

  # Sample size and dimension
  n <- nrow(data)
  d <- ncol(data)

  # Circular or spherical data?
  if (d == 1 | d == 2) {

    avail_stats <- avail_cir_tests
    prefix_stat <- "cir_stat_"
    p <- 2

    # As polar coordinates
    if (d == 2) {

      dim(data) <- c(n, d, 1)
      data <- X_to_Theta(X = data)

    }

  } else {

    avail_stats <- avail_sph_tests
    prefix_stat <- "sph_stat_"
    p <- d
    dim(data) <- c(n, d, 1) # As an array

  }

  # Get the type of statistics
  if (is.character(type)) {

    type <- gsub("-", "_", type)
    type <- unique(type)
    if ("all" %in% type) {

      stats_type <- avail_stats

    } else {

      stats_type <- match.arg(arg = type, choices = avail_stats,
                              several.ok = TRUE)

    }

  } else if (is.numeric(type)) {

    type <- unique(type)
    stats_type <- avail_stats[type]

  } else {

    stop("type must be a character or a numeric vector.")

  }

  # Omit statistics that do not have asymptotic distribution
  if (p_value == "asymp") {

    ind_asymp <- sapply(paste0("p_", prefix_stat, stats_type),
                        exists, where = asNamespace("sphunif"))
    if (!all(ind_asymp)) {

      warning(paste0(paste("Omitting the following statistics with not",
                           "implemented or unknown asymptotic distributions: "),
                     paste(stats_type[!ind_asymp], collapse = ", "), "."))
      stats_type <- stats_type[ind_asymp]
      if (length(stats_type) == 0) {

        stop("No remaining statistics to use.")

      }

    }

  }

  # Number of statistics
  n_stats <- length(stats_type)

  ## Statistics

  # Sample random directions for the CCF09 test
  if ("CCF09" %in% stats_type & is.null(CCF09_dirs)) {

    CCF09_dirs <- r_unif_sph(n = 50, p = p, M = 1)[, , 1]

  }

  # Compute statistics
  stat <- unif_stat(data = data, type = stats_type, data_sorted = data_sorted,
                    Rayleigh_m = Rayleigh_m, cov_a = cov_a,
                    Rothman_t = Rothman_t, Cressie_t = Cressie_t,
                    Pycke_q = Pycke_q, Riesz_s = Riesz_s,
                    CCF09_dirs = CCF09_dirs, K_CCF09 = K_CCF09,
                    CJ12_reg = CJ12_reg)

  ## Calibration

  if (p_value == "crit_val") {

    # Get the critical values
    if (is.null(crit_val)) {

      crit_val <- unif_stat_MC(n = n, type = stats_type, p = p, M = M,
                               r_H1 = NULL, crit_val = NULL, alpha = alpha,
                               return_stats = FALSE, stats_sorted = FALSE,
                               Rayleigh_m = Rayleigh_m, cov_a = cov_a,
                               Rothman_t = Rothman_t, Cressie_t = Cressie_t,
                               Pycke_q = Pycke_q, Riesz_s = Riesz_s,
                               CCF09_dirs = CCF09_dirs, K_CCF09 = K_CCF09,
                               CJ12_reg = CJ12_reg, ...)$crit_val_MC

    } else {

      # Check if there is any missing statistic in crit_val
      names_crit_val <- names(crit_val)
      missing_crit_val <- !(stats_type %in% names_crit_val)
      if (any(missing_crit_val)) {

        stop(paste0("Missing statistics in crit_val: \"",
                    paste(stats_type[missing_crit_val],
                          collapse = "\", \""), "\"."))

      }

      # Match columns of crit_val with stats_type
      crit_val <- crit_val[, pmatch(x = stats_type, table = names_crit_val),
                           drop = FALSE]

    }

    # Rejection?
    reject <- rbind(apply(crit_val, 1, function(x) x > stat))

    # p-values
    p_val <- matrix(NA, nrow = 1, ncol = n_stats)

  } else if (p_value == "MC") {

    # Get the stats_MC
    if (is.null(stats_MC)) {

      stats_MC <- unif_stat_MC(n = n, type = stats_type, p = p, M = M,
                               r_H1 = NULL, crit_val = NULL, alpha = alpha,
                               return_stats = TRUE, stats_sorted = TRUE,
                               Rothman_t = Rothman_t, Pycke_q = Pycke_q,
                               Riesz_s = Riesz_s, CCF09_dirs = CCF09_dirs,
                               CJ12_reg = CJ12_reg, K_CCF09 = K_CCF09,
                               ...)$stats_MC

    }

    # p-values
    p_val <- 1 - as.data.frame(sapply(stats_type, function(distr) {
      ecdf_bin(data = stats_MC[[distr]], sorted_x = stat[[distr]],
               data_sorted = TRUE, efic = TRUE, divide_n = TRUE)
    }, simplify = FALSE))

    # Critical values
    crit_val <- apply(stats_MC, 2, quantile, probs = 1 - alpha, na.rm = TRUE)
    rownames(crit_val) <- alpha

    # Rejection?
    reject <- rbind(apply(crit_val, 1, function(x) stat >= x))

  } else if (p_value == "asymp") {

    # p-values
    p_val <- 1 - unif_stat_distr(x = stat, type = stats_type, p = p, n = n,
                                 approx = "asymp", stats_MC = NULL, M = M,
                                 cov_a = cov_a, Rothman_t = Rothman_t,
                                 Cressie_t = Cressie_t, Pycke_q = Pycke_q,
                                 Riesz_s = Riesz_s, CCF09_dirs = CCF09_dirs,
                                 K_CCF09 = K_CCF09, CJ12_reg = CJ12_reg,
                                 CJ12_beta = CJ12_beta, K_max = K_max,
                                 thre = 0, ...)

    # Critical values
    crit_val <- as.data.frame(matrix(NA, nrow = length(alpha), ncol = n_stats))
    rownames(crit_val) <- alpha

    # Rejection?
    reject <- rbind(sapply(alpha, function(a) p_val < a))
    colnames(reject) <- alpha

  } else {

    stop(paste("Wrong choice for calibration, must be \"MC\", \"asymp\",",
               "or \"crit_val\"."))

  }

  ## Prepare return object

  # Create list of htest objects
  test <- vector(mode = "list", length = n_stats)
  names(test) <- stats_type
  for (i in seq_along(stats_type)) {

    # Type of test
    if (p == 2) {

      method <- switch(stats_type[i],
         "Ajne" = "Ajne test of circular uniformity",
         "Bakshaev" = "Bakshaev (2010) test of circular uniformity",
         "Bingham" = "Bingham test of circular uniformity",
         "CCF09" = paste("Cuesta-Albertos et al. (2009) test of circular",
                         "uniformity with k =", nrow(CCF09_dirs)),
         "Cressie" = paste("Cressie test of circular uniformity with t =",
                           round(Cressie_t, 3)),
         "FG01" = paste("Cramer-von Mises 4-point test of Feltz and",
                        "Goldin (2001)"),
         "Gine_Fn" = "Gine's Fn test of circular uniformity",
         "Gine_Gn" = "Gine's Gn test of circular uniformity",
         "Gini" = "Gini mean difference test of circular uniformity",
         "Gini_squared" = paste("Gini squared mean difference test of",
                                "circular uniformity"),
         "Greenwood" = "Greenwood spacings test of circular uniformity",
         "Hermans_Rasson" = "Hermans-Rasson test of circular uniformity",
         "Hodges_Ajne" = "Hodges-Ajne test of circular uniformity",
         "Kuiper" = "Kuiper test of circular uniformity",
         "Log_gaps" = paste("Darling\'s log-gaps spacings test of",
                            "circular uniformity"),
         "Max_uncover" = paste("Maximum uncovered spacing test of",
                               "circular uniformity with a =",
                               round(cov_a, 3)),
         "Num_uncover" = paste("Number of uncovered spacings test of",
                               "circular uniformity with a =",
                               round(cov_a, 3)),
         "PAD" = paste("Projected Anderson-Darling test of",
                       "circular uniformity"),
         "PCvM" = paste("Projected Cramer-von Mises test of",
                        "circular uniformity"),
         "PRt" = paste("Projected Rothman test of circular uniformity",
                       "with t =", round(Rothman_t, 3)),
         "Pycke" = "Pycke test of circular uniformity",
         "Pycke_q" = paste("Pycke \"q-test\" of spherical uniformity with q =",
                           round(Pycke_q, 3)),
         "Range" = "Range test of circular uniformity",
         "Rao" = "Rao\'s spacings test of circular uniformity",
         "Rayleigh" = paste0("Rayleigh test of circular uniformity",
                             ifelse(Rayleigh_m > 1, paste0(" with m = ",
                                                           Rayleigh_m), "")),
         "Riesz" = "Warning! This is an experimental test not meant to be used",
         "Rothman" = paste("Rothman test of circular uniformity with t =",
                           round(Rothman_t, 3)),
         "Vacancy" = paste("Vacancy test of circular uniformity with a =",
                           round(cov_a, 3)),
         "Watson" = "Watson test of circular uniformity",
         "Watson_1976" = "Watson (1976) test of circular uniformity"
      )

      alternative <- switch(stats_type[i],
         "Ajne" = "any non-axial alternative to circular uniformity",
         "Bakshaev" = "any alternative to circular uniformity",
         "Bingham" = "scatter matrix different from constant",
         "Cressie" = paste("any alternative to circular uniformity",
                           "if t is irrational (conjectured)"),
         "CCF09" = "any alternative to circular uniformity",
         "FG01" = "any alternative to circular uniformity",
         "Gine_Fn" = "any alternative to circular uniformity",
         "Gine_Gn" = "any axial alternative to circular uniformity",
         "Gini" = "any alternative to circular uniformity",
         "Gini_squared" = "any alternative to circular uniformity",
         "Greenwood" = "any alternative to circular uniformity",
         "Hermans_Rasson" = "any alternative to circular uniformity",
         "Hodges_Ajne" = "any non-axial alternative to circular uniformity",
         "Kuiper" = "any alternative to circular uniformity",
         "Log_gaps" = "any alternative to circular uniformity",
         "Max_uncover" = "any alternative to circular uniformity",
         "Num_uncover" = paste("any alternative to circular uniformity",
                               "if a <= 2 * pi"),
         "PAD" = "any alternative to circular uniformity",
         "PCvM" = "any alternative to circular uniformity",
         "PRt" = paste("any alternative to circular uniformity",
                       "if t is irrational"),
         "Pycke" = "any alternative to circular uniformity",
         "Pycke_q" = "any alternative to circular uniformity",
         "Range" = "any alternative to circular uniformity",
         "Rao" = "any alternative to circular uniformity",
         "Rayleigh" = "mean direction different from zero",
         "Rothman" = paste("any alternative to circular uniformity",
                           "if t is irrational"),
         "Riesz" = "unclear, experimental test",
         "Vacancy" = "any alternative to circular uniformity",
         "Watson" = "any alternative to circular uniformity",
         "Watson_1976" = "unclear consistency"
      )

    } else {

      method <- switch(stats_type[i],
         "Ajne" = "Ajne test of spherical uniformity",
         "Bakshaev" = "Bakshaev (2010) test of spherical uniformity",
         "Bingham" = "Bingham test of spherical uniformity",
         "CJ12" = "Cai and Jiang (2012) test of spherical uniformity",
         "CCF09" = paste("Cuesta-Albertos et al. (2009) test of spherical",
                         "uniformity with k =", nrow(CCF09_dirs)),
         "Gine_Fn" = "Gine's Fn test of spherical uniformity",
         "Gine_Gn" = "Gine's Gn test of spherical uniformity",
         "PAD" = paste("Projected Anderson-Darling test of",
                        "spherical uniformity"),
         "PCvM" = paste("Projected Cramer-von Mises test of",
                        "spherical uniformity"),
         "PRt" = paste("Projected Rothman test of spherical uniformity",
                       "with t =", round(Rothman_t, 3)),
         "Pycke" = "Pycke test of spherical uniformity",
         "Rayleigh" = "Rayleigh test of spherical uniformity",
         "Rayleigh_HD" = paste("HD-standardized Rayleigh test of",
                               "spherical uniformity"),
         "Riesz" = "Warning! This is an experimental test not meant to be used"
      )

      alternative <- switch(stats_type[i],
         "Ajne" = "any non-axial alternative to spherical uniformity",
         "Bakshaev" = "any alternative to spherical uniformity",
         "Bingham" = "scatter matrix different from constant",
         "CJ12" = "unclear consistency",
         "CCF09" = "any alternative to spherical uniformity",
         "Gine_Fn" = "any alternative to spherical uniformity",
         "Gine_Gn" = "any axial alternative to spherical uniformity",
         "PCvM" = "any alternative to spherical uniformity",
         "PAD" = "any alternative to spherical uniformity",
         "PRt" = paste("any alternative to spherical uniformity",
                       "if t is irrational"),
         "Pycke" = "any alternative to spherical uniformity",
         "Rayleigh" = "mean direction different from zero",
         "Rayleigh_HD" = "mean direction different from zero",
         "Riesz" = "unclear, experimental test"
      )

    }

    # htest object
    test[[i]] <- list(statistic = c("statistic" = stat[1, i]),
                      p.value = p_val[1, i], alternative = alternative,
                      method = method, data.name = data_name,
                      reject = reject[i, ], crit_val = crit_val[, i])
    class(test[[i]]) <- "htest"

  }

  # If there is only one test, return htest directly
  if (n_stats == 1) {

    test <- test[[1]]

  }
  return(test)

}


#' @title Available circular and (hyper)spherical uniformity tests
#'
#' @description Listing of the tests implemented in the \code{\link{sphunif}}
#' package.
#'
#' @return A character vector whose elements are valid inputs for the
#' \code{type} argument in \code{\link{unif_test}}, \code{\link{unif_stat}},
#' \code{\link{unif_stat_distr}}, and \code{\link{unif_stat_MC}}.
#' \code{avail_cir_tests} provides the available circular tests and
#' \code{avail_sph_tests} the (hyper)spherical tests.
#' @examples
#' # Circular tests
#' avail_cir_tests
#'
#' # Spherical tests
#' avail_sph_tests
#' @name avail_tests


#' @rdname avail_tests
#' @export
avail_cir_tests <- c("Ajne",
                     "Bakshaev",
                     "Bingham",
                     "Cressie",
                     "CCF09",
                     "FG01",
                     "Gine_Fn",
                     "Gine_Gn",
                     "Gini",
                     "Gini_squared",
                     "Greenwood",
                     "Hermans_Rasson",
                     "Hodges_Ajne",
                     "Kuiper",
                     "Log_gaps",
                     "Max_uncover",
                     "Num_uncover",
                     "PAD",
                     "PCvM",
                     "PRt",
                     "Pycke",
                     "Pycke_q",
                     "Range",
                     "Rao",
                     "Rayleigh",
                     "Riesz",
                     "Rothman",
                     "Vacancy",
                     "Watson",
                     "Watson_1976")


#' @rdname avail_tests
#' @export
avail_sph_tests <- c("Ajne",
                     "Bakshaev",
                     "Bingham",
                     "CJ12",
                     "CCF09",
                     "Gine_Fn",
                     "Gine_Gn",
                     "PAD",
                     "PCvM",
                     "PRt",
                     "Pycke",
                     "Rayleigh",
                     "Rayleigh_HD",
                     "Riesz")

#' @title Monte Carlo integration of functions on the (hyper)sphere
#'
#' @description Monte Carlo approximation of the integral
#' \deqn{\int_{S^{p-1}}f(x)\,\mathrm{d}x}{\int_{S^{p-1}} f(x) dx}
#' of a function \eqn{f:S^{p-1} \rightarrow R} defined on the (hyper)sphere
#' \eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{
#' S^{p-1}:=\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}.
#'
#' @param f function to be integrated. Its first argument must be the
#' (hyper)sphere position. Must be vectorized and return a vector of size
#' \code{nrow(x)} for a matrix input \code{x}. See examples.
#' @inheritParams r_unif
#' @param M number of Monte Carlo samples. Defaults to \code{1e4}.
#' @param chunks number of chunks to split the \code{M} Monte Carlo
#' samples. Useful for parallelizing the integration in \code{chunks}
#' tasks containing \code{ceiling(M / chunks)} replications. Useful also for
#' avoiding memory bottlenecks when \code{M} is large. Defaults to
#' \cr\code{ceiling(M / 1e3)}.
#' @param cores number of cores to perform the integration. Defaults to
#' \code{1}.
#' @inheritParams unif_stat_MC
#' @param ... optional arguments to be passed to \code{f} or to
#' \code{\link[foreach]{foreach}} (for example, \code{.export} to export global
#' variables or other functions to the \code{foreach} environment).
#' @return A scalar with the approximate integral.
#' @details
#' It is possible to have a progress bar if \code{int_sph_MC} is wrapped with
#' \code{\link[progressr:with_progress]{progressr::with_progress}} or if
#' \code{progressr::handlers(global = TRUE)} is invoked (once) by the user.
#' See the examples below. The progress bar is updated with the number of
#' finished chunks.
#' @examples
#' ## Sequential simulation
#'
#' # Vectorized functions to be integrated
#' x1 <- function(x) x[, 1]
#' quad <- function(x, a = 0) a + rowSums(x^4)
#'
#' # Approximate \int_{S^{p-1}} x_1 dx = 0
#' int_sph_MC(f = x1, p = 3, M = 1e4, chunks = 2)
#'
#' # Approximate \int_{S^{p-1}} (a + \sum_i x_i^4) dx
#' int_sph_MC(f = quad, p = 2, M = 1e4, a = 0, chunks = 2)
#'
#' # Compare with Gauss--Legendre integration on S^2
#' th_k <- Gauss_Legen_nodes(a = 0, b = 2 * pi, N = 40)
#' w_k <- Gauss_Legen_weights(a = 0, b = 2 * pi, N = 40)
#' sum(w_k * quad(cbind(cos(th_k), sin(th_k)), a = 1))
#'
#' ## Parallel simulation with a progress bar
#' \donttest{
#' # Define a progress bar
#' require(progress)
#' require(progressr)
#' handlers(handler_progress(
#'   format = ":spin [:bar] :percent Total: :elapsedfull End \u2248 :eta",
#'   clear = FALSE))
#'
#' # Call int_sph_MC() within with_progress()
#' with_progress(int_sph_MC(f = x1, p = 3, cores = 2, M = 1e5, chunks = 100))
#'
#' # Instead of using with_progress() each time, it is more practical to run
#' # handlers(global = TRUE)
#' # once to activate progress bars in your R session
#' }
#' @export
int_sph_MC <- function(f, p, M = 1e4, cores = 1, chunks = ceiling(M / 1e3),
                       seeds = NULL, ...) {

  # Chunk large n * M to avoid memory issues
  small_M <- M %/% chunks

  # Extra arguments for foreach::foreach and f
  dots <- list(...)
  foreach_args <- dots[names(dots) %in% names(formals(foreach::foreach))]
  f_args <- dots[names(dots) %in% names(formals(f))]
  name_arg1_f <- names(formals(f))[1]
  names_args_f <- names(f_args)

  # Check seeds
  if (!is.null(seeds) & length(seeds) != chunks) {

    warning(paste("seeds and chunks do not have the same length:",
                  "seeds are ignored."))
    seeds <- NULL

  }

  # Parallel backend
  old_dopar <- doFuture::registerDoFuture()
  old_plan <- future::plan(future::multisession(), workers = cores)
  options(future.rng.onMisuse = "ignore")
  on.exit({

    with(old_dopar, foreach::setDoPar(fun = fun, data = data, info = info))
    future::plan(old_plan)
    options(future.rng.onMisuse = NULL)

  })
  `%op%` <- foreach::`%dopar%`

  # Measure progress?
  if (requireNamespace("progressr", quietly = TRUE)) {

    prog <- progressr::progressor(along = 1:chunks)

  }

  # Monte Carlo
  k <- 0
  int <- do.call(what = foreach::foreach,
                 args = c(list(k = 1:chunks, .combine = "sum",
                               .inorder = FALSE, .multicombine = TRUE,
                               .packages = "sphunif"),
                          foreach_args)) %op% {

    # Sample uniform data
    if (!is.null(seeds)) {

      set.seed(seeds[k], kind = "Mersenne-Twister")

    }
    X <- rbind(r_unif_sph(n = small_M, p = p, M = 1)[, , 1])

    # Evaluate f
    args <- c(list(X), f_args)
    names(args) <- c(name_arg1_f, names_args_f)
    sf <- sum(do.call(what = f, args = args))

    # Remove X and clean memory
    rm(X)
    gc()

    # Signal progress
    if (requireNamespace("progressr", quietly = TRUE)) {

      prog()

    }

    # Return sum
    sf

  }

  # Return integral approximation
  int <- rotasym::w_p(p = p) * int / M
  return(int)

}

# To prevent hanging with parallel computations
Sys.unsetenv("R_TESTS")

one <- function(x) rep(1, nrow(x))
id <- function(x) x[, 1]
quad <- function(x) rowSums(x^2)
cubic <- function(x, a = 0) a + x[, 1] * x[, 2] * x[, 3]
time_consuming <- function(x) {Sys.sleep(0.5); 1}
seeds <- 1:5

test_that("Integration of one", {

  skip_on_cran()
  for (p in c(2, 3, 4, 11)) {
    expect_equal(int_sph_MC(f = one, p = p, M = 10),
                 rotasym::w_p(p = p))
  }

})

test_that("Integration of id", {

  skip_on_cran()
  for (p in c(2, 3, 4, 11)) {
    expect_equal(int_sph_MC(f = id, p = p, M = 1e3, seeds = 1:5, chunks = 5), 0,
                 tolerance = 3e-1)
  }

})

test_that("Integration of quad", {

  skip_on_cran()
  for (p in c(2, 3, 4, 11)) {
    expect_equal(int_sph_MC(f = quad, p = p, M = 10, seeds = 1:5, chunks = 5),
                 rotasym::w_p(p = p))
  }

})

test_that("Integration of cubic", {

  skip_on_cran()
  for (p in c(3, 4, 11)) {
    expect_equal(int_sph_MC(f = cubic, p = p, M = 1e3, seeds = 1:5, chunks = 5),
                 0, tolerance = 1e-1)
  }

})

test_that("Optional argument to f", {

  skip_on_cran()
  expect_equal(int_sph_MC(f = cubic, p = 5, M = 10, seeds = seeds, chunks = 5),
               int_sph_MC(f = cubic, p = 5, M = 10, a = 0, seeds = seeds,
                          chunks = 5))
  expect_equal(int_sph_MC(f = cubic, p = 4, M = 10, a = -5, seeds = seeds,
                          chunks = 5),
               -5 * rotasym::w_p(p = 4) +
                 int_sph_MC(f = cubic, p = 4, M = 10, a = 0, seeds = seeds,
                            chunks = 5))

})

test_that("Edge cases", {

  skip_on_cran()
  expect_warning(int_sph_MC(f = cubic, p = 3, M = 10, seeds = seeds[1:2],
                            chunks = 5))

})

test_that("Progress bars", {

  skip_on_cran()
  o1_silent <- capture.output(i <- int_sph_MC(f = cubic, p = 3, M = 10,
                                              cores = 1, chunks = 10))
  o2_silent <- capture.output(i <- int_sph_MC(f = cubic, p = 3, M = 10,
                                              cores = 2, chunks = 10))
  expect_equal(length(o1_silent), 0)
  expect_equal(length(o2_silent), 0)

})

test_that("Same results with cores = 1 and cores = 2", {

  skip_on_cran()
  expect_equal(int_sph_MC(f = cubic, p = 3, M = 10, cores = 1, seeds = seeds,
                          chunks = 5),
               int_sph_MC(f = cubic, p = 3, M = 10, cores = 2, seeds = seeds,
                          chunks = 5))

})

test_that("Parallelization is faster", {

  skip_on_ci()
  skip_on_cran()
  t1 <- system.time(int_sph_MC(f = time_consuming, p = 3, M = 1e5, cores = 1,
                               chunks = 10))[3]
  t2 <- system.time(int_sph_MC(f = time_consuming, p = 3, M = 1e5, cores = 2,
                               chunks = 10))[3]
  expect_gt(t1, t2)

})

#' @title Comet orbits
#'
#' @description Comet orbits data from the
#' \href{https://ssd.jpl.nasa.gov/sbdb_query.cgi}{
#' JPL Small-Body Database Search Engine}. The normal vector of a comet orbit
#' represents is a vector on \eqn{S^2}.
#'
#' @docType data
#' @format A data frame with 3633 rows and 8 variables:
#' \describe{
#'   \item{id}{database ID.}
#'   \item{full_name}{full name/designation following the
#'   \href{https://www.iau.org/public/themes/naming/#comets}{
#'   IUA naming convention}.}
#'   \item{i}{inclination; the orbit angle with respect to the ecliptic plane,
#'   in radians in \eqn{[0, \pi]}.}
#'   \item{om}{longitude of the ascending node; the angle between the normal
#'   vector of the orbit and the normal vector of the ecliptic plane, in
#'   radians in \eqn{[0, 2\pi)}.}
#'   \item{per_y}{sidereal orbital period (in years).}
#'   \item{class}{orbit classification. A factor with levels given below.}
#'   \item{diameter}{diameter from equivalent sphere (in km).}
#'   \item{ccf2009}{flag indicating if the comet was considered in the data
#'   application in Cuesta-Albertos et al. (2009); see details below.}
#' }
#' @details
#' The normal vector to the ecliptic plane of the comet with inclination
#' \eqn{i} and longitude of the ascending node \eqn{\omega} is
#' \deqn{(\sin(i) \sin(\omega), -\sin(i) \cos(\omega), \cos(i))'.}{
#' (sin(i) sin(\omega), -sin(i) cos(\omega), cos(i))'.}
#'
#' A prograde comet has positive \eqn{\cos(i)}{cos(i)}, negative
#' \eqn{\cos(i)}{cos(i)} represents a retrograde comet.
#'
#' \code{class} has the following levels:
#' \itemize{
#'  \item \code{COM}: comet orbit not matching any defined orbit class.
#'  \item \code{CTc}: Chiron-type comet, as defined by Levison and Duncan
#'  (T_Jupiter > 3; a > a_Jupiter).
#'  \item \code{ETc}: Encke-type comet, as defined by Levison and Duncan
#'  (T_Jupiter > 3; a < a_Jupiter).
#'  \item \code{HTC}: Halley-type comet, classical definition (20y < P < 200y).
#'  \item \code{HYP}: comets on hyperbolic orbits.
#'  \item \code{JFc}: Jupiter-family comet, as defined by Levison and Duncan
#'  (2 < T_Jupiter < 3).
#'  \item \code{JFC}: Jupiter-family comet, classical definition (P < 20y).
#'  \item \code{PAR}: comets on parabolic orbits.
#'}
#' Hyperbolic and parabolic comets are not periodic; only elliptical comets
#' are periodic.
#'
#' The \code{ccf2009} variable gives the observations considered in
#' Cuesta-Albertos et al. (2009) after fetching in the database in 2007-12-14
#' for the comets such that
#' \code{!(class \%in\% c("HYP", "PAR")) & per_y >= 200 & !numbered}. A
#' periodic comet is \code{numbered} by the IUA only after its second
#' perihelion passage, and then its \code{id} starts with \code{c}. Due to the
#' dynamic nature of the data, more comets were added to the database since
#' 2007 and also some observations were updated.
#'
#' The script performing the data preprocessing is available at
#' \href{https://github.com/egarpor/sphunif/blob/master/data-raw/comets.R}{
#' \code{comets.R}}. The data was retrieved on 2020-05-07.
#' @source \url{https://ssd.jpl.nasa.gov/sbdb_query.cgi}
#' @references
#' Cuesta-Albertos, J. A., Cuevas, A., Fraiman, R. (2009) On projection-based
#' tests for directional and compositional data. \emph{Statistics and
#' Computing}, 19:367--380. \doi{10.1007/s11222-008-9098-3}
#' @examples
#' # Load data
#' data("comets")
#'
#' # Add normal vectors
#' comets$normal <- cbind(sin(comets$i) * sin(comets$om),
#'                        -sin(comets$i) * cos(comets$om),
#'                        cos(comets$i))
#'
#' # Add numbered information
#' comets$numbered <- substr(comets$id, 1, 1) == "c"
#'
#' # Tests to be performed
#' type_tests <- c("PCvM", "PAD", "PRt")
#'
#' # Excluding the C/1882 R1-X (Great September comet) records with X = B, C, D
#' comets_ccf2009 <- comets[comets$ccf2009, ][-c(13:15), ]
#'
#' # Sample size
#' nrow(comets_ccf2009)
#'
#' # Tests for the data in Cuesta-Albertos et al. (2009)
#' tests_ccf2009 <- unif_test(data = comets_ccf2009$normal, type = type_tests,
#'                            p_value = "asymp")
#' tests_ccf2009
"comets"


#' @title Planet orbits
#'
#' @description Planet orbits data from the
#' \href{https://ssd.jpl.nasa.gov/?planet_pos}{
#' JPL Keplerian Elements for Approximate Positions of the Major Planets}.
#' The normal vector of a planet orbit represents is a vector on \eqn{S^2}.
#'
#' @docType data
#' @format A data frame with 9 rows and 3 variables:
#' \describe{
#'   \item{planet}{names of the planets and Pluto.}
#'   \item{i}{inclination; the orbit angle with respect to the ecliptic plane,
#'   in radians in \eqn{[0, \pi]}.}
#'   \item{om}{longitude of the ascending node; the angle between the normal
#'   vector of the orbit and the normal vector of the ecliptic plane, in
#'   radians in \eqn{[0, 2\pi)}.}
#' }
#' @details
#' The normal vector to the ecliptic plane of the planet with inclination
#' \eqn{i} and longitude of the ascending node \eqn{\omega} is
#' \deqn{(\sin(i) \sin(\omega), -\sin(i) \cos(\omega), \cos(i))'.}{
#' (sin(i) sin(\omega), -sin(i) cos(\omega), cos(i))'.}
#'
#' The script performing the data preprocessing is available at
#' \href{https://github.com/egarpor/sphunif/blob/master/data-raw/planets.R}{
#' \code{planets.R}}. The data was retrieved on 2020-05-16.
#' @source Table 2b in \url{https://ssd.jpl.nasa.gov/txt/aprx_pos_planets.pdf}
#' @examples
#' # Load data
#' data("planets")
#'
#' # Add normal vectors
#' planets$normal <- cbind(sin(planets$i) * sin(planets$om),
#'                        -sin(planets$i) * cos(planets$om),
#'                        cos(planets$i))
#'
#' # Tests to be performed
#' type_tests <- c("PCvM", "PAD", "PRt")
#'
#' # Tests with Pluto
#' unif_test(data = planets$normal, type = type_tests, p_value = "MC")
#'
#' # Tests without Pluto
#' unif_test(data = planets$normal[-9, ], type = type_tests, p_value = "MC")
"planets"


#' @title Craters named by the IUA
#'
#' @description \emph{Named} craters of the Solar System by the
#' \href{https://planetarynames.wr.usgs.gov}{Gazetteer of Planetary
#' Nomenclature} of the International Astronomical Union (IUA).
#'
#' @docType data
#' @format A data frame with 5235 rows and 7 variables:
#' \describe{
#'   \item{ID}{database ID.}
#'   \item{name}{name of the crater.}
#'   \item{target}{name of the celestial body. A factor with 43 levels,
#'   such as \code{"Moon"}, \code{"Venus"}, or \code{"Europa"}.}
#'   \item{target_type}{type of celestial body. A factor with 3 levels:
#'   \code{"Planet"}, \code{"Moon"}, \code{"Dwarf planet"}, or
#'   \code{"Asteroid"}.}
#'   \item{diameter}{diameter of the crater (in km).}
#'   \item{theta}{longitude angle \eqn{\theta \in [0, 2\pi)} of the
#'   crater center.}
#'   \item{phi}{latitude angle \eqn{\phi \in [-\pi/2, \pi/2]} of the
#'   crater center.}
#' }
#' @details
#' "Craters" are understood in the Gazetteer of Planetary Nomenclature as
#' roughly circular depressions resulting from impact or volcanic activity
#' (the geological origin is
#' \href{https://planetarynames.wr.usgs.gov/DescriptorTerms}{unspecified}).
#'
#' Be aware that the dataset only contains \emph{named} craters by the IUA.
#' Therefore, there is likely a \bold{high uniform bias} on the distribution
#' of craters. Presumably the naming process attempts to cover the planet in
#' a somehow uniform fashion (distant craters are more likely to be named than
#' neighboring craters). Also, there are substantially more craters in the
#' listed bodies than those named by the IUA. See \code{\link{venus}} and
#' \code{\link{rhea}} for more detailed and specific crater datasets.
#'
#' The \eqn{(\theta, \phi)} angles are such their associated planetocentric
#' coordinates are:
#' \deqn{(\cos(\phi) \cos(\theta), \cos(\phi) \sin(\theta), \sin(\phi))',}{
#' (cos(\phi) cos(\theta), cos(\phi) sin(\theta), sin(\phi))',}
#' with \eqn{(0, 0, 1)'} denoting the north pole.
#'
#' The script performing the data preprocessing is available at
#' \href{https://github.com/egarpor/sphunif/blob/master/data-raw/craters.R}{
#' \code{craters.R}}. The data was retrieved on 2020-05-31.
#' @source \url{https://planetarynames.wr.usgs.gov/AdvancedSearch}
#' @examples
#' # Load data
#' data("craters")
#'
#' # Add Cartesian coordinates
#' craters$X <- cbind(cos(craters$theta) * cos(craters$phi),
#'                    sin(craters$theta) * cos(craters$phi),
#'                    sin(craters$phi))
#'
#' # Tests to be performed
#' type_tests <- c("PCvM", "PAD", "PRt")
#'
#' # Tests for Venus and Rhea
#' unif_test(data = craters$X[craters$target == "Venus", ], type = type_tests,
#'           p_value = "asymp")
#' unif_test(data = craters$X[craters$target == "Rhea", ], type = type_tests,
#'           p_value = "asymp")
"craters"


#' @title Venus craters
#'
#' @description Craters on Venus from the
#' \href{https://astrogeology.usgs.gov/search/map/Venus/venuscraters}{
#' USGS Astrogeology Science Center}.
#'
#' @docType data
#' @format A data frame with 967 rows and 4 variables:
#' \describe{
#'   \item{name}{name of the crater (if named).}
#'   \item{diameter}{diameter of the crater (in km).}
#'   \item{theta}{longitude angle \eqn{\theta \in [0, 2\pi)} of the
#'   crater center.}
#'   \item{phi}{latitude angle \eqn{\phi \in [-\pi/2, \pi/2]} of the
#'   crater center.}
#' }
#' @details
#' The \eqn{(\theta, \phi)} angles are such their associated planetocentric
#' coordinates are:
#' \deqn{(\cos(\phi) \cos(\theta), \cos(\phi) \sin(\theta), \sin(\phi))',}{
#' (cos(\phi) cos(\theta), cos(\phi) sin(\theta), sin(\phi))',}
#' with \eqn{(0, 0, 1)'} denoting the north pole.
#'
#' The script performing the data preprocessing is available at
#' \href{https://github.com/egarpor/sphunif/blob/master/data-raw/venus.R}{
#' \code{venus.R}}.
#' @source \url{https://astrogeology.usgs.gov/search/map/Venus/venuscraters}
#' @examples
#' # Load data
#' data("venus")
#'
#' # Add Cartesian coordinates
#' venus$X <- cbind(cos(venus$theta) * cos(venus$phi),
#'                  sin(venus$theta) * cos(venus$phi),
#'                  sin(venus$phi))
#'
#' # Tests
#' unif_test(data = venus$X, type = c("PCvM", "PAD", "PRt"), p_value = "asymp")
"venus"


#' @title Rhea craters from Hirata (2016)
#'
#' @description Craters on Rhea from
#' \href{https://agupubs.onlinelibrary.wiley.com/doi/full/10.1002/2015JE004940}{
#' Hirata (2016)}.
#'
#' @docType data
#' @format A data frame with 3596 rows and 4 variables:
#' \describe{
#'   \item{name}{name of the crater (if named).}
#'   \item{diameter}{diameter of the crater (in km).}
#'   \item{theta}{longitude angle \eqn{\theta \in [0, 2\pi)} of the
#'   crater center.}
#'   \item{phi}{latitude angle \eqn{\phi \in [-\pi/2, \pi/2]} of the
#'   crater center.}
#' }
#' @details
#' The \eqn{(\theta, \phi)} angles are such their associated planetocentric
#' coordinates are:
#' \deqn{(\cos(\phi) \cos(\theta), \cos(\phi) \sin(\theta), \sin(\phi))',}{
#' (cos(\phi) cos(\theta), cos(\phi) sin(\theta), sin(\phi))',}
#' with \eqn{(0, 0, 1)'} denoting the north pole.
#'
#' The script performing the data preprocessing is available at
#' \href{https://github.com/egarpor/sphunif/blob/master/data-raw/rhea.R}{
#' \code{rhea.R}}.
#' @source \url{https://agupubs.onlinelibrary.wiley.com/action/downloadSupplement?doi=10.1002\%2F2015JE004940&file=jgre20485-sup-0002-TableS1.txt}
#' @references
#' Hirata, N. (2016) Differential impact cratering of Saturn's satellites by
#' heliocentric impactors. \emph{Journal of Geophysical Research: Planets},
#' 121:111--117. \doi{10.1002/2015JE004940}
#' @examples
#' # Load data
#' data("rhea")
#'
#' # Add Cartesian coordinates
#' rhea$X <- cbind(cos(rhea$theta) * cos(rhea$phi),
#'                 sin(rhea$theta) * cos(rhea$phi),
#'                 sin(rhea$phi))
#'
#' # Tests
#' unif_test(data = rhea$X[rhea$diam > 15 & rhea$diam < 20, ],
#'           type = c("PCvM", "PAD", "PRt"), p_value = "asymp")
"rhea"


#' #' @title TODO
#' #'
#' #' @description TODO.
#' #'
#' #' @docType data
#' #' @format TODO
#' #' \describe{
#' #'   TODO
#' #' }
#' #' @details
#' #' TODO
#' #' @source TODO
#' #' @references
#' #' TODO
#' #' @examples
#' #' # TODO
#' "ambrosia"
#'
#'
#' #' @rdname ambrosia
#' "angustifolia"
#'
#'
#' #' @rdname ambrosia
#' "chenopodium"

#' @title Asymptotic distributions for circular uniformity statistics
#'
#' @description Computation of the asymptotic null distributions of circular
#' uniformity statistics.
#'
#' @param x a vector of size \code{nx} or a matrix of size \code{c(nx, 1)}.
#' @inheritParams Sobolev
#' @param thre error threshold for the tail probability given by the
#' the first terms of the truncated series of a Sobolev test. Defaults to
#' \code{0} (no further truncation).
#' @param K_Kolmogorov,K_Kuiper,K_Watson,K_Watson_1976,K_Ajne integer giving
#' the truncation of the series present in the null asymptotic distributions.
#' For the Kolmogorov-Smirnov-related series defaults to \code{25}; for the
#' others series defaults to a smaller number.
#' @inheritParams unif_stat_distr
#' @inheritParams cir_stat
#' @param alternating use the alternating series expansion for the distribution
#' of the Kolmogorov-Smirnov statistic? Defaults to \code{TRUE}.
#' @param second_term use the second-order series expansion for the
#' distribution of the Kuiper statistic? Defaults to \code{TRUE}.
#' @param N number of points used in the
#' \link[=Gauss_Legen_nodes]{Gauss-Legendre quadrature}. Defaults to \code{40}.
#' @param exact use the exact distribution for the Hodges-Ajne statistic?
#' Defaults to \code{TRUE}.
#' @param asymp_std compute the distribution associated to the normalized
#' Hodges-Ajne statistic? Defaults to \code{FALSE}.
#' @param max_gap compute the distribution associated to the maximum gap for
#' the range statistic? Defaults to \code{TRUE}.
#' @param abs_val compute the distribution associated to the absolute value of
#' the Darling's log gaps statistic? Defaults to \code{TRUE}.
#' @inheritParams sph_stat_distr
#' @param ... further parameters passed to \code{\link{p_Sobolev}} or
#' \code{\link{d_Sobolev}} (such as \code{x_tail}).
#' @return A matrix of size \code{c(nx, 1)} with the evaluation of the
#' distribution or density function at \code{x}.
#' @references García-Portugués, E. and Verdebout, T. (2018) An overview of
#' uniformity tests on the hypersphere. \emph{arXiv:1804.00286}.
#' \url{https://arxiv.org/abs/1804.00286}.
#' @details
#' Descriptions and references for most of the tests are available
#' in García-Portugués and Verdebout (2018).
#' @examples
#' # Ajne
#' curve(d_cir_stat_Ajne(x), to = 1.5, n = 2e2, ylim = c(0, 4))
#' curve(p_cir_stat_Ajne(x), n = 2e2, col = 2, add = TRUE)
#'
#' # Bakshaev
#' curve(d_cir_stat_Bakshaev(x, method = "HBE"), to = 6, n = 2e2,
#'       ylim = c(0, 1))
#' curve(p_cir_stat_Bakshaev(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)
#'
#' # Bingham
#' curve(d_cir_stat_Bingham(x), to = 12, n = 2e2, ylim = c(0, 1))
#' curve(p_cir_stat_Bingham(x), n = 2e2, col = 2, add = TRUE)
#
#' # Greenwood
#' curve(d_cir_stat_Greenwood(x), from = -6, to = 6, n = 2e2, ylim = c(0, 1))
#' curve(p_cir_stat_Greenwood(x), n = 2e2, col = 2, add = TRUE)
#'
#' # Hermans-Rasson
#' curve(p_cir_stat_Hermans_Rasson(x, method = "HBE"), to = 10, n = 2e2,
#'       ylim = c(0, 1))
#' curve(d_cir_stat_Hermans_Rasson(x, method = "HBE"), n = 2e2, add = TRUE,
#'       col = 2)
#'
#' # Hodges-Ajne
#' plot(25:45, d_cir_stat_Hodges_Ajne(cbind(25:45), n = 50), type = "h",
#'      lwd = 2, ylim = c(0, 1))
#' lines(25:45, p_cir_stat_Hodges_Ajne(cbind(25:45), n = 50), type = "s",
#'       col = 2)
#'
#' # Kolmogorov-Smirnov
#' curve(d_Kolmogorov(x), to = 3, n = 2e2, ylim = c(0, 2))
#' curve(p_Kolmogorov(x), n = 2e2, col = 2, add = TRUE)
#'
#' # Kuiper
#' curve(d_cir_stat_Kuiper(x, n = 50), to = 3, n = 2e2, ylim = c(0, 2))
#' curve(p_cir_stat_Kuiper(x, n = 50), n = 2e2, col = 2, add = TRUE)
#'
#' # Kuiper and Watson with Stephens modification
#' curve(d_cir_stat_Kuiper(x, n = 8, Stephens = TRUE), to = 2.5, n = 2e2,
#'       ylim = c(0, 10))
#' curve(d_cir_stat_Watson(x, n = 8, Stephens = TRUE), n = 2e2, lty = 2,
#'       add = TRUE)
#' n <- c(10, 20, 30, 40, 50, 100, 500)
#' col <- rainbow(length(n))
#' for (i in seq_along(n)) {
#'   curve(d_cir_stat_Kuiper(x, n = n[i], Stephens = TRUE), n = 2e2,
#'         col = col[i], add = TRUE)
#'   curve(d_cir_stat_Watson(x, n = n[i], Stephens = TRUE), n = 2e2,
#'         col = col[i], lty = 2, add = TRUE)
#' }
#'
#' # Maximum uncovered spacing
#' curve(d_cir_stat_Max_uncover(x), from = -3, to = 6, n = 2e2, ylim = c(0, 1))
#' curve(p_cir_stat_Max_uncover(x), n = 2e2, col = 2, add = TRUE)
#'
#' # Number of uncovered spacing
#' curve(d_cir_stat_Num_uncover(x), from = -4, to = 4, n = 2e2, ylim = c(0, 1))
#' curve(p_cir_stat_Num_uncover(x), n = 2e2, col = 2, add = TRUE)
#'
#' # Log gaps
#' curve(d_cir_stat_Log_gaps(x), from = -1, to = 4, n = 2e2, ylim = c(0, 1))
#' curve(p_cir_stat_Log_gaps(x), n = 2e2, col = 2, add = TRUE)
#'
#' # Gine Fn
#' curve(d_cir_stat_Gine_Fn(x, method = "HBE"), to = 2.5, n = 2e2,
#'       ylim = c(0, 2))
#' curve(p_cir_stat_Gine_Fn(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)
#'
#' # Gine Gn
#' curve(d_cir_stat_Gine_Gn(x, method = "HBE"), to = 2.5, n = 2e2,
#'       ylim = c(0, 2))
#' curve(p_cir_stat_Gine_Gn(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)
#'
#' # Gini mean difference
#' curve(d_cir_stat_Gini(x), from = -4, to = 4, n = 2e2, ylim = c(0, 1))
#' curve(p_cir_stat_Gini(x), n = 2e2, col = 2, add = TRUE)
#'
#' # Gini mean squared difference
#' curve(d_cir_stat_Gini_squared(x), from = -10, to = 10, n = 2e2,
#'       ylim = c(0, 1))
#' curve(p_cir_stat_Gini_squared(x), n = 2e2, col = 2, add = TRUE)
#'
#' # PAD
#' curve(d_cir_stat_PAD(x, method = "HBE"), to = 3, n = 2e2, ylim = c(0, 1.5))
#' curve(p_cir_stat_PAD(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)
#'
#' # PCvM
#' curve(d_cir_stat_PCvM(x, method = "HBE"), to = 4, n = 2e2, ylim = c(0, 2))
#' curve(p_cir_stat_PCvM(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)
#'
#' # PRt
#' curve(d_cir_stat_PRt(x, method = "HBE"), n = 2e2, ylim = c(0, 5))
#' curve(p_cir_stat_PRt(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)
#'
#' # Pycke
#' curve(d_cir_stat_Pycke(x), from = -5, to = 10, n = 2e2, ylim = c(0, 1))
#' curve(p_cir_stat_Pycke(x), n = 2e2, col = 2, add = TRUE)
#'
#' # Pycke q
#' curve(d_cir_stat_Pycke_q(x, method = "HBE"), to = 15, n = 2e2,
#'       ylim = c(0, 1))
#' curve(p_cir_stat_Pycke_q(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)
#'
#' # Range
#' curve(d_cir_stat_Range(x, n = 50), to = 2, n = 2e2, ylim = c(0, 4))
#' curve(p_cir_stat_Range(x, n = 50), n = 2e2, col = 2, add = TRUE)
#'
#' # Rao
#' curve(d_cir_stat_Rao(x), from = -6, to = 6, n = 2e2, ylim = c(0, 1))
#' curve(p_cir_stat_Rao(x), n = 2e2, col = 2, add = TRUE)
#'
#' # Rayleigh
#' curve(d_cir_stat_Rayleigh(x), to = 12, n = 2e2, ylim = c(0, 1))
#' curve(p_cir_stat_Rayleigh(x), n = 2e2, col = 2, add = TRUE)
#'
#' # Riesz
#' curve(d_cir_stat_Riesz(x, method = "HBE"), to = 6, n = 2e2,
#'       ylim = c(0, 1))
#' curve(p_cir_stat_Riesz(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)
#'
#' # Rothman
#' curve(d_cir_stat_Rothman(x, method = "HBE"), n = 2e2, ylim = c(0, 5))
#' curve(p_cir_stat_Rothman(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)
#'
#' # Vacancy
#' curve(d_cir_stat_Vacancy(x), from = -4, to = 4, n = 2e2, ylim = c(0, 1))
#' curve(p_cir_stat_Vacancy(x), n = 2e2, col = 2, add = TRUE)
#'
#' # Watson
#' curve(d_cir_stat_Watson(x), to = 0.5, n = 2e2, ylim = c(0, 15))
#' curve(p_cir_stat_Watson(x), n = 2e2, col = 2, add = TRUE)
#'
#' # Watson (1976)
#' curve(d_cir_stat_Watson_1976(x), to = 1.5, n = 2e2, ylim = c(0, 3))
#' curve(p_cir_stat_Watson_1976(x), n = 2e2, col = 2, add = TRUE)
#' @name cir_stat_distr
NULL


#' @rdname cir_stat_distr
#' @export
p_cir_stat_Bakshaev <- function(x, K_max = 1e3, thre = 0, ...) {

  p_sph_stat_Bakshaev(x = x, p = 2, K_max = K_max, thre = thre, ...)

}


#' @rdname cir_stat_distr
#' @export
d_cir_stat_Bakshaev <- function(x, K_max = 1e3, thre = 0, ...) {

  d_sph_stat_Bakshaev(x = x, p = 2, K_max = K_max, thre = thre, ...)

}


#' @rdname cir_stat_distr
#' @export
p_cir_stat_Gine_Fn <- function(x, K_max = 1e3, thre = 0, ...) {

  p_sph_stat_Gine_Fn(x = x, p = 2, K_max = K_max, thre = thre, ...)

}


#' @rdname cir_stat_distr
#' @export
d_cir_stat_Gine_Fn <- function(x, K_max = 1e3, thre = 0, ...) {

  d_sph_stat_Gine_Fn(x = x, p = 2, K_max = K_max, thre = thre, ...)

}


#' @rdname cir_stat_distr
#' @export
p_cir_stat_Gine_Gn <- function(x, K_max = 1e3, thre = 0, ...) {

  p_sph_stat_Gine_Gn(x = x, p = 2, K_max = K_max, thre = thre, ...)

}


#' @rdname cir_stat_distr
#' @export
d_cir_stat_Gine_Gn <- function(x, K_max = 1e3, thre = 0, ...) {

  d_sph_stat_Gine_Gn(x = x, p = 2, K_max = K_max, thre = thre, ...)

}



#' @rdname cir_stat_distr
#' @export
p_cir_stat_Hermans_Rasson <- function(x, K_max = 1e3, thre = 0, ...) {

  cbind(p_Sobolev(x = x, p = 2, type = "Hermans_Rasson", K_max = K_max,
                  thre = thre, ...))

}


#' @rdname cir_stat_distr
#' @export
d_cir_stat_Hermans_Rasson <- function(x, K_max = 1e3, thre = 0, ...) {

  cbind(d_Sobolev(x = x, p = 2, type = "Hermans_Rasson", K_max = K_max,
                  thre = thre, ...))

}


#' @rdname cir_stat_distr
#' @export
p_cir_stat_PAD <- function(x, K_max = 1e3, thre = 0, ...) {

  p_sph_stat_PAD(x = x, p = 2, K_max = K_max, thre = thre, ...)

}


#' @rdname cir_stat_distr
#' @export
d_cir_stat_PAD <- function(x, K_max = 1e3, thre = 0, ...) {

  d_sph_stat_PAD(x = x, p = 2, K_max = K_max, thre = thre, ...)

}


#' @rdname cir_stat_distr
#' @export
p_cir_stat_PCvM <- function(x, K_max = 1e3, thre = 0, ...) {

  p_sph_stat_PCvM(x = x, p = 2, K_max = K_max, thre = thre, ...)

}


#' @rdname cir_stat_distr
#' @export
d_cir_stat_PCvM <- function(x, K_max = 1e3, thre = 0, ...) {

  d_sph_stat_PCvM(x = x, p = 2, K_max = K_max, thre = thre, ...)

}


#' @rdname cir_stat_distr
#' @export
p_cir_stat_PRt <- function(x, t = 1 / 3, K_max = 1e3, thre = 0, ...) {

  p_sph_stat_PRt(x = x, p = 2, t = t, K_max = K_max, thre = thre, ...)

}


#' @rdname cir_stat_distr
#' @export
d_cir_stat_PRt <- function(x, t = 1 / 3, K_max = 1e3, thre = 0, ...) {

  d_sph_stat_PRt(x = x, p = 2, t = t, K_max = K_max, thre = thre, ...)

}


#' @rdname cir_stat_distr
#' @export
p_cir_stat_Pycke_q <- function(x, q = 0.5, K_max = 1e3, thre = 0, ...) {

  cbind(p_Sobolev(x = x, p = 2, type = "Pycke_q", Pycke_q = q,
                  K_max = K_max, thre = thre, ...))

}


#' @rdname cir_stat_distr
#' @export
d_cir_stat_Pycke_q <- function(x, q = 0.5, K_max = 1e3, thre = 0, ...) {

  cbind(d_Sobolev(x = x, p = 2, type = "Pycke_q", Pycke_q = q,
                  K_max = K_max, thre = thre, ...))

}


#' @rdname cir_stat_distr
#' @export
p_cir_stat_Rothman <- function(x, t = 1 / 3, K_max = 1e3, thre = 0, ...) {

  cbind(p_Sobolev(x = x, p = 2, type = "Rothman", Rothman_t = t,
                  K_max = K_max, thre = thre, ...))

}


#' @rdname cir_stat_distr
#' @export
d_cir_stat_Rothman <- function(x, t = 1 / 3, K_max = 1e3, thre = 0, ...) {

  cbind(d_Sobolev(x = x, p = 2, type = "Rothman", Rothman_t = t,
                  K_max = K_max, thre = thre, ...))

}


#' @rdname cir_stat_distr
#' @export
p_cir_stat_Riesz <- function(x, s = 1, K_max = 1e3, thre = 0, ...) {

  p_sph_stat_Riesz(x = x, p = 2, s = s, K_max = K_max, thre = thre, ...)

}


#' @rdname cir_stat_distr
#' @export
d_cir_stat_Riesz <- function(x, s = 1, K_max = 1e3, thre = 0, ...) {

  d_sph_stat_Riesz(x = x, p = 2, s = s, K_max = K_max, thre = thre, ...)

}

## Retrieve data from https://planetarynames.wr.usgs.gov/AdvancedSearch

# The source is the Gazetteer of Planetary Nomenclature of the
# International Astronomical Union (IAU)

# The database was accessed on the 2020-05-31 using the following query:
# System: All.
# Target: All.
# Coordinate System: +East. 0 - 360. Planetocentric.
# Latitude Longitude: Do not enter anything.
# Feature Type: Crater, craters.
# Feature Name: Do not enter anything.
# Approval Status: Adopted by IAU.
# Feature Diameter (km): Do not select anything.
# Approval Date: Do not enter anything.
# Continent: All.
# Ethnic/Cultural Group or Country: All.
# Reference: All.
# Columns to include:	
# [x] Feature ID
# [x] Feature Name
# [ ] Clean Feature Name
# [x] Target
# [x] Diameter
# [x] Center Lat/Lon
# [x] Lat/Lon Boundaries
# [ ] Coordinate System
# [ ] Continent/Ethnicity
# [ ] Feature Type
# [ ] Feature Type Code
# [ ] Quad
# [ ] Approval Status
# [ ] Approval Date
# [ ] Reference
# [ ] Origin
# [ ] Additional Info
# [ ] Last Updated
# Sorted by: Feature ID. Ascending.
# Output Format: CSV - (Comma Separated Values) for importing into Excel

# Read data
craters <- read.csv(file = "craters-accessed-2020-05-31.csv", header = TRUE)
craters$X <- NULL

## Preprocess data

# Transform degrees to radians
craters$phi <- craters$Center_Latitude / 180 * pi
craters$theta <- (craters$Center_Longitude / 180 * pi) %% (2 * pi)
craters$Center_Latitude <- NULL
craters$Center_Longitude <- NULL

# Rename
names(craters)[1:4] <- c("ID", "name", "target", "diameter")

# Type of body (planet, moon, asteroid, ...)
body_type <- read.csv(file = "bodies-type.csv", sep = ",", dec = ".")[, 1:2]
craters$target_type <- body_type$Type[match(craters$target, body_type$Body)]

# Reorder
craters <- craters[, c(1:3, 7, 4:6)]

# Save object
save(list = "craters", file = "craters.rda", compress = "bzip2")

#' @title Weighted sums of non-central chi squared random variables
#'
#' @description Approximated density, distribution, and quantile functions for
#' weighted sums of non-central chi squared random variables:
#' \deqn{Q_K = \sum_{i = 1}^K w_i \chi^2_{d_i}(\lambda_i),}
#' where \eqn{w_1, \ldots, w_n} are positive weights, \eqn{d_1, \ldots, d_n}
#' are positive degrees of freedom, and \eqn{\lambda_1, \ldots, \lambda_n}
#' are non-negative non-centrality parameters. Also, simulation of \eqn{Q_K}.
#'
#' @inheritParams r_unif
#' @param x vector of quantiles.
#' @param u vector of probabilities.
#' @param weights vector with the positive weights of the sum. Must have the
#' same length as \code{dfs}.
#' @param dfs vector with the positive degrees of freedom of the chi squared
#' random variables. Must have the same length as \code{weights}.
#' @param ncps non-centrality parameters. Either \code{0} (default) or a
#' vector with the same length as \code{weights}.
#' @param method method for approximating the density, distribution, or
#' quantile function. Must be \code{"I"} (Imhof), \code{"SW"}
#' (Satterthwaite--Welch), \code{"HBE"} (Hall--Buckley--Eagleson), or
#' \code{"MC"} (Monte Carlo; only for distribution or quantile functions).
#' Defaults to \code{"I"}.
#' @param exact_chisq if \code{weights} and \code{dfs} have length one, shall
#' the \code{\link[stats]{Chisquare}} functions be called? Otherwise, the
#' approximations are computed for this exact case. Defaults to \code{TRUE}.
#' @param imhof_epsabs,imhof_epsrel,imhof_limit precision parameters passed to
#' \code{\link[CompQuadForm]{imhof}}'s \code{epsabs}, \code{epsrel}, and
#' \code{limit}, respectively. They default to \code{1e-6}, \code{1e-6},
#' and \code{1e4}.
#' @param grad_method,grad_method.args numerical differentiation parameters
#' passed to \code{\link[numDeriv]{grad}}'s \code{method} and
#' \code{method.args}, respectively. They default to \code{"simple"},
#' and \code{list(eps = 1e-7)} (better precision than \code{imhof_epsabs} to
#' avoid numerical artifacts).
#' @param M number of Monte Carlo samples for approximating the distribution if
#' \code{method = "MC"}. Defaults to \code{1e4}.
#' @param MC_sample if provided, it is employed when \code{method = "MC"}. If
#' not, it is computed internally.
#' @param nlm_gradtol,nlm_iterlim convergence control parameters passed to
#' \code{\link[stats]{nlm}}'s \code{gradtol} and \code{iterlim}, respectively.
#' They default to \code{1e-6} and \code{1e3}.
#' @param thre vector with the error thresholds of the tail probability and
#' mean/variance explained by the first terms of the series. Defaults to
#' \code{1e-4}. See details.
#' @param log are \code{weights} and \code{dfs} given in log-scale? Defaults to
#' \code{FALSE}.
#' @param x_tail scalar evaluation point for determining the upper tail
#' probability. If \code{NULL}, set to the \code{0.90} quantile of the whole
#' series, computed by the \code{"HBE"} approximation.
#' @return
#' \itemize{
#'   \item \code{d_wschisq}: density function evaluated at \code{x}, a vector.
#'   \item \code{p_wschisq}: distribution function evaluated at \code{x},
#'   a vector.
#'   \item \code{q_wschisq}: quantile function evaluated at \code{u}, a vector.
#'   \item \code{r_wschisq}: a vector of size \code{n} containing a random
#'   sample.
#'   \item \code{cutoff_wschisq}: a data frame with the indexes up to which the
#'   truncated series explains the tail probability with absolute error
#'   \code{thre}, or the proportion of the mean/variance of the whole series
#'   that is \emph{not} explained by the truncated series.
#' }
#' @author Eduardo García-Portugués and Paula Navarro-Esteban.
#' @details
#' Four methods are implemented for approximating the distribution of a
#' weighted sum of chi squared random variables:
#' \itemize{
#'   \item \code{"I"}: Imhof's approximation (Imhof, 1961) for the evaluation
#'   of the distribution function. If this method is selected, the function
#'   is simply a wrapper to \code{\link[CompQuadForm]{imhof}} from the
#'   \code{CompQuadForm} package (Duchesne and Lafaye De Micheaux, 2010).
#'   \item \code{"SW"}: Satterthwaite--Welch (Satterthwaite, 1946; Welch, 1938)
#'   approximation, consisting in matching the first \emph{two} moments of
#'   \eqn{Q_K} with a gamma distribution.
#'   \item \code{"HBE"}: Hall--Buckley--Eagleson (Hall, 1983; Buckley and
#'   Eagleson, 1988) approximation, consisting in matching the first
#'   \emph{three} moments of \eqn{Q_K} with a gamma distribution.
#'   \item \code{"MC"}: Monte Carlo approximation using the empirical
#'   cumulative distribution function with \code{M} simulated samples.
#' }
#' The Imhof method is exact up to the prescribed numerical accuracy. It is
#' also the most time-consuming method. The density and quantile functions
#' for this approximation are obtained by numerical differentiation and
#' inversion, respectively, of the approximated distribution.
#'
#' For the methods based on gamma matching, the \code{\link[stats]{GammaDist}}
#' density, distribution, and quantile functions are invoked. The
#' Hall--Buckley--Eagleson approximation tends to overperform the
#' Satterthwaite--Welch approximation.
#'
#' The Monte Carlo method is relatively inaccurate and slow, but serves as an
#' unbiased reference of the true distribution function. The inversion of the
#' empirical cumulative distribution is done by \code{\link[stats]{quantile}}.
#'
#' An empirical comparison of these and other approximation methods is
#' given in Bodenham and Adams (2016).
#'
#' \code{cutoff_wschisq} removes \code{NA}s/\code{NaN}s in \code{weights} or
#' \code{dfs} with a message. The threshold \code{thre} ensures that the
#' tail probability of the truncated and whole series differ less than
#' \code{thre} at \code{x_tail}, or that \code{thre} is the proportion of the
#' mean/variance of the whole series that is \emph{not} retained. The (upper)
#' tail probabilities for evaluating truncation are computed using the
#' Hall--Buckley--Eagleson approximation at \code{x_tail}.
#' @references
#' Bodenham, D. A. and Adams, N. M. (2016). A comparison of efficient
#' approximations for a weighted sum of chi-squared random variables.
#' \emph{Statistics and Computing}, 26(4):917--928.
#' \doi{10.1007/s11222-015-9583-4}
#'
#' Buckley, M. J. and Eagleson, G. K. (1988). An approximation to the
#' distribution of quadratic forms in normal random variables.
#' \emph{Australian Journal of Statistics}, 30(1):150--159.
#' \doi{10.1111/j.1467-842X.1988.tb00471.x}
#'
#' Duchesne, P. and Lafaye De Micheaux, P. (2010) Computing the distribution
#' of quadratic forms: Further comparisons between the Liu--Tang--Zhang
#' approximation and exact methods. \emph{Computational Statistics and Data
#' Analysis}, 54(4):858--862. \doi{10.1016/j.csda.2009.11.025}
#'
#' Hall, P. (1983). Chi squared approximations to the distribution of a sum
#' of independent random variables. \emph{Annals of Probability},
#' 11(4):1028--1036. \doi{10.1214/aop/1176993451}
#'
#' Imhof, J. P. (1961). Computing the distribution of quadratic forms in normal
#' variables. \emph{Biometrika}, 48(3/4):419--426.
#' \doi{10.2307/2332763}
#'
#' Satterthwaite, F. E. (1946). An approximate distribution of estimates of
#' variance components. \emph{Biometrics Bulletin}, 2(6):110--114.
#' \doi{10.2307/3002019}
#'
#' Welch, B. L. (1938). The significance of the difference between two means
#' when the population variances are unequal. \emph{Biometrika},
#' 29(3/4):350--362. \doi{10.2307/2332010}
#' @examples
#' # Plotting functions for the examples
#' add_approx_dens <- function(x, dfs, weights, ncps) {
#'
#'   lines(x, d_wschisq(x, weights = weights, dfs = dfs, ncps = ncps,
#'                      method = "SW", exact_chisq = FALSE), col = 3)
#'   lines(x, d_wschisq(x, weights = weights, dfs = dfs, ncps = ncps,
#'                      method = "HBE", exact_chisq = FALSE), col = 4)
#'   lines(x, d_wschisq(x, weights = weights, dfs = dfs, ncps = ncps,
#'                      method = "I", exact_chisq = TRUE), col = 2)
#'   legend("topright", legend = c("True", "SW", "HBE", "I"), lwd = 2,
#'          col = c(1, 3:4, 2))
#'
#' }
#' add_approx_distr <- function(x, dfs, weights, ncps, ...) {
#'
#'   lines(x, p_wschisq(x, weights = weights, dfs = dfs, ncps = ncps,
#'                      method = "SW", exact_chisq = FALSE), col = 3)
#'   lines(x, p_wschisq(x, weights = weights, dfs = dfs, ncps = ncps,
#'                      method = "HBE", exact_chisq = FALSE), col = 4)
#'   lines(x, p_wschisq(x, weights = weights, dfs = dfs, ncps = ncps,
#'                      method = "MC", exact_chisq = FALSE), col = 5,
#'                      type = "s")
#'   lines(x, p_wschisq(x, weights = weights, dfs = dfs, ncps = ncps,
#'                      method = "I", exact_chisq = TRUE), col = 2)
#'   legend("bottomright", legend = c("True", "SW", "HBE", "MC", "I"), lwd = 2,
#'          col = c(1, 3:5, 2))
#'
#' }
#' add_approx_quant <- function(u, dfs, weights, ncps, ...) {
#'
#'   lines(u, q_wschisq(u, weights = weights, dfs = dfs, ncps = ncps,
#'                      method = "SW", exact_chisq = FALSE), col = 3)
#'   lines(u, q_wschisq(u, weights = weights, dfs = dfs, ncps = ncps,
#'                      method = "HBE", exact_chisq = FALSE), col = 4)
#'   lines(u, q_wschisq(u, weights = weights, dfs = dfs, ncps = ncps,
#'                      method = "MC", exact_chisq = FALSE), col = 5,
#'                      type = "s")
#'   lines(u, q_wschisq(u, weights = weights, dfs = dfs, ncps = ncps,
#'                      method = "I", exact_chisq = TRUE), col = 2)
#'   legend("topleft", legend = c("True", "SW", "HBE", "MC", "I"), lwd = 2,
#'          col = c(1, 3:5, 2))
#'
#' }
#'
#' # Validation plots for density, distribution, and quantile functions
#' u <- seq(0.01, 0.99, l = 100)
#' old_par <- par(mfrow = c(1, 3))
#'
#' # Case 1: 1 * ChiSq_3(0) + 1 * ChiSq_3(0) = ChiSq_6(0)
#' weights <- c(1, 1)
#' dfs <- c(3, 3)
#' ncps <- 0
#' x <- seq(-1, 30, l = 100)
#' main <- expression(1 * chi[3]^2 * (0) + 1 * chi[3]^2 * (0))
#' plot(x, dchisq(x, df = 6), type = "l", main = main, ylab = "Density")
#' add_approx_dens(x = x, weights = weights, dfs = dfs, ncps = ncps)
#' plot(x, pchisq(x, df = 6), type = "l", main = main, ylab = "Distribution")
#' add_approx_distr(x = x, weights = weights, dfs = dfs, ncps = ncps)
#' plot(u, qchisq(u, df = 6), type = "l", main = main, ylab = "Quantile")
#' add_approx_quant(u = u, weights = weights, dfs = dfs, ncps = ncps)
#' \donttest{
#' # Case 2: 2 * ChiSq_3(1) + 1 * ChiSq_6(0.5) + 0.5 * ChiSq_12(0.25)
#' weights <- c(2, 1, 0.5)
#' dfs <- c(3, 6, 12)
#' ncps <- c(1, 0.5, 0.25)
#' x <- seq(0, 70, l = 100)
#' main <- expression(2 * chi[3]^2 * (1)+ 1 * chi[6]^2 * (0.5) +
#'                    0.5 * chi[12]^2 * (0.25))
#' samp <- r_wschisq(n = 1e4, weights = weights, dfs = dfs, ncps = ncps)
#' hist(samp, breaks = 50, freq = FALSE, main = main, ylab = "Density",
#'      xlim = range(x), xlab = "x"); box()
#' add_approx_dens(x = x, weights = weights, dfs = dfs, ncps = ncps)
#' plot(x, ecdf(samp)(x), main = main, ylab = "Distribution", type = "s")
#' add_approx_distr(x = x, weights = weights, dfs = dfs, ncps = ncps)
#' plot(u, quantile(samp, probs = u), type = "s", main = main,
#'      ylab = "Quantile")
#' add_approx_quant(u = u, weights = weights, dfs = dfs, ncps = ncps)
#'
#' # Case 3: \sum_{k = 1}^K k^(-3) * ChiSq_{5k}(1 / k^2)
#' K <- 1e2
#' weights<- 1 / (1:K)^3
#' dfs <- 5 * 1:K
#' ncps <- 1 / (1:K)^2
#' x <- seq(0, 25, l = 100)
#' main <- substitute(sum(k^(-3) * chi[5 * k]^2 * (1 / k^2), k == 1, K),
#'                    list(K = K))
#' samp <- r_wschisq(n = 1e4, weights = weights, dfs = dfs, ncps = ncps)
#' hist(samp, breaks = 50, freq = FALSE, main = main, ylab = "Density",
#'      xlim = range(x), xlab = "x"); box()
#' add_approx_dens(x = x, weights = weights, dfs = dfs, ncps = ncps)
#' plot(x, ecdf(samp)(x), main = main, ylab = "Distribution", type = "s")
#' add_approx_distr(x = x, weights = weights, dfs = dfs, ncps = ncps)
#' plot(u, quantile(samp, probs = u), type = "s", main = main,
#'      ylab = "Quantile")
#' add_approx_quant(u = u, weights = weights, dfs = dfs, ncps = ncps)
#' par(old_par)
#'
#' # Cutoffs for infinite series of the last example
#' K <- 1e7
#' log_weights<- -3 * log(1:K)
#' log_dfs <- log(5) + log(1:K)
#' (cutoff <- cutoff_wschisq(thre = 10^(-(1:4)), weights = log_weights,
#'                           dfs = log_dfs, log = TRUE))
#'
#' # Approximation
#' x <- seq(0, 25, l = 100)
#' l <- length(cutoff$mean)
#' main <- expression(sum(k^(-3) * chi[5 * k]^2, k == 1, K))
#' col <- viridisLite::viridis(l)
#' plot(x, d_wschisq(x, weights = exp(log_weights[1:cutoff$mean[l]]),
#'                   dfs = exp(log_dfs[1:cutoff$mean[l]])), type = "l",
#'      ylab = "Density", col = col[l], lwd = 3)
#' for(i in rev(seq_along(cutoff$mean)[-l])) {
#'   lines(x, d_wschisq(x, weights = exp(log_weights[1:cutoff$mean[i]]),
#'                      dfs = exp(log_dfs[1:cutoff$mean[i]])), col = col[i])
#' }
#' legend("topright", legend = paste0(rownames(cutoff), " (", cutoff$mean, ")"),
#'        lwd = 2, col = col)
#' }
#' @name wschisq


#' @rdname wschisq
#' @export
d_wschisq <- function(x, weights, dfs, ncps = 0,
                      method = c("I", "SW", "HBE")[1], exact_chisq = TRUE,
                      imhof_epsabs = 1e-6, imhof_epsrel = 1e-6,
                      imhof_limit = 1e4, grad_method = "simple",
                      grad_method.args = list(eps = 1e-7)) {

  # Check lengths and ncps
  stopifnot(length(weights) == length(dfs), length(ncps) %in% c(1, length(dfs)))
  stopifnot(all(ncps >= 0))

  # Vector of ncps
  l_weights <- length(weights)
  if (length(ncps) == 1) {

    ncps <- rep(ncps, l_weights)

  }

  # Remove zero weights
  zero_weights <- weights == 0
  if (any(zero_weights)) {

    weights <- weights[!zero_weights]
    dfs <- dfs[!zero_weights]
    ncps <- ncps[!zero_weights]

  }

  # Is it a sum or not?
  if (exact_chisq & l_weights == 1) {

    dens <- dchisq(x = x / weights[1], df = dfs[1], ncp = ncps[1]) / weights[1]

  } else {

    # Approximation method
    if (method == "I") {

      # Numerical derivative of the Imhof's cdf approximation
      p <- function(x) {

        p_wschisq(x = x, weights = weights, dfs = dfs, ncps = ncps,
                  method = "I", exact_chisq = exact_chisq,
                  imhof_epsabs = imhof_epsabs, imhof_epsrel = imhof_epsrel,
                  imhof_limit = imhof_limit)

      }
      dens <- numDeriv::grad(func = p, x = x, method = grad_method,
                             method.args = grad_method.args)
      dens <- pmax(dens, 0)

    } else if (method == "SW") {

      # Cumulants
      kappa1 <- sum(weights * (dfs + ncps))
      kappa2 <- 2 * sum(weights^2 * (dfs + 2 * ncps))

      # Gamma moment matching
      k <- kappa1^2 / kappa2
      theta <- kappa2 / kappa1

      # Gamma(k, theta)'s pdf
      dens <- dgamma(x = x, shape = k, scale = theta)

    } else if (method == "HBE") {

      # Cumulants
      kappa1 <- sum(weights * (dfs + ncps))
      kappa2 <- 2 * sum(weights^2 * (dfs + 2 * ncps))
      kappa3 <- 8 * sum(weights^3 * (dfs + 3 * ncps))

      # Gamma moment matching
      nu <- 8 * kappa2^3 / kappa3^2
      k <- nu / 2
      theta <- 2
      z <- sqrt(2 * nu / kappa2) * (x - kappa1) + nu

      # Gamma(k, theta)'s pdf
      dens <- dgamma(x = z, shape = k, scale = theta) * sqrt(2 * nu / kappa2)

    } else {

      stop("method must be \"I\", \"SW\", or \"HBE\".")

    }

  }

  return(dens)

}


#' @rdname wschisq
#' @export
p_wschisq <- function(x, weights, dfs, ncps = 0,
                      method = c("I", "SW", "HBE", "MC")[1], exact_chisq = TRUE,
                      imhof_epsabs = 1e-6, imhof_epsrel = 1e-6,
                      imhof_limit = 1e4, M = 1e4, MC_sample = NULL) {

  # Check lengths and ncps
  stopifnot(length(weights) == length(dfs), length(ncps) %in% c(1, length(dfs)))
  stopifnot(all(ncps >= 0))

  # Vector of ncps
  l_weights <- length(weights)
  if (length(ncps) == 1) {

    ncps <- rep(ncps, l_weights)

  }

  # Remove zero weights
  zero_weights <- weights == 0
  if (any(zero_weights)) {

    weights <- weights[!zero_weights]
    dfs <- dfs[!zero_weights]
    ncps <- ncps[!zero_weights]

  }

  # Is it a sum or not?
  if (exact_chisq & l_weights == 1) {

    prob <- pchisq(q = x / weights[1], df = dfs[1], ncp = ncps[1])

  } else {

    # Approximation method
    if (method == "I") {

      # Imhof's method
      prob <- numeric(length(x))
      prob[x == Inf] <- 1
      ind_safe <- x > 0 & is.finite(x)
      if (any(ind_safe)) {

        prob[ind_safe] <- 1 - unlist(sapply(x[ind_safe], CompQuadForm::imhof,
                                            lambda = weights, h = dfs,
                                            delta = ncps, epsabs = imhof_epsabs,
                                            epsrel = imhof_epsrel,
                                            limit = imhof_limit)[1, ])

      }

      # Avoid numerical problems
      prob <- pmin(pmax(prob, 0), 1)

    } else if (method == "SW") {

      # Cumulants
      kappa1 <- sum(weights * (dfs + ncps))
      kappa2 <- 2 * sum(weights^2 * (dfs + 2 * ncps))

      # Gamma moment matching
      k <- kappa1^2 / kappa2
      theta <- kappa2 / kappa1

      # Gamma(k, theta)'s cdf
      prob <- pgamma(q = x, shape = k, scale = theta)

    } else if (method == "HBE") {

      # Cumulants
      kappa1 <- sum(weights * (dfs + ncps))
      kappa2 <- 2 * sum(weights^2 * (dfs + 2 * ncps))
      kappa3 <- 8 * sum(weights^3 * (dfs + 3 * ncps))

      # Gamma moment matching
      nu <- 8 * kappa2^3 / kappa3^2
      k <- nu / 2
      theta <- 2
      z <- sqrt(2 * nu / kappa2) * (x - kappa1) + nu

      # Gamma(k, theta)'s cdf
      prob <- pgamma(q = z, shape = k, scale = theta)

    } else if (method == "MC") {

      # Monte Carlo approximation with efficient ecdf evaluation
      if (!is.null(MC_sample)) {

        prob <- p_wschisq_MC(x = x, dfs = 0, weights = 0, ncps = 0, M = M,
                             sample = MC_sample, use_sample = TRUE,
                             x_sorted = !is.unsorted(x))

      } else {

        prob <- p_wschisq_MC(x = x, weights = weights, dfs = dfs, ncps = ncps,
                             M = M, sample = 0, use_sample = FALSE,
                             x_sorted = !is.unsorted(x))

      }

    } else {

      stop("method must be \"I\", \"SW\", \"HBE\", or \"MC\".")

    }

  }

  return(prob)

}


#' @rdname wschisq
#' @export
q_wschisq <- function(u, weights, dfs, ncps = 0,
                      method = c("I", "SW", "HBE", "MC")[1], exact_chisq = TRUE,
                      imhof_epsabs = 1e-6, imhof_epsrel = 1e-6,
                      imhof_limit = 1e4, nlm_gradtol = 1e-6, nlm_iterlim = 1e3,
                      M = 1e4, MC_sample = NULL) {

  # Check lengths and ncps
  stopifnot(length(weights) == length(dfs), length(ncps) %in% c(1, length(dfs)))
  stopifnot(all(ncps >= 0))

  # Vector of ncps
  l_weights <- length(weights)
  if (length(ncps) == 1) {

    ncps <- rep(ncps, l_weights)

  }
  # Remove zero weights
  zero_weights <- weights == 0
  if (any(zero_weights)) {

    weights <- weights[!zero_weights]
    dfs <- dfs[!zero_weights]
    ncps <- ncps[!zero_weights]

  }


  # Is it a sum or not?
  if (exact_chisq & l_weights == 1) {

    q <- weights[1] * qchisq(p = u, df = dfs[1], ncp = ncps[1])

  } else {

    # Approximation method
    if (method == "SW") {

      # Cumulants
      kappa1 <- sum(weights * (dfs + ncps))
      kappa2 <- 2 * sum(weights^2 * (dfs + 2 * ncps))

      # Gamma moment matching
      k <- kappa1^2 / kappa2
      theta <- kappa2 / kappa1

      # Gamma(k, theta)'s quantile function
      q <- qgamma(p = u, shape = k, scale = theta)

    } else if (method == "HBE") {

      # Cumulants
      kappa1 <- sum(weights * (dfs + ncps))
      kappa2 <- 2 * sum(weights^2 * (dfs + 2 * ncps))
      kappa3 <- 8 * sum(weights^3 * (dfs + 3 * ncps))

      # Gamma moment matching
      nu <- 8 * kappa2^3 / kappa3^2
      k <- nu / 2
      theta <- 2

      # Gamma(k, theta)'s quantile function
      z <- qgamma(p = u, shape = k, scale = theta)
      q <- (z - nu) / sqrt(2 * nu / kappa2) + kappa1

    } else if (method %in% c("I", "MC")) {

      # Monte Carlo approximation
      if (is.null(MC_sample)) {

        # For Imhof, do a less demanding Monte Carlo to get good starting
        # values for inverting the approximated cdf
        if (method == "I") {

          M <- 1e3

        }
        MC_sample <- r_wschisq(n = M, weights = weights, dfs = dfs, ncps = ncps)

      }
      q <- drop(quantile(MC_sample, probs = u))

      if (method == "I") {

        # Work in safe log-scale for calling afterwards nlm()
        q <- pmin(pmax(log(q), -6), 6)

        # Numerical inversion of the Imhof's cdf approximation
        F_v <- function(x, v) {

          (p_wschisq(x = exp(x), weights = weights, dfs = dfs, ncps = ncps,
                     imhof_epsabs = imhof_epsabs, imhof_epsrel = imhof_epsrel,
                     imhof_limit = imhof_limit) - v)^2

        }
        q <- sapply(seq_along(u), function(i) {
          exp(nlm(f = F_v, p = q[i], v = u[i], fscale = 0,
                  gradtol = nlm_gradtol, iterlim = nlm_iterlim)$estimate)
        })

      }

    } else {

      stop("method must be \"I\", \"SW\", \"HBE\", or \"MC\".")

    }

  }

  return(q)

}


#' @rdname wschisq
#' @export
r_wschisq <- function(n, weights, dfs, ncps = 0) {

  # Check lengths and ncps
  stopifnot(length(weights) == length(dfs), length(ncps) %in% c(1, length(dfs)))
  stopifnot(all(ncps >= 0))

  # Vector of ncps
  l_weights <- length(weights)
  if (length(ncps) == 1) {

    ncps <- rep(ncps, l_weights)

  }

  # Remove zero weights
  zero_weights <- weights == 0
  if (any(zero_weights)) {

    weights <- weights[!zero_weights]
    dfs <- dfs[!zero_weights]
    ncps <- ncps[!zero_weights]

  }

  # Sample
  samp <- numeric(n)
  for (i in seq_along(weights)) {

    samp <- samp + weights[i] * rchisq(n = n, df = dfs[i], ncp = ncps[i])

  }
  return(samp)

}


#' @rdname wschisq
#' @export
cutoff_wschisq <- function(thre = 1e-4, weights, dfs, ncps = 0, log = FALSE,
                           x_tail = NULL) {

  # thre must be in [0, 1]
  stopifnot(all(0 <= thre & thre <= 1))

  # Check lengths and ncps
  l_weights <- length(weights)
  stopifnot(l_weights == length(dfs), length(ncps) %in% c(1, length(dfs)))
  stopifnot(all(ncps >= 0))

  # Vector of ncps
  l_weights <- length(weights)
  if (length(ncps) == 1) {

    ncps <- rep(ncps, l_weights)

  }

  # Remove NAs and display message
  if (anyNA(weights) || anyNA(dfs) || anyNA(ncps)) {

    trunc <- which(is.na(weights) | is.na(dfs) | is.na(ncps))[1]
    if (trunc > 1) {

      weights <- weights[1:(trunc - 1)]
      dfs <- dfs[1:(trunc - 1)]
      ncps <- ncps[1:(trunc - 1)]
      message("NAs present in weights, dfs, or ncps. ",
              "Truncated from ", l_weights,
              " elements to the last common non-NA entry (",
              trunc - 1, ").")
      l_weights <- trunc - 1

    } else {

      stop("No weights, dfs, or ncps without NAs.")

    }

  }

  # No thresholding at all
  if (all(thre == 0)) {

    cutoffs <- t(sapply(thre, function(r) {
      c("prob" = l_weights, "mean" = l_weights, "var" = l_weights)
    }))
    rownames(cutoffs) <- thre
    return(as.data.frame(cutoffs))

  }

  # If x_tail is not given, set it to the 0.90 quantile
  stopifnot(length(x_tail) <= 1)
  if (is.null(x_tail)) {

    # Avoid overflows
    w <- switch(log + 1, weights, exp(weights))
    d <- switch(log + 1, dfs, exp(dfs))
    fin <- which(!(is.finite(w) & is.finite(d)))
    fin <- ifelse(length(fin) == 0, l_weights, fin[1] - 1)

    # Quantile
    x_tail <- q_wschisq(u = 0.90, weights = w[1:fin], dfs = d[1:fin],
                        ncps = ncps[1:fin], method = "HBE")

  }

  # Cumulants
  kappa1 <- sum(weights * (dfs + ncps))
  kappa2 <- 2 * sum(weights^2 * (dfs + 2 * ncps))
  kappa3 <- 8 * sum(weights^3 * (dfs + 3 * ncps))

  # Gamma moment matching
  kappa1 <- cumsum(switch(log + 1, weights * (dfs + ncps),
                          exp(weights + log(exp(dfs) + exp(ncps)))))
  kappa2 <- 2 * cumsum(switch(log + 1, weights^2 * (dfs + 2 * ncps),
                              exp(2 * weights + log(exp(dfs) + 2 * exp(ncps)))))
  kappa3 <- 8 * cumsum(switch(log + 1, weights^3 * (dfs + 3 * ncps),
                              exp(3 * weights + log(exp(dfs) + 3 * exp(ncps)))))
  nu <- 8 * kappa2^3 / kappa3^2
  k <- nu / 2
  theta <- 2
  z <- sqrt(2 * nu / kappa2) * (x_tail - kappa1) + nu

  # Tail probabilities using the HBE method
  prob_HBE <- 1 - pgamma(q = z, shape = k, scale = theta)
  dist_prob <- abs(prob_HBE - prob_HBE[l_weights])

  # Proportion of cumulated mean
  means <- switch(log + 1, weights * (dfs + ncps),
                  exp(weights + log(exp(dfs) + exp(ncps))))
  prop_mean <- cumsum(means) / sum(means, na.rm = TRUE)

  # Proportion of cumulated variance
  vars <- switch(log + 1, weights^2 * 2 * (dfs + 2 * ncps),
                 exp(2 * weights + log(2) + log(exp(dfs) + 2 * exp(ncps))))
  prop_var <- cumsum(vars) / sum(vars, na.rm = TRUE)

  # Cutoffs
  cutoffs <- t(sapply(thre, function(r) {
    c("prob" = which(dist_prob <= r)[1],
      "mean" = which(prop_mean >= 1 - r)[1],
      "var" = which(prop_var >= 1 - r)[1])
  }))
  rownames(cutoffs) <- thre
  return(as.data.frame(cutoffs))

}

set.seed(14323)
K <- rpois(n = 1, lambda = 9) + 1
x <- seq(0, 3, l = 10)
weights <- abs(rnorm(n = K))
weights0 <- c(K:2, 0)
weights1 <- c(K:2, 1e-15)
dfs <- rpois(n = K, lambda = 4) + 1
ncps <- runif(n = K, min = 0, max = 2)^2
M <- 1e4

test_that("r_wschisq vs. r_wschisq_Cpp and definition", {

  samp1 <- r_wschisq(n = M, weights = weights, dfs = dfs, ncps = ncps)
  samp2 <- r_wschisq_Cpp(n = M, weights = weights, dfs = dfs, ncps = ncps)
  expect_gt(ks.test(samp1, samp2)$p.value, 0.10)
  samp3 <- colSums(weights * t(sapply(1:K, function(i) {
    lambda <- sqrt(ncps[i] / dfs[i])
    rowSums(matrix((rnorm(n = M * dfs[i]) + lambda)^2,
                   nrow = M, ncol = dfs[i]))
  })))
  expect_gt(ks.test(samp1, samp3)$p.value, 0.10)

})

test_that("Wrong method", {

  expect_error(d_wschisq(x = 1, weights = weights, dfs = dfs, method = "A"))
  expect_error(p_wschisq(x = 1, weights = weights, dfs = dfs, method = "A"))
  expect_error(q_wschisq(u = 0.5, weights = weights, dfs = dfs, method = "A"))

})

test_that("Single chi square", {

  expect_equal(d_wschisq(x = 1, weights = 2, dfs = 3, ncps = 1),
               drop(d_chisq(x = 1 / 2, df = 3, ncp = 1) / 2))
  expect_equal(p_wschisq(x = 1, weights = 2, dfs = 3, ncps = 1),
               drop(p_chisq(x = 1 / 2, df = 3, ncp = 1)))
  expect_equal(q_wschisq(u = 0.5, weights = 2, dfs = 3, ncps = 1),
               drop(2 * qchisq(p = 0.5, df = 3, ncp = 1)))

})

test_that("Zero weights method", {

  expect_equal(p_wschisq(x = 2:1, weights = weights0, dfs = dfs, ncps = ncps,
                         method = "HBE"),
               p_wschisq(x = 2:1, weights = weights1, dfs = dfs, ncps = ncps,
                         method = "HBE"),
               tolerance = 1e-6)
  expect_equal(q_wschisq(u = 0.5, weights = weights0, dfs = dfs, ncps = ncps,
                         method = "HBE"),
               q_wschisq(u = 0.5, weights = weights1, dfs = dfs, ncps = ncps,
                         method = "HBE"),
               tolerance = 1e-6)
  expect_equal(d_wschisq(x = 2:1, weights = weights0, dfs = dfs, ncps = ncps,
                         method = "HBE"),
               d_wschisq(x = 2:1, weights = weights1, dfs = dfs, ncps = ncps,
                         method = "HBE"))
  expect_equal({set.seed(1); r_wschisq(n = 5, weights = c(5, 0),
                                       dfs = 2:1, ncps = 2:1)},
               {set.seed(1); r_wschisq(n = 5, weights = c(5, 1e-10),
                                       dfs = 2:1, ncps = 2:1)})

})

test_that("p_wschisq methods", {

  expect_lt(max(abs(p_wschisq(x = x, weights = weights, dfs = dfs,
                              ncps = ncps, method = "HBE") -
                      p_wschisq(x = x, weights = weights, dfs = dfs,
                                ncps = ncps, method = "MC")
  )), 1e-5)
  expect_lt(max(abs(p_wschisq(x = x, weights = weights, dfs = dfs,
                              ncps = ncps, method = "MC") -
                      p_wschisq(x = x, weights = weights, dfs = dfs,
                                ncps = ncps, method = "MC", MC_sample =
                                  r_wschisq(n = 1e4, weights = weights,
                                            dfs = dfs, ncps = ncps))
  )), 1e-5)
  expect_lt(max(abs(p_wschisq(x = x, weights = weights, dfs = dfs,
                              ncps = ncps, method = "HBE") -
                      p_wschisq(x = x, weights = weights, dfs = dfs,
                                ncps = ncps, method = "SW")
  )), 1e-5)

})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' @title Surface area of the intersection of two hyperspherical caps
#'
#' @description Computation of
#' \deqn{A_x(\theta_{ij}) := \frac{1}{\omega_p}
#' \int_{S^{p - 1}} 1_{\{{\bf X}_i'\boldsymbol\gamma \le x,
#' {\bf X}_j'\boldsymbol\gamma \le x\}}\,\mathrm{d}\boldsymbol\gamma,}{
#' A_x(\theta_{ij}) := \frac{1}{\omega_p} \int_{S^{p - 1}}
#' 1_{X_i'\gamma \le x, X_j'\gamma \le x} d\gamma,}
#' where \eqn{\theta_{ij} := \cos^{-1}({\bf X}_i'{\bf X}_j)
#' \in [0, \pi]}{\theta_{ij} := \cos^{-1}(X_i'X_j) \in [0, \pi]},
#' \eqn{x \in [-1, 1]}, and \eqn{\omega_{p}} is the surface area of
#' \eqn{S^{p - 1}}. \eqn{A_x(\theta_{ij})} is the proportion of surface area
#' of \eqn{S^{p - 1}} covered by the intersection of two hyperspherical caps
#' centered at \eqn{{\bf X}_i}{X_i} and \eqn{{\bf X}_j}{X_j} and with
#' common solid angle \eqn{\pi - \cos^{-1}(x)}.
#'
#' @inheritParams Gegenbauer
#' @param x vector with values in \eqn{[-1, 1]}.
#' @inheritParams r_unif
#' @param N number of points used in the
#' \link[=Gauss_Legen_nodes]{Gauss-Legendre quadrature}. Defaults to
#' \code{160}.
#' @param as_matrix return a matrix with the values of \eqn{A_x(\theta)} on
#' the grid formed by \code{theta} and \code{x}? If \code{FALSE},
#' \eqn{A_x(\theta)} is evaluated on \code{theta} and \code{x} if they equal
#' in size. Defaults to \code{TRUE}.
#' @return A matrix of size \code{c(length(theta), length(x))} containing the
#' evaluation of \eqn{A_x(\theta)} if \code{as_matrix = TRUE}. Otherwise,
#' a vector of size \code{c(length(theta)} if \code{theta} and \code{x} equal
#' in size.
#' @details
#' See García-Portugués et al. (2020) for more details about the
#' \eqn{A_x(\theta)} function.
#' @references
#' García-Portugués, E., Navarro-Esteban, P., Cuesta-Albertos, J. A. (2020)
#' On a projection-based class of uniformity tests on the hypersphere.
#' \emph{arXiv:2008.09897}. \url{https://arxiv.org/abs/2008.09897}
#' @examples
#' # Plot A_x(theta) for several dimensions and x's
#' A_lines <- function(x, th = seq(0, pi, l = 200)) {
#'
#'   plot(th, A_theta_x(theta = th, x = x, p = 2), type = "l",
#'        col = 1, ylim = c(0, 1.25), main = paste("x =", x),
#'        ylab = expression(A[x](theta)),
#'        xlab = expression(theta), axes = FALSE)
#'   axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
#'        labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
#'   axis(2); box()
#'   abline(h = c(0, 1), lty = 2)
#'   lines(th, A_theta_x(theta = th, x = x, p = 3), col = 2)
#'   lines(th, A_theta_x(theta = th, x = x, p = 4), col = 3)
#'   lines(th, A_theta_x(theta = th, x = x, p = 5), col = 4)
#'   legend("top", lwd = 2, legend = paste("p =", 2:5),
#'          col = 1:4, cex = 0.75, horiz = TRUE)
#'
#' }
#' old_par <- par(mfrow = c(2, 3))
#' A_lines(x = -0.75)
#' A_lines(x = -0.25)
#' A_lines(x = 0)
#' A_lines(x = 0.25)
#' A_lines(x = 0.5)
#' A_lines(x = 0.75)
#' par(old_par)
#'
#' # As surface of (theta, x) for several dimensions
#' A_surf <- function(p, x = seq(-1, 1, l = 201), th = seq(0, pi, l = 201)) {
#'
#'   col <- c("white", viridisLite::viridis(20))
#'   breaks <- c(-1, seq(1e-15, 1, l = 21))
#'   A <- A_theta_x(theta = th, x = x, p = p)
#'   image(th, x, A, main = paste("p =", p), col = col, breaks = breaks,
#'         xlab = expression(theta), axes = FALSE)
#'   axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
#'        labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
#'   axis(2); box()
#'   contour(th, x, A, levels = breaks, add = TRUE)
#'
#' }
#' old_par <- par(mfrow = c(2, 2))
#' A_surf(p = 2)
#' A_surf(p = 3)
#' A_surf(p = 4)
#' A_surf(p = 5)
#' par(old_par)
#'
#' # No matrix return
#' th <- seq(0, pi, l = 5)
#' x <- seq(-1, 1, l = 5)
#' diag(A_theta_x(theta = th, x = x, p = 2))
#' A_theta_x(theta = th, x = x, p = 2, as_matrix = FALSE)
#' @export
A_theta_x <- function(theta, x, p, N = 160L, as_matrix = TRUE) {
    .Call('_sphunif_A_theta_x', PACKAGE = 'sphunif', theta, x, p, N, as_matrix)
}

#' @title Shortest angles matrix
#'
#' @description Efficient computation of the shortest angles matrix
#' \eqn{\boldsymbol\Psi}{\Psi}, defined as
#' \deqn{\Psi_{ij}:=\cos^{-1}({\bf X}_i'{\bf X}_j),\quad
#' i,j=1,\ldots,n,}{\Psi_{ij} = \cos^{-1}(X_i'X_j), i, j = 1, \ldots, n,}
#' for a sample \eqn{{\bf X}_1,\ldots,{\bf X}_n\in S^{p-1}:=\{{\bf x}\in
#' R^p:||{\bf x}||=1\}}{X_1, \ldots, X_n \in
#' S^{p - 1} := \{x \in R^p : ||x|| = 1\}}, \eqn{p\ge 2}.
#'
#' For a circular sample \eqn{\Theta_1, \ldots, \Theta_n \in [0, 2\pi)},
#' \eqn{\boldsymbol\Psi}{\Psi} can be expressed as
#' \deqn{\Psi_{ij}=\pi-|\pi-|\Theta_i-\Theta_j||,\quad
#' i,j=1,\ldots,n.}{\Psi_{ij}=\pi-|\pi-|\Theta_i-\Theta_j||, i,j=1,\ldots,n.}
#'
#' @param data an array of size \code{c(n, p, M)} containing the Cartesian
#' coordinates of \code{M} samples of size \code{n} of directions on
#' \eqn{S^{p-1}}. Alternatively if \code{p = 2}, an array of size
#' \code{c(n, 1, M)} containing the angles on \eqn{[0, 2\pi)} of the \code{M}
#' circular samples of size \code{n} on \eqn{S^{1}}. Must not contain
#' \code{NA}'s.
#' @param ind_tri if \code{use_ind_tri = TRUE}, the vector of 0-based indexes
#' provided by \code{upper_tri_ind(n)}, which allows to extract the upper
#' triangular part of the matrix \eqn{\boldsymbol\Psi}{\Psi}. See the examples.
#' @param use_ind_tri use the already computed vector index \code{ind_tri}? If
#' \code{FALSE} (default), \code{ind_tri} is computed internally.
#' @param scalar_prod return the scalar products
#' \eqn{{\bf X}_i'{\bf X}}{X_i'X_j} instead of the shortest angles? Only taken
#' into account for data in \emph{Cartesian} form. Defaults to
#' \code{FALSE}.
#' @param angles_diff return the (unwrapped) angles difference
#' \eqn{\Theta_i-\Theta_j} instead of the shortest angles? Only taken into
#' account for data in \emph{angular} form. Defaults to \code{FALSE}.
#' @param n sample size, used to determine the index vector that gives the
#' upper triangular part of \eqn{\boldsymbol\Psi}{\Psi}.
#' @return
#' \itemize{
#'   \item \code{Psi_mat}: a matrix of size
#'   \code{c(n * (n - 1) / 2, M)} containing, for each column, the vector
#'   half of \eqn{\boldsymbol\Psi}{\Psi} for each of the \code{M} samples.
#'   \item \code{upper_tri_ind}: a matrix of size \code{n * (n - 1) / 2}
#'   containing the 0-based linear indexes for extracting the upper triangular
#'   matrix of a matrix of size \code{c(n, n)}, diagonal excluded, assuming
#'   column-major order.
#' }
#' @section Warning:
#' Be careful on avoiding the next bad usages of \code{Psi_mat}, which will
#' produce spurious results:
#' \itemize{
#'   \item The directions in \code{data} do \emph{not} have unit norm when
#'   Cartesian coordinates are employed.
#'   \item The entries of \code{data} are \emph{not} in \eqn{[0, 2\pi)} when
#'   polar coordinates are employed.
#'   \item \code{ind_tri} is a vector of size \code{n * (n - 1) / 2} that
#'   does \emph{not} contain the indexes produced by \code{upper_tri_ind(n)}.
#' }
#' @examples
#' # Shortest angles
#' n <- 5
#' X <- r_unif_sph(n = n, p = 2, M = 2)
#' Theta <- X_to_Theta(X)
#' dim(Theta) <- c(n, 1, 2)
#' Psi_mat(X)
#' Psi_mat(Theta)
#'
#' # Precompute ind_tri
#' ind_tri <- upper_tri_ind(n)
#' Psi_mat(X, ind_tri = ind_tri, use_ind_tri = TRUE)
#'
#' # Compare with R
#' A <- acos(tcrossprod(X[, , 1]))
#' ind <- upper.tri(A)
#' A[ind]
#'
#' # Reconstruct matrix
#' Psi_vec <- Psi_mat(Theta[, , 1, drop = FALSE])
#' Psi <- matrix(0, nrow = n, ncol = n)
#' Psi[upper.tri(Psi)] <- Psi_vec
#' Psi <- Psi + t(Psi)
#' @name Psi
NULL

#' @rdname Psi
#' @export
Psi_mat <- function(data, ind_tri = 0L, use_ind_tri = FALSE, scalar_prod = FALSE, angles_diff = FALSE) {
    .Call('_sphunif_Psi_mat', PACKAGE = 'sphunif', data, ind_tri, use_ind_tri, scalar_prod, angles_diff)
}

#' @rdname Psi
#' @export
upper_tri_ind <- function(n) {
    .Call('_sphunif_upper_tri_ind', PACKAGE = 'sphunif', n)
}

#' @title Sort the columns of a matrix
#'
#' @description Convenience functions to sort the columns of a matrix in an
#' increasing way.
#'
#' @param A a matrix of arbitrary dimensions.
#' @return A matrix with the same dimensions as \code{A} such that each
#' column is sorted increasingly (for \code{sort_each_col}) or contains the
#' sorting indexes (for \code{sort_index_each_col}).
#' \code{sort_index_each_col}.
#' @keywords internal
sort_each_col <- function(A) {
    .Call('_sphunif_sort_each_col', PACKAGE = 'sphunif', A)
}

#' @rdname sort_each_col
#' @keywords internal
sort_index_each_col <- function(A) {
    .Call('_sphunif_sort_index_each_col', PACKAGE = 'sphunif', A)
}

#' @title Transforming between polar and Cartesian coordinates
#'
#' @description Transformation between a matrix \code{Theta} containing
#' \code{M} circular samples of size \code{n} on \eqn{[0, 2\pi)} and an array
#' \code{X} containing the associated Cartesian coordinates on
#' \eqn{S^1:=\{{\bf x}\in R^2:||{\bf x}||=1\}}{S^1:=\{x\in R^2:||x||=1\}}.
#'
#' @inheritParams cir_stat
#' @param X an \bold{array} of size \code{c(n, 2, M)} containing the Cartesian
#' coordinates of \code{M} samples of size \code{n} of directions on
#' \eqn{S^{1}}. Must not contain \code{NA}'s.
#' @return
#' \itemize{
#'   \item \code{Theta_to_X}: the corresponding \code{X}.
#'   \item \code{X_to_Theta}: the corresponding \code{Theta}.
#' }
#' @examples
#' # Sample
#' Theta <- r_unif_cir(n = 10, M = 2)
#' X <- r_unif_sph(n = 10, p = 2, M = 2)
#'
#' # Check equality
#' sum(abs(X - Theta_to_X(X_to_Theta(X))))
#' sum(abs(Theta - X_to_Theta(Theta_to_X(Theta))))
#' @name cir_coord_conv
NULL

#' @title Low-level utilities for \pkg{sphunif}
#'
#' @description Internal and undocumented low-level utilities for
#' \pkg{sphunif}.
#'
#' @param n_dist a positive integer \eqn{(n - 1) * n / 2} for which \eqn{n}
#' is to be recovered.
#' @param t a vector to evaluate \eqn{t / \sqrt{1 - t^2}}.
#' @name utils
NULL

#' @rdname cir_coord_conv
#' @export
Theta_to_X <- function(Theta) {
    .Call('_sphunif_Theta_to_X', PACKAGE = 'sphunif', Theta)
}

#' @rdname cir_coord_conv
#' @export
X_to_Theta <- function(X) {
    .Call('_sphunif_X_to_Theta', PACKAGE = 'sphunif', X)
}

#' @title Circular gaps
#'
#' @description Computation of the circular gaps of an angular sample
#' \eqn{\Theta_1,\ldots,\Theta_n} on \eqn{[0, 2\pi)}, defined as
#' \deqn{\Theta_{(2)} - \Theta_{(1)},\ldots,\Theta_{(n)} - \Theta_{(n - 1)},
#' 2\pi - \Theta_{(n)} - \Theta_{(1)},}
#' where
#' \deqn{0 \le \Theta_{(1)} \le \Theta_{(2)} \le \ldots \le
#' \Theta_{(n)} \le 2\pi.}
#'
#' @inheritParams cir_stat
#' @param sorted are the columns of \code{Theta} sorted increasingly? If
#' \code{TRUE}, performance is improved. If \code{FALSE} (default), each
#' column of \code{Theta} is sorted internally.
#' @return A matrix of size \code{c(n, M)} containing the \code{n} circular
#' gaps for each of the \code{M} circular samples.
#' @section Warning:
#' Be careful on avoiding the next bad usages of \code{cir_gaps}, which will
#' produce spurious results:
#' \itemize{
#'   \item The entries of \code{Theta} are \emph{not} in \eqn{[0, 2\pi)}.
#'   \item \code{Theta} is \emph{not} sorted increasingly when
#'   \code{data_sorted = TRUE}.
#' }
#' @examples
#' Theta <- cbind(c(pi, 0, 3 * pi / 2), c(0, 3 * pi / 2, pi), c(5, 3, 1))
#' cir_gaps(Theta)
#' @export
cir_gaps <- function(Theta, sorted = FALSE) {
    .Call('_sphunif_cir_gaps', PACKAGE = 'sphunif', Theta, sorted)
}

#' @title Efficient evaluation of the empirical cumulative distribution
#' function
#'
#' @description Evaluates the empirical cumulative distribution function
#' (ecdf) of a sample \code{data} at the evaluation points \code{sorted_x}.
#' This is done through binary search.
#'
#' @param data a vector or column matrix containing the sample.
#' @param sorted_x a vector or column matrix with the evaluation points
#' \bold{sorted increasingly}.
#' @param data_sorted is \code{data} is already sorted increasingly?
#' This avoids sorting the data internally.
#' @param efic use the more efficient version of the ecdf evaluation? Set to
#' \code{FALSE} only for debugging purposes.
#' @param divide_n if \code{FALSE}, returns the absolute frequencies instead
#' of the relative frequencies. Defaults to \code{TRUE}.
#' @return The ecdf evaluated at \code{sorted_x}.
#' @author Original code from Douglas Bates'
#' \url{https://github.com/dmbates/ecdfExample}. Minor adaptations by Eduardo
#' García-Portugués.
#' @section Warning:
#' Be careful on avoiding the next bad usages of the function, which will
#' produce spurious results:
#' \itemize{
#'   \item \code{sorted_x} is not sorted increasingly.
#'   \item \code{data} is not sorted increasingly when
#'   \code{data_sorted = TRUE}-
#' }
#' @keywords internal
ecdf_bin <- function(data, sorted_x, data_sorted = FALSE, efic = TRUE, divide_n = TRUE) {
    .Call('_sphunif_ecdf_bin', PACKAGE = 'sphunif', data, sorted_x, data_sorted, efic, divide_n)
}

#' @title The incomplete beta function and its inverse
#'
#' @description Computes the incomplete beta function
#' \deqn{I_x(a,b):=\int_0^x u^{a-1}(1-u)^{b-1}\,d\mathrm{u},\quad a,b>0}{
#' I_x(a,b):=\int_0^x u^{a-1}(1-u)^{b-1}du, a,b>0}
#' and its inverse function.
#'
#' @inheritParams cir_stat_distr
#' @param u a vector of probabilities of size \code{nu} or a matrix of size
#' \code{c(nu, 1)}.
#' @param a,b scalars giving the parameters of the beta function.
#' @param lower_tail accumulate the probability from the lower tail? If
#' \code{FALSE}, the probability is accumulated from the \emph{upper} tail.
#' Defaults to \code{FALSE}.
#' @param log use log-scale? If \code{TRUE}, returns the logarithm of the
#' incomplete beta function and uses log-scale for \code{u} in \code{beta_inc}.
#' Defaults to \code{FALSE}.
#' @return
#' \itemize{
#'   \item \code{beta_inc}: a matrix of size \code{c(nx, 1)} with the
#'   evaluation of the incomplete beta function at \code{x}.
#'   \item \code{beta_inc_inv}: a matrix of size \code{c(nu, 1)} with the
#'   evaluation of the inverse incomplete beta function at \code{u}.
#' }
#' @details
#' The functions are mere wrappers to R's internal \code{pbeta} and
#' \code{qbeta} functions.
#' @keywords internal
beta_inc <- function(x, a, b, lower_tail = TRUE, log = FALSE) {
    .Call('_sphunif_beta_inc', PACKAGE = 'sphunif', x, a, b, lower_tail, log)
}

#' @rdname beta_inc
beta_inc_inv <- function(u, a, b, lower_tail = TRUE, log = FALSE) {
    .Call('_sphunif_beta_inc_inv', PACKAGE = 'sphunif', u, a, b, lower_tail, log)
}

#' @rdname utils
#' @keywords internal
n_from_dist_vector <- function(n_dist) {
    .Call('_sphunif_n_from_dist_vector', PACKAGE = 'sphunif', n_dist)
}

#' @rdname utils
#' @keywords internal
t_inv_sqrt_one <- function(t) {
    .Call('_sphunif_t_inv_sqrt_one', PACKAGE = 'sphunif', t)
}

#' @rdname cir_stat
#' @export
cir_stat_Kuiper <- function(Theta, sorted = FALSE, KS = FALSE, Stephens = FALSE) {
    .Call('_sphunif_cir_stat_Kuiper', PACKAGE = 'sphunif', Theta, sorted, KS, Stephens)
}

#' @rdname cir_stat
#' @export
cir_stat_Watson <- function(Theta, sorted = FALSE, CvM = FALSE, Stephens = FALSE) {
    .Call('_sphunif_cir_stat_Watson', PACKAGE = 'sphunif', Theta, sorted, CvM, Stephens)
}

#' @rdname cir_stat
#' @export
cir_stat_Watson_1976 <- function(Theta, sorted = FALSE, minus = FALSE) {
    .Call('_sphunif_cir_stat_Watson_1976', PACKAGE = 'sphunif', Theta, sorted, minus)
}

#' @rdname cir_stat
#' @export
cir_stat_Range <- function(Theta, sorted = FALSE, gaps_in_Theta = FALSE, max_gap = TRUE) {
    .Call('_sphunif_cir_stat_Range', PACKAGE = 'sphunif', Theta, sorted, gaps_in_Theta, max_gap)
}

#' @rdname cir_stat
#' @export
cir_stat_Rao <- function(Theta, sorted = FALSE, gaps_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_Rao', PACKAGE = 'sphunif', Theta, sorted, gaps_in_Theta)
}

#' @rdname cir_stat
#' @export
cir_stat_Greenwood <- function(Theta, sorted = FALSE, gaps_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_Greenwood', PACKAGE = 'sphunif', Theta, sorted, gaps_in_Theta)
}

#' @rdname cir_stat
#' @export
cir_stat_Log_gaps <- function(Theta, sorted = FALSE, gaps_in_Theta = FALSE, abs_val = TRUE) {
    .Call('_sphunif_cir_stat_Log_gaps', PACKAGE = 'sphunif', Theta, sorted, gaps_in_Theta, abs_val)
}

#' @rdname cir_stat
#' @export
cir_stat_Vacancy <- function(Theta, a = 2 * pi, sorted = FALSE, gaps_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_Vacancy', PACKAGE = 'sphunif', Theta, a, sorted, gaps_in_Theta)
}

#' @rdname cir_stat
#' @export
cir_stat_Max_uncover <- function(Theta, a = 2 * pi, sorted = FALSE, gaps_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_Max_uncover', PACKAGE = 'sphunif', Theta, a, sorted, gaps_in_Theta)
}

#' @rdname cir_stat
#' @export
cir_stat_Num_uncover <- function(Theta, a = 2 * pi, sorted = FALSE, gaps_in_Theta = FALSE, minus_val = TRUE) {
    .Call('_sphunif_cir_stat_Num_uncover', PACKAGE = 'sphunif', Theta, a, sorted, gaps_in_Theta, minus_val)
}

#' @rdname cir_stat
#' @export
cir_stat_Gini <- function(Theta, sorted = FALSE, gaps_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_Gini', PACKAGE = 'sphunif', Theta, sorted, gaps_in_Theta)
}

#' @rdname cir_stat
#' @export
cir_stat_Gini_squared <- function(Theta, sorted = FALSE, gaps_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_Gini_squared', PACKAGE = 'sphunif', Theta, sorted, gaps_in_Theta)
}

#' @rdname cir_stat
#' @export
cir_stat_Ajne <- function(Theta, Psi_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_Ajne', PACKAGE = 'sphunif', Theta, Psi_in_Theta)
}

#' @keywords internal
cir_stat_An_Psi <- function(Psi, n) {
    .Call('_sphunif_cir_stat_An_Psi', PACKAGE = 'sphunif', Psi, n)
}

#' @rdname cir_stat
#' @export
cir_stat_Rothman <- function(Theta, t = 1.0 / 3.0, Psi_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_Rothman', PACKAGE = 'sphunif', Theta, t, Psi_in_Theta)
}

#' @keywords internal
cir_stat_Rothman_Psi <- function(Psi, t_m2, t_min2, n) {
    .Call('_sphunif_cir_stat_Rothman_Psi', PACKAGE = 'sphunif', Psi, t_m2, t_min2, n)
}

#' @rdname cir_stat
#' @export
cir_stat_Hodges_Ajne <- function(Theta, asymp_std = FALSE, sorted = FALSE, use_Cressie = TRUE) {
    .Call('_sphunif_cir_stat_Hodges_Ajne', PACKAGE = 'sphunif', Theta, asymp_std, sorted, use_Cressie)
}

#' @rdname cir_stat
#' @export
cir_stat_Cressie <- function(Theta, t = 1.0 / 3.0, sorted = FALSE) {
    .Call('_sphunif_cir_stat_Cressie', PACKAGE = 'sphunif', Theta, t, sorted)
}

#' @rdname cir_stat
#' @export
cir_stat_FG01 <- function(Theta, sorted = FALSE) {
    .Call('_sphunif_cir_stat_FG01', PACKAGE = 'sphunif', Theta, sorted)
}

#' @rdname cir_stat
#' @export
cir_stat_Rayleigh <- function(Theta, m = 1L) {
    .Call('_sphunif_cir_stat_Rayleigh', PACKAGE = 'sphunif', Theta, m)
}

#' @rdname cir_stat
#' @export
cir_stat_Bingham <- function(Theta) {
    .Call('_sphunif_cir_stat_Bingham', PACKAGE = 'sphunif', Theta)
}

#' @rdname cir_stat
#' @export
cir_stat_Hermans_Rasson <- function(Theta, Psi_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_Hermans_Rasson', PACKAGE = 'sphunif', Theta, Psi_in_Theta)
}

#' @keywords internal
cir_stat_Hermans_Rasson_Psi <- function(Psi, n) {
    .Call('_sphunif_cir_stat_Hermans_Rasson_Psi', PACKAGE = 'sphunif', Psi, n)
}

#' @rdname cir_stat
#' @export
cir_stat_Gine_Gn <- function(Theta, Psi_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_Gine_Gn', PACKAGE = 'sphunif', Theta, Psi_in_Theta)
}

#' @rdname cir_stat
#' @export
cir_stat_Gine_Fn <- function(Theta, Psi_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_Gine_Fn', PACKAGE = 'sphunif', Theta, Psi_in_Theta)
}

#' @rdname cir_stat
#' @export
cir_stat_Pycke <- function(Theta, Psi_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_Pycke', PACKAGE = 'sphunif', Theta, Psi_in_Theta)
}

#' @keywords internal
cir_stat_Pycke_Psi <- function(Psi, n) {
    .Call('_sphunif_cir_stat_Pycke_Psi', PACKAGE = 'sphunif', Psi, n)
}

#' @rdname cir_stat
#' @export
cir_stat_Pycke_q <- function(Theta, Psi_in_Theta = FALSE, q = 0.5) {
    .Call('_sphunif_cir_stat_Pycke_q', PACKAGE = 'sphunif', Theta, Psi_in_Theta, q)
}

#' @keywords internal
cir_stat_Pycke_q_Psi <- function(Psi, n, q = 0.5) {
    .Call('_sphunif_cir_stat_Pycke_q_Psi', PACKAGE = 'sphunif', Psi, n, q)
}

#' @rdname cir_stat
#' @export
cir_stat_Bakshaev <- function(Theta, Psi_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_Bakshaev', PACKAGE = 'sphunif', Theta, Psi_in_Theta)
}

#' @rdname cir_stat
#' @export
cir_stat_Riesz <- function(Theta, Psi_in_Theta = FALSE, s = 1.0) {
    .Call('_sphunif_cir_stat_Riesz', PACKAGE = 'sphunif', Theta, Psi_in_Theta, s)
}

#' @rdname cir_stat
#' @export
cir_stat_PCvM <- function(Theta, Psi_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_PCvM', PACKAGE = 'sphunif', Theta, Psi_in_Theta)
}

#' @rdname cir_stat
#' @export
cir_stat_PRt <- function(Theta, t = 1.0 / 3.0, Psi_in_Theta = FALSE) {
    .Call('_sphunif_cir_stat_PRt', PACKAGE = 'sphunif', Theta, t, Psi_in_Theta)
}

#' @rdname cir_stat
#' @export
cir_stat_PAD <- function(Theta, Psi_in_Theta = FALSE, AD = FALSE, sorted = FALSE) {
    .Call('_sphunif_cir_stat_PAD', PACKAGE = 'sphunif', Theta, Psi_in_Theta, AD, sorted)
}

#' @rdname cir_stat
#' @export
cir_stat_CCF09 <- function(Theta, dirs, K_CCF09 = 25L, original = FALSE) {
    .Call('_sphunif_cir_stat_CCF09', PACKAGE = 'sphunif', Theta, dirs, K_CCF09, original)
}

#' @rdname cir_stat_distr
#' @export
p_Kolmogorov <- function(x, K_Kolmogorov = 25L, alternating = TRUE) {
    .Call('_sphunif_p_Kolmogorov', PACKAGE = 'sphunif', x, K_Kolmogorov, alternating)
}

#' @rdname cir_stat_distr
#' @export
d_Kolmogorov <- function(x, K_Kolmogorov = 25L, alternating = TRUE) {
    .Call('_sphunif_d_Kolmogorov', PACKAGE = 'sphunif', x, K_Kolmogorov, alternating)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Ajne <- function(x, K_Ajne = 15L) {
    .Call('_sphunif_p_cir_stat_Ajne', PACKAGE = 'sphunif', x, K_Ajne)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Ajne <- function(x, K_Ajne = 15L) {
    .Call('_sphunif_d_cir_stat_Ajne', PACKAGE = 'sphunif', x, K_Ajne)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Bingham <- function(x) {
    .Call('_sphunif_p_cir_stat_Bingham', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Bingham <- function(x) {
    .Call('_sphunif_d_cir_stat_Bingham', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Greenwood <- function(x) {
    .Call('_sphunif_p_cir_stat_Greenwood', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Greenwood <- function(x) {
    .Call('_sphunif_d_cir_stat_Greenwood', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Gini <- function(x) {
    .Call('_sphunif_p_cir_stat_Gini', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Gini <- function(x) {
    .Call('_sphunif_d_cir_stat_Gini', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Gini_squared <- function(x) {
    .Call('_sphunif_p_cir_stat_Gini_squared', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Gini_squared <- function(x) {
    .Call('_sphunif_d_cir_stat_Gini_squared', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Hodges_Ajne2 <- function(x, n, asymp_std = FALSE) {
    .Call('_sphunif_p_cir_stat_Hodges_Ajne2', PACKAGE = 'sphunif', x, n, asymp_std)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Hodges_Ajne <- function(x, n, exact = TRUE, asymp_std = FALSE) {
    .Call('_sphunif_p_cir_stat_Hodges_Ajne', PACKAGE = 'sphunif', x, n, exact, asymp_std)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Hodges_Ajne <- function(x, n, exact = TRUE, asymp_std = FALSE) {
    .Call('_sphunif_d_cir_stat_Hodges_Ajne', PACKAGE = 'sphunif', x, n, exact, asymp_std)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Kuiper <- function(x, n, K_Kuiper = 12L, second_term = TRUE, Stephens = FALSE) {
    .Call('_sphunif_p_cir_stat_Kuiper', PACKAGE = 'sphunif', x, n, K_Kuiper, second_term, Stephens)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Kuiper <- function(x, n, K_Kuiper = 12L, second_term = TRUE, Stephens = FALSE) {
    .Call('_sphunif_d_cir_stat_Kuiper', PACKAGE = 'sphunif', x, n, K_Kuiper, second_term, Stephens)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Log_gaps <- function(x, abs_val = TRUE) {
    .Call('_sphunif_p_cir_stat_Log_gaps', PACKAGE = 'sphunif', x, abs_val)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Log_gaps <- function(x, abs_val = TRUE) {
    .Call('_sphunif_d_cir_stat_Log_gaps', PACKAGE = 'sphunif', x, abs_val)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Max_uncover <- function(x) {
    .Call('_sphunif_p_cir_stat_Max_uncover', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Max_uncover <- function(x) {
    .Call('_sphunif_d_cir_stat_Max_uncover', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Num_uncover <- function(x) {
    .Call('_sphunif_p_cir_stat_Num_uncover', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Num_uncover <- function(x) {
    .Call('_sphunif_d_cir_stat_Num_uncover', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Pycke <- function(x) {
    .Call('_sphunif_p_cir_stat_Pycke', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Pycke <- function(x) {
    .Call('_sphunif_d_cir_stat_Pycke', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Vacancy <- function(x) {
    .Call('_sphunif_p_cir_stat_Vacancy', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Vacancy <- function(x) {
    .Call('_sphunif_d_cir_stat_Vacancy', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Watson <- function(x, n = 0L, K_Watson = 25L, Stephens = FALSE) {
    .Call('_sphunif_p_cir_stat_Watson', PACKAGE = 'sphunif', x, n, K_Watson, Stephens)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Watson <- function(x, n = 0L, K_Watson = 25L, Stephens = FALSE) {
    .Call('_sphunif_d_cir_stat_Watson', PACKAGE = 'sphunif', x, n, K_Watson, Stephens)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Watson_1976 <- function(x, K_Watson_1976 = 8L, N = 40L) {
    .Call('_sphunif_p_cir_stat_Watson_1976', PACKAGE = 'sphunif', x, K_Watson_1976, N)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Watson_1976 <- function(x, K_Watson_1976 = 8L) {
    .Call('_sphunif_d_cir_stat_Watson_1976', PACKAGE = 'sphunif', x, K_Watson_1976)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Range <- function(x, n, max_gap = TRUE) {
    .Call('_sphunif_p_cir_stat_Range', PACKAGE = 'sphunif', x, n, max_gap)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Range <- function(x, n, max_gap = TRUE) {
    .Call('_sphunif_d_cir_stat_Range', PACKAGE = 'sphunif', x, n, max_gap)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Rao <- function(x) {
    .Call('_sphunif_p_cir_stat_Rao', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Rao <- function(x) {
    .Call('_sphunif_d_cir_stat_Rao', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
p_cir_stat_Rayleigh <- function(x) {
    .Call('_sphunif_p_cir_stat_Rayleigh', PACKAGE = 'sphunif', x)
}

#' @rdname cir_stat_distr
#' @export
d_cir_stat_Rayleigh <- function(x) {
    .Call('_sphunif_d_cir_stat_Rayleigh', PACKAGE = 'sphunif', x)
}

#' @title Projection of the spherical uniform distribution
#'
#' @description Density, distribution, and quantile functions of the
#' projection of the spherical uniform random variable on an arbitrary
#' direction, that is, the random variable
#' \eqn{\boldsymbol{\gamma}'{\bf X}}{\gamma'X}, where \eqn{{\bf X}}{X}
#' is uniformly distributed on the (hyper)sphere
#' \eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{S^{p-1}:=
#' \{x\in R^p:||x||=1\}}, \eqn{p\ge 2}, and
#' \eqn{\boldsymbol{\gamma}\in S^{p-1}}{\gamma\in S^{p-1}} is an
#' \emph{arbitrary} projection direction. Note that the distribution is
#' invariant to the choice of \eqn{\boldsymbol{\gamma}}{\gamma}. Also,
#' efficient simulation of \eqn{\boldsymbol{\gamma}'{\bf X}}{\gamma'X}.
#'
#' @inheritParams cir_stat_distr
#' @inheritParams r_unif
#' @param u vector of probabilities.
#' @param log compute the logarithm of the density or distribution?
#' @return A matrix of size \code{c(nx, 1)} with the evaluation of the
#' density, distribution, or quantile function at \code{x} or \code{u}.
#' For \code{r_proj_unif}, a random vector of size \code{n}.
#' @author Eduardo García-Portugués and Paula Navarro-Esteban.
#' @examples
#' # Density function
#' curve(d_proj_unif(x, p = 2), from = -2, to = 2, n = 2e2, ylim = c(0, 2))
#' curve(d_proj_unif(x, p = 3), n = 2e2, col = 2, add = TRUE)
#' curve(d_proj_unif(x, p = 4), n = 2e2, col = 3, add = TRUE)
#' curve(d_proj_unif(x, p = 5), n = 2e2, col = 4, add = TRUE)
#' curve(d_proj_unif(x, p = 6), n = 2e2, col = 5, add = TRUE)
#'
#' # Distribution function
#' curve(p_proj_unif(x, p = 2), from = -2, to = 2, n = 2e2, ylim = c(0, 1))
#' curve(p_proj_unif(x, p = 3), n = 2e2, col = 2, add = TRUE)
#' curve(p_proj_unif(x, p = 4), n = 2e2, col = 3, add = TRUE)
#' curve(p_proj_unif(x, p = 5), n = 2e2, col = 4, add = TRUE)
#' curve(p_proj_unif(x, p = 6), n = 2e2, col = 5, add = TRUE)
#'
#' # Quantile function
#' curve(q_proj_unif(u = x, p = 2), from = 0, to = 1, n = 2e2, ylim = c(-1, 1))
#' curve(q_proj_unif(u = x, p = 3), n = 2e2, col = 2, add = TRUE)
#' curve(q_proj_unif(u = x, p = 4), n = 2e2, col = 3, add = TRUE)
#' curve(q_proj_unif(u = x, p = 5), n = 2e2, col = 4, add = TRUE)
#' curve(q_proj_unif(u = x, p = 6), n = 2e2, col = 5, add = TRUE)
#'
#' # Sampling
#' hist(r_proj_unif(n = 1e4, p = 4), freq = FALSE, breaks = 50)
#' curve(d_proj_unif(x, p = 4), n = 2e2, col = 3, add = TRUE)
#' @name proj_unif
NULL

#' @title Sample uniformly distributed circular and spherical data
#'
#' @description Simulation of the uniform distribution on \eqn{[0, 2\pi)} and
#' \eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{
#' S^{p-1}:=\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}.
#'
#' @param n sample size.
#' @param M number of samples of size \code{n}. Defaults to \code{1}.
#' @param p integer giving the dimension of the ambient space \eqn{R^p} that
#' contains \eqn{S^{p-1}}.
#' @param sorted return each circular sample sorted? Defaults to \code{FALSE}.
#' @return
#' \itemize{
#'   \item \code{r_unif_cir}: a \bold{matrix} of size \code{c(n, M)} with
#'   \code{M} random samples of size \code{n} of uniformly-generated circular
#'   data on \eqn{[0, 2\pi)}.
#'   \item \code{r_unif_sph}: an \bold{array} of size \code{c(n, p, M)} with
#'   \code{M} random samples of size \code{n} of uniformly-generated
#'   directions on \eqn{S^{p-1}}.
#' }
#' @examples
#' # A sample on [0, 2*pi)
#' n <- 5
#' r_unif_cir(n = n)
#'
#' # A sample on S^1
#' p <- 2
#' samp <- r_unif_sph(n = n, p = p)
#' samp
#' rowSums(samp^2)
#'
#' # A sample on S^2
#' p <- 3
#' samp <- r_unif_sph(n = n, p = p)
#' samp
#' rowSums(samp^2)
#' @name r_unif
NULL

#' @title Utilities for weighted sums of non-central chi squared random
#' variables
#'
#' @description Simulation from a weighted sum of non-central chi squared
#' random variables and Monte Carlo approximation of its distribution function.
#'
#' @inheritParams r_unif
#' @inheritParams wschisq
#' @param ncps non-negative non-centrality parameters. A vector with the same
#' length as \code{weights}.
#' @param M number of Monte Carlo samples for approximating the distribution.
#' Defaults to \code{1e4}.
#' @param sample if \code{use_sample = TRUE}, the Monte Carlo sample to
#' approximate the distribution. If not, it is computed internally. Defaults
#' to \code{1e4}.
#' @param use_sample use the already computed \code{sample}? If \code{FALSE}
#' (default), \code{sample} is computed internally.
#' @return
#' \itemize{
#'   \item \code{r_wschisq_Cpp}: a matrix of size \code{c(n, 1)}
#'   containing a random sample.
#'   \item \code{p_wschisq_MC}: a matrix of size \code{c(nx, 1)}
#'   with the evaluation of the distribution function at \code{x}.
#' }
#' @name wschisq_utils
NULL

#' @title Density and distribution of a chi squared
#'
#' @description Computation of the density and distribution functions of a chi
#' squared.
#'
#' @inheritParams cir_stat_distr
#' @param df degrees of freedom.
#' @param ncp non-centrality parameter.
#' @return A matrix of size \code{c(nx, 1)} with the evaluation of the
#' density or distribution function at \code{x}.
#' @name chisq
NULL

#' @rdname proj_unif
#' @export
d_proj_unif <- function(x, p, log = FALSE) {
    .Call('_sphunif_d_proj_unif', PACKAGE = 'sphunif', x, p, log)
}

#' @rdname proj_unif
#' @export
p_proj_unif <- function(x, p, log = FALSE) {
    .Call('_sphunif_p_proj_unif', PACKAGE = 'sphunif', x, p, log)
}

#' @rdname proj_unif
#' @export
q_proj_unif <- function(u, p) {
    .Call('_sphunif_q_proj_unif', PACKAGE = 'sphunif', u, p)
}

#' @rdname r_unif
#' @export
r_unif_cir <- function(n, M = 1L, sorted = FALSE) {
    .Call('_sphunif_r_unif_cir', PACKAGE = 'sphunif', n, M, sorted)
}

#' @rdname r_unif
#' @export
r_unif_sph <- function(n, p, M = 1L) {
    .Call('_sphunif_r_unif_sph', PACKAGE = 'sphunif', n, p, M)
}

#' @rdname wschisq_utils
#' @keywords internal
r_wschisq_Cpp <- function(n, weights, dfs, ncps) {
    .Call('_sphunif_r_wschisq_Cpp', PACKAGE = 'sphunif', n, weights, dfs, ncps)
}

#' @rdname wschisq_utils
#' @keywords internal
p_wschisq_MC <- function(x, weights = 0L, dfs = 0L, ncps = 0L, M = 1e4L, sample = 0L, use_sample = FALSE, x_sorted = FALSE) {
    .Call('_sphunif_p_wschisq_MC', PACKAGE = 'sphunif', x, weights, dfs, ncps, M, sample, use_sample, x_sorted)
}

#' @rdname chisq
#' @keywords internal
d_chisq <- function(x, df, ncp = 0L) {
    .Call('_sphunif_d_chisq', PACKAGE = 'sphunif', x, df, ncp)
}

#' @rdname chisq
#' @keywords internal
p_chisq <- function(x, df, ncp = 0L) {
    .Call('_sphunif_p_chisq', PACKAGE = 'sphunif', x, df, ncp)
}

#' @title Gauss--Legendre quadrature
#'
#' @description Convenience for computing the nodes \eqn{x_k} and weights
#' \eqn{w_k} of the \emph{Gauss--Legendre} quadrature formula
#' in \eqn{(a, b)}:
#' \deqn{\int_a^b f(x) w(x)\,\mathrm{d}x\approx\sum_{k=1}^N w_k f(x_k).}{
#' \int_a^b f(x) dx\approx\sum_{k=1}^N w_k f(x_k)}.
#'
#' @param a,b scalars giving the interval \eqn{(a, b)}. Defaults to
#' \eqn{(-1, 1)}.
#' @param N number of points used in the Gauss--Legendre quadrature. The
#' following choices are supported: \code{5}, \code{10}, \code{20},
#' \code{40}, \code{80}, \code{160}, \code{320}, \code{640}, \code{1280},
#' \code{2560}, and \code{5120}. Defaults to \code{40}.
#' @return A matrix of size \code{c(N, 1)} with the nodes \eqn{x_k}
#' (\code{Gauss_Legen_nodes}) or the corresponding weights \eqn{w_k}
#' (\code{Gauss_Legen_weights}).
#' @details For \eqn{C^\infty} functions, Gauss--Legendre quadrature
#' can be very efficient. It is exact for polynomials up to degree
#' \eqn{2N - 1}.
#'
#' The nodes and weights up to \eqn{N = 80} were retrieved from
#' \href{https://dlmf.nist.gov/3.5#v}{NIST} and have \eqn{10^{-21}} precision.
#' For \eqn{N = 160} onwards, the nodes and weights were computed with the
#' \code{gauss.quad} function from the \href{https://CRAN.R-project.org/package=statmod}{
#' \code{statmod}} package (Smyth, 1998), and have \eqn{10^{-15}} precision.
#' @references
#' \emph{NIST Digital Library of Mathematical Functions}. Release
#' 1.0.20 of 2018-09-15. F. W. J. Olver, A. B. Olde Daalhuis, D. W. Lozier,
#' B. I. Schneider, R. F. Boisvert, C. W. Clark, B. R. Miller,
#' and B. V. Saunders, eds. \url{https://dlmf.nist.gov/}
#'
#' Smyth, G. K. (1998). Numerical integration. In: \emph{Encyclopedia of
#' Biostatistics}, P. Armitage and T. Colton (eds.), Wiley, London,
#' pp. 3088-3095.
#' @examples
#' ## Integration of a smooth function in (1, 10)
#'
#' # Weights and nodes for integrating
#' x_k <- Gauss_Legen_nodes(a = 1, b = 10, N = 40)
#' w_k <- Gauss_Legen_weights(a = 1, b = 10, N = 40)
#'
#' # Check quadrature
#' f <- function(x) sin(x) * x^2 - log(x + 1)
#' integrate(f, lower = 1, upper = 10, rel.tol = 1e-12)
#' sum(w_k * f(x_k))
#'
#' # Exact for polynomials up to degree 2 * N - 1
#' f <- function(x) (((x + 0.5) / 1e3)^5 - ((x - 0.5)/ 5)^4 +
#'   ((x - 0.25) / 10)^2 + 1)^20
#' sum(w_k * f(x_k))
#' integrate(f, lower = -1, upper = 1, rel.tol = 1e-12)
#'
#' ## Integration on (0, pi)
#'
#' # Weights and nodes for integrating
#' th_k <- Gauss_Legen_nodes(a = 0, b = pi, N = 40)
#' w_k <- Gauss_Legen_weights(a = 0, b = pi, N = 40)
#'
#' # Check quadrature
#' p <- 4
#' psi <- function(th) -sin(th / 2)
#' w <- function(th) sin(th)^(p - 2)
#' integrate(function(th) psi(th) * w(th), lower = 0, upper = pi,
#'           rel.tol = 1e-12)
#' sum(w_k * psi(th_k) * w(th_k))
#'
#' # Integral with Gegenbauer polynomial
#' k <- 3
#' C_k <- function(th) drop(Gegen_polyn(theta = th, k = k, p = p))
#' integrate(function(th) psi(th) * C_k(th) * w(th), lower = 0, upper = pi,
#'           rel.tol = 1e-12)
#' th_k <- drop(Gauss_Legen_nodes(a = 0, b = pi, N = 80))
#' w_k <- drop(Gauss_Legen_weights(a = 0, b = pi, N = 80))
#' sum(w_k * psi(th_k) * C_k(th_k) * w(th_k))
#' @name Gauss_Legen
NULL

#' @rdname Gauss_Legen
#' @export
Gauss_Legen_nodes <- function(a = -1, b = 1, N = 40L) {
    .Call('_sphunif_Gauss_Legen_nodes', PACKAGE = 'sphunif', a, b, N)
}

#' @rdname Gauss_Legen
#' @export
Gauss_Legen_weights <- function(a = -1, b = 1, N = 40L) {
    .Call('_sphunif_Gauss_Legen_weights', PACKAGE = 'sphunif', a, b, N)
}

#' @rdname sph_stat_distr
#' @export
p_sph_stat_Bingham <- function(x, p) {
    .Call('_sphunif_p_sph_stat_Bingham', PACKAGE = 'sphunif', x, p)
}

#' @rdname sph_stat_distr
#' @export
d_sph_stat_Bingham <- function(x, p) {
    .Call('_sphunif_d_sph_stat_Bingham', PACKAGE = 'sphunif', x, p)
}

#' @rdname sph_stat_distr
#' @export
p_sph_stat_CJ12 <- function(x, regime = 1L, beta = 0) {
    .Call('_sphunif_p_sph_stat_CJ12', PACKAGE = 'sphunif', x, regime, beta)
}

#' @rdname sph_stat_distr
#' @export
d_sph_stat_CJ12 <- function(x, regime = 3L, beta = 0) {
    .Call('_sphunif_d_sph_stat_CJ12', PACKAGE = 'sphunif', x, regime, beta)
}

#' @rdname sph_stat_distr
#' @export
p_sph_stat_Rayleigh <- function(x, p) {
    .Call('_sphunif_p_sph_stat_Rayleigh', PACKAGE = 'sphunif', x, p)
}

#' @rdname sph_stat_distr
#' @export
d_sph_stat_Rayleigh <- function(x, p) {
    .Call('_sphunif_d_sph_stat_Rayleigh', PACKAGE = 'sphunif', x, p)
}

#' @rdname sph_stat_distr
#' @export
p_sph_stat_Rayleigh_HD <- function(x, p) {
    .Call('_sphunif_p_sph_stat_Rayleigh_HD', PACKAGE = 'sphunif', x, p)
}

#' @rdname sph_stat_distr
#' @export
d_sph_stat_Rayleigh_HD <- function(x, p) {
    .Call('_sphunif_d_sph_stat_Rayleigh_HD', PACKAGE = 'sphunif', x, p)
}

#' @rdname sph_stat
#' @export
sph_stat_Rayleigh <- function(X) {
    .Call('_sphunif_sph_stat_Rayleigh', PACKAGE = 'sphunif', X)
}

#' @rdname sph_stat
#' @export
sph_stat_Bingham <- function(X) {
    .Call('_sphunif_sph_stat_Bingham', PACKAGE = 'sphunif', X)
}

#' @rdname sph_stat
#' @export
sph_stat_Ajne <- function(X, Psi_in_X = FALSE) {
    .Call('_sphunif_sph_stat_Ajne', PACKAGE = 'sphunif', X, Psi_in_X)
}

#' @rdname sph_stat
#' @export
sph_stat_Gine_Gn <- function(X, Psi_in_X = FALSE, p = 0L) {
    .Call('_sphunif_sph_stat_Gine_Gn', PACKAGE = 'sphunif', X, Psi_in_X, p)
}

#' @keywords internal
sph_stat_Gine_Gn_Psi <- function(Psi, n, p) {
    .Call('_sphunif_sph_stat_Gine_Gn_Psi', PACKAGE = 'sphunif', Psi, n, p)
}

#' @rdname sph_stat
#' @export
sph_stat_Gine_Fn <- function(X, Psi_in_X = FALSE, p = 0L) {
    .Call('_sphunif_sph_stat_Gine_Fn', PACKAGE = 'sphunif', X, Psi_in_X, p)
}

#' @keywords internal
sph_stat_Gine_Fn_Psi <- function(Psi, n, p) {
    .Call('_sphunif_sph_stat_Gine_Fn_Psi', PACKAGE = 'sphunif', Psi, n, p)
}

#' @rdname sph_stat
#' @export
sph_stat_Pycke <- function(X, Psi_in_X = FALSE, p = 0L) {
    .Call('_sphunif_sph_stat_Pycke', PACKAGE = 'sphunif', X, Psi_in_X, p)
}

#' @keywords internal
sph_stat_Pycke_Psi <- function(Psi, n, p) {
    .Call('_sphunif_sph_stat_Pycke_Psi', PACKAGE = 'sphunif', Psi, n, p)
}

#' @rdname sph_stat
#' @export
sph_stat_Bakshaev <- function(X, Psi_in_X = FALSE, p = 0L) {
    .Call('_sphunif_sph_stat_Bakshaev', PACKAGE = 'sphunif', X, Psi_in_X, p)
}

#' @rdname sph_stat
#' @export
sph_stat_Riesz <- function(X, Psi_in_X = FALSE, p = 0L, s = 1.0) {
    .Call('_sphunif_sph_stat_Riesz', PACKAGE = 'sphunif', X, Psi_in_X, p, s)
}

#' @keywords internal
sph_stat_Riesz_Psi <- function(Psi, n, s) {
    .Call('_sphunif_sph_stat_Riesz_Psi', PACKAGE = 'sphunif', Psi, n, s)
}

#' @rdname sph_stat
#' @export
sph_stat_PCvM <- function(X, Psi_in_X = FALSE, p = 0L, N = 160L, L = 1e3L) {
    .Call('_sphunif_sph_stat_PCvM', PACKAGE = 'sphunif', X, Psi_in_X, p, N, L)
}

#' @keywords internal
sph_stat_PCvM_Psi <- function(Psi, n, p, th_grid, int_grid) {
    .Call('_sphunif_sph_stat_PCvM_Psi', PACKAGE = 'sphunif', Psi, n, p, th_grid, int_grid)
}

#' @rdname sph_stat
#' @export
sph_stat_PRt <- function(X, t = 1.0 / 3.0, Psi_in_X = FALSE, p = 0L, N = 160L, L = 1e3L) {
    .Call('_sphunif_sph_stat_PRt', PACKAGE = 'sphunif', X, t, Psi_in_X, p, N, L)
}

#' @keywords internal
sph_stat_PRt_Psi <- function(Psi, t_m, theta_t_m, n, p, th_grid, int_grid) {
    .Call('_sphunif_sph_stat_PRt_Psi', PACKAGE = 'sphunif', Psi, t_m, theta_t_m, n, p, th_grid, int_grid)
}

#' @rdname sph_stat
#' @export
sph_stat_PAD <- function(X, Psi_in_X = FALSE, p = 0L, N = 160L, L = 1e3L) {
    .Call('_sphunif_sph_stat_PAD', PACKAGE = 'sphunif', X, Psi_in_X, p, N, L)
}

#' @keywords internal
sph_stat_PAD_Psi <- function(Psi, n, p, th_grid, int_grid) {
    .Call('_sphunif_sph_stat_PAD_Psi', PACKAGE = 'sphunif', Psi, n, p, th_grid, int_grid)
}

#' @rdname sph_stat
#' @export
sph_stat_CCF09 <- function(X, dirs, K_CCF09 = 25L, original = FALSE) {
    .Call('_sphunif_sph_stat_CCF09', PACKAGE = 'sphunif', X, dirs, K_CCF09, original)
}

#' @rdname sph_stat
#' @export
sph_stat_Rayleigh_HD <- function(X) {
    .Call('_sphunif_sph_stat_Rayleigh_HD', PACKAGE = 'sphunif', X)
}

#' @rdname sph_stat
#' @export
sph_stat_CJ12 <- function(X, regime = 3L, Psi_in_X = FALSE, p = 0L) {
    .Call('_sphunif_sph_stat_CJ12', PACKAGE = 'sphunif', X, regime, Psi_in_X, p)
}

#' @keywords internal
sph_stat_CJ12_Psi <- function(Psi, n, p) {
    .Call('_sphunif_sph_stat_CJ12_Psi', PACKAGE = 'sphunif', Psi, n, p)
}

set.seed(123456)
x <- cbind(seq(-1, 1, l = 20))
u <- cbind(seq(0, 1, l = 20))
K <- rpois(n = 1, lambda = 9) + 1
weights <- rnorm(n = K)
dfs <- rpois(n = K, lambda = 4) + 1
ncps <- runif(n = K, min = 0, max = 2)^2

test_that("r_proj_unif vs. p_proj_unif", {

  for (p in c(2:4, 11)) {
    expect_gt(ks.test(r_proj_unif(n = 1e3, p = p),
                      "p_proj_unif", p = p)$p.value, 0.05)
  }

})

test_that("r_proj_unif vs. r_unif_sph", {

  for (p in c(2:4, 11)) {
    expect_gt(ks.test(r_proj_unif(n = 1e3, p = p),
                      r_unif_sph(n = 1e3, p = p)[, 1, 1])$p.value, 0.05)
  }

})

test_that("sorting in r_unif_cir", {

  expect_equal({set.seed(1); drop(r_unif_cir(n = 10, sorted = TRUE))},
               {set.seed(1); sort(r_unif_cir(n = 10, sorted = FALSE))})

})

test_that("p_proj_unif vs. q_proj_unif", {

  for (p in c(2:4, 11)) {
    expect_equal(q_proj_unif(u = p_proj_unif(x = x, p = p), p = p), x)
    expect_equal(p_proj_unif(x = q_proj_unif(u = u, p = p), p = p), u)
  }

})

test_that("Wrong dimensions", {

  expect_error(d_proj_unif(x = 0:1, p = 1))
  expect_error(p_proj_unif(x = 0:1, p = 1))
  expect_error(q_proj_unif(u = 0:1, p = 1))
  expect_error(r_proj_unif(n = 1, p = 1))

})

test_that("d_chisq and p_chisq", {

  expect_equal(d_chisq(x = x, df = 3), dchisq(x = x, df = 3))
  expect_equal(p_chisq(x = x, df = 3), pchisq(q = x, df = 3))

})

test_that("r_wschisq_Cpp vs. p_wschisq_MC", {

  samp1 <- r_wschisq_Cpp(n = 1e3, weights = weights, dfs = dfs, ncps = ncps)
  samp2 <- r_wschisq_Cpp(n = 1e4, weights = weights, dfs = dfs, ncps = ncps)
  expect_gt(ks.test(samp1, "p_wschisq_MC", weights = weights, dfs = dfs,
                    ncps = ncps)$p.value, 0.10)
  expect_gt(ks.test(samp1, "p_wschisq_MC", sample = samp2,
                    use_sample = TRUE)$p.value, 0.10)
  expect_gt(ks.test(samp1, "p_wschisq_MC", sample = samp2,
                    use_sample = TRUE, x_sorted = TRUE)$p.value, 0.10)

})

p <- 4
f <- function(th) drop(-sin(th / 2) * sin(th)^(p - 2))
I <- integrate(f, lower = 0, upper = pi, rel.tol = 1e-12)$value

test_that("Gauss_Legen for N = 5, 10, 20, ..., 2560, 5120", {

  expect_equal(sum(Gauss_Legen_weights(a = 0, b = pi, N = 5) *
                     f(Gauss_Legen_nodes(a = 0, b = pi, N = 5))), I,
               tolerance = 1e-3)
  expect_equal(sum(Gauss_Legen_weights(a = 0, b = pi, N = 10) *
                     f(Gauss_Legen_nodes(a = 0, b = pi, N = 10))), I,
               tolerance = 1e-12)
  for (N in c(10 * 2^(1:9))) {
    expect_equal(sum(Gauss_Legen_weights(a = 0, b = pi, N = N) *
                       f(Gauss_Legen_nodes(a = 0, b = pi, N = N))), I,
                 tolerance = 3e-14)
  }

})

test_that("Gauss_Legen edge cases", {

  expect_error(Gauss_Legen_weights(a = 0, b = pi, N = 19))
  expect_error(Gauss_Legen_nodes(a = 0, b = pi, N = 19))

})

# To prevent hanging with parallel computations
Sys.unsetenv("R_TESTS")

n <- 50
set.seed(1242333)
cir_0 <- unif_stat_MC(n = n, M = 5e2, type = "all", p = 2, seeds = 5)
sph_0 <- unif_stat_MC(n = n, M = 5e2, type = "all", p = 3, seeds = 5)
crit_val_bad <- sph_0$crit_val_MC
colnames(crit_val_bad) <- paste0(colnames(crit_val_bad), "_bad")

# Circular
dirs_2 <- r_unif_sph(n = 20, p = 2, M = 1)[, , 1]
cir_pow <- unif_stat_MC(n = n, M = 5e2, p = 2, r_H1 = r_alt, alt = "MvMF",
                        kappa = 0.5, crit_val = cir_0$crit_val_MC,
                        CCF09_dirs = dirs_2)

# Spherical
dirs_3 <- r_unif_sph(n = 20, p = 3, M = 1)[, , 1]
sph_pow <- unif_stat_MC(n = n, M = 5e2, p = 3, r_H1 = r_alt, alt = "MvMF",
                        kappa = 0.5, crit_val = sph_0$crit_val_MC,
                        CCF09_dirs = dirs_3)

test_that("Rejections for MvMF", {

  expect_true(all(sph_pow$power_MC[1, ] > 0.01))
  expect_true(all(cir_pow$power_MC[1, ] > 0.01))

})

test_that("Edge cases", {

  expect_error(unif_stat_MC(n = n, M = 1e2, type = "all", p = 1))
  expect_error(unif_stat_MC(n = n, M = 1e2, type = "all", p = 3,
                            crit_val = sph_0$crit_val_MC[, 1:3]))
  expect_error(unif_stat_MC(n = n, M = 1e2, type = "all", p = 3,
                            crit_val = crit_val_bad[, 1:3]))
  suppressWarnings(expect_warning(unif_stat_MC(n = n, M = 1e2, type = "all",
                                               p = 5, seeds = 1:3,
                                               chunks = 2)))
  expect_equal(unif_stat_MC(n = n, M = 1e2, type = c("PAD", "Ajne"), p = 3,
                            seeds = 5)$stats$PAD,
               unif_stat_MC(n = n, M = 1e2, type = "PAD", p = 3, seeds = 5,
                            crit_val = sph_0$crit_val_MC)$stats$PAD)

})

test_that("Several options", {

  expect_equal(unif_stat_MC(n = n, M = 10, type = "all", p = 2,
                            return_stats = FALSE)$stats, NA)
  suppressWarnings(expect_equal(unif_stat_MC(n = n, M = 10, type = "all", p = 3,
                                             return_stats = FALSE)$stats, NA))
  suppressWarnings(
    expect_equal(unname(as.matrix(unif_stat_MC(n = n, M = 10, type = "all",
                                               p = 5, stats_sorted = TRUE,
                                               seeds = 1)$stats)),
                 sort_each_col(as.matrix(unif_stat_MC(n = n, M = 10,
                                                      type = "all", p = 5,
                                                      seeds = 1)$stats)))
  )
  set.seed(1)
  CCF09_dirs <- r_unif_sph(n = 50, p = 3, M = 1)[, , 1]
  expect_equal(unif_stat_MC(n = n, M = 5, type = "CCF09", p = 3,
                            stats_sorted = TRUE, seeds = 1, chunks = 1)$CCF09,
               unif_stat_MC(n = n, M = 5, type = "CCF09", p = 3,
                            stats_sorted = TRUE, seeds = 1, chunks = 1,
                            CCF09_dirs = CCF09_dirs)$CCF09)

})

test_that("Progress bars", {

  skip_on_cran()
  o1_silent <- capture.output(s <- unif_stat_MC(n = n, M = 10, type = "all",
                                                p = 2, cores = 1, chunks = 10))
  o2_silent <- capture.output(s <- unif_stat_MC(n = n, M = 10, type = "all",
                                                p = 2, cores = 2, chunks = 10))
  expect_equal(length(o1_silent), 0)
  expect_equal(length(o2_silent), 0)

})

test_that("Same results with cores = 1 and cores = 2", {

  skip_on_cran()
  expect_equal(unif_stat_MC(n = n, M = 10, type = "all",
                            p = 2, chunks = 10, cores = 1, seeds = 1:10),
               unif_stat_MC(n = n, M = 10, type = "all",
                            p = 2, chunks = 10, cores = 2, seeds = 1:10))

})

test_that("Parallelization is faster", {

  skip_on_ci()
  skip_on_cran()
  t1 <- system.time(unif_stat_MC(n = 100, M = 1e4, type = "all", p = 2,
                                 chunks = 10, cores = 1))[3]
  t2 <- system.time(unif_stat_MC(n = 100, M = 1e4, type = "all", p = 2,
                                 chunks = 10, cores = 2))[3]
  expect_gt(t1, t2)

})

set.seed(131121)
Theta_1 <- r_unif_cir(n = 33, M = 1)
Theta_2 <- r_unif_cir(n = 10, M = 2)
X_2 <- r_unif_sph(n = 33, p = 2, M = 1)
X_3 <- r_unif_sph(n = 10, p = 3, M = 2)
X_4 <- r_unif_sph(n = 10, p = 4, M = 2)
X_9 <- r_unif_sph(n = 10, p = 9, M = 2)
r_d_2 <- r_unif_sph(n = 5, p = 2, M = 1)[, , 1]
r_d_3 <- r_unif_sph(n = 5, p = 3, M = 1)[, , 1]
r_d_4 <- r_unif_sph(n = 5, p = 4, M = 1)[, , 1]
r_d_9 <- r_unif_sph(n = 5, p = 9, M = 1)[, , 1]
avail_cir_tests_s1 <- c("Gine_Fn", "Pycke", "Watson", "PCvM", "Cressie")
avail_cir_tests_s2 <- c("Gine_Gn", "Rothman", "Watson_1976", "PRt")
avail_sph_tests_s1 <- c("Gine_Fn", "Pycke", "PAD", "PCvM", "Cai")
avail_sph_tests_s2 <- c("Gine_Gn", "CCF09", "Ajne", "PRt")

test_that("unif_stat doing all or separate statistics", {

  expect_equal(as.numeric(unif_stat(data = Theta_1, type = "all",
                                    CCF09_dirs = r_d_2)),
               as.numeric(sapply(avail_cir_tests, function(test)
                 unif_stat(data = Theta_1, type = test,
                           CCF09_dirs = r_d_2))))
  expect_equal(as.matrix(unif_stat(data = Theta_2, type = "all",
                                   CCF09_dirs = r_d_2)),
               sapply(avail_cir_tests, function(test)
                 as.matrix(unif_stat(data = Theta_2, type = test,
                                     CCF09_dirs = r_d_2))))
  expect_equal(as.numeric(unif_stat(data = X_2, type = "all",
                                   CCF09_dirs = r_d_2)),
               as.numeric(sapply(avail_cir_tests, function(test)
                 unif_stat(data = X_2, type = test,
                           CCF09_dirs = r_d_2))))
  expect_equal(as.matrix(unif_stat(data = X_3, type = "all",
                                   CCF09_dirs = r_d_3)),
               sapply(avail_sph_tests, function(test)
                 as.matrix(unif_stat(data = X_3, type = test,
                                     CCF09_dirs = r_d_3))))
  suppressWarnings(expect_warning(
    expect_equal(as.matrix(unif_stat(data = X_4, type = "all",
                                     CCF09_dirs = r_d_4)),
                 sapply(avail_sph_tests, function(test)
                   as.matrix(unif_stat(data = X_4, type = test,
                                       CCF09_dirs = r_d_4))))
  ))
  suppressWarnings(expect_warning(
    expect_equal(as.matrix(unif_stat(data = X_9, type = "all",
                                     CCF09_dirs = r_d_9)),
                 sapply(avail_sph_tests, function(test)
                   as.matrix(unif_stat(data = X_9, type = test,
                                       CCF09_dirs = r_d_9))))
  ))

})

test_that("unif_stat with randomly-chosen and separate statistics", {

  expect_equal(as.numeric(unif_stat(data = Theta_1, type = avail_cir_tests_s1,
                                    CCF09_dirs = r_d_2)),
               as.numeric(sapply(avail_cir_tests_s1, function(test)
                 unif_stat(data = Theta_1, type = test,
                           CCF09_dirs = r_d_2))))
  expect_equal(as.numeric(unif_stat(data = Theta_1, type = avail_cir_tests_s2,
                                    CCF09_dirs = r_d_2)),
               as.numeric(sapply(avail_cir_tests_s2, function(test)
                 unif_stat(data = Theta_1, type = test,
                           CCF09_dirs = r_d_2))))
  expect_equal(as.numeric(unif_stat(data = X_2, type = avail_sph_tests_s2,
                                   CCF09_dirs = r_d_2)),
               as.numeric(sapply(avail_sph_tests_s2, function(test)
                 unif_stat(data = X_2, type = test,
                           CCF09_dirs = r_d_2))))
  expect_equal(as.matrix(unif_stat(data = X_3, type = avail_sph_tests_s2,
                                   CCF09_dirs = r_d_3)),
               sapply(avail_sph_tests_s2, function(test)
                 as.matrix(unif_stat(data = X_3, type = test,
                                     CCF09_dirs = r_d_3))))
  expect_equal(as.matrix(unif_stat(data = X_4, type = avail_sph_tests_s2,
                                   CCF09_dirs = r_d_4)),
               sapply(avail_sph_tests_s2, function(test)
                 as.matrix(unif_stat(data = X_4, type = test,
                                     CCF09_dirs = r_d_4))))

})

test_that("unif_stat with data_sorted = TRUE and CCF09 = NULL", {

  expect_equal(as.numeric(unif_stat(data = sort_each_col(Theta_1),
                                    type = avail_cir_tests_s1,
                                    CCF09_dirs = r_d_2,
                                    data_sorted = TRUE)),
               as.numeric(sapply(avail_cir_tests_s1, function(test)
                 unif_stat(data = sort_each_col(Theta_1), type = test,
                           CCF09_dirs = r_d_2,
                           data_sorted = TRUE))))
  expect_equal({
    set.seed(131312)
    rd <-  CCF09_dirs <- r_unif_sph(n = 50, p = 2, M = 1)[, , 1]
    as.matrix(unif_stat(data = Theta_2, type = c("Range", "Rao",
                                                 "CCF09"),
                         CCF09_dirs = rd))
    }, {
    set.seed(131312)
    as.matrix(unif_stat(data = Theta_2, type = c("Range", "Rao",
                                                 "CCF09")))
    })
  expect_equal({
    set.seed(131312)
    r_d <-  CCF09_dirs <- r_unif_sph(n = 50, p = 3, M = 1)[, , 1]
    as.matrix(unif_stat(data = X_3, type = c("PCvM", "Bakshaev",
                                             "CCF09"),
                        CCF09_dirs = r_d))
    }, {
    set.seed(131312)
    as.matrix(unif_stat(data = X_3, type = c("PCvM", "Bakshaev",
                                             "CCF09")))
    })

})

test_that("unif_stat equality for data = Theta and data = Theta_to_X(Theta)", {

  expect_equal(unif_stat(data = Theta_1, type = avail_cir_tests,
                         CCF09_dirs = r_d_2),
               unif_stat(data = Theta_to_X(Theta_1), type = avail_cir_tests,
                         CCF09_dirs = r_d_2))
  expect_equal(unif_stat(data = X_2, type = avail_sph_tests,
                         CCF09_dirs = r_d_2),
               unif_stat(data = X_to_Theta(X_2), type = avail_sph_tests,
                         CCF09_dirs = r_d_2))

})

test_that("Errors in edge cases", {

  expect_error(unif_stat(data = rbind(Theta_2, c(0, NA))))
  expect_error(unif_stat(data = rbind(X_3, c(0, NA, 1))))
  expect_error(unif_stat(data = rbind(c(1, 0))))
  expect_error(unif_stat(data = 1))
  expect_error(unif_stat(data = 1:3, type = function(x) x^2))

})

test_that("Passing edge cases", {

  expect_equal(unif_stat(data = 1:6, type = c("Ajne", "Cressie")),
               unif_stat(data = cbind(1:6), type = c("Ajne", "Cressie")))
  expect_equal(unif_stat(data = cbind(1:5, 2:6), type = c("Ajne", "Cressie")),
               {A <- array(dim = c(5, 1, 2)); A[, 1, ] <- cbind(1:5, 2:6);
               unif_stat(data = A, type = c("Ajne", "Cressie"))})
  expect_equal(c(as.matrix(unif_stat(data = Theta_2,
                                     type = avail_cir_tests[2:5],
                                     CCF09_dirs = r_d_2))),
               c(sapply(2:5, function(test)
                 as.matrix(unif_stat(data = Theta_2, type = test,
                                     CCF09_dirs = r_d_2)))))
  expect_equal(c(as.matrix(unif_stat(data = X_3, type = avail_sph_tests[2:5],
                                     CCF09_dirs = r_d_3))),
               c(sapply(2:5, function(test)
                 as.matrix(unif_stat(data = X_3, type = test,
                                     CCF09_dirs = r_d_3)))))
  expect_equal(unif_stat(data = 1:6, type = c("Ajne", "Ajne", "Watson")),
               unif_stat(data = 1:6, type = c("Ajne", "Watson")))

})

test_that("KS, CvM, and AD", {

  expect_equal(as.numeric(unif_stat(data = Theta_1, type = "KS")),
               as.numeric(cir_stat_Kuiper(Theta = Theta_1, KS = TRUE)))
  expect_equal(as.numeric(unif_stat(data = Theta_1, type = "CvM")),
               as.numeric(cir_stat_Watson(Theta = Theta_1, CvM = TRUE)))
  expect_equal(as.numeric(unif_stat(data = Theta_1, type = "AD")),
               as.numeric(goftest::ad.test(Theta_1, null = "punif",
                                           min = 0, max = 2 * pi)$statistic))
  expect_equal(unif_stat(data = Theta_2, type = "KS")$KS,
               cir_stat_Kuiper(Theta = Theta_2, KS = TRUE))
  expect_equal(unif_stat(data = Theta_2, type = "CvM")$CvM,
               cir_stat_Watson(Theta = Theta_2, CvM = TRUE))
  expect_equal(as.numeric(unif_stat(data = Theta_2, type = "AD")$AD),
               apply(Theta_2, 2, function(x)
                 as.numeric(goftest::ad.test(x, null = "punif", min = 0,
                                             max = 2 * pi)$statistic)))

})

test_that("Riesz vs. Rayleigh", {

  expect_equal(
    as.numeric(unif_stat(data = Theta_2, type = c("Riesz", "Rayleigh"),
                         Riesz_s = 2)$Riesz),
    as.numeric(unif_stat(data = Theta_2, type = "Riesz", Riesz_s = 2)$Riesz)
  )
  expect_equal(
    as.numeric(unif_stat(data = Theta_2, type = c("Riesz", avail_sph_tests_s2),
                         Riesz_s = 2)$Riesz),
    as.numeric(unif_stat(data = Theta_2, type = "Riesz", Riesz_s = 2)$Riesz)
  )
  expect_equal(
    as.numeric(unif_stat(data = X_2, type = c("Riesz", "Rayleigh"),
                         Riesz_s = 2)$Riesz),
    as.numeric(unif_stat(data = X_2, type = "Riesz", Riesz_s = 2)$Riesz)
  )
  expect_equal(
    as.numeric(unif_stat(data = X_2, type = c("Riesz", avail_sph_tests_s2),
                         Riesz_s = 2)$Riesz),
    as.numeric(unif_stat(data = X_2, type = "Riesz", Riesz_s = 2)$Riesz)
  )
  expect_equal(
    as.numeric(unif_stat(data = X_3, type = c("Riesz", "Rayleigh"),
                         Riesz_s = 2)$Riesz),
    as.numeric(unif_stat(data = X_3, type = "Riesz", Riesz_s = 2)$Riesz)
  )
  expect_equal(
    as.numeric(unif_stat(data = X_3, type = c("Riesz", avail_sph_tests_s2),
                         Riesz_s = 2)$Riesz),
    as.numeric(unif_stat(data = X_3, type = "Riesz", Riesz_s = 2)$Riesz)
  )
  expect_equal(
    as.numeric(unif_stat(data = X_4, type = c("Riesz", "Rayleigh"),
                         Riesz_s = 2)$Riesz),
    as.numeric(unif_stat(data = X_4, type = "Riesz", Riesz_s = 2)$Riesz)
  )
  expect_equal(
    as.numeric(unif_stat(data = X_4, type = c("Riesz", avail_sph_tests_s2),
                         Riesz_s = 2)$Riesz),
    as.numeric(unif_stat(data = X_4, type = "Riesz", Riesz_s = 2)$Riesz)
  )
  expect_equal(
    as.numeric(unif_stat(data = X_9, type = c("Riesz", "Rayleigh"),
                         Riesz_s = 2)$Riesz),
    as.numeric(unif_stat(data = X_9, type = "Riesz", Riesz_s = 2)$Riesz)
  )
  expect_equal(
    as.numeric(unif_stat(data = X_9, type = c("Riesz", avail_sph_tests_s2),
                         Riesz_s = 2)$Riesz),
    as.numeric(unif_stat(data = X_9, type = "Riesz", Riesz_s = 2)$Riesz)
  )

})

test_that("Riesz vs. Pycke", {

  expect_equal(
    as.numeric(unif_stat(data = Theta_2, type = c("Riesz", "Pycke"),
                         Riesz_s = 0)$Pycke),
    as.numeric(unif_stat(data = Theta_2, type = "Pycke")$Pycke)
  )
  expect_equal(
    as.numeric(unif_stat(data = X_2, type = c("Riesz", "Pycke"),
                         Riesz_s = 0)$Pycke),
    as.numeric(unif_stat(data = X_2, type = "Pycke")$Pycke)
  )
  expect_equal(
    as.numeric(unif_stat(data = X_3, type = c("Riesz", "Pycke"),
                         Riesz_s = 0)$Pycke),
    as.numeric(unif_stat(data = X_3, type = "Pycke")$Pycke)
  )
  suppressWarnings(expect_warning(
    expect_equal(
      as.numeric(unif_stat(data = X_4, type = c("Riesz", "Pycke"),
                           Riesz_s = 0)$Riesz),
      as.numeric(unif_stat(data = X_4, type = "Pycke")$Pycke)
    )))
  suppressWarnings(expect_warning(
    expect_equal(
      as.numeric(unif_stat(data = X_9, type = c("Riesz", "Pycke"),
                           Riesz_s = 0)$Riesz),
      as.numeric(unif_stat(data = X_9, type = "Pycke")$Pycke)
    )))

})

n <- 10
set.seed(123456789)
X2 <- r_unif_sph(n = n, p = 2)
X3 <- r_unif_sph(n = n, p = 3)
X4 <- r_unif_sph(n = n, p = 4)
X5 <- r_unif_sph(n = n, p = 5)
X9 <- r_unif_sph(n = n, p = 9)
X200 <- r_unif_sph(n = n, p = 200)
Psi2 <- Psi_mat(X2)
Psi3 <- Psi_mat(X3)
Psi4 <- Psi_mat(X4)
Psi5 <- Psi_mat(X5)
Psi9 <- Psi_mat(X9)
Psi200 <- Psi_mat(X200)
dim(Psi2) <- c(dim(Psi2), 1)
dim(Psi3) <- c(dim(Psi3), 1)
dim(Psi4) <- c(dim(Psi4), 1)
dim(Psi5) <- c(dim(Psi5), 1)
dim(Psi9) <- c(dim(Psi9), 1)
dim(Psi200) <- c(dim(Psi200), 1)
Th2 <- X_to_Theta(X2)
rd3 <- X3[1:5, , 1]

test_that("PAD with X", {

  expect_equal(drop(sph_stat_PAD(X2)), 0.5752646, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PAD(X3)), 0.7093266, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PAD(X4)), 1.4088836, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PAD(X5)), 0.7106791, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PAD(X9)), 1.5825283, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PAD(X200)), 0.9412612, tolerance = 1e-6)

})

test_that("PAD with Psi", {

  expect_equal(drop(sph_stat_PAD(Psi2, Psi_in_X = TRUE, p = 2)), 0.5752646,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PAD(Psi3, Psi_in_X = TRUE, p = 3)), 0.7093266,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PAD(Psi4, Psi_in_X = TRUE, p = 4)), 1.4088836,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PAD(Psi5, Psi_in_X = TRUE, p = 5)), 0.7106791,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PAD(Psi9, Psi_in_X = TRUE, p = 9)), 1.5825283,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PAD(Psi200, Psi_in_X = TRUE, p = 200)), 0.9412612,
               tolerance = 1e-6)

})

test_that("PCvM with X", {

  expect_equal(drop(sph_stat_PCvM(X2)), 0.0751836, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PCvM(X3)), 0.1027005, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PCvM(X4)), 0.2406067, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PCvM(X5)), 0.1124227, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PCvM(X9)), 0.2766211, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PCvM(X200)), 0.1553891, tolerance = 1e-6)

})

test_that("PCvM with Psi", {

  expect_equal(drop(sph_stat_PCvM(Psi2, Psi_in_X = TRUE, p = 2)), 0.0751836,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PCvM(Psi3, Psi_in_X = TRUE, p = 3)), 0.1027005,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PCvM(Psi4, Psi_in_X = TRUE, p = 4)), 0.2406067,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PCvM(Psi5, Psi_in_X = TRUE, p = 5)), 0.1124227,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PCvM(Psi9, Psi_in_X = TRUE, p = 9)), 0.2766211,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PCvM(Psi200, Psi_in_X = TRUE, p = 200)), 0.1553891,
               tolerance = 1e-6)

})

test_that("PCvM vs. Bakshaev", {

  expect_equal(sph_stat_PCvM(X3), 0.125 * sph_stat_Bakshaev(X3))
  expect_equal(sph_stat_PCvM(Psi3, Psi_in_X = TRUE, p = 3),
               0.125 * sph_stat_Bakshaev(Psi3, Psi_in_X = TRUE, p = 3))

})

test_that("PRt with X", {

  expect_equal(drop(sph_stat_PRt(X2)), 0.0994200, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PRt(X3)), 0.1188291, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PRt(X4)), 0.3213707, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PRt(X5)), 0.1477722, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PRt(X9)), 0.3794311, tolerance = 1e-6)
  expect_equal(drop(sph_stat_PRt(X200)), 0.2059544, tolerance = 1e-6)

})

test_that("PRt with Psi", {

  expect_equal(drop(sph_stat_PRt(Psi2, Psi_in_X = TRUE, p = 2)), 0.0994200,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PRt(Psi3, Psi_in_X = TRUE, p = 3)), 0.1188291,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PRt(Psi4, Psi_in_X = TRUE, p = 4)), 0.3213707,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PRt(Psi5, Psi_in_X = TRUE, p = 5)), 0.1477722,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PRt(Psi9, Psi_in_X = TRUE, p = 9)), 0.3794311,
               tolerance = 1e-6)
  expect_equal(drop(sph_stat_PRt(Psi200, Psi_in_X = TRUE, p = 200)), 0.2059544,
               tolerance = 1e-6)

})

test_that("PRt vs. Ajne", {

  expect_equal(sph_stat_PRt(X2, t = 0.5), sph_stat_Ajne(X2))
  expect_equal(sph_stat_PRt(X3, t = 0.5), sph_stat_Ajne(X3))
  expect_equal(sph_stat_PRt(X4, t = 0.5), sph_stat_Ajne(X4))
  expect_equal(sph_stat_PRt(X5, t = 0.5), sph_stat_Ajne(X5))
  expect_equal(sph_stat_PRt(X9, t = 0.5), sph_stat_Ajne(X9))
  expect_equal(sph_stat_PRt(X200, t = 0.5), sph_stat_Ajne(X200))

})

test_that("Riesz vs. Bakshaev", {

  expect_equal(sph_stat_Bakshaev(X2), sph_stat_Riesz(X2, s = 1),
               tolerance = 1e-5)
  expect_equal(sph_stat_Bakshaev(X3), sph_stat_Riesz(X3, s = 1),
               tolerance = 1e-5)
  expect_equal(sph_stat_Bakshaev(X4), sph_stat_Riesz(X4, s = 1),
               tolerance = 1e-5)
  expect_equal(sph_stat_Bakshaev(X5), sph_stat_Riesz(X5, s = 1),
               tolerance = 1e-5)
  expect_equal(sph_stat_Bakshaev(X9), sph_stat_Riesz(X9, s = 1),
               tolerance = 1e-5)
  expect_equal(sph_stat_Bakshaev(X200), sph_stat_Riesz(X200, s = 1),
               tolerance = 1e-5)

})

test_that("Riesz vs. Pycke", {

  expect_equal((n - 1) / (2 * n) * sph_stat_Pycke(X2),
               sph_stat_Riesz(X2, s = 0))
  expect_equal((log(4) - 1) / 2 + (2 * pi * (n - 1) / n) * sph_stat_Pycke(X3),
               sph_stat_Riesz(X3, s = 0))
  expect_warning(expect_equal(sph_stat_Pycke(X4), sph_stat_Riesz(X4, s = 0)))
  expect_warning(expect_equal(sph_stat_Pycke(X5), sph_stat_Riesz(X5, s = 0)))
  expect_warning(expect_equal(sph_stat_Pycke(X9), sph_stat_Riesz(X9, s = 0)))
  expect_warning(expect_equal(sph_stat_Pycke(X200),
                              sph_stat_Riesz(X200, s = 0)))
  expect_warning(sph_stat_Pycke(X4))
  expect_warning(sph_stat_Pycke(X5))
  expect_warning(sph_stat_Pycke(X9))
  expect_warning(sph_stat_Pycke(X200))

})

test_that("sph_stat with psi_in_X = TRUE", {

  expect_equal(sph_stat_Ajne(Psi2, Psi_in_X = TRUE), sph_stat_Ajne(X2))
  expect_equal(sph_stat_Ajne(Psi3, Psi_in_X = TRUE), sph_stat_Ajne(X3))
  expect_equal(sph_stat_Gine_Gn(Psi2, Psi_in_X = TRUE, p = 2),
               sph_stat_Gine_Gn(X2))
  expect_equal(sph_stat_Gine_Gn(Psi3, Psi_in_X = TRUE, p = 3),
               sph_stat_Gine_Gn(X3))
  expect_equal(sph_stat_Gine_Fn(Psi2, Psi_in_X = TRUE, p = 2),
               sph_stat_Gine_Fn(X2))
  expect_equal(sph_stat_Gine_Fn(Psi3, Psi_in_X = TRUE, p = 3),
               sph_stat_Gine_Fn(X3))
  expect_equal(sph_stat_Pycke(cos(Psi2), Psi_in_X = TRUE, p = 2),
               sph_stat_Pycke(X2))
  expect_equal(sph_stat_Pycke(cos(Psi3), Psi_in_X = TRUE, p = 3),
               sph_stat_Pycke(X3))

})

test_that("CJ12", {

  expect_equal(sph_stat_CJ12(X3, regime = 1), sph_stat_CJ12(X3, regime = 2))
  pn <- 3
  expect_equal(drop(sph_stat_CJ12(X3, regime = 1) -
                      sph_stat_CJ12(X3, regime = 3)),
               4 * log(n) - log(log(n)) -
                 (4 * pn * (pn - 2) * log(n) - log(pn)))

})

test_that("CCF09", {

  expect_equal(1 - p_Kolmogorov(sph_stat_CCF09(X3, dirs = rd3)),
               sph_stat_CCF09(X3, dirs = rd3, original = TRUE))
  expect_error(sph_stat_CCF09(X4, dirs = rd3))

})

test_that("Edge cases", {

  expect_error(sph_stat_Gine_Gn(Psi2, Psi_in_X = TRUE))
  expect_error(sph_stat_Gine_Fn(Psi2, Psi_in_X = TRUE))
  expect_error(sph_stat_Pycke(Psi2, Psi_in_X = TRUE))
  expect_error(sph_stat_Bakshaev(Psi2, Psi_in_X = TRUE))
  expect_error(sph_stat_PCvM(Psi2, Psi_in_X = TRUE))
  expect_error(sph_stat_PAD(Psi2, Psi_in_X = TRUE))
  expect_error(sph_stat_PRt(Psi2, Psi_in_X = TRUE))
  expect_error(sph_stat_CJ12(Psi2, Psi_in_X = TRUE))

})

test_that("Edge cases psi", {

  expect_error(sphunif:::sph_stat_PCvM_Psi(cbind(drop(Psi2)), n = n, p = 1,
                                           th_grid = 1:10, int_grid = 1:10))
  expect_error(sphunif:::sph_stat_PAD_Psi(cbind(drop(Psi2)), n = n, p = 1,
                                          th_grid = 1:10, int_grid = 1:10))
  expect_error(sphunif:::sph_stat_PRt_Psi(cbind(drop(Psi2)), n = n, p = 1,
                                          t_m = 0.5, theta_t_m = 1,
                                          th_grid = 1:10, int_grid = 1:10))
  expect_error(sphunif:::sph_stat_Pycke_Psi(cbind(drop(Psi2)), n = n, p = 1))
  expect_error(sphunif:::sph_stat_Pycke_Psi(cbind(drop(Psi2)), n = n, p = 5))

})

# Circular projected statistic with weight w
cir_stat_Pn <- function(samp, w, N = 2560, ...) {

  # Integration nodes and weighs
  x_k <- Gauss_Legen_nodes(a = -1, b = 1, N = N)
  wx_k <- Gauss_Legen_weights(a = -1, b = 1, N = N)
  s <- sort(x_k, index.return = TRUE)
  x_k <- x_k[s$ix]
  wx_k <- wx_k[s$ix]
  th_k <- Gauss_Legen_nodes(a = 0, b = 2 * pi, N = N)
  wth_k <- Gauss_Legen_weights(a = 0, b = 2 * pi, N = N)

  # Integral on x
  int <- function(gamma) {

    Fn <- sphunif:::ecdf_bin(data = drop(samp[, , 1] %*% gamma), sorted_x = x_k)
    Fp <- p_proj_unif(x = x_k, p = 2)
    fp <- d_proj_unif(x = x_k, p = 2)
    sum(wx_k * (Fn - Fp)^2 * w(Fp, ...) * fp)

  }

  # integral on gamma
  n * sum(wth_k * apply(Theta_to_X(th_k)[, , 1], 1, int)) / (2 * pi)

}

# Parameters
n <- 50
set.seed(1223213)
samp <- r_unif_sph(n = n, p = 2)
w_PCvM <- function(x) rep(1, length(x))
w_k <- function(x, k = 1) -2 * k^2 * pi^2 * cos(2 * pi * k * x)
w_k_plus <- function(x, k = 1) pmax(w_k(x, k = k), 0)
w_k_minus <- function(x, k = 1) pmax(-w_k(x, k = k), 0)

test_that("cir_stat_Pn equals PCvM", {

  skip_on_cran()
  expect_equal(cir_stat_Pn(samp, w = w_PCvM), drop(sph_stat_PCvM(samp)),
               tolerance = 1e-4)

})

test_that("cir_stat_Pn equals Rayleigh", {

  skip_on_cran()
  expect_equal(cir_stat_Pn(samp, w = w_k, k = 1),
               drop(sph_stat_Rayleigh(samp) / 2),
               tolerance  = 1e-4)
  expect_equal(cir_stat_Pn(samp, w = w_k_plus, k = 1) -
                 cir_stat_Pn(samp, w = w_k_minus, k = 1),
               drop(sph_stat_Rayleigh(samp) / 2),
               tolerance = 1e-4)

})

test_that("cir_stat_Pn equals Bingham", {

  skip_on_cran()
  expect_equal(cir_stat_Pn(samp, w = w_k, k = 2),
               drop(sph_stat_Bingham(samp) / 2),
               tolerance  = 1e-4)
  expect_equal(cir_stat_Pn(samp, w = w_k_plus, k = 2) -
                 cir_stat_Pn(samp, w = w_k_minus, k = 2),
               drop(sph_stat_Bingham(samp) / 2), tolerance = 1e-4)

})

set.seed(1231323)
Theta <- r_unif_cir(n = 10, M = 2)
X <- r_unif_sph(n = 10, p = 2, M = 2)
Theta_s <- sphunif:::sort_each_col(Theta)
n <- 200
samp <- rnorm(n)
samp_s <- sort(samp)
x <- seq(-1, 1, l = 10)
y <- seq(0, 1, l = 20)

test_that("Theta_to_X and X_to_Theta", {

  expect_equal(Theta, X_to_Theta(Theta_to_X(Theta)))
  expect_equal(X, Theta_to_X(X_to_Theta(X)))
  expect_error(Theta_to_X(Theta[, 1]))
  expect_error(X_to_Theta(X[, , 1]))
  expect_error(X_to_Theta(r_unif_sph(n = 10, p = 3, M = 2)))
  expect_error(Theta_to_X(X))
  expect_error(X_to_Theta(Theta))

})

test_that("cir_gaps", {

  expect_equal(cir_gaps(Theta), cbind(cir_gaps(Theta[, 1, drop = FALSE]),
                                      cir_gaps(Theta[, 2, drop = FALSE])))
  expect_equal(cir_gaps(Theta), cir_gaps(Theta_s, sorted = TRUE))
  expect_equal(cir_gaps(Theta), rbind(diff(Theta_s),
                                      2 * pi - (Theta_s[10, ] - Theta_s[1, ])))

})

test_that("ecdf_bin", {

  expect_equal(sphunif:::ecdf_bin(samp, x),
               sphunif:::ecdf_bin(samp_s, x, data_sorted = TRUE))
  expect_equal(sphunif:::ecdf_bin(samp, x),
               sphunif:::ecdf_bin(samp_s, x, data_sorted = TRUE))
  expect_equal(ecdf(samp)(x), drop(sphunif:::ecdf_bin(samp, x, efic = TRUE)))
  expect_equal(n * ecdf(samp)(x),
               drop(sphunif:::ecdf_bin(samp, x, divide_n = FALSE)))
  expect_equal(sphunif:::ecdf_bin(samp, x, efic = FALSE),
               sphunif:::ecdf_bin(samp, x, efic = TRUE))

})

test_that("beta_inc and beta_inc_inv", {

  expect_equal(y, drop(sphunif:::beta_inc_inv(
    sphunif:::beta_inc(y, a = 1, b = 2), a = 1, b = 2)))
  expect_equal(pmax(x, 0), drop(sphunif:::beta_inc_inv(
    sphunif:::beta_inc(x, a = 1, b = 2), a = 1, b = 2)))
  expect_equal(y, drop(sphunif:::beta_inc(
    sphunif:::beta_inc_inv(y, a = 1, b = 2), a = 1, b = 2)))
  expect_equal(y, drop(sphunif:::beta_inc(
    sphunif:::beta_inc_inv(y, a = 1, b = 2, lower_tail = TRUE),
    a = 1, b = 2, lower_tail = TRUE)))
  expect_equal(exp(sphunif:::beta_inc(y, a = 1, b = 2, log = TRUE)),
               sphunif:::beta_inc(y, a = 1, b = 2))
  expect_equal(sphunif:::beta_inc_inv(log(y), a = 1, b = 2, log = TRUE),
               sphunif:::beta_inc_inv(y, a = 1, b = 2))
  expect_equal(drop(sphunif:::beta_inc(x, 0.75, 2)), pbeta(x, 0.75, 2))
  expect_equal(drop(sphunif:::beta_inc_inv(y, 0.75, 2)), qbeta(y, 0.75, 2))

})

test_that("t_inv_sqrt_one and n_from_dist_vector", {

  expect_equal(drop(sphunif:::t_inv_sqrt_one(x)), x / sqrt(1 - x^2))
  expect_equal(drop(sphunif:::n_from_dist_vector(
    sum(lower.tri(tcrossprod(1:n), diag = FALSE)))), n)

})

k <- 0:5
k1 <- 1:5
psi_PCvM_2 <- function(th) psi_Pn(theta = th, q = 1, type = "PCvM")
psi_PCvM_3 <- function(th) psi_Pn(theta = th, q = 2, type = "PCvM")
psi_PCvM_4 <- function(th) psi_Pn(theta = th, q = 3, type = "PCvM")
psi_PCvM_11 <- function(th) psi_Pn(theta = th, q = 10, type = "PCvM")
psi_PAD_2 <- function(th) psi_Pn(theta = th, q = 1, type = "PAD")
psi_PAD_3 <- function(th) psi_Pn(theta = th, q = 2, type = "PAD")
psi_PAD_4 <- function(th) psi_Pn(theta = th, q = 3, type = "PAD")
psi_PAD_11 <- function(th) psi_Pn(theta = th, q = 10, type = "PAD")
psi_PRt_2 <- function(th) psi_Pn(theta = th, q = 1, type = "PRt")
psi_PRt_3 <- function(th) psi_Pn(theta = th, q = 2, type = "PRt")
psi_PRt_4 <- function(th) psi_Pn(theta = th, q = 3, type = "PRt")
psi_PRt_11 <- function(th) psi_Pn(theta = th, q = 10, type = "PRt")

test_that("Gegen_coefs (Gauss = TRUE) vs. *_Pn (Gauss = FALSE) for PCvM", {

  expect_equal(Gegen_coefs(psi = psi_PCvM_2, k = k, p = 2, Gauss = TRUE),
               Gegen_coefs_Pn(k = k, p = 2, type = "PCvM", Gauss = FALSE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PCvM_3, k = k, p = 3, Gauss = TRUE),
               Gegen_coefs_Pn(k = k, p = 3, type = "PCvM", Gauss = FALSE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PCvM_4, k = k, p = 4, Gauss = TRUE),
               Gegen_coefs_Pn(k = k, p = 4, type = "PCvM", Gauss = FALSE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PCvM_11, k = k, p = 11, Gauss = TRUE),
               Gegen_coefs_Pn(k = k, p = 11, type = "PCvM", Gauss = FALSE),
               tolerance = 1e-6)

})

test_that("Gegen_coefs (Gauss = FALSE) vs. *_Pn (Gauss = TRUE) for PCvM", {

  expect_equal(Gegen_coefs(psi = psi_PCvM_2, k = k, p = 2, Gauss = FALSE),
               Gegen_coefs_Pn(k = k, p = 2, type = "PCvM", Gauss = TRUE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PCvM_3, k = k, p = 3, Gauss = FALSE),
               Gegen_coefs_Pn(k = k, p = 3, type = "PCvM", Gauss = TRUE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PCvM_4, k = k, p = 4, Gauss = FALSE),
               Gegen_coefs_Pn(k = k, p = 4, type = "PCvM", Gauss = TRUE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PCvM_11, k = k, p = 11, Gauss = FALSE),
               Gegen_coefs_Pn(k = k, p = 11, type = "PCvM", Gauss = TRUE),
               tolerance = 1e-6)

})

test_that("Gegen_coefs (Gauss = TRUE) vs. *_Pn (Gauss = FALSE) for PAD", {

  expect_equal(Gegen_coefs(psi = psi_PAD_2, k = k, p = 2, Gauss = TRUE),
               Gegen_coefs_Pn(k = k, p = 2, type = "PAD", Gauss = FALSE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PAD_3, k = k, p = 3, Gauss = TRUE),
               Gegen_coefs_Pn(k = k, p = 3, type = "PAD", Gauss = FALSE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PAD_4, k = k, p = 4, Gauss = TRUE),
               Gegen_coefs_Pn(k = k, p = 4, type = "PAD", Gauss = FALSE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PAD_11, k = k, p = 11, Gauss = TRUE),
               Gegen_coefs_Pn(k = k, p = 11, type = "PAD", Gauss = FALSE),
               tolerance = 1e-6)

})

test_that("Gegen_coefs (Gauss = FALSE) vs. *_Pn (Gauss = TRUE) for PAD", {

  expect_equal(Gegen_coefs(psi = psi_PAD_2, k = k, p = 2, Gauss = FALSE),
               Gegen_coefs_Pn(k = k, p = 2, type = "PAD", Gauss = TRUE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PAD_3, k = k, p = 3, Gauss = FALSE),
               Gegen_coefs_Pn(k = k, p = 3, type = "PAD", Gauss = TRUE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PAD_4, k = k, p = 4, Gauss = FALSE),
               Gegen_coefs_Pn(k = k, p = 4, type = "PAD", Gauss = TRUE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PAD_11, k = k, p = 11, Gauss = FALSE),
               Gegen_coefs_Pn(k = k, p = 11, type = "PAD", Gauss = TRUE),
               tolerance = 1e-6)

})

test_that("Gegen_coefs (Gauss = TRUE) vs. *_Pn (Gauss = FALSE) for PRt", {

  expect_equal(Gegen_coefs(psi = psi_PRt_2, k = k, p = 2, Gauss = TRUE),
               Gegen_coefs_Pn(k = k, p = 2, type = "PRt", Gauss = FALSE),
               tolerance = 1e-5)
  expect_equal(Gegen_coefs(psi = psi_PRt_3, k = k, p = 3, Gauss = TRUE),
               Gegen_coefs_Pn(k = k, p = 3, type = "PRt", Gauss = FALSE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PRt_4, k = k, p = 4, Gauss = TRUE),
               Gegen_coefs_Pn(k = k, p = 4, type = "PRt", Gauss = FALSE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PRt_11, k = k, p = 11, Gauss = TRUE),
               Gegen_coefs_Pn(k = k, p = 11, type = "PRt", Gauss = FALSE),
               tolerance = 1e-6)

})

test_that("Gegen_coefs (Gauss = FALSE) vs. *_Pn (Gauss = TRUE) for PRt", {

  expect_equal(Gegen_coefs(psi = psi_PRt_2, k = k, p = 2, Gauss = FALSE),
               Gegen_coefs_Pn(k = k, p = 2, type = "PRt", Gauss = TRUE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PRt_3, k = k, p = 3, Gauss = FALSE),
               Gegen_coefs_Pn(k = k, p = 3, type = "PRt", Gauss = TRUE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PRt_4, k = k, p = 4, Gauss = FALSE),
               Gegen_coefs_Pn(k = k, p = 4, type = "PRt", Gauss = TRUE),
               tolerance = 1e-6)
  expect_equal(Gegen_coefs(psi = psi_PRt_11, k = k, p = 11, Gauss = FALSE),
               Gegen_coefs_Pn(k = k, p = 11, type = "PRt", Gauss = TRUE),
               tolerance = 1e-6)

})

test_that("Gegen_coefs_Pn edge cases", {

  expect_error(Gegen_coefs_Pn(k = k, p = 11, type = "PRt2", Gauss = TRUE))

})

test_that("Gegen_coefs_Pn verbose", {

  expect_message(Gegen_coefs_Pn(k = k, p = 2, type = "PCvM",
                                verbose = TRUE), NA)
  expect_message(Gegen_coefs_Pn(k = k, p = 3, type = "PCvM",
                                verbose = TRUE), NA)
  expect_message(Gegen_coefs_Pn(k = k, p = 4, type = "PCvM",
                                verbose = TRUE), NA)
  expect_message(Gegen_coefs_Pn(k = k, p = 5, type = "PCvM",
                                verbose = TRUE))
  expect_message(Gegen_coefs_Pn(k = k, p = 2, type = "PAD",
                                verbose = TRUE), NA)
  expect_message(Gegen_coefs_Pn(k = k, p = 4, type = "PAD",
                                verbose = TRUE))
  expect_message(Gegen_coefs_Pn(k = k, p = 2, type = "PRt",
                                verbose = TRUE), NA)
  expect_message(Gegen_coefs_Pn(k = k, p = 3, type = "PRt",
                                verbose = TRUE), NA)

})

n <- 5
set.seed(123456789)
X2 <- r_unif_sph(n = n, p = 2)
X3 <- r_unif_sph(n = n, p = 3)
X4 <- r_unif_sph(n = n, p = 4)
X5 <- r_unif_sph(n = n, p = 5)
X11 <- r_unif_sph(n = n, p = 11)
Psi2 <- Psi_mat(X2)
Psi3 <- Psi_mat(X3)
Psi4 <- Psi_mat(X4)
Psi5 <- Psi_mat(X5)
Psi11 <- Psi_mat(X11)
dim(Psi2) <- c(dim(Psi2), 1)
dim(Psi3) <- c(dim(Psi3), 1)
dim(Psi4) <- c(dim(Psi4), 1)
dim(Psi5) <- c(dim(Psi5), 1)
dim(Psi11) <- c(dim(Psi11), 1)
Th2 <- X_to_Theta(X2)
Psi_to_mat <- function(Psi) {

  Psi_mat <- matrix(0, nrow = n, ncol = n)
  Psi_mat[upper.tri(Psi_mat)] <- Psi
  return(Psi_mat + t(Psi_mat))

}

test_that("PCvM vs. psi_Pn", {

  expect_equal(2 / n * sum(psi_Pn(theta = Psi2, q = 1, type = "PCvM")) +
                 (3 - 2 * n) / 6,
               drop(sph_stat_PCvM(Psi2, Psi_in_X = TRUE, p = 2)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi3, q = 2, type = "PCvM")) +
                 (3 - 2 * n) / 6,
               drop(sph_stat_PCvM(Psi3, Psi_in_X = TRUE, p = 3)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi4, q = 3, type = "PCvM")) +
                 (3 - 2 * n) / 6,
               drop(sph_stat_PCvM(Psi4, Psi_in_X = TRUE, p = 4)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi5, q = 4, type = "PCvM")) +
                 (3 - 2 * n) / 6,
               drop(sph_stat_PCvM(Psi5, Psi_in_X = TRUE, p = 5)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi11, q = 10, type = "PCvM")) +
                 (3 - 2 * n) / 6,
               drop(sph_stat_PCvM(Psi11, Psi_in_X = TRUE, p = 11)),
               tolerance = 5e-5)

})

test_that("PCvM vs. psi_Pn (psi_Gauss = FALSE)", {

  expect_equal(2 / n * sum(psi_Pn(theta = Psi5, q = 4, type = "PCvM",
                                  psi_Gauss = FALSE)) +
                 (3 - 2 * n) / 6,
               drop(sph_stat_PCvM(Psi5, Psi_in_X = TRUE, p = 5)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi11, q = 10, type = "PCvM",
                                  psi_Gauss = FALSE)) +
                 (3 - 2 * n) / 6,
               drop(sph_stat_PCvM(Psi11, Psi_in_X = TRUE, p = 11)),
               tolerance = 5e-5)

})

test_that("PAD vs. psi_Pn", {

  expect_equal(2 / n * sum(psi_Pn(theta = Psi2, q = 1, type = "PAD")) + n,
               drop(sph_stat_PAD(Psi2, Psi_in_X = TRUE, p = 2)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi3, q = 2, type = "PAD")) + n,
               drop(sph_stat_PAD(Psi3, Psi_in_X = TRUE, p = 3)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi4, q = 3, type = "PAD")) + n,
               drop(sph_stat_PAD(Psi4, Psi_in_X = TRUE, p = 4)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi5, q = 4, type = "PAD")) + n,
               drop(sph_stat_PAD(Psi5, Psi_in_X = TRUE, p = 5)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi11, q = 10, type = "PAD")) + n,
               drop(sph_stat_PAD(Psi11, Psi_in_X = TRUE, p = 11)),
               tolerance = 1e-5)

})

test_that("PAD vs. psi_Pn (psi_Gauss = FALSE)", {

  expect_equal(2 / n * sum(psi_Pn(theta = Psi3, q = 2, type = "PAD",
                                  psi_Gauss = FALSE)) + n,
               drop(sph_stat_PAD(Psi3, Psi_in_X = TRUE, p = 3)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi4, q = 3, type = "PAD",
                                  psi_Gauss = FALSE)) + n,
               drop(sph_stat_PAD(Psi4, Psi_in_X = TRUE, p = 4)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi5, q = 4, type = "PAD",
                                  psi_Gauss = FALSE)) + n,
               drop(sph_stat_PAD(Psi5, Psi_in_X = TRUE, p = 5)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi11, q = 10, type = "PAD",
                                  psi_Gauss = FALSE)) + n,
               drop(sph_stat_PAD(Psi11, Psi_in_X = TRUE, p = 11)),
               tolerance = 1e-5)

})

test_that("PRt vs. psi_Pn", {

  expect_equal(2 / n * sum(psi_Pn(theta = Psi2, q = 1, type = "PRt")) +
                 (1 - n) / 2 + n * (1 / 3 * (1 - 1 / 3)),
               drop(sph_stat_PRt(Psi2, Psi_in_X = TRUE, p = 2)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi3, q = 2, type = "PRt")) +
                 (1 - n) / 2 + n * (1 / 3 * (1 - 1 / 3)),
               drop(sph_stat_PRt(Psi3, Psi_in_X = TRUE, p = 3)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi4, q = 3, type = "PRt")) +
                 (1 - n) / 2 + n * (1 / 3 * (1 - 1 / 3)),
               drop(sph_stat_PRt(Psi4, Psi_in_X = TRUE, p = 4)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi5, q = 4, type = "PRt")) +
                 (1 - n) / 2 + n * (1 / 3 * (1 - 1 / 3)),
               drop(sph_stat_PRt(Psi5, Psi_in_X = TRUE, p = 5)),
               tolerance = 1e-5)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi11, q = 10, type = "PRt")) +
                 (1 - n) / 2 + n * (1 / 3 * (1 - 1 / 3)),
               drop(sph_stat_PRt(Psi11, Psi_in_X = TRUE, p = 11)),
               tolerance = 1e-5)

})

test_that("PRt vs. psi_Pn (psi_Gauss = FALSE)", {

  expect_equal(2 / n * sum(psi_Pn(theta = Psi5, q = 4, type = "PRt",
                                  psi_Gauss = FALSE)) +
                 (1 - n) / 2 + n * (1 / 3 * (1 - 1 / 3)),
               drop(sph_stat_PRt(Psi5, Psi_in_X = TRUE, p = 5)),
               tolerance = 1e-6)
  expect_equal(2 / n * sum(psi_Pn(theta = Psi11, q = 10, type = "PRt",
                                  psi_Gauss = FALSE)) +
                 (1 - n) / 2 + n * (1 / 3 * (1 - 1 / 3)),
               drop(sph_stat_PRt(Psi11, Psi_in_X = TRUE, p = 11)),
               tolerance = 1e-6)

})

test_that("Gegen_coefs_Pn PRt for p = 2)", {

  expect_equal(Gegen_coefs_Pn(k = k1, p = 2, type = "PRt"),
               exp(log(2) - 2 * log(k1 * pi) + log(sin(k1 * pi * 1 / 3)^2)))
  expect_equal(Gegen_coefs_Pn(k = k1, p = 2, type = "PRt", Rothman_t = 1 / 2),
               exp(log(2) - 2 * log(k1 * pi) + log(sin(k1 * pi * 1 / 2)^2)))

})

test_that("PCvM vs. psi_Pn tilde", {

  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi2)), q = 1, type = "PCvM",
                          tilde = TRUE)) / n,
               drop(sph_stat_PCvM(Psi2, Psi_in_X = TRUE, p = 2)),
               tolerance = 1e-5)
  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi3)), q = 2, type = "PCvM",
                          tilde = TRUE)) / n,
               drop(sph_stat_PCvM(Psi3, Psi_in_X = TRUE, p = 3)),
               tolerance = 1e-5)
  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi4)), q = 3, type = "PCvM",
                          tilde = TRUE)) / n,
               drop(sph_stat_PCvM(Psi4, Psi_in_X = TRUE, p = 4)),
               tolerance = 1e-5)
  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi5)), q = 4, type = "PCvM",
                          tilde = TRUE)) / n,
               drop(sph_stat_PCvM(Psi5, Psi_in_X = TRUE, p = 5)),
               tolerance = 1e-5)
  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi11)), q = 10, type = "PCvM",
                          tilde = TRUE)) / n,
               drop(sph_stat_PCvM(Psi11, Psi_in_X = TRUE, p = 11)),
               tolerance = 5e-5)

})

test_that("PAD vs. psi_Pn tilde", {

  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi2)), q = 1, type = "PAD",
                          tilde = TRUE)) / n,
               drop(sph_stat_PAD(Psi2, Psi_in_X = TRUE, p = 2)),
               tolerance = 1e-5)
  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi3)), q = 2, type = "PAD",
                          tilde = TRUE)) / n,
               drop(sph_stat_PAD(Psi3, Psi_in_X = TRUE, p = 3)),
               tolerance = 5e-5)
  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi4)), q = 3, type = "PAD",
                          tilde = TRUE)) / n,
               drop(sph_stat_PAD(Psi4, Psi_in_X = TRUE, p = 4)),
               tolerance = 1e-5)
  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi5)), q = 4, type = "PAD",
                          tilde = TRUE)) / n,
               drop(sph_stat_PAD(Psi5, Psi_in_X = TRUE, p = 5)),
               tolerance = 1e-5)
  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi11)), q = 10, type = "PAD",
                          tilde = TRUE)) / n,
               drop(sph_stat_PAD(Psi11, Psi_in_X = TRUE, p = 11)),
               tolerance = 1e-5)

})

test_that("PRt vs. psi_Pn tilde", {

  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi2)), q = 1, type = "PRt",
                          tilde = TRUE)) / n,
               drop(sph_stat_PRt(Psi2, Psi_in_X = TRUE, p = 2)),
               tolerance = 1e-5)
  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi3)), q = 2, type = "PRt",
                          tilde = TRUE)) / n,
               drop(sph_stat_PRt(Psi3, Psi_in_X = TRUE, p = 3)),
               tolerance = 1e-5)
  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi4)), q = 3, type = "PRt",
                          tilde = TRUE)) / n,
               drop(sph_stat_PRt(Psi4, Psi_in_X = TRUE, p = 4)),
               tolerance = 1e-5)
  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi5)), q = 4, type = "PRt",
                          tilde = TRUE)) / n,
               drop(sph_stat_PRt(Psi5, Psi_in_X = TRUE, p = 5)),
               tolerance = 1e-5)
  expect_equal(sum(psi_Pn(theta = c(Psi_to_mat(Psi11)), q = 10, type = "PRt",
                          tilde = TRUE)) / n,
               drop(sph_stat_PRt(Psi11, Psi_in_X = TRUE, p = 11)),
               tolerance = 1e-5)

})

x <- seq(-1, 1, l = 5)
k0 <- 0

test_that("Gegen_coefs_Pn edge cases", {

  expect_equal(Gegen_coefs_Pn(k = k0, p = 2, type = "PCvM", Gauss = TRUE),
               1 / 3)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 2, type = "PCvM", Gauss = FALSE),
               1 / 3)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 3, type = "PCvM", Gauss = TRUE),
               1 / 3)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 3, type = "PCvM", Gauss = FALSE),
               1 / 3)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 5, type = "PCvM", Gauss = TRUE),
               1 / 3)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 5, type = "PCvM", Gauss = FALSE),
               1 / 3)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 2, type = "PAD", Gauss = TRUE), -1)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 2, type = "PAD", Gauss = FALSE), -1)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 3, type = "PAD", Gauss = TRUE), -1)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 3, type = "PAD", Gauss = FALSE), -1)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 5, type = "PAD", Gauss = TRUE), -1)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 5, type = "PAD", Gauss = FALSE), -1)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 2, type = "PRt", Gauss = TRUE),
               5 / 18)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 2, type = "PRt", Gauss = FALSE),
               5 / 18)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 3, type = "PRt", Gauss = TRUE),
               5 / 18)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 3, type = "PRt", Gauss = FALSE),
               5 / 18)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 5, type = "PRt", Gauss = TRUE),
               5 / 18)
  expect_equal(Gegen_coefs_Pn(k = k0, p = 5, type = "PRt", Gauss = FALSE),
               5 / 18)

})

test_that("psi_Pn edge cases", {

  expect_error(psi_Pn(theta = Psi2, q = 1, type = "PRt2"))

})

test_that("axk edge cases", {

  expect_equal(akx(x = x, k = k, p = 2, sqr = TRUE)^2,
               akx(x = x, k = k, p = 2, sqr = FALSE))
  expect_equal(akx(x = x, k = k, p = 5, sqr = TRUE)^2,
               akx(x = x, k = k, p = 5, sqr = FALSE))
  expect_equal(akx(x = x[1], k = k, p = 2),
               akx(x = x, k = k, p = 2)[1, , drop = FALSE])
  expect_equal(akx(x = x[1], k = k, p = 5),
               akx(x = x, k = k, p = 5)[1, , drop = FALSE])
  expect_equal(akx(x = x, k = k[1], p = 2),
               akx(x = x, k = k, p = 2)[, 1, drop = FALSE])
  expect_equal(akx(x = x, k = k[1], p = 5),
               akx(x = x, k = k, p = 5)[, 1, drop = FALSE])
  expect_equal(akx(x = x[1], k = k[1], p = 2),
               akx(x = x, k = k, p = 2)[1, 1, drop = FALSE])
  expect_equal(akx(x = x[1], k = k[1], p = 5),
               akx(x = x, k = k, p = 5)[1, 1, drop = FALSE])

})

f_PRt_2 <- f_locdev_Pn(p = 2, type = "PRt")
f_PRt_3 <- f_locdev_Pn(p = 3, type = "PRt")
f_PRt_4 <- f_locdev_Pn(p = 4, type = "PRt")
f_PRt_11 <- f_locdev_Pn(p = 11, type = "PRt")
f_PCvM_2 <- f_locdev_Pn(p = 2, type = "PCvM")
f_PCvM_3 <- f_locdev_Pn(p = 3, type = "PCvM")
f_PCvM_4 <- f_locdev_Pn(p = 4, type = "PCvM")
f_PCvM_11 <- f_locdev_Pn(p = 11, type = "PCvM")
f_PAD_2 <- f_locdev_Pn(p = 2, type = "PAD")
f_PAD_3 <- f_locdev_Pn(p = 3, type = "PAD")
f_PAD_4 <- f_locdev_Pn(p = 4, type = "PAD")
f_PAD_11 <- f_locdev_Pn(p = 11, type = "PAD")

test_that("Integral one for f_locdev_Pn with PRt", {

  expect_equal(int_sph_MC(f = function(x) f_PRt_2(x[, 1]), p = 2, M = 1e3,
                          chunks = 1, seeds = 1), 1,
               tolerance = 3e-2)
  expect_equal(int_sph_MC(f = function(x) f_PRt_3(x[, 1]), p = 3, M = 1e3,
                          chunks = 1, seeds = 1), 1,
               tolerance = 3e-2)
  expect_equal(int_sph_MC(f = function(x) f_PRt_4(x[, 1]), p = 4, M = 1e3,
                          chunks = 1, seeds = 1), 1,
               tolerance = 3e-2)
  expect_equal(int_sph_MC(f = function(x) f_PRt_11(x[, 1]), p = 11, M = 1e3,
                          chunks = 1, seeds = 1), 1,
               tolerance = 3e-2)

})

test_that("Integral one for f_locdev_Pn with PCvM", {

  expect_equal(int_sph_MC(f = function(x) f_PCvM_2(x[, 1]), p = 2, M = 1e3,
                          chunks = 1, seeds = 1), 1,
               tolerance = 3e-2)
  expect_equal(int_sph_MC(f = function(x) f_PCvM_3(x[, 1]), p = 3, M = 1e3,
                          chunks = 1, seeds = 1), 1,
               tolerance = 3e-2)
  expect_equal(int_sph_MC(f = function(x) f_PCvM_4(x[, 1]), p = 4, M = 1e3,
                          chunks = 1, seeds = 1), 1,
               tolerance = 3e-2)
  expect_equal(int_sph_MC(f = function(x) f_PCvM_11(x[, 1]), p = 11, M = 1e3,
                          chunks = 1, seeds = 1), 1,
               tolerance = 3e-2)

})

test_that("Integral one for f_locdev_Pn with PAD", {

  expect_equal(int_sph_MC(f = function(x) f_PAD_2(x[, 1]), p = 2, M = 1e3,
                          chunks = 1, seeds = 1), 1,
               tolerance = 7e-2)
  expect_equal(int_sph_MC(f = function(x) f_PAD_3(x[, 1]), p = 3, M = 1e3,
                          chunks = 1, seeds = 1), 1,
               tolerance = 7e-2)
  expect_equal(int_sph_MC(f = function(x) f_PAD_4(x[, 1]), p = 4, M = 1e3,
                          chunks = 1, seeds = 1), 1,
               tolerance = 7e-2)
  expect_equal(int_sph_MC(f = function(x) f_PAD_11(x[, 1]), p = 11, M = 1e3,
                          chunks = 1, seeds = 1), 1,
               tolerance = 7e-2)

})

n <- 5
X <- r_unif_sph(n = n, p = 2, M = 2)
Theta <- X_to_Theta(X)
dim(Theta) <- c(n, 1, 2)
set.seed(98720222)
A <- matrix(rnorm(4 * 7), nrow = 4, ncol = 7)
B1 <- tcrossprod(X[, , 1])
B2 <- tcrossprod(X[, , 2])
b <- acos(c(B1[upper.tri(B1)], B2[upper.tri(B2)]))

test_that("upper_tri_ind", {

  expect_equal(drop(upper_tri_ind(5)) + 1,
               {I <- matrix(1:5^2, 5, 5); I[upper.tri(I)]})
  expect_equal(drop(upper_tri_ind(7)) + 1,
               {I <- matrix(1:7^2, 7, 7); I[upper.tri(I)]})

})

test_that("sort_each_col and sort_index_each_col", {

  expect_equal(sphunif:::sort_each_col(A), apply(A, 2, sort))
  expect_equal(sphunif:::sort_index_each_col(A), apply(A, 2, order))
  expect_equal(sphunif:::sort_index_each_col(
    sphunif:::sort_index_each_col(A)), apply(A, 2, rank))

})

test_that("Psi with X/Theta and use_ind_tri = TRUE/FALSE", {

  expect_equal(Psi_mat(X, use_ind_tri = FALSE),
               Psi_mat(Theta, use_ind_tri = FALSE))
  expect_equal(Psi_mat(X, use_ind_tri = FALSE),
               Psi_mat(X, ind_tri = upper_tri_ind(n), use_ind_tri = TRUE))
  expect_equal(Psi_mat(Theta, use_ind_tri = FALSE),
               Psi_mat(Theta, ind_tri = upper_tri_ind(n), use_ind_tri = TRUE))
  expect_equal(Psi_mat(X, use_ind_tri = FALSE),
               Psi_mat(Theta, ind_tri = upper_tri_ind(n), use_ind_tri = TRUE))
  expect_equal(Psi_mat(Theta, use_ind_tri = FALSE),
               Psi_mat(X, ind_tri = upper_tri_ind(n), use_ind_tri = TRUE))

})

test_that("Psi with angle_diff = TRUE/FALSE", {

  expect_equal(Psi_mat(X, angles_diff = TRUE),
               Psi_mat(X, angles_diff = FALSE))
  expect_equal(Psi_mat(X, angles_diff = TRUE),
               Psi_mat(Theta, angles_diff = FALSE))
  expect_equal(acos(cos(Psi_mat(Theta, angles_diff = TRUE))),
               Psi_mat(Theta, angles_diff = FALSE))
  expect_equal(acos(cos(Psi_mat(Theta, angles_diff = TRUE))),
               Psi_mat(X))
  expect_equal(acos(cos(Psi_mat(Theta, angles_diff = TRUE,
                                ind_tri =  upper_tri_ind(n),
                                use_ind_tri = TRUE))),
               Psi_mat(X))

})

test_that("Psi with scalar_prod = TRUE/FALSE", {

  expect_equal(Psi_mat(Theta, scalar_prod = TRUE),
               Psi_mat(Theta, scalar_prod = FALSE))
  expect_equal(Psi_mat(X, scalar_prod = FALSE),
               Psi_mat(Theta, scalar_prod = TRUE))
  expect_equal(acos(Psi_mat(X, scalar_prod = TRUE)),
               Psi_mat(X, scalar_prod = FALSE))
  expect_equal(acos(Psi_mat(X, scalar_prod = TRUE)),
               Psi_mat(Theta))
  expect_equal(acos(Psi_mat(X, scalar_prod = TRUE,
                            ind_tri =  upper_tri_ind(n),
                            use_ind_tri = TRUE)),
               Psi_mat(X))

})

test_that("Psi reconstruction", {

  expect_equal(c(Psi_mat(Theta)), b)
  expect_equal(c(Psi_mat(X)), b)
  expect_equal(Psi_mat(X)[, 1],
               {D <- acos(1 - as.matrix(dist(X[, , 1]))^2 / 2);
               D[upper.tri(D)]})
  expect_equal(Psi_mat(X)[, 2],
               {D <- acos(1 - as.matrix(dist(X[, , 2]))^2 / 2);
               D[upper.tri(D)]})
  expect_equal(Psi_mat(Theta)[, 1],
               {D <- acos(1 - as.matrix(dist(X[, , 1]))^2 / 2);
               D[upper.tri(D)]})
  expect_equal(Psi_mat(Theta)[, 2],
               {D <- acos(1 - as.matrix(dist(X[, , 2]))^2 / 2);
               D[upper.tri(D)]})
})

n <- 10
set.seed(123456789)
Theta_1 <- r_unif_cir(n = n, M = 1)
Theta_2 <- r_unif_cir(n = n, M = 2)
Psi_1 <- Psi_mat(Theta_to_X(Theta_1))
Psi_2 <- Psi_mat(Theta_to_X(Theta_2))
h <- function(th) 0.5 * (th^2 / (4 * pi^2) - th / (2 * pi) + 1 / 6)

test_that("Kolmogorov-Smirnov", {

  expect_equal(drop(cir_stat_Kuiper(Theta_1, KS = TRUE)),
               as.numeric(sqrt(n) * ks.test(Theta_1, y = "punif",
                                            min = 0, max = 2 * pi)$statistic))
  expect_equal(drop(cir_stat_Kuiper(Theta_2, KS = TRUE)),
               sqrt(n) * as.numeric(c(
                   ks.test(Theta_2[, 1], y = "punif",
                           min = 0, max = 2 * pi)$statistic,
                   ks.test(Theta_2[, 2], y = "punif",
                           min = 0, max = 2 * pi)$statistic)))

})

test_that("Cramér-von Mises", {

  expect_equal(drop(cir_stat_Watson(Theta_1, CvM = TRUE)),
               as.numeric(goftest::cvm.test(Theta_1, null = "punif",
                                            min = 0, max = 2 * pi)$statistic))
  expect_equal(drop(cir_stat_Watson(Theta_2, CvM = TRUE)),
               as.numeric(c(
                 goftest::cvm.test(Theta_2[, 1], null = "punif",
                                   min = 0, max = 2 * pi)$statistic,
                 goftest::cvm.test(Theta_2[, 2], null = "punif",
                                   min = 0, max = 2 * pi)$statistic)))

})

test_that("Anderson-Darling", {

  expect_equal(drop(cir_stat_PAD(Theta_1, AD = TRUE)),
               as.numeric(goftest::ad.test(Theta_1, null = "punif",
                                            min = 0, max = 2 * pi)$statistic))
  expect_equal(drop(cir_stat_PAD(Theta_2, AD = TRUE)),
               as.numeric(c(
                 goftest::ad.test(Theta_2[, 1], null = "punif",
                                  min = 0, max = 2 * pi)$statistic,
                 goftest::ad.test(Theta_2[, 2], null = "punif",
                                  min = 0, max = 2 * pi)$statistic)))

})

test_that("Stephens modifications", {

  expect_equal(drop(cir_stat_Kuiper(Theta_1, KS = TRUE, Stephens = TRUE)),
               drop(cir_stat_Kuiper(Theta_1, KS = TRUE, Stephens = FALSE)) *
                 (1 + 0.12 / sqrt(n) + 0.21 / n))
  expect_equal(drop(cir_stat_Kuiper(Theta_1, Stephens = TRUE)),
               drop(cir_stat_Kuiper(Theta_1, Stephens = FALSE)) *
                 (1 + 0.155 / sqrt(n) + 0.24 / n))
  expect_equal(drop(cir_stat_Watson(Theta_1, CvM = TRUE, Stephens = TRUE)),
               (drop(cir_stat_Watson(Theta_1, CvM = TRUE, Stephens = FALSE)) -
                  0.4 / n + 0.6 / n^2) * (1 + 1 / n))
  expect_equal(drop(cir_stat_Watson(Theta_1, Stephens = TRUE)),
               (drop(cir_stat_Watson(Theta_1, Stephens = FALSE)) -
                  0.1 / n + 0.1 / n^2) * (1 + 0.8 / n))

})

test_that("PCvM vs. Watson", {

  expect_equal(cir_stat_PCvM(Theta_2), 2 * cir_stat_Watson(Theta_2))
  expect_equal(sph_stat_PCvM(Theta_to_X(Theta_1)), 2 * cir_stat_Watson(Theta_1))

})


test_that("Watson form in MJ (2000, page 111)", {

  expect_equal(drop(cir_stat_Watson(Theta_1)), 2 * sum(h(Psi_1)) / n + 1 / 12)

})

test_that("Hodges_Ajne use_Cressie and asymp_std", {

  expect_equal(cir_stat_Hodges_Ajne(Theta_1, use_Cressie = TRUE),
               cir_stat_Hodges_Ajne(Theta_1, use_Cressie = FALSE))
  expect_equal(cir_stat_Hodges_Ajne(Theta_2, use_Cressie = TRUE),
               cir_stat_Hodges_Ajne(Theta_2, use_Cressie = FALSE))
  expect_equal(cir_stat_Hodges_Ajne(Theta_2, asymp_std = TRUE),
               (2 * cir_stat_Hodges_Ajne(Theta_2) - n) / sqrt(n))

})

test_that("cir_stat_Range max_gap", {

  expect_equal(2 * pi - cir_stat_Range(Theta_1, max_gap = FALSE),
               cir_stat_Range(Theta_1, max_gap = TRUE))

})

test_that("Watson_1976 minus", {

  expect_equal(drop(cir_stat_Watson_1976(Theta = Theta_1, minus = TRUE)),
               as.numeric(sqrt(n) * ks.test(Theta_1, y = "punif",
                                            min = 0, max = 2 * pi,
                                            alternative = "less")$statistic -
                            sqrt(n) * (mean(Theta_1 / (2 * pi)) - 1 / 2)))
  expect_equal(drop(cir_stat_Watson_1976(Theta = Theta_2, minus = TRUE)),
               apply(Theta_2, 2, function(x) {
                 as.numeric(sqrt(n) * ks.test(x, y = "punif",
                                              min = 0, max = 2 * pi,
                                              alternative = "less")$statistic -
                              sqrt(n) * (mean(x / (2 * pi)) - 1 / 2))
                 }))

})

test_that("Gine_Fn constructed from Gine_Gn and Ajne", {

  expect_equal(cir_stat_Gine_Fn(Theta_1),
               cir_stat_Gine_Gn(Theta_1) + 4 * cir_stat_Ajne(Theta_1))
  expect_equal(cir_stat_Gine_Fn(Psi_2, Psi_in_Theta = TRUE),
               cir_stat_Gine_Gn(Psi_2, Psi_in_Theta = TRUE) +
                 4 * cir_stat_Ajne(Psi_2, Psi_in_Theta = TRUE))

})

test_that("Bakshaev is a particular case of Riesz", {

  expect_equal(cir_stat_Riesz(Theta_2, s = 1), cir_stat_Bakshaev(Theta_2))
  expect_equal(cir_stat_Riesz(Psi_2, s = 1, Psi_in_Theta = TRUE),
               cir_stat_Bakshaev(Psi_2, Psi_in_Theta = TRUE))
  expect_equal(cir_stat_Bakshaev(Theta_2),
               cir_stat_Bakshaev(Psi_2, Psi_in_Theta = TRUE))

})

## Retrieve data from https://astrogeology.usgs.gov/search/map/Venus/venuscraters

# The source is https://agupubs.onlinelibrary.wiley.com/doi/full/10.1002/2015JE004940

# Download
download.file(url = paste0("https://agupubs.onlinelibrary.wiley.com/action/",
                           "downloadSupplement?doi=10.1002%2F2015JE004940",
                           "&file=jgre20485-sup-0002-TableS1.txt"),
              destfile = "jgre20485-sup-0002-tables1.txt")

## Import and preprocess data

# Read data
rhea <- read.table(file = "jgre20485-sup-0002-TableS1.txt", skip = 2,
                   sep = "\t", dec = ".")
names(rhea) <- c("Center_Longitude", "Center_Latitude", "diameter", "name")

# Transform degrees to radians
rhea$theta <- (rhea$Center_Longitude / 180 * pi) %% (2 * pi)
rhea$phi <- rhea$Center_Latitude / 180 * pi
rhea$Center_Latitude <- NULL
rhea$Center_Longitude <- NULL

# Reorder
rhea <- rhea[, c(2, 1, 3:4)]

# Save object
save(list = "rhea", file = "rhea.rda", compress = "bzip2")

#' @title Conversion between angular and Cartesian coordinates of the
#' (hyper)sphere
#'
#' @description Transforms the angles \eqn{(\theta_1,\ldots,\theta_{p-1})'} in
#' \eqn{[0,\pi)^{p-2}\times[-\pi,\pi)} into the Cartesian coordinates
#' \deqn{(\cos(x_1),\sin(x_1)\cos(x_2),\ldots,
#' \sin(x_1)\cdots\sin(x_{p-2})\cos(x_{p-1}),
#' \sin(x_1)\cdots\sin(x_{p-2})\sin(x_{p-1}))'}
#' of \eqn{S^{p-1}}, and vice versa.
#'
#' @param theta matrix of size \code{c(n, p - 1)} with the angles.
#' @param x matrix of size \code{c(n, p)} with the Cartesian coordinates.
#' Assumed to be of unit norm by rows.
#' @return For \code{angles_to_sphere}, the matrix \code{x}. For
#' \code{sphere_to_angles}, the matrix \code{theta}.
#' @examples
#' # Check changes of coordinates
#' sphere_to_angles(angles_to_sphere(c(pi / 2, 0, pi)))
#' sphere_to_angles(angles_to_sphere(rbind(c(pi / 2, 0, pi), c(pi, pi / 2, 0))))
#' angles_to_sphere(sphere_to_angles(c(0, sqrt(0.5), sqrt(0.1), sqrt(0.4))))
#' angles_to_sphere(sphere_to_angles(
#'   rbind(c(0, sqrt(0.5), sqrt(0.1), sqrt(0.4)),
#'         c(0, sqrt(0.5), sqrt(0.5), 0),
#'         c(0, 1, 0, 0),
#'         c(0, 0, 0, -1),
#'         c(0, 0, 1, 0))))
#'
#' # Circle
#' sphere_to_angles(angles_to_sphere(0))
#' sphere_to_angles(angles_to_sphere(cbind(0:3)))
#' angles_to_sphere(cbind(sphere_to_angles(rbind(c(0, 1), c(1, 0)))))
#' angles_to_sphere(cbind(sphere_to_angles(rbind(c(0, 1)))))
#' @export
angles_to_sphere <- function(theta) {

  # Convert to matrix
  if (!is.matrix(theta)) {

    theta <- rbind(theta)

  }

  # Apply transformation
  q <- ncol(theta)
  x <- matrix(0, nrow = nrow(theta), ncol = q + 1)
  cos_theta <- cos(theta)
  sin_theta <- sin(theta)
  if (q == 1) {

    x <- cbind(cos_theta, sin_theta)

  } else {

    prod_sin_theta <- t(apply(sin_theta, 1, cumprod))
    x <- cbind(cos_theta[, 1, drop = FALSE],
               prod_sin_theta[, -q, drop = FALSE] *
                 cos_theta[, -1, drop = FALSE],
               prod_sin_theta[, q, drop = FALSE])

  }
  return(unname(x))

}


#' @rdname angles_to_sphere
#' @export
sphere_to_angles <- function(x) {

  # Convert to matrix
  if (!is.matrix(x)) {

    x <- rbind(x)

  }

  # Check unit norm
  x <- rotasym::check_unit_norm(x)

  # Apply transformation
  p <- ncol(x)
  i_norm <- t(apply(x^2, 1, function(x) rev(sqrt(cumsum(rev(x))))))[, -p]
  theta <- x[, -p, drop = FALSE] / i_norm
  theta[is.nan(theta)] <- 1 # Avoid NaNs
  theta <- acos(theta)
  xp <- (x[, p] < 0)
  theta[, p - 1] <- (2 * pi) * xp + (1 - 2 * xp) * theta[, p - 1]
  return(unname(theta))

}


#' @title (Hyper)spherical harmonics
#'
#' @description Computation of a certain explicit representation of
#' (hyper)spherical harmonics on
#' \eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{
#' S^{p-1}:=\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}. Details are available in
#' García-Portugués et al. (2021).
#'
#' @param x locations in \eqn{S^{p-1}} to evaluate \eqn{g_{i,k}}. Either a
#' matrix of size \code{c(nx, p)} or a vector of size \code{p}. Normalized
#' internally if required (with a \code{warning} message).
#' @param i,k alternative indexing to refer to the \code{i}-th (hyper)spherical
#' harmonic of order \code{k}. \code{i} is a positive integer smaller than
#' \code{\link[=Sobolev]{d_p_k}} and \code{k} is a non-negative integer.
#'
#' @param m (hyper)spherical harmonic index, as used in Proposition 2.1. The
#' index is computed internally from \code{i} and \code{k}. Defaults to
#' \code{NULL}.
#' @param show_m flag to print \code{m} if computed internally when
#' \code{m = NULL}.
#' @return A vector of size \code{nrow(x)}.
#' @details
#' The implementation uses Proposition 2.1 in García-Portugués et al. (2021),
#' which adapts Theorem 1.5.1 in Dai and Xu (2013) with the correction of
#' typos in the normalizing constant \eqn{h_\alpha} and in the definition of
#' the function \eqn{g_\alpha} of the latter theorem.
#' @references
#' Dai, F. and Xu, Y. (2013). \emph{Approximation Theory and Harmonic Analysis
#' on Spheres and Balls}. Springer, New York. \doi{10.1007/978-1-4614-6660-4}
#'
#' García-Portugués, E., Paindaveine, D., and Verdebout, T. (2021). On the
#' power of Sobolev tests for isotropy under local rotationally symmetric
#' alternatives. \emph{arXiv:2108.09874}. \url{https://arxiv.org/abs/2108.09874}
#' @examples
#' n <- 3e3
#' old_par <- par(mfrow = c(2, 3))
#' k <- 2
#' for (i in 1:d_p_k(p = 3, k = k)) {
#'   X <- r_unif_sph(n = n, p = 3, M = 1)[, , 1]
#'   col <- rainbow(n)[rank(g_i_k(x = X, k = k, i = i, show_m = TRUE))]
#'   scatterplot3d::scatterplot3d(X[, 1], X[, 2], X[, 3], color = col,
#'                                axis = FALSE, pch = 19)
#' }
#' for (k in 0:5) {
#'   X <- r_unif_sph(n = n, p = 3, M = 1)[, , 1]
#'   col <- rainbow(n)[rank(g_i_k(x = X, k = k, i = 1, show_m = TRUE))]
#'   scatterplot3d::scatterplot3d(X[, 1], X[, 2], X[, 3], color = col,
#'                                axis = FALSE, pch = 19)
#' }
#' par(old_par)
#' @export
#' @name harmonics


#' @rdname harmonics
g_i_k <- function(x, i = 1, k = 1, m = NULL, show_m = FALSE) {

  # Set x as a matrix
  if (!is.matrix(x)) {

    x <- rbind(x)

  }

  # Check unit norm
  x <- rotasym::check_unit_norm(x)

  # Get dimension
  p <- ncol(x)

  # Obtain m if not supplied
  if (is.null(m)) {

    # Initial check
    dpk <- d_p_k(p = p, k = k)
    stopifnot(1 <= i && i <= dpk + 0.1)

    # Possible m's for k
    ms <- as.matrix(do.call(expand.grid, c(rep(list(0:k), p - 1), list(0:1))))
    ms <- ms[rowSums(ms) == k, , drop = FALSE]
    stopifnot(abs(dpk - nrow(ms)) < 0.1)

    # Selected m
    m <- ms[i, ]
    if (show_m) message(paste0("m = (", paste0(m, collapse = ", "), ")"))

  } else {

    # Retrieve k for p = 2
    k <- sum(m)

    # Initial checks
    stopifnot(p == length(m))
    stopifnot(all(m >= 0) && m[p] %in% c(0, 1))

  }

  # Obtain angles
  theta <- sphere_to_angles(x = x)[, (p - 1):1, drop = FALSE]

  # Treat differently the p = 2 and p >= 3 cases
  if (p == 2) {

    if (m[2] == 0) {

      g <- cos(k * theta[, 1])

    } else if (m[2] == 1) {

      g <- sin(k * theta[, 1])

    }
    g <- sqrt(2) * g

  } else if (p >= 3) {

    # zeta_m
    if (m[p] == 0) {

      zeta_m <- cos(m[p - 1] * theta[, 1])

    } else if (m[p] == 1) {

      zeta_m <- sin((m[p - 1] + 1) * theta[, 1])

    }

    # b_m
    b_m <- ifelse(m[p - 1] + m[p] > 0, 2, 1)

    # Special notation
    abs_j <- function(m, j) sum(m[j:length(m)])
    lpoch <- function(a, n) lgamma(a + n) - lgamma(a)
    lamb_j <- function(m, j, p) abs_j(m = m, j = j + 1) + (p - j - 1) / 2

    # B_m
    B_m <- 0
    for (j in 1:(p - 2)) {

      la_j <- lamb_j(m = m, j = j, p = p)
      m_j1 <- abs_j(m = m, j = j + 1)
      B_m <- B_m + lfactorial(m[j]) + lpoch(a = (p - j + 1) / 2, n = m_j1) +
        log(m[j] + la_j) - lpoch(a = 2 * la_j, n = m[j]) -
        lpoch(a = (p - j) / 2, n = m_j1) - log(la_j)

    }
    B_m <- exp(log(b_m) + B_m)

    # g
    g <- 1
    for (j in 1:(p - 2)) {

      g <- g * sin(theta[, p - j])^abs_j(m = m, j = j + 1) *
        gsl::gegenpoly_n(n = m[j], lambda = lamb_j(m = m, j = j, p = p),
                         x = cos(theta[, p - j]))

    }
    g <- g * sqrt(B_m) * zeta_m

  }

  return(unname(g))

}

set.seed(12314)
int_prod <- function(x, i1, i2, k1, k2) {
  g_i_k(x = x, k = k1, i = i1) * g_i_k(x = x, k = k2, i = i2)
}

test_that("angles_to_sphere vs. sphere_to_angles", {

  n <- 10
  for (p in c(2:4, 11)) {
    X <- r_unif_sph(n = n, p = p, M = 1)[, , 1]
    th <- t(matrix(runif(n * (p - 1)), nrow = p - 1, ncol = n) *
              c(rep(pi, p - 2), 2 * pi))
    expect_equal(angles_to_sphere(sphere_to_angles(x = X)), X)
    expect_equal(sphere_to_angles(angles_to_sphere(theta = th)), th)
    expect_equal(angles_to_sphere(sphere_to_angles(x = X[1, ])),
                 X[1, , drop = FALSE])
    expect_equal(sphere_to_angles(angles_to_sphere(theta = th[1, ])),
                 th[1, , drop = FALSE])

  }

})

test_that("Harmonics with (i,k) vs. m", {

  skip_on_cran()
  for (p in 2:5) {
    repeat {
      m <- c(rpois(n = p - 1, lambda = 2), sample(c(0, 1), size = 1))
      if (sum(m) > 0) break
    }
    k <- sum(m)
    ms <- as.matrix(do.call(expand.grid, c(rep(list(0:k), p - 1), list(0:1))))
    ms <- ms[rowSums(ms) == k, ]
    i <- which(apply(ms, 1, function(x) all(x == m)))
    X <- r_unif_sph(n = 1, p = p, M = 1)[, , 1]
    expect_message(expect_equal(
      g_i_k(x = X, i = i, k = k, show_m = TRUE), g_i_k(x = X, m = m)))
  }

})

test_that("Orthonormality of harmonics among i's with common k", {

  skip_on_cran()
  M <- 5e3
  for (p in 2:5) {
    for (k in 1:2) {

      dpk <- d_p_k(p = p, k = k)
      intprods <- diag(rep(1, dpk))
      for (i1 in 1:dpk) {
        for (i2 in i1:dpk) {
          intprods[i1, i2] <- int_sph_MC(f = int_prod, i1 = i1, i2 = i2,
                                         k1 = k, k2 = k, p = p, M = M,
                                        cores = 1) /
            rotasym::w_p(p = p)
          image(abs(intprods), zlim = c(0, 1.5))
          expect_lt(max(abs(intprods - diag(rep(1, dpk)))), 0.1)
        }
      }

    }
  }

})

test_that("Orthonormality of harmonics among i's with different k's = 1,2", {

  skip_on_cran()
  M <- 5e3
  for (p in 2:5) {

    dpk1 <- d_p_k(p = p, k = 1)
    dpk2 <- d_p_k(p = p, k = 2)
    intprods <- matrix(0, nrow = dpk1, ncol = dpk2)
    for (i1 in 1:dpk1) {
      for (i2 in 1:dpk2) {
        intprods[i1, i2] <- int_sph_MC(f = int_prod, i1 = i1, i2 = i2,
                                       k1 = 1, k2 = 2, p = p, M = M,
                                       cores = 1) /
          rotasym::w_p(p = p)
        image(abs(intprods), zlim = c(0, 1.5))
        expect_lt(max(abs(intprods)), 0.1)
      }
    }

  }

})

test_that("Orthonormality of harmonics among i's with different k's = 0,1", {

  skip_on_cran()
  M <- 5e3
  for (p in 2:5) {

    dpk0 <- d_p_k(p = p, k = 0)
    dpk1 <- d_p_k(p = p, k = 1)
    intprods <- matrix(0, nrow = dpk0, ncol = dpk1)
    for (i0 in 1:dpk0) {
      for (i1 in 1:dpk1) {
        intprods[i0, i1] <- int_sph_MC(f = int_prod, i1 = i0, i2 = i1,
                                       k1 = 0, k2 = 1, p = p, M = M,
                                       cores = 1) /
          rotasym::w_p(p = p)
        image(abs(intprods), zlim = c(0, 1.5))
        expect_lt(max(abs(intprods)), 0.1)
      }
    }

  }

})

#' @title Statistics for testing circular uniformity
#'
#' @description Low-level implementation of several statistics for assessing
#' circular uniformity on \eqn{[0, 2\pi)} or, equivalently,
#' \eqn{S^1:=\{{\bf x}\in R^2:||{\bf x}||=1\}}{S^1:=\{x\in R^2:||x||=1\}}.
#'
#' @param Theta a \bold{matrix} of size \code{c(n, M)} with \code{M} samples
#' of size \code{n} of circular data on \eqn{[0, 2\pi)}. Must not contain
#' \code{NA}'s.
#' @param gaps_in_Theta does \code{Theta} contain the matrix of
#' \emph{circular gaps} that is obtained with \cr\code{\link{cir_gaps}(Theta)}?
#' If \code{FALSE} (default), the circular gaps are computed internally.
#' @inheritParams cir_gaps
#' @param Psi_in_Theta does \code{Theta} contain the shortest angles matrix
#' \eqn{\boldsymbol\Psi}{\Psi} that is obtained with
#' \cr\code{\link{Psi_mat}(array(Theta, dim = c(n, 1, M)))}? If \code{FALSE}
#' (default), \eqn{\boldsymbol\Psi}{\Psi} is computed internally.
#' @param KS compute the Kolmogorov-Smirnov statistic (which is
#' \emph{not} invariant under origin shifts) instead of the Kuiper statistic?
#' Defaults to \code{FALSE}.
#' @param CvM compute the Cramér-von Mises statistic (which is \emph{not}
#' invariant under origin shifts) instead of the Watson statistic? Defaults to
#' \code{FALSE}.
#' @param AD compute the Anderson-Darling statistic (which is \emph{not}
#' invariant under origin shifts) instead of the Projected Anderson-Darling
#' statistic? Defaults to \code{FALSE}.
#' @param m integer \eqn{m} for the \eqn{m}-modal Rayleigh test. Defaults to
#' \code{m = 1} (the standard Rayleigh test).
#' @param max_gap compute the maximum gap for the range statistic? If
#' \code{TRUE} (default), rejection happens for \emph{large} values of
#' the statistic, which is consistent with the rest of tests. Otherwise,
#' the minimum gap is computed and rejection happens for \emph{low} values.
#' @param asymp_std normalize the Hodges-Ajne statistic in terms of its
#' asymptotic distribution? Defaults to \code{FALSE}.
#' @param use_Cressie compute the Hodges-Ajne statistic as a particular case
#' of the Cressie statistic? Defaults to \code{TRUE} as it is more efficient.
#' If \code{FALSE}, the geometric construction in Ajne (1968) is employed.
#' @param minus compute the invariant \eqn{D_n^-} instead of \eqn{D_n^+}?
#' Defaults to \code{FALSE}.
#' @param a \eqn{a_n = a / n} parameter used in the length of the arcs
#' of the coverage-based tests. Must be positive. Defaults to \code{2 * pi}.
#' @param t \eqn{t} parameter for the Rothman and Cressie tests, a real in
#' \eqn{(0, 1)}. Defaults to \code{1 / 3}.
#' @param q \eqn{q} parameter for the Pycke "\eqn{q}-test", a real in
#' \eqn{(0, 1)}. Defaults to \code{1 / 2}.
#' @param s \eqn{s} parameter for the \eqn{s}-Riesz test, a real in
#' \eqn{(0, 2)}. Defaults to \code{1}.
#' @param abs_val return the absolute value of the Darling's log gaps
#' statistic? If \code{TRUE} (default), rejection happens for \emph{large}
#' values of the statistic, which is consistent with the rest of tests.
#' Otherwise, the signed statistic is computed and rejection happens for
#' large \emph{absolute} values.
#' @param minus_val return the negative value of the (standardized) number of
#' uncovered spacings? If \code{TRUE} (default), rejection happens for
#' \emph{large} values of the statistic, which is consistent with the rest of
#' tests. Otherwise, rejection happens for \emph{low} values.
#' @param dirs a matrix of size \code{c(n_proj, 2)} containing \code{n_proj}
#' random directions (in Cartesian coordinates) on \eqn{S^1} to perform the
#' CCF09 test.
#' @param K_CCF09 integer giving the truncation of the series present in the
#' asymptotic distribution of the Kolmogorov-Smirnov statistic. Defaults to
#' \code{25}.
#' @param original return the CCF09 statistic as originally defined?
#' If \code{FALSE} (default), a faster and equivalent statistic is computed,
#' and rejection happens for \emph{large} values of the statistic, which is
#' consistent with the rest of tests. Otherwise, rejection happens for
#' \emph{low} values.
#' @param Stephens compute Stephens (1970) modification so that the null
#' distribution of the is less dependent on the sample size? The modification
#' does not alter the test decision.
#' @return A matrix of size \code{c(M, 1)} containing the statistics for each
#' of the \code{M} samples.
#' @section Warning:
#' Be careful on avoiding the next bad usages of the functions, which will
#' produce spurious results:
#' \itemize{
#'   \item The entries of \code{Theta} are \emph{not} in \eqn{[0, 2\pi)}.
#'   \item \code{Theta} does \emph{not} contain the circular gaps when
#'   \code{gaps_in_Theta = TRUE}.
#'   \item \code{Theta} is \emph{not} sorted increasingly when
#'   \code{data_sorted = TRUE}.
#'   \item \code{Theta} does \emph{not} contain
#'   \code{\link{Psi_mat}(array(Theta, dim = c(n, 1, M)))} when
#'   \cr\code{Psi_in_Theta = TRUE}.
#'   \item The directions in \code{dirs} do \emph{not} have unit norm.
#' }
#' @details
#' Descriptions and references for most of the statistics are available
#' in García-Portugués and Verdebout (2018).
#'
#' The statistics \code{cir_stat_PCvM} and \code{cir_stat_PRt} are provided
#' for the sake of completion, but they equal the more efficiently-implemented
#' statistics \code{2 * cir_stat_Watson} and \code{cir_stat_Rothman},
#' respectively.
#' @references
#' García-Portugués, E. and Verdebout, T. (2018) An overview of uniformity
#' tests on the hypersphere. \emph{arXiv:1804.00286}.
#' \url{https://arxiv.org/abs/1804.00286}.
#' @examples
#' ## Sample uniform circular data
#'
#' M <- 2
#' n <- 100
#' set.seed(987202226)
#' Theta <- r_unif_cir(n = n, M = M)
#'
#' ## Tests based on the empirical cumulative distribution function
#'
#' # Kuiper
#' cir_stat_Kuiper(Theta)
#' cir_stat_Kuiper(Theta, Stephens = TRUE)
#'
#' # Watson
#' cir_stat_Watson(Theta)
#' cir_stat_Watson(Theta, Stephens = TRUE)
#'
#' # Watson (1976)
#' cir_stat_Watson_1976(Theta)
#'
#' ## Partition-based tests
#'
#' # Ajne
#' Theta_array <- Theta
#' dim(Theta_array) <- c(nrow(Theta), 1, ncol(Theta))
#' Psi <- Psi_mat(Theta_array)
#' cir_stat_Ajne(Theta)
#' cir_stat_Ajne(Psi, Psi_in_Theta = TRUE)
#'
#' # Rothman
#' cir_stat_Rothman(Theta, t = 0.5)
#' cir_stat_Rothman(Theta)
#' cir_stat_Rothman(Psi, Psi_in_Theta = TRUE)
#'
#' # Hodges-Ajne
#' cir_stat_Hodges_Ajne(Theta)
#' cir_stat_Hodges_Ajne(Theta, use_Cressie = FALSE)
#'
#' # Cressie
#' cir_stat_Cressie(Theta, t = 0.5)
#' cir_stat_Cressie(Theta)
#'
#' # FG01
#' cir_stat_FG01(Theta)
#'
#' ## Spacings-based tests
#'
#' # Range
#' cir_stat_Range(Theta)
#'
#' # Rao
#' cir_stat_Rao(Theta)
#'
#' # Greenwood
#' cir_stat_Greenwood(Theta)
#'
#' # Log gaps
#' cir_stat_Log_gaps(Theta)
#'
#' # Vacancy
#' cir_stat_Vacancy(Theta)
#'
#' # Maximum uncovered spacing
#' cir_stat_Max_uncover(Theta)
#'
#' # Number of uncovered spacings
#' cir_stat_Num_uncover(Theta)
#'
#' # Gini mean difference
#' cir_stat_Gini(Theta)
#'
#' # Gini mean squared difference
#' cir_stat_Gini_squared(Theta)
#'
#' ## Sobolev tests
#'
#' # Rayleigh
#' cir_stat_Rayleigh(Theta)
#' cir_stat_Rayleigh(Theta, m = 2)
#'
#' # Bingham
#' cir_stat_Bingham(Theta)
#'
#' # Hermans-Rasson
#' cir_stat_Hermans_Rasson(Theta)
#' cir_stat_Hermans_Rasson(Psi, Psi_in_Theta = TRUE)
#'
#' # Gine Fn
#' cir_stat_Gine_Fn(Theta)
#' cir_stat_Gine_Fn(Psi, Psi_in_Theta = TRUE)
#'
#' # Gine Gn
#' cir_stat_Gine_Gn(Theta)
#' cir_stat_Gine_Gn(Psi, Psi_in_Theta = TRUE)
#'
#' # Pycke
#' cir_stat_Pycke(Theta)
#' cir_stat_Pycke(Psi, Psi_in_Theta = TRUE)
#'
#' # Pycke q
#' cir_stat_Pycke_q(Theta)
#' cir_stat_Pycke_q(Psi, Psi_in_Theta = TRUE)
#'
#' # Bakshaev
#' cir_stat_Bakshaev(Theta)
#' cir_stat_Bakshaev(Psi, Psi_in_Theta = TRUE)
#'
#' # Riesz
#' cir_stat_Riesz(Theta, s = 1)
#' cir_stat_Riesz(Psi, Psi_in_Theta = TRUE, s = 1)
#
#' # Projected Cramér-von Mises
#' cir_stat_PCvM(Theta)
#' cir_stat_PCvM(Psi, Psi_in_Theta = TRUE)
#'
#' # Projected Rothman
#' cir_stat_PRt(Theta, t = 0.5)
#' cir_stat_PRt(Theta)
#' cir_stat_PRt(Psi, Psi_in_Theta = TRUE)
#'
#' # Projected Anderson-Darling
#' cir_stat_PAD(Theta)
#' cir_stat_PAD(Psi, Psi_in_Theta = TRUE)
#'
#' ## Other tests
#'
#' # CCF09
#' dirs <- r_unif_sph(n = 3, p = 2, M = 1)[, , 1]
#' cir_stat_CCF09(Theta, dirs = dirs)
#'
#' ## Connection of Kuiper and Watson statistics with KS and CvM, respectively
#'
#' # Rotate sample for KS and CvM
#' alpha <- seq(0, 2 * pi, l = 1e4)
#' KS_alpha <- sapply(alpha, function(a) {
#'   cir_stat_Kuiper((Theta[, 2, drop = FALSE] + a) %% (2 * pi), KS = TRUE)
#' })
#' CvM_alpha <- sapply(alpha, function(a) {
#'   cir_stat_Watson((Theta[, 2, drop = FALSE] + a) %% (2 * pi), CvM = TRUE)
#' })
#' AD_alpha <- sapply(alpha, function(a) {
#'   cir_stat_PAD((Theta[, 2, drop = FALSE] + a) %% (2 * pi), AD = TRUE)
#' })
#'
#' # Kuiper is the maximum rotated KS
#' plot(alpha, KS_alpha, type = "l")
#' abline(h = cir_stat_Kuiper(Theta[, 2, drop = FALSE]), col = 2)
#' points(alpha[which.max(KS_alpha)], max(KS_alpha), col = 2, pch = 16)
#'
#' # Watson is the minimum rotated CvM
#' plot(alpha, CvM_alpha, type = "l")
#' abline(h = cir_stat_Watson(Theta[, 2, drop = FALSE]), col = 2)
#' points(alpha[which.min(CvM_alpha)], min(CvM_alpha), col = 2, pch = 16)
#'
#' # Anderson-Darling is the average rotated AD?
#' plot(alpha, AD_alpha, type = "l")
#' abline(h = cir_stat_PAD(Theta[, 2, drop = FALSE]), col = 2)
#' abline(h = mean(AD_alpha), col = 3)
#' @name cir_stat
NULL

#' @title Gegenbauer polynomials and coefficients
#'
#' @description The \href{https://dlmf.nist.gov/18.3}{Gegenbauer polynomials}
#' \eqn{\{C_k^{(\lambda)}(x)\}_{k = 0}^\infty}{
#' {C_k^(\lambda)(x)}_{k = 0}^\infty}
#' form a family of orthogonal polynomials on the interval \eqn{[-1, 1]}
#' with respect to the weight function \eqn{(1 - x^2)^{\lambda - 1/2}},
#' for \eqn{\lambda > -1/2}, \eqn{\lambda \neq 0}. They usually appear
#' when dealing with functions defined on
#' \eqn{S^{p-1} := \{{\bf x} \in R^p : ||{\bf x}|| = 1\}}{
#' S^{p-1} := \{x \in R^p : ||x|| = 1\}} with index \eqn{\lambda = p / 2 - 1}.
#'
#' The Gegenbauer polynomials are somehow simpler to evaluate for
#' \eqn{x = \cos(\theta)}, with \eqn{\theta \in [0, \pi]}. This simplifies
#' also the connection with the Chebyshev polynomials
#' \eqn{\{T_k(x)\}_{k = 0}^\infty}{{T_k(x)}_{k = 0}^\infty}, which admit
#' the \href{https://dlmf.nist.gov/18.5.E1}{explicit expression}
#' \eqn{T_k(\cos(\theta)) = \cos(k\theta)}. The Chebyshev polynomials
#' appear as the limit of the Gegenbauer polynomials
#' (divided by \eqn{\lambda}) when \eqn{\lambda} goes to \eqn{0}, so they
#' can be regarded as the extension by continuity of
#' \eqn{\{C_k^{(p/2 - 1)}(x)\}_{k = 0}^\infty}{
#' {C_k^(p/2 - 1)(x)}_{k = 0}^\infty} to the case \eqn{p = 2}.
#'
#' For a \href{https://dlmf.nist.gov/18.18.i}{reasonably smooth} function
#' \eqn{\psi} defined on \eqn{[0, \pi]},
#' \deqn{\psi(\theta) = \sum_{k = 0}^\infty b_{k, p}
#' C_k^{(p/2 - 1)}(\cos(\theta)),}{\psi(\theta) = \sum_{k = 0}^\infty b_{k, p}
#' C_k^(p/2 - 1)(\cos(\theta)),}
#' provided that the coefficients
#' \deqn{b_{k, p} := \frac{1}{c_{k, p}} \int_0^\pi \psi(\theta)
#' C_k^{(p/2 - 1)}(\cos(\theta)) (\sin(\theta))^{p - 2}\,\mathrm{d}\theta}{
#' b_{k, p} := \frac{1}{c_{k, p}} \int_0^\pi \psi(\theta)
#' C_k^(p/2 - 1)(\cos(\theta)) (\sin(\theta))^{p - 2} d\theta}
#' are finite, where the normalizing constants are
#' \deqn{c_{k, p} := \int_0^\pi (C_k^{(p/2 - 1)}(\cos(\theta)))^2
#' (\sin(\theta))^{p - 2} \,\mathrm{d}\theta.}{
#' c_{k, p} := \int_0^\pi (C_k^(p/2 - 1)(\cos(\theta)))^2
#' (\sin(\theta))^{p - 2} d\theta.}
#' The (squared) "Gegenbauer norm" of \eqn{\psi} is
#' \deqn{\|\psi\|_{G, p}^2 := \int_0^\pi \psi(\theta)^2
#' C_k^{(p/2 - 1)}(\cos(\theta)) (\sin(\theta))^{p - 2}\,\mathrm{d}\theta.}{
#' ||\psi||_{G, p}^2 := \int_0^\pi \psi(\theta)^2
#' C_k^(p/2 - 1)(\cos(\theta)) (\sin(\theta))^{p - 2} d\theta.}
#'
#' The previous expansion can be generalized for a 2-dimensional function
#' \eqn{\psi} defined on \eqn{[0, \pi] \times [0, \pi]}:
#' \deqn{\psi(\theta_1, \theta_2) = \sum_{k = 0}^\infty \sum_{m = 0}^\infty
#' b_{k, m, p} C_k^{(p/2 - 1)}(\cos(\theta_1))
#' C_k^{(p/2 - 1)}(\cos(\theta_2)),}{
#' \psi(\theta_1, \theta_2) = \sum_{k = 0}^\infty \sum_{m = 0}^\infty
#' b_{k, m, p} C_k^(p/2 - 1)(\cos(\theta_1)) C_k^(p/2 - 1)(\cos(\theta_2))}
#' with coefficients
#' \deqn{b_{k, m, p} := \frac{1}{c_{k, p} c_{m, p}} \int_0^\pi\int_0^\pi
#' \psi(\theta_1, \theta_2) C_k^{(p/2 - 1)}(\cos(\theta_1))
#' C_k^{(p/2 - 1)}(\cos(\theta_2)) (\sin(\theta_1))^{p - 2}
#' (\sin(\theta_2))^{p - 2}\,\mathrm{d}\theta_1\,\mathrm{d}\theta_2.}{
#' b_{k, m, p} := \frac{1}{c_{k, p} c_{m, p}} \int_0^\pi \int_0^\pi
#' \psi(\theta_1, \theta_2) C_k^(p/2 - 1)(\cos(\theta_1))
#' C_k^(p/2 - 1)(\cos(\theta_2)) (\sin(\theta_1))^{p - 2}
#' (\sin(\theta_2))^{p - 2} d\theta_1 d\theta_2.}
#' The (squared) "Gegenbauer norm" of \eqn{\psi} is
#' \deqn{\|\psi\|_{G, p}^2 := \int_0^\pi\int_0^\pi \psi(\theta_1, \theta_2)^2
#' C_k^{(p/2 - 1)}(\cos(\theta_1)) C_k^{(p/2 - 1)}(\cos(\theta_2))
#' (\sin(\theta_1))^{p - 2} (\sin(\theta_2))^{p - 2}
#' \,\mathrm{d}\theta_1\,\mathrm{d}\theta_2.}{
#' ||\psi||_{G, p}^2 := \int_0^\pi\int_0^\pi \psi(\theta_1, \theta_2)^2
#' C_k^(p/2 - 1)(\cos(\theta_1)) C_k^(p/2 - 1)(\cos(\theta_2))
#' (\sin(\theta_1))^{p - 2} (\sin(\theta_2))^{p - 2} d\theta_1 d\theta_2.}
#'
#' @param theta,theta_1,theta_2 vectors with values in \eqn{[0, \pi]}.
#' @param k,m vectors with the orders of the Gegenbauer polynomials. Must
#' be integers larger or equal than \code{0}.
#' @inheritParams r_unif
#' @param psi function defined in \eqn{[0, \pi]} and whose Gegenbauer
#' coefficients are to be computed. Must be vectorized. For
#' \code{Gegen_coefs_2d}, it must return a matrix of size
#' \code{c(length(theta_1), length(theta_2))}.
#' @param coefs for \code{Gegen_series} and \code{Gegen_norm}, a vector of
#' coefficients \eqn{b_{k, p}} with length \code{length(k)}. For
#' \code{Gegen_series_2d} and \code{Gegen_norm_2d}, a matrix of coefficients
#' \eqn{b_{k, m, p}} with size \code{c(length(k), length(m))}. The
#' order of the coefficients is given by \code{k} and \code{m}.
#' @param Gauss use a Gauss--Legendre quadrature rule of \code{N} nodes
#' in the computation of the Gegenbauer coefficients? Otherwise, call
#' \code{\link{integrate}}. Defaults to \code{TRUE}.
#' @param N number of points used in the \link[=Gauss_Legen_nodes]{
#' Gauss--Legendre quadrature} for computing the Gegenbauer coefficients.
#' Defaults to \code{320}.
#' @param normalize consider normalized coefficients (divided by
#' \eqn{c_{k, p}})? Defaults to \code{TRUE}.
#' @param only_const return only the normalizing constants \eqn{c_{k, p}}?
#' Defaults to \code{FALSE}.
#' @param tol tolerance passed to \code{\link{integrate}}'s \code{rel.tol} and
#' \code{abs.tol} if \code{Gauss = FALSE}. Defaults to \code{1e-6}.
#' @param ... further arguments to be passed to \code{psi}.
#' @param cumulative return the cumulative norm for increasing truncation of
#' the series? Defaults to \code{FALSE}.
#' @return
#' \itemize{
#'   \item \code{Gegen_polyn}: a matrix of size
#'   \code{c(length(theta), length(k))} containing the evaluation of the
#'   \code{length(k)} Gegenbauer polynomials at \code{theta}.
#'   \item \code{Gegen_coefs}: a vector of size \code{length(k)} containing
#'   the coefficients \eqn{b_{k, p}}.
#'   \item \code{Gegen_series}: the evaluation of the truncated series
#'   expansion, a vector of size \code{length(theta)}.
#'   \item \code{Gegen_norm}: the Gegenbauer norm of the truncated series,
#'   a scalar if \code{cumulative = FALSE}, otherwise a vector of size
#'   \code{length(k)}.
#'   \item \code{Gegen_polyn_2d}: a 4-dimensional array of size
#'   \code{c(length(theta_1), length(theta_2), length(k), length(m))}
#'   containing the evaluation of the \code{length(k) * length(m)}
#'   2-dimensional Gegenbauer polynomials at the bivariate grid
#'   spanned by \code{theta_1} and \code{theta_2}.
#'   \item \code{Gegen_coefs_2d}: a matrix of size
#'   \code{c(length(k), length(m))} containing the coefficients
#'   \eqn{b_{k, m, p}}.
#'   \item \code{Gegen_series_2d}: the evaluation of the truncated series
#'   expansion, a matrix of size \code{c(length(theta_1), length(theta_2))}.
#'   \item \code{Gegen_norm_2d}: the 2-dimensional Gegenbauer norm of the
#'   truncated series, a scalar.
#' }
#' @details
#' The \code{Gegen_polyn} function is a wrapper to the functions
#' \code{\link[gsl:Gegenbauer]{gegenpoly_n}} and
#' \code{\link[gsl:Gegenbauer]{gegenpoly_array}} in the
#' \code{\link[gsl]{gsl-package}}, which they interface the functions
#' defined in the header file \code{gsl_sf_gegenbauer.h} (documented
#' \href{https://www.gnu.org/software/gsl/doc/html/specfunc.html#gegenbauer-functions}{
#' here}) of the \href{https://www.gnu.org/software/gsl/}{
#' GNU Scientific Library}.
#'
#' Note that the function \code{Gegen_polyn} computes the regular
#' \emph{unnormalized} Gegenbauer polynomials.
#'
#' For the case \eqn{p = 2}, the Chebyshev polynomials are considered.
#' @references
#' Galassi, M., Davies, J., Theiler, J., Gough, B., Jungman, G., Alken, P.,
#' Booth, M., and Rossi, F. (2009) \emph{GNU Scientific Library Reference
#' Manual}. Network Theory Ltd. \url{http://www.gnu.org/software/gsl/}
#'
#' \emph{NIST Digital Library of Mathematical Functions}. Release
#' 1.0.20 of 2018-09-15. F. W. J. Olver, A. B. Olde Daalhuis, D. W. Lozier,
#' B. I. Schneider, R. F. Boisvert, C. W. Clark, B. R. Miller,
#' and B. V. Saunders, eds. \url{https://dlmf.nist.gov/}
#' @examples
#' ## Representation of Gegenbauer polynomials (Chebyshev polynomials for p = 2)
#'
#' th <- seq(0, pi, l = 500)
#' k <- 0:3
#' old_par <- par(mfrow = c(2, 2))
#' for (p in 2:5) {
#'   matplot(th, t(Gegen_polyn(theta = th, k = k, p = p)), lty = 1,
#'           type = "l", main = substitute(p == d, list(d = p)),
#'           axes = FALSE, xlab = expression(theta), ylab = "")
#'   axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
#'        labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
#'   axis(2); box()
#'   mtext(text = expression({C[k]^{p/2 - 1}}(cos(theta))), side = 2,
#'         line = 2, cex = 0.75)
#'   legend("bottomleft", legend = paste("k =", k), lwd = 2, col = seq_along(k))
#' }
#' par(old_par)
#'
#' ## Coefficients and series in p = 2
#'
#' # Function in [0, pi] to be projected in Chebyshev polynomials
#' psi <- function(th) -sin(th / 2)
#'
#' # Coefficients
#' p <- 2
#' k <- 0:4
#' (coefs <- Gegen_coefs(k = k, p = p, psi = psi))
#'
#' # Series
#' plot(th, psi(th), type = "l", axes = FALSE, xlab = expression(theta),
#'       ylab = "", ylim = c(-1.25, 0))
#' axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
#'      labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
#' axis(2); box()
#' col <- viridisLite::viridis(length(coefs))
#' for (i in seq_along(coefs)) {
#'   lines(th, Gegen_series(theta = th, coefs = coefs[1:(i + 1)], k = 0:i,
#'                          p = p), col = col[i])
#' }
#' lines(th, psi(th), lwd = 2)
#'
#' ## Coefficients and series in p = 3
#'
#' # Function in [0, pi] to be projected in Gegenbauer polynomials
#' psi <- function(th) tan(th / 3)
#'
#' # Coefficients
#' p <- 3
#' k <- 0:10
#' (coefs <- Gegen_coefs(k = k, p = p, psi = psi))
#'
#' # Series
#' plot(th, psi(th), type = "l", axes = FALSE, xlab = expression(theta),
#'       ylab = "", ylim = c(0, 2))
#' axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
#'      labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
#' axis(2); box()
#' col <- viridisLite::viridis(length(coefs))
#' for (i in seq_along(coefs)) {
#'   lines(th, Gegen_series(theta = th, coefs = coefs[1:(i + 1)], k = 0:i,
#'                          p = p), col = col[i])
#' }
#' lines(th, psi(th), lwd = 2)
#' \donttest{
#' ## Surface representation
#'
#' # Surface in [0, pi]^2 to be projected in Gegenbauer polynomials
#' p <- 3
#' psi <- function(th_1, th_2) A_theta_x(theta = th_1, x = cos(th_2),
#'                                       p = p, as_matrix = TRUE)
#'
#' # Coefficients
#' k <- 0:20
#' m <- 0:10
#' coefs <- Gegen_coefs_2d(k = k, m = m, p = p, psi = psi)
#'
#' # Series
#' th <- seq(0, pi, l = 100)
#' col <- viridisLite::viridis(20)
#' old_par <- par(mfrow = c(2, 2))
#' image(th, th, A_theta_x(theta = th, x = cos(th), p = p), axes = FALSE,
#'       col = col, zlim = c(0, 1), xlab = expression(theta[1]),
#'       ylab = expression(theta[2]), main = "Original")
#' axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
#'      labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
#' axis(2, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
#'      labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
#' box()
#' for(K in c(5, 10, 20)) {
#'   A <- Gegen_series_2d(theta_1 = th, theta_2 = th,
#'                        coefs = coefs[1:(K + 1), ], k = 0:K, m = m, p = p)
#'   image(th, th, A, axes = FALSE, col = col, zlim = c(0, 1),
#'         xlab = expression(theta[1]), ylab = expression(theta[2]),
#'         main = paste(K, "x", m[length(m)], "coefficients"))
#'   axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
#'        labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
#'   axis(2, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
#'        labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
#'   box()
#' }
#' par(old_par)
#' }
#' @name Gegenbauer


#' @rdname Gegenbauer
#' @export
Gegen_polyn <- function(theta, k, p) {

  # Check orders
  stopifnot(all(k >= 0))

  # Check dimension
  stopifnot(p >= 2)

  # Check that theta is in range
  stopifnot(all(0 <= theta & theta <= pi))

  # Handle the special case p = 2 (Chebyshev series) separately
  p <- as.integer(p)
  if (p == 2) {

    return(cos(k %o% theta))

  } else {

    # Vectorization on k or not
    if (length(k) == 1) {

      return(rbind(gsl::gegenpoly_n(n = k, lambda = 0.5 * p - 1,
                                    x = cos(theta))))

    } else {

      return(gsl::gegenpoly_array(nmax = max(k), lambda = 0.5 * p - 1,
                                  x = cos(theta))[k + 1, , drop = FALSE])

    }

  }

}


#' @rdname Gegenbauer
#' @export
Gegen_coefs <- function(k, p, psi, Gauss = TRUE, N = 320, normalize = TRUE,
                        only_const = FALSE, tol = 1e-6, ...) {

  # Check orders
  stopifnot(all(k >= 0))

  # Check dimension
  stopifnot(p >= 2)

  # Log-normalizing constants for each projection
  if (normalize) {

    if (p == 2) {

      # Beware if p = 2: the normalizing constant for the zeroth-order Chebyshev
      # term (constant term) is pi / 2, whereas for the rest is pi
      # Check table https://dlmf.nist.gov/18.3 in NIST
      log_const <- log(ifelse(k > 0, pi / 2, pi))

    } else {

      # Normalizing constants worked out from table https://dlmf.nist.gov/18.3
      # in NIST (with lambda = p / 2 - 1)
      log_const <- (4 - p) * log(2) + log(pi) + lgamma(p + k - 2) -
        (log(p + 2 * (k - 1)) + lfactorial(k) + 2 * lgamma(p / 2 - 1))

    }

  } else {

    log_const <- 0

  }

  # Return only the normalizing constants?
  if (only_const) {

    return(exp(log_const))

  }

  # Compute projections
  if (Gauss) {

    # Gauss--Legendre nodes and weigths
    th_k <- drop(Gauss_Legen_nodes(a = 0, b = pi, N = N))
    w_k <- drop(Gauss_Legen_weights(a = 0, b = pi, N = N))

    # Integral approximation, using implicit column recycling for
    # multiplication by vectors
    projs <- colSums((w_k * psi(th_k, ...) * sin(th_k)^(p - 2)) *
                       t(Gegen_polyn(theta = th_k, k = k, p = p)),
                     na.rm = TRUE)

  } else {

    # Use intergrate()
    projs <- sapply(k, function(k_i) {
      f <- function(th) psi(th, ...) * sin(th)^(p - 2) *
        drop(Gegen_polyn(theta = th, k = k_i, p = p))
      integrate(f = f, lower = 0, upper = pi, subdivisions = 1e3,
                rel.tol = tol, abs.tol = tol,
                stop.on.error = FALSE)$value
    })

  }

  # Normalization of the projections
  return(sign(projs) * exp(log(abs(projs)) - log_const))

}


#' @rdname Gegenbauer
#' @export
Gegen_series <- function(theta, coefs, k, p, normalize = TRUE) {

  # Check orders
  stopifnot(all(k >= 0))

  # Check dimension
  stopifnot(p >= 2)

  # Check leength of coefficients
  stopifnot(length(coefs) == length(k))

  # Normalize unnormalized coefficients?
  if (!normalize) {

    coefs <- coefs / Gegen_coefs(k = k, p = p, only_const = TRUE)

  }

  # Series
  return(drop(coefs %*% Gegen_polyn(theta = theta, k = k, p = p)))

}


#' @rdname Gegenbauer
#' @export
Gegen_norm <- function(coefs, k, p, normalize = TRUE, cumulative = FALSE) {

  # Normalizing constants (required for both cases)
  c_kp <- Gegen_coefs(k = k, p = p, only_const = TRUE)

  # Norm
  op <- ifelse(cumulative, cumsum, sum)
  if (normalize) {

    return(sqrt(op(coefs^2 * c_kp)))

  } else {

    return(sqrt(op(coefs^2 / c_kp)))

  }

}


#' @rdname Gegenbauer
#' @export
Gegen_polyn_2d <- function(theta_1, theta_2, k, m, p) {

  # Check orders
  stopifnot(all(k >= 0))
  stopifnot(all(m >= 0))

  # Check dimension
  stopifnot(p >= 2)

  # Check that theta_1 and theta_2 are in range
  stopifnot(all(0 <= theta_1 & theta_1 <= pi))
  stopifnot(all(0 <= theta_2 & theta_2 <= pi))

  # Compute a 4-dimensional array of size length(theta_1) x length(theta_2) x
  # length(k) x length(m) with the k x m surfaces of the 2-dimensional
  # Gegenbauer polynomials
  polyn_2d <- aperm(a = Gegen_polyn(theta = theta_1, k = k, p = p) %o%
                      Gegen_polyn(theta = theta_2, k = m, p = p),
                    perm = c(2, 4, 1, 3))
  return(polyn_2d)

}


#' @rdname Gegenbauer
#' @export
Gegen_coefs_2d <- function(k, m, p, psi, Gauss = TRUE, N = 320,
                           normalize = TRUE, only_const = FALSE, tol = 1e-6,
                           ...) {

  # Check orders
  stopifnot(all(k >= 0))
  stopifnot(all(m >= 0))

  # Check dimension
  stopifnot(p >= 2)

  # Log-normalizing constants for each projection
  if (normalize) {

    if (p == 2) {

      # Beware if p = 2: the normalizing constant for the zeroth-order Chebyshev
      # term (constant term) is pi / 2, whereas for the rest is pi
      # Check table https://dlmf.nist.gov/18.3 in NIST
      log_const_k <- log(ifelse(k > 0, pi / 2, pi))
      log_const_m <- log(ifelse(m > 0, pi / 2, pi))

    } else {

      # Normalizing constants worked out from table https://dlmf.nist.gov/18.3
      # in NIST (with lambda = p / 2 - 1)
      log_const_k <- (4 - p) * log(2) + log(pi) + lgamma(p + k - 2) -
        (log(p + 2 * (k - 1)) + lfactorial(k) + 2 * lgamma(p / 2 - 1))
      log_const_m <- (4 - p) * log(2) + log(pi) + lgamma(p + m - 2) -
        (log(p + 2 * (m - 1)) + lfactorial(m) + 2 * lgamma(p / 2 - 1))

    }

    # Products of univariate normalizing constants
    log_const <- outer(log_const_k, log_const_m, `+`)

  } else {

    log_const <- matrix(0, nrow = length(k), ncol = length(m))

  }

  # Return only the normalizing constants?
  if (only_const) {

    return(exp(log_const))

  }

  # Compute projections
  if (Gauss) {

    # Gauss--Legendre nodes and weigths
    th_k <- drop(Gauss_Legen_nodes(a = 0, b = pi, N = N))
    w_k <- drop(Gauss_Legen_weights(a = 0, b = pi, N = N))

    # Compute polynomials
    polyn_2d <- Gegen_polyn_2d(theta_1 = th_k, theta_2 = th_k, k = k, m = m,
                               p = p)

    # Jacobian
    sin_q <- tcrossprod(sin(th_k)^(p - 2))

    # Expand weights
    w_k <- tcrossprod(w_k)

    # psi evaluated on the bivariate grid
    psi_q <- matrix(psi(th_k, th_k, ...), nrow = N, ncol = N)

    # Integral approximation
    projs <- apply(polyn_2d, c(3, 4),
                   function(C_km) sum(w_k * psi_q * C_km * sin_q, na.rm = TRUE))

  } else {

    # Use intergrate()
    projs <- sapply(m, function(m_i) {
      sapply(k, function(k_i) {

        # Integrand on th_1
        f_1 <- function(th_1) {
          sapply(th_1, function(th) {

            # Integrand on th_2
            f_2 <- function(th_2) {
              drop(psi(th, th_2, ...)) * sin(th)^(p - 2) * sin(th_2)^(p - 2) *
                drop(Gegen_polyn_2d(theta_1 = th, theta_2 = th_2, k = k_i,
                                    m = m_i, p = p))
            }

            # Integral on th_2
            integrate(f = f_2, lower = 0, upper = pi, subdivisions = 1e3,
                      rel.tol = tol, abs.tol = tol, stop.on.error = FALSE)$value

          })
        }

        # Integral on th_1
        integrate(f = f_1, lower = 0, upper = pi, subdivisions = 1e3,
                  rel.tol = tol, abs.tol = tol, stop.on.error = FALSE)$value
      })
    })

  }

  # Normalization of the projections
  return(sign(projs) * exp(log(abs(projs)) - log_const))

}


#' @rdname Gegenbauer
#' @export
Gegen_series_2d <- function(theta_1, theta_2, coefs, k, m, p,
                            normalize = TRUE) {

  # Check orders
  stopifnot(all(k >= 0))
  stopifnot(all(m >= 0))

  # Check dimension
  stopifnot(p >= 2)

  # Check size of coefficients
  stopifnot(nrow(coefs) == length(k) & ncol(coefs) == length(m))

  # Normalize unnormalized coefficients?
  if (!normalize) {

    coefs <- coefs / Gegen_coefs_2d(k = k, m = m, p = p, only_const = TRUE)

  }

  # Compute polynomials
  polyn_2d <- Gegen_polyn_2d(theta_1 = theta_1, theta_2 = theta_2,
                             k = k, m = m, p = p)

  # Add them into the series
  return(apply(apply(polyn_2d, c(1, 2),
                     function(C_km) c(coefs * C_km)), 2:3, sum))

}


#' @rdname Gegenbauer
#' @export
Gegen_norm_2d <- function(coefs, k, m, p, normalize = TRUE) {

  # Normalizing constants (required for both cases)
  c_kmp <- Gegen_coefs_2d(k = k, m = m, p = p, only_const = TRUE)

  # Norm
  if (normalize) {

    return(sqrt(sum(coefs^2 * c_kmp)))

  } else {

    return(sqrt(sum(coefs^2 / c_kmp)))

  }

}

theta <- seq(0, pi, l = 5)
k <- 0:2
coefs_1 <- 4:2
f_1_2 <- function(th) Gegen_series(theta = th, coefs = coefs_1, k = k, p = 2)
f_1_3 <- function(th) Gegen_series(theta = th, coefs = coefs_1, k = k, p = 3)
f_1_4 <- function(th) Gegen_series(theta = th, coefs = coefs_1, k = k, p = 4)
f_1_11 <- function(th) Gegen_series(theta = th, coefs = coefs_1, k = k, p = 11)
I <- diag(1, nrow = length(k), ncol = length(k))

the_1 <- seq(0, pi, l = 4)
the_2 <- seq(0, pi, l = 7)
m <- 0:1
coefs_2 <- 3:1 %*% t(-1:0)
f_2_2 <- function(th_1, th_2) Gegen_series_2d(theta_1 = th_1, theta_2 = th_2,
                                              coefs = coefs_2, k = k, m = m,
                                              p = 2)
f_2_3 <- function(th_1, th_2) Gegen_series_2d(theta_1 = th_1, theta_2 = th_2,
                                              coefs = coefs_2, k = k, m = m,
                                              p = 3)
f_2_4 <- function(th_1, th_2) Gegen_series_2d(theta_1 = th_1, theta_2 = th_2,
                                              coefs = coefs_2, k = k, m = m,
                                              p = 4)
f_2_11 <- function(th_1, th_2) Gegen_series_2d(theta_1 = th_1, theta_2 = th_2,
                                               coefs = coefs_2, k = k, m = m,
                                               p = 11)

th_k <- drop(Gauss_Legen_nodes(a = 0, b = pi, N = 80))
w_k <- drop(Gauss_Legen_weights(a = 0, b = pi, N = 80))
w_k <- tcrossprod(w_k)
sin_k <- sin(th_k) %o% sin(th_k)

test_that("Gegen_coefs orthonormality", {

  for (p in c(2, 3, 4, 11)) {
    expect_equal(sapply(k, function(l) {
      Gegen_coefs(k = k, p = p, psi = function(th)
        drop(Gegen_polyn(theta = th, k = l, p = p)))
    }), I)
  }

})

test_that("Gegen_coefs normalizing constants", {

  for (p in c(2, 3, 4, 11)) {
    expect_equal(Gegen_coefs(k = k, p = p, only_const = TRUE),
                 sapply(k, function(m) {
                   f <- function(th)
                     drop(Gegen_polyn(theta = th, k = m, p = p))^2 *
                     sin(th)^(p - 2)
                   integrate(f = f, lower = 0, upper = pi)$value
                 }))
  }

})

test_that("Gegen_coefs_2d orthonormality", {

  for (p in c(2, 3, 4, 11)) {
    expect_equal(sapply(m, function(l) {
      sum(Gegen_coefs_2d(k = k, m = m, p = p, psi = function(th_1, th_2)
        Gegen_polyn_2d(theta_1 = th_1, theta_2 = th_2, k = l, m = l, p = p),
        N = 40))
    }), rep(1, length(m)), tolerance = 1e-4)
  }

})

test_that("Gegen_coefs_2d normalizing constants", {

  for (p in c(2, 3, 4, 11)) {
    expect_equal(Gegen_coefs_2d(k = k, m = m, p = p, only_const = TRUE),
                 Gegen_coefs(k = k, p = p, only_const = TRUE) %*%
                   t(Gegen_coefs(k = m, p = p, only_const = TRUE)))
  }

})

test_that("Gegen_series (Gauss = TRUE)", {

  expect_equal(Gegen_coefs(k = k, p = 2, psi = f_1_2), coefs_1)
  expect_equal(Gegen_coefs(k = k, p = 3, psi = f_1_3), coefs_1)
  expect_equal(Gegen_coefs(k = k, p = 4, psi = f_1_4), coefs_1)
  expect_equal(Gegen_coefs(k = k, p = 11, psi = f_1_11), coefs_1)

})

test_that("Gegen_series (Gauss = TRUE) with unnormalized coefficients", {

  expect_equal(Gegen_series(theta = theta, p = 2, k = k, normalize = FALSE,
                            coefs = Gegen_coefs(k = k, p = 2, psi = f_1_2,
                                                normalize = FALSE)),
               f_1_2(theta))
  expect_equal(Gegen_series(theta = theta, p = 3, k = k, normalize = FALSE,
                            coefs = Gegen_coefs(k = k, p = 3, psi = f_1_3,
                                                normalize = FALSE)),
               f_1_3(theta))
  expect_equal(Gegen_series(theta = theta, p = 4, k = k, normalize = FALSE,
                            coefs = Gegen_coefs(k = k, p = 4, psi = f_1_4,
                                                normalize = FALSE)),
               f_1_4(theta))
  expect_equal(Gegen_series(theta = theta, p = 11, k = k, normalize = FALSE,
                            coefs = Gegen_coefs(k = k, p = 11, psi = f_1_11,
                                                normalize = FALSE)),
               f_1_11(theta))

})

test_that("Gegen_series (Gauss = FALSE)", {

  expect_equal(Gegen_coefs(k = k, p = 2, psi = f_1_2, Gauss = FALSE), coefs_1)
  expect_equal(Gegen_coefs(k = k, p = 3, psi = f_1_3, Gauss = FALSE), coefs_1)
  expect_equal(Gegen_coefs(k = k, p = 4, psi = f_1_4, Gauss = FALSE), coefs_1)
  expect_equal(Gegen_coefs(k = k, p = 11, psi = f_1_11, Gauss = FALSE), coefs_1)

})

test_that("Gegen_series_2d (Gauss = TRUE)", {

  expect_equal(Gegen_coefs_2d(k = k, m = m, p = 2, psi = f_2_2, N = 40),
               coefs_2)
  expect_equal(Gegen_coefs_2d(k = k, m = m, p = 3, psi = f_2_3, N = 40),
               coefs_2)
  expect_equal(Gegen_coefs_2d(k = k, m = m, p = 4, psi = f_2_4, N = 40),
               coefs_2)
  expect_equal(Gegen_coefs_2d(k = k, m = m, p = 11, psi = f_2_11, N = 40),
               coefs_2)

})

test_that("Gegen_series_2d (Gauss = FALSE)", {

  expect_equal(Gegen_coefs_2d(k = k, m = m, p = 2, psi = f_2_2,
                              Gauss = FALSE, tol = 1e-2), coefs_2)
  expect_equal(Gegen_coefs_2d(k = k, m = m, p = 3, psi = f_2_3,
                              Gauss = FALSE, tol = 1e-2), coefs_2)
  expect_equal(Gegen_coefs_2d(k = k, m = m, p = 4, psi = f_2_4,
                              Gauss = FALSE, tol = 1e-2), coefs_2)
  expect_equal(Gegen_coefs_2d(k = k, m = m, p = 11, psi = f_2_11,
                              Gauss = FALSE, tol = 1e-2), coefs_2)

})

test_that("Gegen_series_2d (Gauss = TRUE) with unnormalized coefficients", {

  expect_equal(Gegen_series_2d(theta_1 = the_1, theta_2 = the_2, p = 2,
                               k = k, m = m, normalize = FALSE,
                               coefs = Gegen_coefs_2d(k = k, m = m, p = 2,
                                                      psi = f_2_2,
                                                      normalize = FALSE)),
               f_2_2(the_1, the_2))
  expect_equal(Gegen_series_2d(theta_1 = the_1, theta_2 = the_2, p = 3,
                               k = k, m = m, normalize = FALSE,
                               coefs = Gegen_coefs_2d(k = k, m = m, p = 3,
                                                      psi = f_2_3,
                                                      normalize = FALSE)),
               f_2_3(the_1, the_2))
  expect_equal(Gegen_series_2d(theta_1 = the_1, theta_2 = the_2, p = 4,
                               k = k, m = m, normalize = FALSE,
                               coefs = Gegen_coefs_2d(k = k, m = m, p = 4,
                                                      psi = f_2_4,
                                                      normalize = FALSE)),
               f_2_4(the_1, the_2))
  expect_equal(Gegen_series_2d(theta_1 = the_1, theta_2 = the_2, p = 11,
                               k = k, m = m, normalize = FALSE,
                               coefs = Gegen_coefs_2d(k = k, m = m, p = 11,
                                                      psi = f_2_11,
                                                      normalize = FALSE)),
               f_2_11(the_1, the_2))

})

test_that("Gegen_norm", {

  expect_equal(Gegen_norm(coefs = coefs_1, k = k, p = 2)^2,
               integrate(f = function(th) f_1_2(th)^2, lower = 0,
                         upper = pi)$value)
  expect_equal(Gegen_norm(coefs = coefs_1, k = k, p = 3)^2,
               integrate(f = function(th) f_1_3(th)^2 * sin(th), lower = 0,
                         upper = pi)$value)
  expect_equal(Gegen_norm(coefs = coefs_1, k = k, p = 4)^2,
               integrate(f = function(th) f_1_4(th)^2 * sin(th)^2, lower = 0,
                         upper = pi)$value)
  expect_equal(Gegen_norm(coefs = coefs_1, k = k, p = 11)^2,
               integrate(f = function(th) f_1_11(th)^2 * sin(th)^9, lower = 0,
                         upper = pi)$value)

})

test_that("Gegen_norm with unnormalized coefficients", {

  expect_equal(Gegen_norm(coefs = Gegen_coefs(k = k, p = 2, psi = f_1_2,
                                              normalize = FALSE),
                          k = k, p = 2, normalize = FALSE)^2,
               integrate(f = function(th) f_1_2(th)^2, lower = 0,
                         upper = pi)$value)
  expect_equal(Gegen_norm(coefs = Gegen_coefs(k = k, p = 3, psi = f_1_3,
                                              normalize = FALSE),
                          k = k, p = 3, normalize = FALSE)^2,
               integrate(f = function(th) f_1_3(th)^2 * sin(th), lower = 0,
                         upper = pi)$value)
  expect_equal(Gegen_norm(coefs = Gegen_coefs(k = k, p = 4, psi = f_1_4,
                                              normalize = FALSE),
                          k = k, p = 4, normalize = FALSE)^2,
               integrate(f = function(th) f_1_4(th)^2 * sin(th)^2, lower = 0,
                         upper = pi)$value)
  expect_equal(Gegen_norm(coefs = Gegen_coefs(k = k, p = 11, psi = f_1_11,
                                              normalize = FALSE),
                          k = k, p = 11, normalize = FALSE)^2,
               integrate(f = function(th) f_1_11(th)^2 * sin(th)^9, lower = 0,
                         upper = pi)$value)

})

test_that("Gegen_norm_2d", {

  expect_equal(Gegen_norm_2d(coefs = coefs_2, k = k, m = m, p = 2)^2,
               sum(w_k * f_2_2(th_k, th_k)^2))
  expect_equal(Gegen_norm_2d(coefs = coefs_2, k = k, m = m, p = 3)^2,
               sum(w_k * f_2_3(th_k, th_k)^2 * sin_k))
  expect_equal(Gegen_norm_2d(coefs = coefs_2, k = k, m = m, p = 4)^2,
               sum(w_k * f_2_4(th_k, th_k)^2 * sin_k^2))
  expect_equal(Gegen_norm_2d(coefs = coefs_2, k = k, m = m, p = 11)^2,
               sum(w_k * f_2_11(th_k, th_k)^2 * sin_k^9))

})

test_that("Gegen_norm_2d with unnormalized coefficients", {

  expect_equal(Gegen_norm_2d(coefs = Gegen_coefs_2d(k = k, m = m, p = 2,
                                                    psi = f_2_2,
                                                    normalize = FALSE,
                                                    N = 80),
                             k = k, m = m, p = 2, normalize = FALSE)^2,
               sum(w_k * f_2_2(th_k, th_k)^2))
  expect_equal(Gegen_norm_2d(coefs = Gegen_coefs_2d(k = k, m = m, p = 3,
                                                    psi = f_2_3,
                                                    normalize = FALSE,
                                                    N = 80),
                             k = k, m = m, p = 3, normalize = FALSE)^2,
               sum(w_k * f_2_3(th_k, th_k)^2 * sin_k))
  expect_equal(Gegen_norm_2d(coefs = Gegen_coefs_2d(k = k, m = m, p = 4,
                                                    psi = f_2_4,
                                                    normalize = FALSE,
                                                    N = 80),
                             k = k, m = m, p = 4, normalize = FALSE)^2,
               sum(w_k * f_2_4(th_k, th_k)^2 * sin_k^2))
  expect_equal(Gegen_norm_2d(coefs = Gegen_coefs_2d(k = k, m = m, p = 11,
                                                    psi = f_2_11,
                                                    normalize = FALSE,
                                                    N = 80),
                             k = k, m = m, p = 11, normalize = FALSE)^2,
               sum(w_k * f_2_11(th_k, th_k)^2 * sin_k^9))
})

#' @title Null distributions for circular and (hyper)spherical uniformity
#' statistics
#'
#' @description Approximate computation of the null distributions of several
#' statistics for assessing uniformity on the (hyper)sphere
#' \eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{S^{p-1}:=
#' \{x\in R^p:||x||=1\}}, \eqn{p\ge 2}. The approximation is done either by
#' means of the asymptotic distribution or by Monte Carlo.
#'
#' @param x evaluation points for the null distribution(s). Either a vector of
#' size \code{nx}, if the evaluation points are common for the tests in
#' \code{type}, or a matrix of size \code{c(nx, length(type))} with columns
#' containing the evaluation points for each test. Must not contain \code{NA}'s.
#' @inheritParams unif_test
#' @inheritParams r_unif
#' @param n sample size employed for computing the statistic.
#' @param approx type of approximation to the null distribution, either
#' \code{"asymp"} (default) for employing the asymptotic null distribution, if
#' available, or \code{"MC"}, for employing the Monte Carlo approximation of
#' the exact null distribution.
#' @param M number of Monte Carlo replications for approximating the null
#' distribution when \code{approx = "MC"}. Also, number of Monte Carlo samples
#' for approximating the asymptotic distributions based on weighted sums of chi
#' squared random variables. Defaults to \code{1e4}.
#' @param stats_MC a data frame of size \code{c(M, length(type))}, with column
#' names containing the character vector \code{type}, that results from
#' extracting \code{$stats_MC} from a call to \code{\link{unif_stat_MC}}. If
#' provided, the computation of Monte Carlo statistics when \code{approx = "MC"}
#' is skipped. \code{stats_MC} is checked internally to see if it is sorted.
#' Internally computed if \code{NULL} (default).
#' @inheritParams unif_stat
#' @inheritParams cir_stat
#' @inheritParams cir_stat_distr
#' @inheritParams sph_stat_distr
#' @param CJ12_beta \eqn{\beta} parameter in the exponential regime of CJ12 test,
#' a positive real.
#' @param K_Kuiper,K_Watson,K_Watson_1976,K_Ajne integer giving the truncation
#' of the series present in the null asymptotic distributions. For the
#' Kolmogorov-Smirnov-related series defaults to \code{25}.
#' @param K_max integer giving the truncation of the series that compute the
#' asymptotic p-value of a Sobolev test. Defaults to \code{1e4}.
#' @param ... if \code{approx = "MC"}, optional performance parameters to be
#' passed to \cr\code{\link{unif_stat_MC}}: \code{chunks}, \code{cores},
#' and \code{seed}.
#' @return A data frame of size \code{c(nx, length(type))}, with column names
#' given by \code{type}, that contains the values of the null distributions of
#' the statistics evaluated at \code{x}.
#' @details
#' When \code{approx = "asymp"}, statistics that do not have an implemented or
#' known asymptotic are omitted, and a warning is generated.
#'
#' For Sobolev tests, \code{K_max = 1e4} produces probabilities uniformly
#' accurate with three digits for the \code{"PCvM"}, \code{"PAD"}, and
#' \code{"PRt"} tests, for dimensions \eqn{p \le 11}. With \code{K_max = 5e4},
#' these probabilities are uniformly accurate in the fourth digit. With
#' \code{K_max = 1e3}, only two-digit uniform accuracy is obtained. Uniform
#' accuracy deteriorates when \eqn{p} increases, e.g., a digit accuracy is lost
#' when \eqn{p = 51}.
#'
#' Descriptions and references on most of the asymptotic distributions
#' are available in García-Portugués and Verdebout (2018).
#' @references
#' García-Portugués, E. and Verdebout, T. (2018) An overview of uniformity
#' tests on the hypersphere. \emph{arXiv:1804.00286}.
#' \url{https://arxiv.org/abs/1804.00286}.
#' @examples
#' ## Asymptotic distribution
#'
#' # Circular statistics
#' x <- seq(0, 1, l = 5)
#' unif_stat_distr(x = x, type = "Kuiper", p = 2, n = 10)
#' unif_stat_distr(x = x, type = c("Ajne", "Kuiper"), p = 2, n = 10)
#' unif_stat_distr(x = x, type = c("Ajne", "Kuiper"), p = 2, n = 10, K_Ajne = 5)
#'\donttest{
#' # All circular statistics
#' unif_stat_distr(x = x, type = avail_cir_tests, p = 2, n = 10, K_max = 1e3)
#' }
#' # Spherical statistics
#' unif_stat_distr(x = cbind(x, x + 1), type = c("Rayleigh", "Bingham"),
#'                 p = 3, n = 10)
#' unif_stat_distr(x = cbind(x, x + 1), type = c("Rayleigh", "Bingham"),
#'                 p = 3, n = 10, M = 100)
#'\donttest{
#' # All spherical statistics
#' unif_stat_distr(x = x, type = avail_sph_tests, p = 3, n = 10, K_max = 1e3)
#'
#' ## Monte Carlo distribution
#'
#' # Circular statistics
#' x <- seq(0, 5, l = 10)
#' unif_stat_distr(x = x, type = avail_cir_tests, p = 2, n = 10, approx = "MC")
#' unif_stat_distr(x = x, type = "Kuiper", p = 2, n = 10, approx = "MC")
#' unif_stat_distr(x = x, type = c("Ajne", "Kuiper"), p = 2, n = 10,
#'                 approx = "MC")
#'
#' # Spherical statistics
#' unif_stat_distr(x = x, type = avail_sph_tests, p = 3, n = 10,
#'                 approx = "MC")
#' unif_stat_distr(x = cbind(x, x + 1), type = c("Rayleigh", "Bingham"),
#'                 p = 3, n = 10, approx = "MC")
#' unif_stat_distr(x = cbind(x, x + 1), type = c("Rayleigh", "Bingham"),
#'                 p = 3, n = 10, approx = "MC")
#'
#' ## Specific arguments
#'
#' # Rothman
#' unif_stat_distr(x = x, type = "Rothman", p = 2, n = 10, Rothman_t = 0.5,
#'                 approx = "MC")
#'
#' # CCF09
#' dirs <- r_unif_sph(n = 5, p = 3, M = 1)[, , 1]
#' x <- seq(0, 1, l = 10)
#' unif_stat_distr(x = x, type = "CCF09", p = 3, n = 10, approx = "MC",
#'                 CCF09_dirs = dirs)
#' unif_stat_distr(x = x, type = "CCF09", p = 3, n = 10, approx = "MC")
#'
#' # CJ12
#' unif_stat_distr(x = x, type = "CJ12", p = 3, n = 100, CJ12_reg = 3)
#' unif_stat_distr(x = x, type = "CJ12", p = 3, n = 100, CJ12_reg = 2,
#'                CJ12_beta = 0.01)
#' unif_stat_distr(x = x, type = "CJ12", p = 3, n = 100, CJ12_reg = 1)
#' }
#' @export
unif_stat_distr <- function(x, type, p, n, approx = "asymp", M = 1e4,
                            stats_MC = NULL, Rothman_t = 1 / 3, Pycke_q = 0.5,
                            Riesz_s = 1, CCF09_dirs = NULL, CJ12_reg = 3,
                            CJ12_beta = 0, Stephens = FALSE, K_Kuiper = 25,
                            K_Watson = 25, K_Watson_1976 = 5, K_Ajne = 5e2,
                            K_CCF09 = 25, K_max = 1e4, ...) {

  # Stop if NA's
  if (anyNA(x)) {

    stop("NAs present in x, please remove them.")

  }

  # As a matrix
  if (is.vector(x)) {

    x <- matrix(x, ncol = 1)

  }

  # Dimension
  if (p == 2) {

    avail_stats <- avail_cir_tests
    prefix_distr <- "p_cir_stat_"

  } else {

    avail_stats <- avail_sph_tests
    prefix_distr <- "p_sph_stat_"

  }

  # Get the type of statistics
  if (is.character(type)) {

    type <- gsub("-", "_", type)
    type <- unique(type)
    if ("all" %in% type) {

      stats_type <- avail_stats

    } else {

      stats_type <- match.arg(arg = type, choices = avail_stats,
                              several.ok = TRUE)

    }

  } else if (is.numeric(type)) {

    type <- unique(type)
    stats_type <- avail_stats[type]

  } else {

    stop("type must be a character or a numeric vector.")

  }

  # Omit statistics that do not have asymptotic distribution
  if (approx == "asymp") {

    ind_asymp <- sapply(paste0(prefix_distr, stats_type),
                        exists, where = asNamespace("sphunif"))
    if (!all(ind_asymp)) {

      warning(paste0(paste("Omitting the following statistics with not",
                           "implemented or known asymptotic distributions: "),
                     paste(stats_type[!ind_asymp], collapse = ", "), "."))
      stats_type <- stats_type[ind_asymp]

    }

  }

  # Number of statistics
  n_stats <- length(stats_type)

  # Check dimension
  n_col_x <- ncol(x)
  if (n_col_x != n_stats) {

    if (n_col_x == 1) {

      x <- matrix(x, nrow = nrow(x), ncol = n_stats, byrow = FALSE)

    } else {

      stop(paste("Check x and type, ncol(x) must equal the number of valid",
                 "type values."))
    }

  }

  # Put x as a data frame
  x <- as.data.frame(x)
  names(x) <- stats_type

  # Asymptotic or Monte Carlo distribution?
  if (approx == "asymp") {

    # Optional arguments
    args <- list("t" = Rothman_t, "q" = Pycke_q, "s" = Riesz_s,
                 "dirs" = CCF09_dirs, "regime" = CJ12_reg, "beta" = CJ12_beta,
                 "Stephens" = Stephens, "K_Kuiper" = K_Kuiper,
                 "K_Watson" = K_Watson, "K_Watson_1976" = K_Watson_1976,
                 "K_Ajne" = K_Ajne, "K_CCF09" = K_CCF09, "K_max" = K_max,
                 "thre" = 0, "n" = n, "p" = p)
    names_args <- names(args)

    # Evaluate distributions
    distrs <- sapply(stats_type, function(distr) {

      # Additional arguments
      name_distr <- paste0(prefix_distr, distr)
      distr_args <- args[names_args %in% names(formals(name_distr))]

      # Call distribution
      do.call(what = name_distr, args = c(list(x = x[[distr]]), distr_args))

    }, simplify = FALSE)

  } else if (approx == "MC") {

    # Sample statistics under the null
    if (is.null(stats_MC)) {

      stats_MC <- unif_stat_MC(n = n, type = stats_type, p = p, M = M,
                               r_H1 = NULL, crit_val = NULL,
                               alpha = c(0.10, 0.05, 0.01),
                               return_stats = TRUE, stats_sorted = TRUE,
                               Rothman_t = Rothman_t, Pycke_q = Pycke_q,
                               CCF09_dirs = CCF09_dirs, CJ12_reg = CJ12_reg,
                               K_CCF09 = K_CCF09, ...)$stats_MC

    } else {

      # Check if there is any missing statistic in stats_MC
      names_stats_MC <- names(stats_MC)
      missing_stats_MC <- !(stats_type %in% names_stats_MC)
      if (any(missing_stats_MC)) {

        stop(paste0("Missing statistics in stats_MC: \"",
                    paste(stats_type[missing_stats_MC],
                          collapse = "\", \""), "\"."))

      }

      # Sort for faster comparisons
      if (any(apply(stats_MC, 2, is.unsorted))) {

        stats_MC <- as.data.frame(sort_each_col(as.matrix(stats_MC)))
        names(stats_MC) <- names_stats_MC

      }

      # Match columns of stats_MC with stats_type
      stats_MC <- stats_MC[, pmatch(x = stats_type, table = names_stats_MC),
                           drop = FALSE]

    }

    # Is it required to sort x?
    if (any(apply(x, 2, is.unsorted))) {

      # Indexes for sorting and then unsorting
      ind_1 <- sort_index_each_col(x);
      ind_2 <- sort_index_each_col(ind_1);

      # Approximate distributions
      distrs <- sapply(stats_type, function(distr) {
        ecdf_bin(data = stats_MC[[distr]], sorted_x = x[[distr]][ind_1],
                 data_sorted = TRUE, efic = TRUE, divide_n = TRUE)
      }, simplify = FALSE)
      distrs <- distrs[ind_2, ]

    } else {

      # Approximate distributions
      distrs <- sapply(stats_type, function(distr) {
        ecdf_bin(data = stats_MC[[distr]], sorted_x = x[[distr]],
                 data_sorted = TRUE, efic = TRUE, divide_n = TRUE)
      }, simplify = FALSE)

    }

  } else {

    stop("Wrong choice for approx.")

  }

  # Return data frame
  distrs <- as.data.frame(lapply(lapply(distrs, pmin, 1), pmax, 0))
  names(distrs) <- stats_type
  return(distrs)

}

---
title: "`sphunif`: Uniformity Tests on the Circle, Sphere, and Hypersphere"
author: "Eduardo García-Portugués and Thomas Verdebout"
date: "`r Sys.Date()`, v`r packageVersion('sphunif')`"
bibliography: sphunif.bib
biblio-style: apalike-custom
output:
  rmarkdown::html_vignette:
    fig_caption: yes
    toc: yes
vignette: >
  %\VignetteIndexEntry{sphunif: Uniformity Tests on the Circle, Sphere, and Hypersphere}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
options(rmarkdown.html_vignette.check_title = FALSE)
knitr::opts_chunk$set(
  collapse = TRUE,
  message = FALSE,
  comment = "#>"
)
```

## Just give me a quick example!

### Circular data

Suppose that you want to test if a sample of *circular* data is uniformly distributed. For example, the following circular uniform sample in **radians**:

```{r, cir_data}
set.seed(987202226)
cir_data <- runif(n = 30, min = 0, max = 2 * pi)
```

Then you can call the main function in the `sphunif` package, `unif_test()`, specifying the `type` of test to be performed. For example, the Watson (omnibus) test:

```{r, unif_test_cir}
library(sphunif)
unif_test(data = cir_data, type = "Watson") # An htest object
```

By default, the calibration of the test statistic is done by Monte Carlo. This can be changed with `p_value = "asymp"` to use the asymptotic distribution:

```{r, asymp}
unif_test(data = cir_data, type = "Watson", p_value = "MC") # Monte Carlo
unif_test(data = cir_data, type = "Watson", p_value = "asymp") # Asymp. distr.
```

You can perform *several* tests within a *single* call to `unif_test()`. Choose the available circular uniformity tests from

```{r, avail_cir}
avail_cir_tests
```

For example, you can use the more powerful Projected Anderson--Darling test (@Garcia-Portugues2023, also omnibus) and the Watson test:

```{r, unif_test_avail_cir}
# A *list* of htest objects
unif_test(data = cir_data, type = c("PAD", "Watson"))
```

@Garcia-Portugues2018 gives a review of different tests of uniformity on the circle. Section 5.1 in @Pewsey2021 also gives an overview of recent advances.

### Spherical data

Suppose now that you want to test if a sample of *spherical* data is uniformly distributed. Consider a *non*-uniformly-generated^[Uniformly-distributed polar coordinates do not translate into uniformly-distributed Cartesian coordinates!] sample of directions in **Cartesian coordinates**, such as:

```{r, sph_data}
# Sample data on S^2
set.seed(987202226)
theta <- runif(n = 30, min = 0, max = 2 * pi)
phi <- runif(n = 30, min = 0, max = pi)
sph_data <- cbind(cos(theta) * sin(phi), sin(theta) * sin(phi), cos(phi))
```

The available spherical uniformity tests:

```{r, avail_sph}
avail_sph_tests
```

See again @Garcia-Portugues2018 for a review of tests of uniformity on the sphere and complementary Section 5.1 in @Pewsey2021.

The default `type = "all"` equals `type = avail_sph_tests`, which might be too much for standard analysis:

```{r, unif_test_avail_sph}
unif_test(data = sph_data, type = "all", p_value = "MC", M = 1e2)
unif_test(data = sph_data, type = "Rayleigh", p_value = "asymp")
```

### Higher dimensions

The *hyperspherical* setting is treated analogously to the spherical setting, and the available tests are exactly the same (`avail_sph_tests`). Here is an example of testing uniformity with a sample of size `100` on the $9$-sphere:

```{r, hyp_data}
# Sample data on S^9
hyp_data <- r_unif_sph(n = 30, p = 10)

# Test
unif_test(data = hyp_data, type = "Rayleigh", p_value = "asymp")
```

## A data example: are Venusian craters uniformly distributed?

The following snippet partially reproduces the data application in @Garcia-Portugues2021 on testing the uniformity of known Venusian craters. Incidentally, it also illustrates how to monitor the progress of a Monte Carlo simulation when `p_value = "MC"` using [progressr](https://progressr.futureverse.org).

```{r, venus}
# Load spherical data
data(venus)
head(venus)
nrow(venus)

# Compute Cartesian coordinates from polar coordinates
venus$X <- cbind(cos(venus$theta) * cos(venus$phi),
                 sin(venus$theta) * cos(venus$phi),
                 sin(venus$phi))

# Test uniformity using the Projected Cramér-von Mises and Projected
# Anderson-Darling tests on S^2, both using asymptotic distributions
unif_test(data = venus$X, type = c("PCvM", "PAD"), p_value = "asymp")

# Define a handler for progressr
require(progress)
require(progressr)
handlers(handler_progress(
  format = paste("(:spin) [:bar] :percent Iter: :current/:total Rate:",
                 ":tick_rate iter/sec ETA: :eta Elapsed: :elapsedfull"),
  clear = FALSE))

# Test uniformity using Monte-Carlo approximated null distributions
with_progress(
  unif_test(data = venus$X, type = c("PCvM", "PAD"),
            p_value = "MC", chunks = 1e2, M = 5e2, cores = 2)
)
```

## Simulation studies done simple

`unif_stat()` allows computing several statistics to different samples within a single call, thus facilitating Monte Carlo experiments:

```{r, unif_stat_vec}
# M samples of size n on S^2
samps_sph <- r_unif_sph(n = 30, p = 3, M = 10)

# Apply all statistics to the M samples
unif_stat(data = samps_sph, type = "all")
```

Additionally, `unif_stat_MC()` is an utility for performing the previous simulation through a convenient parallel wrapper:

```{r, MC}
# Break the simulation in 10 chunks of tasks to be divided between 2 cores
sim <- unif_stat_MC(n = 30, type = "all", p = 3, M = 1e2, cores = 2,
                    chunks = 10)

# Critical values for test statistics
sim$crit_val_MC

# Test statistics
head(sim$stats_MC)

# Power computation using a pre-built sampler for the alternative
pow <- unif_stat_MC(n = 30, type = "all", p = 3, M = 1e2, cores = 2,
                    chunks = 10, r_H1 = r_alt, crit_val = sim$crit_val_MC,
                    alt = "vMF", kappa = 1)
pow$power_MC

# How to use a custom sampler according to ?unif_stat_MC
r_H1 <- function(n, p, M, l = 1) {

  samp <- array(dim = c(n, p, M))
  for (j in 1:M) {

    samp[, , j] <- mvtnorm::rmvnorm(n = n, mean = c(l, rep(0, p - 1)),
                                    sigma = diag(rep(1, p)))
    samp[, , j] <- samp[, , j] / sqrt(rowSums(samp[, , j]^2))

  }
  return(samp)

}
pow <- unif_stat_MC(n = 30, type = "all", p = 3, M = 1e2, cores = 2,
                    chunks = 5, r_H1 = r_H1, crit_val = sim$crit_val_MC)
pow$power_MC
```

`unif_stat_MC()` can be used for constructing the Monte Carlo calibration necessary for `unif_test()`, either for providing a rejection rule based on `$crit_val_MC` or for approximating the $p$-value by `$stats_MC`.

```{r, crit_val}
# Using precomputed critical values
ht1 <- unif_test(data = samps_sph[, , 1], type = "Rayleigh",
                 p_value = "crit_val", crit_val = sim$crit_val_MC)
ht1
ht1$reject

# Using precomputed Monte Carlo statistics
ht2 <- unif_test(data = samps_sph[, , 1], type = "Rayleigh",
                 p_value = "MC", stats_MC = sim$stats_MC)
ht2
ht2$reject

# Faster than
unif_test(data = samps_sph[, , 1], type = "Rayleigh", p_value = "MC")
```

## Citation

If you use `sphunif`, please cite it!

```{r}
citation("sphunif")
```

## References

## Test environments

* local R installation, R 4.1.0
* ubuntu 16.04 (on travis-ci), R 3.5.0
* win-builder (release, devel)
* Windows Server 2008 R2 SP1, R-release, 32/64 bit
* Windows Server 2008 R2 SP1, R-devel, 32/64 bit
* Windows Server 2008 R2 SP1, R-oldrel, 32/64 bit
* Windows Server 2008 R2 SP1, R-patched, 32/64 bit
* macOS 10.13.6 High Sierra, R-release, brew
* macOS 10.13.6 High Sierra, R-release, CRAN's setup
* Ubuntu Linux 20.04.1 LTS, R-release, GCC
* Ubuntu Linux 20.04.1 LTS, R-devel, GCC
* Debian Linux, R-release, GCC
* Debian Linux, R-devel, GCC

## R CMD check results

0 errors | 0 warnings | 0 notes

* This is a resubmission where I have
    - added missing @return in r_alt;
    - removed the uses of ::: in documentation examples.
    - (I have also changed the parallel backend from doParallel to doFuture.)
* "Possibly mis-spelled words in DESCRIPTION" -> Double-checked, they are correctly spelled.
* "Found the following (possibly) invalid URLs" -> Double-checked, they are correct. They come from the usage of \doi{} with correct doi's. The redirections might be causing timeouts.
* Please observe that the use of `set.seed` in `unif_stat_MC` and `int_sph_MC` serves the very important purpose of ensuring full reproducibility and facilitating debugging when running simulations in parallel. Since the default is **not** to set any seeds (`seeds = NULL`), unless the user explicitly demands otherwise by overriding `seeds`, I believe the package is compliant with CRAN's policy.

---
title: "`sphunif`: Uniformity Tests on the Circle, Sphere, and Hypersphere"
author: "Eduardo García-Portugués and Thomas Verdebout"
date: "`r Sys.Date()`, v`r packageVersion('sphunif')`"
bibliography: sphunif.bib
biblio-style: apalike-custom
output:
  rmarkdown::html_vignette:
    fig_caption: yes
    toc: yes
vignette: >
  %\VignetteIndexEntry{sphunif: Uniformity Tests on the Circle, Sphere, and Hypersphere}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
options(rmarkdown.html_vignette.check_title = FALSE)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Just give me a quick example!

### Circular data

Suppose that you want to test if a sample of *circular* data is uniformly distributed. For example, the following circular uniform sample in **radians**:

```{r, cir_data}
set.seed(987202226)
cir_data <- runif(n = 30, min = 0, max = 2 * pi)
```

Then you can call the main function in the `sphunif` package, `unif_test`, specifying the `type` of test to be performed. For example, the Watson (omnibus) test:

```{r, unif_test_cir}
library(sphunif)
unif_test(data = cir_data, type = "Watson") # An htest object
```

By default, the calibration of the test statistic is done by Monte Carlo. This can be changed with `p_value = "asymp"` to use the asymptotic distribution:

```{r, asymp}
unif_test(data = cir_data, type = "Watson", p_value = "MC") # Monte Carlo
unif_test(data = cir_data, type = "Watson", p_value = "asymp") # Asymp. distr.
```

You can perform *several* tests within a *single* call to `unif_test`. Choose the available circular uniformity tests from

```{r, avail_cir}
avail_cir_tests
```

For example, you can use the Projected Anderson--Darling (@Garcia-Portugues2020b, also an omnibus test) test and the Watson test:

```{r, unif_test_avail_cir}
# A *list* of htest objects
unif_test(data = cir_data, type = c("PAD", "Watson"))
```

@Garcia-Portugues2020a gives a review of different tests of uniformity on the circle. Section 5.1 in @Pewsey2021 also gives an overview of recent advances.

### Spherical data

Suppose now that you want to test if a sample of *spherical* data is uniformly distributed. Consider a *non*-uniformly-generated^[Uniformly-distributed polar coordinates do not translate into uniformly-distributed Cartesian coordinates!] sample of directions in **Cartesian coordinates**, such as:

```{r, sph_data}
# Sample data on S^2
set.seed(987202226)
theta <- runif(n = 30, min = 0, max = 2 * pi)
phi <- runif(n = 30, min = 0, max = pi)
sph_data <- cbind(cos(theta) * sin(phi), sin(theta) * sin(phi), cos(phi))
```

The available spherical uniformity tests:

```{r, avail_sph}
avail_sph_tests
```

See again @Garcia-Portugues2020a for a review of tests of uniformity on the sphere and complementary Section 5.1 in @Pewsey2021.

The default `type = "all"` equals `type = avail_sph_tests`, which might be too much for standard analysis:

```{r, unif_test_avail_sph}
unif_test(data = sph_data, type = "all", p_value = "MC", M = 1e3)
unif_test(data = sph_data, type = "Rayleigh", p_value = "asymp")
```

### Higher dimensions

The *hyperspherical* setting is treated analogously to the spherical setting, and the available tests are exactly the same (`avail_sph_tests`). Here is an example of testing uniformity with a sample of size `100` on the $9$-sphere:

```{r, hyp_data}
# Sample data on S^9
hyp_data <- r_unif_sph(n = 30, p = 10)

# Test
unif_test(data = hyp_data, type = "Rayleigh", p_value = "asymp")
```

## A data example: are Venusian craters uniformly distributed?

The following snippet partially reproduces the data application in @Garcia-Portugues2021 on testing the uniformity of known Venusian craters. Incidentally, it also illustrates how to monitor the progress of a Monte Carlo simulation when `p_value = "MC"` using [progressr](https://progressr.futureverse.org).

```{r, venus}
# Load spherical data
data(venus)
head(venus)
nrow(venus)

# Compute Cartesian coordinates from polar coordinates
venus$X <- cbind(cos(venus$theta) * cos(venus$phi),
                 sin(venus$theta) * cos(venus$phi),
                 sin(venus$phi))

# Test uniformity using the Projected Cramér-von Mises and Projected
# Anderson-Darling tests on S^2, both using asymptotic distributions
unif_test(data = venus$X, type = c("PCvM", "PAD"), p_value = "asymp")

# Define a handler for progressr
require(progress)
require(progressr)
handlers(handler_progress(
  format = ":spin [:bar] :percent Total: :elapsedfull End \u2248 :eta",
  clear = FALSE))

# Test uniformity using Monte-Carlo approximated null distributions
with_progress(
  unif_test(data = venus$X, type = c("PCvM", "PAD"),
            p_value = "MC", chunks = 1e2, M = 5e2, cores = 2)
)
```

## Simulation studies done simple

`unif_stat` allows to compute several statistics to different samples within a single call, thus facilitating Monte Carlo experiments:

```{r, unif_stat_vec}
# M samples of size n on S^2
samps_sph <- r_unif_sph(n = 30, p = 3, M = 10)

# Apply all statistics to the M samples
unif_stat(data = samps_sph, type = "all")
```

Additionally, `unif_stat_MC` is an utility for performing the previous simulation through a convenient parallel wrapper:

```{r, MC}
# Break the simulation in 10 chunks of tasks to be divided between 2 cores
sim <- unif_stat_MC(n = 30, type = "all", p = 3, M = 1e4, cores = 2,
                    chunks = 10)

# Critical values for test statistics
sim$crit_val_MC

# Test statistics
head(sim$stats_MC)

# Power computation using a pre-built sampler for the alternative
pow <- unif_stat_MC(n = 30, type = "all", p = 3, M = 1e4, cores = 2,
                    chunks = 10, r_H1 = r_alt, crit_val = sim$crit_val_MC,
                    alt = "vMF", kappa = 1)
pow$power_MC

# How to use a custom sampler according to ?unif_stat_MC
r_H1 <- function(n, p, M, l = 1) {

  samp <- array(dim = c(n, p, M))
  for (j in 1:M) {

    samp[, , j] <- mvtnorm::rmvnorm(n = n, mean = c(l, rep(0, p - 1)),
                                    sigma = diag(rep(1, p)))
    samp[, , j] <- samp[, , j] / sqrt(rowSums(samp[, , j]^2))

  }
  return(samp)

}
pow <- unif_stat_MC(n = 30, type = "all", p = 3, M = 1e2, cores = 2,
                    chunks = 5, r_H1 = r_H1, crit_val = sim$crit_val_MC)
pow$power_MC
```

`unif_stat_MC` can be used for constructing the Monte Carlo calibration necessary for `unif_test`, either for providing a rejection rule based on `$crit_val_MC` or for approximating the $p$-value by `$stats_MC`.

```{r, crit_val}
# Using precomputed critical values
ht1 <- unif_test(data = samps_sph[, , 1], type = "Rayleigh",
                 p_value = "crit_val", crit_val = sim$crit_val_MC)
ht1
ht1$reject

# Using precomputed Monte Carlo statistics
ht2 <- unif_test(data = samps_sph[, , 1], type = "Rayleigh",
                 p_value = "MC", stats_MC = sim$stats_MC)
ht2
ht2$reject

# Faster than
unif_test(data = samps_sph[, , 1], type = "Rayleigh", p_value = "MC")
```

## References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Pn.R
\name{Pn}
\alias{Pn}
\alias{psi_Pn}
\alias{Gegen_coefs_Pn}
\alias{akx}
\alias{f_locdev_Pn}
\title{Utilities for projected-ecdf statistics of spherical uniformity}
\usage{
psi_Pn(theta, q, type, Rothman_t = 1/3, tilde = FALSE, psi_Gauss = TRUE,
  psi_N = 320, tol = 1e-06)

Gegen_coefs_Pn(k, p, type, Rothman_t = 1/3, Gauss = TRUE, N = 320,
  tol = 1e-06, verbose = FALSE)

akx(x, p, k, sqr = FALSE)

f_locdev_Pn(p, type, K = 1000, N = 320, K_max = 10000, thre = 0.001,
  Rothman_t = 1/3, verbose = FALSE)
}
\arguments{
\item{theta}{vectors with values in \eqn{[0, \pi]}.}

\item{q}{integer giving the dimension of the sphere \eqn{S^q}.}

\item{type}{type of projected-ecdf test statistic. Must be either
\code{"PCvM"} (Cramér--von Mises), \code{"PAD"} (Anderson--Darling), or
\code{"PRt"} (Rothman).}

\item{Rothman_t}{\eqn{t} parameter for the Rothman test, a real in
\eqn{(0, 1)}. Defaults to \code{1 / 3}.}

\item{tilde}{include the constant and bias term? Defaults to \code{FALSE}.}

\item{psi_Gauss}{use a \link[=Gauss_Legen_nodes]{Gauss--Legendre quadrature}
rule with \code{psi_N} nodes in the computation of the kernel function?
Defaults to \code{TRUE}.}

\item{psi_N}{number of points used in the Gauss--Legendre quadrature for
computing the kernel function. Defaults to \code{320}.}

\item{tol}{tolerance passed to \code{\link{integrate}}'s \code{rel.tol} and
\code{abs.tol} if \code{Gauss = FALSE}. Defaults to \code{1e-6}.}

\item{k}{vectors with the orders of the Gegenbauer polynomials. Must
be integers larger or equal than \code{0}.}

\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}

\item{Gauss}{use a Gauss--Legendre quadrature rule of \code{N} nodes
in the computation of the Gegenbauer coefficients? Otherwise, call
\code{\link{integrate}}. Defaults to \code{TRUE}.}

\item{N}{number of points used in the \link[=Gauss_Legen_nodes]{
Gauss--Legendre quadrature} for computing the Gegenbauer coefficients.
Defaults to \code{320}.}

\item{verbose}{flag to print informative messages. Defaults to \code{FALSE}.}

\item{x}{evaluation points for \eqn{a_{k, p}^x}, a vector with values in
\eqn{[-1, 1]}.}

\item{sqr}{return the \emph{signed} square root of \eqn{a_{k, p}^x}?
Defaults to \code{FALSE}.}

\item{K}{number of equispaced points on \eqn{[-1, 1]} used for evaluating
\eqn{f} and then interpolating. Defaults to \code{1e3}.}

\item{K_max}{integer giving the truncation of the series. Defaults to
\code{1e4}.}

\item{thre}{proportion of norm \emph{not} explained by the first terms of the
truncated series. Defaults to \code{1e-3}.}
}
\value{
\itemize{
  \item \code{psi_Pn}: a vector of size \code{length(theta)} with the
  evaluation of \eqn{\psi}.
  \item \code{Gegen_coefs_Pn}: a vector of size \code{length(k)} containing
  the coefficients \eqn{b_{k, p}^W}.
  \item \code{akx}: a matrix of size \code{c(length(x), length(k))}
  containing the coefficients \eqn{a_{k, p}^x}.
  \item \code{f_locdev_Pn}: the projected alternative \eqn{f} as a function
  ready to be evaluated.
}
}
\description{
Computation of the kernels
\deqn{\psi_p^W(\theta) := \int_{-1}^1 A_x(\theta)\,\mathrm{d}W(F_p(x)),}{
\psi_p^W(\theta) := \int_{-1}^1 A_x(\theta) dW(F_p(x)),
}
where \eqn{A_x(\theta)} is the proportion of area surface of
\eqn{S^{p - 1}} covered by the
\link[=A_theta_x]{intersection of two hyperspherical caps} with common solid
angle \eqn{\pi - \cos^{-1}(x)} and centers separated by
an angle \eqn{\theta \in [0, \pi]}, \eqn{F_p} is the distribution function
of the \link[=p_proj_unif]{projected spherical uniform distribution},
and \eqn{W} is a measure on \eqn{[0, 1]}.

Also, computation of the \link[=Gegen_coefs]{Gegenbauer coefficients} of
\eqn{\psi_p^W}:
\deqn{b_{k, p}^W := \frac{1}{c_{k, p}}\int_0^\pi \psi_p^W(\theta)
C_k^{p / 2 - 1}(\cos\theta)\,\mathrm{d}\theta.}{
b_{k, p}^W := \frac{1}{c_{k, p}} \int_0^\pi \psi_p^W(\theta)
C_k^(p / 2 - 1)(\cos\theta) d\theta.}
These coefficients can also be computed via
\deqn{b_{k, p}^W = \int_{-1}^1 a_{k, p}^x\,\mathrm{d}W(F_p(x))}{
b_{k, p}^W = \int_{-1}^1 a_{k, p}^x dW(F_p(x))}
for a certain function \eqn{x\rightarrow a_{k, p}^x}. They serve to define
\link[=locdev]{projected alternatives to uniformity}.
}
\details{
The evaluation of \eqn{\psi_p^W} and \eqn{b_{k, p}^W} depends on the type of
projected-ecdf statistic:
\itemize{
\item PCvM: closed-form expressions for \eqn{\psi_p^W} and \eqn{b_{k, p}^W}
with \eqn{p = 2, 3, 4}, numerical integration required for \eqn{p \ge 5}.
\item PAD: closed-form expressions for \eqn{\psi_2^W} and \eqn{b_{k, 3}^W},
numerical integration required for \eqn{\psi_p^W} with \eqn{p \ge 3} and
\eqn{b_{k, p}^W} with \eqn{p = 2} and \eqn{p \ge 4}.
\item PRt: closed-form expressions for \eqn{\psi_p^W} and \eqn{b_{k, p}^W}
for any \eqn{p \ge 2}.
}
See García-Portugués et al. (2020) for more details.
}
\examples{
# Kernels in the projected-ecdf test statistics
k <- 0:10
coefs <- list()
(coefs$PCvM <- t(sapply(2:5, function(p)
  Gegen_coefs_Pn(k = k, p = p, type = "PCvM"))))
(coefs$PAD <- t(sapply(2:5, function(p)
  Gegen_coefs_Pn(k = k, p = p, type = "PAD"))))
(coefs$PRt <- t(sapply(2:5, function(p)
  Gegen_coefs_Pn(k = k, p = p, type = "PRt"))))

# Gegenbauer expansion
th <- seq(0, pi, length.out = 501)[-501]
old_par <- par(mfrow = c(3, 4))
for (type in c("PCvM", "PAD", "PRt")) {

  for (p in 2:5) {

    plot(th, psi_Pn(theta = th, q = p - 1, type = type), type = "l",
         main = paste0(type, ", p = ", p), xlab = expression(theta),
         ylab = expression(psi(theta)), axes = FALSE, ylim = c(-1.5, 1))
    axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
         labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
    axis(2); box()
    lines(th, Gegen_series(theta = th, coefs = coefs[[type]][p - 1, ],
                           k = k, p = p), col = 2)

  }

}
par(old_par)

# Analytical coefficients vs. numerical integration
test_coef <- function(type, p, k = 0:20) {

  plot(k, log1p(abs(Gegen_coefs_Pn(k = k, p = p, type = type))),
       ylab = "Coefficients", main = paste0(type, ", p = ", p))
  points(k, log1p(abs(Gegen_coefs(k = k, p = p, psi = psi_Pn, type = type,
                                  q = p - 1))), col = 2)
  legend("topright", legend = c("log(1 + Gegen_coefs_Pn))",
                                "log(1 + Gegen_coefs(psi_Pn))"),
         lwd = 2, col = 1:2)

}

# PCvM statistic
old_par <- par(mfrow = c(2, 2))
for (p in 2:5) test_coef(type = "PCvM", p = p)
par(old_par)

# PAD statistic
old_par <- par(mfrow = c(2, 2))
for (p in 2:5) test_coef(type = "PAD", p = p)
par(old_par)

# PRt statistic
old_par <- par(mfrow = c(2, 2))
for (p in 2:5) test_coef(type = "PRt", p = p)
par(old_par)

# akx
akx(x = seq(-1, 1, l = 5), k = 1:4, p = 2)
akx(x = 0, k = 1:4, p = 3)

# PRt alternative to uniformity
z <- seq(-1, 1, l = 1e3)
p <- c(2:5, 10, 15, 17)
col <- viridisLite::viridis(length(p))
plot(z, f_locdev_Pn(p = p[1], type = "PRt")(z), type = "s",
     col = col[1], ylim = c(0, 0.6), ylab = expression(f[Rt](z)))
for (k in 2:length(p)) {
  lines(z, f_locdev_Pn(p = p[k], type = "PRt")(z), type = "s", col = col[k])
}
legend("topleft", legend = paste("p =", p), col = col, lwd = 2)
}
\references{
García-Portugués, E., Navarro-Esteban, P., Cuesta-Albertos, J. A. (2020)
On a projection-based class of uniformity tests on the hypersphere.
\emph{arXiv:2008.09897}. \url{https://arxiv.org/abs/2008.09897}
}
\author{
Eduardo García-Portugués and Paula Navarro-Esteban.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Gegenbauer.R
\name{Gegenbauer}
\alias{Gegenbauer}
\alias{Gegen_polyn}
\alias{Gegen_coefs}
\alias{Gegen_series}
\alias{Gegen_norm}
\alias{Gegen_polyn_2d}
\alias{Gegen_coefs_2d}
\alias{Gegen_series_2d}
\alias{Gegen_norm_2d}
\title{Gegenbauer polynomials and coefficients}
\usage{
Gegen_polyn(theta, k, p)

Gegen_coefs(k, p, psi, Gauss = TRUE, N = 320, normalize = TRUE,
  only_const = FALSE, tol = 1e-06, ...)

Gegen_series(theta, coefs, k, p, normalize = TRUE)

Gegen_norm(coefs, k, p, normalize = TRUE, cumulative = FALSE)

Gegen_polyn_2d(theta_1, theta_2, k, m, p)

Gegen_coefs_2d(k, m, p, psi, Gauss = TRUE, N = 320, normalize = TRUE,
  only_const = FALSE, tol = 1e-06, ...)

Gegen_series_2d(theta_1, theta_2, coefs, k, m, p, normalize = TRUE)

Gegen_norm_2d(coefs, k, m, p, normalize = TRUE)
}
\arguments{
\item{theta, theta_1, theta_2}{vectors with values in \eqn{[0, \pi]}.}

\item{k, m}{vectors with the orders of the Gegenbauer polynomials. Must
be integers larger or equal than \code{0}.}

\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}

\item{psi}{function defined in \eqn{[0, \pi]} and whose Gegenbauer
coefficients are to be computed. Must be vectorized. For
\code{Gegen_coefs_2d}, it must return a matrix of size
\code{c(length(theta_1), length(theta_2))}.}

\item{Gauss}{use a Gauss--Legendre quadrature rule of \code{N} nodes
in the computation of the Gegenbauer coefficients? Otherwise, call
\code{\link{integrate}}. Defaults to \code{TRUE}.}

\item{N}{number of points used in the \link[=Gauss_Legen_nodes]{
Gauss--Legendre quadrature} for computing the Gegenbauer coefficients.
Defaults to \code{320}.}

\item{normalize}{consider normalized coefficients (divided by
\eqn{c_{k, p}})? Defaults to \code{TRUE}.}

\item{only_const}{return only the normalizing constants \eqn{c_{k, p}}?
Defaults to \code{FALSE}.}

\item{tol}{tolerance passed to \code{\link{integrate}}'s \code{rel.tol} and
\code{abs.tol} if \code{Gauss = FALSE}. Defaults to \code{1e-6}.}

\item{...}{further arguments to be passed to \code{psi}.}

\item{coefs}{for \code{Gegen_series} and \code{Gegen_norm}, a vector of
coefficients \eqn{b_{k, p}} with length \code{length(k)}. For
\code{Gegen_series_2d} and \code{Gegen_norm_2d}, a matrix of coefficients
\eqn{b_{k, m, p}} with size \code{c(length(k), length(m))}. The
order of the coefficients is given by \code{k} and \code{m}.}

\item{cumulative}{return the cumulative norm for increasing truncation of
the series? Defaults to \code{FALSE}.}
}
\value{
\itemize{
  \item \code{Gegen_polyn}: a matrix of size
  \code{c(length(theta), length(k))} containing the evaluation of the
  \code{length(k)} Gegenbauer polynomials at \code{theta}.
  \item \code{Gegen_coefs}: a vector of size \code{length(k)} containing
  the coefficients \eqn{b_{k, p}}.
  \item \code{Gegen_series}: the evaluation of the truncated series
  expansion, a vector of size \code{length(theta)}.
  \item \code{Gegen_norm}: the Gegenbauer norm of the truncated series,
  a scalar if \code{cumulative = FALSE}, otherwise a vector of size
  \code{length(k)}.
  \item \code{Gegen_polyn_2d}: a 4-dimensional array of size
  \code{c(length(theta_1), length(theta_2), length(k), length(m))}
  containing the evaluation of the \code{length(k) * length(m)}
  2-dimensional Gegenbauer polynomials at the bivariate grid
  spanned by \code{theta_1} and \code{theta_2}.
  \item \code{Gegen_coefs_2d}: a matrix of size
  \code{c(length(k), length(m))} containing the coefficients
  \eqn{b_{k, m, p}}.
  \item \code{Gegen_series_2d}: the evaluation of the truncated series
  expansion, a matrix of size \code{c(length(theta_1), length(theta_2))}.
  \item \code{Gegen_norm_2d}: the 2-dimensional Gegenbauer norm of the
  truncated series, a scalar.
}
}
\description{
The \href{https://dlmf.nist.gov/18.3}{Gegenbauer polynomials}
\eqn{\{C_k^{(\lambda)}(x)\}_{k = 0}^\infty}{
{C_k^(\lambda)(x)}_{k = 0}^\infty}
form a family of orthogonal polynomials on the interval \eqn{[-1, 1]}
with respect to the weight function \eqn{(1 - x^2)^{\lambda - 1/2}},
for \eqn{\lambda > -1/2}, \eqn{\lambda \neq 0}. They usually appear
when dealing with functions defined on
\eqn{S^{p-1} := \{{\bf x} \in R^p : ||{\bf x}|| = 1\}}{
S^{p-1} := \{x \in R^p : ||x|| = 1\}} with index \eqn{\lambda = p / 2 - 1}.

The Gegenbauer polynomials are somehow simpler to evaluate for
\eqn{x = \cos(\theta)}, with \eqn{\theta \in [0, \pi]}. This simplifies
also the connection with the Chebyshev polynomials
\eqn{\{T_k(x)\}_{k = 0}^\infty}{{T_k(x)}_{k = 0}^\infty}, which admit
the \href{https://dlmf.nist.gov/18.5.E1}{explicit expression}
\eqn{T_k(\cos(\theta)) = \cos(k\theta)}. The Chebyshev polynomials
appear as the limit of the Gegenbauer polynomials
(divided by \eqn{\lambda}) when \eqn{\lambda} goes to \eqn{0}, so they
can be regarded as the extension by continuity of
\eqn{\{C_k^{(p/2 - 1)}(x)\}_{k = 0}^\infty}{
{C_k^(p/2 - 1)(x)}_{k = 0}^\infty} to the case \eqn{p = 2}.

For a \href{https://dlmf.nist.gov/18.18.i}{reasonably smooth} function
\eqn{\psi} defined on \eqn{[0, \pi]},
\deqn{\psi(\theta) = \sum_{k = 0}^\infty b_{k, p}
C_k^{(p/2 - 1)}(\cos(\theta)),}{\psi(\theta) = \sum_{k = 0}^\infty b_{k, p}
C_k^(p/2 - 1)(\cos(\theta)),}
provided that the coefficients
\deqn{b_{k, p} := \frac{1}{c_{k, p}} \int_0^\pi \psi(\theta)
C_k^{(p/2 - 1)}(\cos(\theta)) (\sin(\theta))^{p - 2}\,\mathrm{d}\theta}{
b_{k, p} := \frac{1}{c_{k, p}} \int_0^\pi \psi(\theta)
C_k^(p/2 - 1)(\cos(\theta)) (\sin(\theta))^{p - 2} d\theta}
are finite, where the normalizing constants are
\deqn{c_{k, p} := \int_0^\pi (C_k^{(p/2 - 1)}(\cos(\theta)))^2
(\sin(\theta))^{p - 2} \,\mathrm{d}\theta.}{
c_{k, p} := \int_0^\pi (C_k^(p/2 - 1)(\cos(\theta)))^2
(\sin(\theta))^{p - 2} d\theta.}
The (squared) "Gegenbauer norm" of \eqn{\psi} is
\deqn{\|\psi\|_{G, p}^2 := \int_0^\pi \psi(\theta)^2
C_k^{(p/2 - 1)}(\cos(\theta)) (\sin(\theta))^{p - 2}\,\mathrm{d}\theta.}{
||\psi||_{G, p}^2 := \int_0^\pi \psi(\theta)^2
C_k^(p/2 - 1)(\cos(\theta)) (\sin(\theta))^{p - 2} d\theta.}

The previous expansion can be generalized for a 2-dimensional function
\eqn{\psi} defined on \eqn{[0, \pi] \times [0, \pi]}:
\deqn{\psi(\theta_1, \theta_2) = \sum_{k = 0}^\infty \sum_{m = 0}^\infty
b_{k, m, p} C_k^{(p/2 - 1)}(\cos(\theta_1))
C_k^{(p/2 - 1)}(\cos(\theta_2)),}{
\psi(\theta_1, \theta_2) = \sum_{k = 0}^\infty \sum_{m = 0}^\infty
b_{k, m, p} C_k^(p/2 - 1)(\cos(\theta_1)) C_k^(p/2 - 1)(\cos(\theta_2))}
with coefficients
\deqn{b_{k, m, p} := \frac{1}{c_{k, p} c_{m, p}} \int_0^\pi\int_0^\pi
\psi(\theta_1, \theta_2) C_k^{(p/2 - 1)}(\cos(\theta_1))
C_k^{(p/2 - 1)}(\cos(\theta_2)) (\sin(\theta_1))^{p - 2}
(\sin(\theta_2))^{p - 2}\,\mathrm{d}\theta_1\,\mathrm{d}\theta_2.}{
b_{k, m, p} := \frac{1}{c_{k, p} c_{m, p}} \int_0^\pi \int_0^\pi
\psi(\theta_1, \theta_2) C_k^(p/2 - 1)(\cos(\theta_1))
C_k^(p/2 - 1)(\cos(\theta_2)) (\sin(\theta_1))^{p - 2}
(\sin(\theta_2))^{p - 2} d\theta_1 d\theta_2.}
The (squared) "Gegenbauer norm" of \eqn{\psi} is
\deqn{\|\psi\|_{G, p}^2 := \int_0^\pi\int_0^\pi \psi(\theta_1, \theta_2)^2
C_k^{(p/2 - 1)}(\cos(\theta_1)) C_k^{(p/2 - 1)}(\cos(\theta_2))
(\sin(\theta_1))^{p - 2} (\sin(\theta_2))^{p - 2}
\,\mathrm{d}\theta_1\,\mathrm{d}\theta_2.}{
||\psi||_{G, p}^2 := \int_0^\pi\int_0^\pi \psi(\theta_1, \theta_2)^2
C_k^(p/2 - 1)(\cos(\theta_1)) C_k^(p/2 - 1)(\cos(\theta_2))
(\sin(\theta_1))^{p - 2} (\sin(\theta_2))^{p - 2} d\theta_1 d\theta_2.}
}
\details{
The \code{Gegen_polyn} function is a wrapper to the functions
\code{\link[gsl:Gegenbauer]{gegenpoly_n}} and
\code{\link[gsl:Gegenbauer]{gegenpoly_array}} in the
\code{\link[gsl]{gsl-package}}, which they interface the functions
defined in the header file \code{gsl_sf_gegenbauer.h} (documented
\href{https://www.gnu.org/software/gsl/doc/html/specfunc.html#gegenbauer-functions}{
here}) of the \href{https://www.gnu.org/software/gsl/}{
GNU Scientific Library}.

Note that the function \code{Gegen_polyn} computes the regular
\emph{unnormalized} Gegenbauer polynomials.

For the case \eqn{p = 2}, the Chebyshev polynomials are considered.
}
\examples{
## Representation of Gegenbauer polynomials (Chebyshev polynomials for p = 2)

th <- seq(0, pi, l = 500)
k <- 0:3
old_par <- par(mfrow = c(2, 2))
for (p in 2:5) {
  matplot(th, t(Gegen_polyn(theta = th, k = k, p = p)), lty = 1,
          type = "l", main = substitute(p == d, list(d = p)),
          axes = FALSE, xlab = expression(theta), ylab = "")
  axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
       labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
  axis(2); box()
  mtext(text = expression({C[k]^{p/2 - 1}}(cos(theta))), side = 2,
        line = 2, cex = 0.75)
  legend("bottomleft", legend = paste("k =", k), lwd = 2, col = seq_along(k))
}
par(old_par)

## Coefficients and series in p = 2

# Function in [0, pi] to be projected in Chebyshev polynomials
psi <- function(th) -sin(th / 2)

# Coefficients
p <- 2
k <- 0:4
(coefs <- Gegen_coefs(k = k, p = p, psi = psi))

# Series
plot(th, psi(th), type = "l", axes = FALSE, xlab = expression(theta),
      ylab = "", ylim = c(-1.25, 0))
axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
     labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
axis(2); box()
col <- viridisLite::viridis(length(coefs))
for (i in seq_along(coefs)) {
  lines(th, Gegen_series(theta = th, coefs = coefs[1:(i + 1)], k = 0:i,
                         p = p), col = col[i])
}
lines(th, psi(th), lwd = 2)

## Coefficients and series in p = 3

# Function in [0, pi] to be projected in Gegenbauer polynomials
psi <- function(th) tan(th / 3)

# Coefficients
p <- 3
k <- 0:10
(coefs <- Gegen_coefs(k = k, p = p, psi = psi))

# Series
plot(th, psi(th), type = "l", axes = FALSE, xlab = expression(theta),
      ylab = "", ylim = c(0, 2))
axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
     labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
axis(2); box()
col <- viridisLite::viridis(length(coefs))
for (i in seq_along(coefs)) {
  lines(th, Gegen_series(theta = th, coefs = coefs[1:(i + 1)], k = 0:i,
                         p = p), col = col[i])
}
lines(th, psi(th), lwd = 2)
\donttest{
## Surface representation

# Surface in [0, pi]^2 to be projected in Gegenbauer polynomials
p <- 3
psi <- function(th_1, th_2) A_theta_x(theta = th_1, x = cos(th_2),
                                      p = p, as_matrix = TRUE)

# Coefficients
k <- 0:20
m <- 0:10
coefs <- Gegen_coefs_2d(k = k, m = m, p = p, psi = psi)

# Series
th <- seq(0, pi, l = 100)
col <- viridisLite::viridis(20)
old_par <- par(mfrow = c(2, 2))
image(th, th, A_theta_x(theta = th, x = cos(th), p = p), axes = FALSE,
      col = col, zlim = c(0, 1), xlab = expression(theta[1]),
      ylab = expression(theta[2]), main = "Original")
axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
     labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
axis(2, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
     labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
box()
for(K in c(5, 10, 20)) {
  A <- Gegen_series_2d(theta_1 = th, theta_2 = th,
                       coefs = coefs[1:(K + 1), ], k = 0:K, m = m, p = p)
  image(th, th, A, axes = FALSE, col = col, zlim = c(0, 1),
        xlab = expression(theta[1]), ylab = expression(theta[2]),
        main = paste(K, "x", m[length(m)], "coefficients"))
  axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
       labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
  axis(2, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
       labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
  box()
}
par(old_par)
}
}
\references{
Galassi, M., Davies, J., Theiler, J., Gough, B., Jungman, G., Alken, P.,
Booth, M., and Rossi, F. (2009) \emph{GNU Scientific Library Reference
Manual}. Network Theory Ltd. \url{http://www.gnu.org/software/gsl/}

\emph{NIST Digital Library of Mathematical Functions}. Release
1.0.20 of 2018-09-15. F. W. J. Olver, A. B. Olde Daalhuis, D. W. Lozier,
B. I. Schneider, R. F. Boisvert, C. W. Clark, B. R. Miller,
and B. V. Saunders, eds. \url{https://dlmf.nist.gov/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unif-alts.R
\name{F_from_f}
\alias{F_from_f}
\alias{F_inv_from_f}
\title{Distribution and quantile functions from angular function}
\usage{
F_from_f(f, p, Gauss = TRUE, N = 320, K = 1000, tol = 1e-06, ...)

F_inv_from_f(f, p, Gauss = TRUE, N = 320, K = 1000, tol = 1e-06, ...)
}
\arguments{
\item{f}{angular function defined on \eqn{[-1, 1]}. Must be vectorized.}

\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}

\item{Gauss}{use a \link[=Gauss_Legen_nodes]{Gauss--Legendre quadrature}
rule to integrate \eqn{f} with \code{N} nodes? Otherwise, rely on
\code{\link{integrate}} Defaults to \code{TRUE}.}

\item{N}{number of points used in the Gauss--Legendre quadrature. Defaults
to \code{320}.}

\item{K}{number of equispaced points on \eqn{[-1, 1]} used for evaluating
\eqn{F^{-1}} and then interpolating. Defaults to \code{1e3}.}

\item{tol}{tolerance passed to \code{\link{uniroot}} for the inversion of
\eqn{F}. Also, passed to \code{\link{integrate}}'s \code{rel.tol} and
\code{abs.tol} if \code{Gauss = FALSE}. Defaults to \code{1e-6}.}

\item{...}{further parameters passed to \code{f}.}
}
\value{
A \code{\link{splinefun}} object ready to evaluate \eqn{F} or
\eqn{F^{-1}}, as specified.
}
\description{
Numerical computation of the distribution function \eqn{F} and
the quantile function \eqn{F^{-1}} for an \link[=locdev]{angular function}
\eqn{f} in a \link[=tang-norm-decomp]{tangent-normal decomposition}.
\eqn{F^{-1}(x)} results from the inversion of
\deqn{F(x) = \int_{-1}^x \omega_{p - 1}c_f f(z) (1 - z^2)^{(p - 3) / 2}
\,\mathrm{d}z}{F(x) = \int_{-1}^x \omega_{p - 1}c_f f(z)
(1 - z^2)^{(p - 3) / 2} dz}
for \eqn{x\in [-1, 1]}, where \eqn{c_f} is a normalizing constant and
\eqn{\omega_{p - 1}} is the surface area of \eqn{S^{p - 2}}.
}
\details{
The normalizing constant \eqn{c_f} is such that \eqn{F(1) = 1}. It does not
need to be part of \code{f} as it is computed internally.

Interpolation is performed by a monotone cubic spline. \code{Gauss = TRUE}
yields more accurate results, at expenses of a heavier computation.

If \code{f} yields negative values, these are silently truncated to zero.
}
\examples{
f <- function(x) rep(1, length(x))
plot(F_from_f(f = f, p = 4, Gauss = TRUE), ylab = "F(x)", xlim = c(-1, 1))
plot(F_from_f(f = f, p = 4, Gauss = FALSE), col = 2, add = TRUE,
     xlim = c(-1, 1))
curve(p_proj_unif(x = x, p = 4), col = 3, add = TRUE, n = 300)
plot(F_inv_from_f(f = f, p = 4, Gauss = TRUE), ylab = "F^{-1}(x)")
plot(F_inv_from_f(f = f, p = 4, Gauss = FALSE), col = 2, add = TRUE)
curve(q_proj_unif(u = x, p = 4), col = 3, add = TRUE, n = 300)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{comets}
\alias{comets}
\title{Comet orbits}
\format{
A data frame with 3633 rows and 8 variables:
\describe{
  \item{id}{database ID.}
  \item{full_name}{full name/designation following the
  \href{https://www.iau.org/public/themes/naming/#comets}{
  IUA naming convention}.}
  \item{i}{inclination; the orbit angle with respect to the ecliptic plane,
  in radians in \eqn{[0, \pi]}.}
  \item{om}{longitude of the ascending node; the angle between the normal
  vector of the orbit and the normal vector of the ecliptic plane, in
  radians in \eqn{[0, 2\pi)}.}
  \item{per_y}{sidereal orbital period (in years).}
  \item{class}{orbit classification. A factor with levels given below.}
  \item{diameter}{diameter from equivalent sphere (in km).}
  \item{ccf2009}{flag indicating if the comet was considered in the data
  application in Cuesta-Albertos et al. (2009); see details below.}
}
}
\source{
\url{https://ssd.jpl.nasa.gov/sbdb_query.cgi}
}
\usage{
comets
}
\description{
Comet orbits data from the
\href{https://ssd.jpl.nasa.gov/sbdb_query.cgi}{
JPL Small-Body Database Search Engine}. The normal vector of a comet orbit
represents is a vector on \eqn{S^2}.
}
\details{
The normal vector to the ecliptic plane of the comet with inclination
\eqn{i} and longitude of the ascending node \eqn{\omega} is
\deqn{(\sin(i) \sin(\omega), -\sin(i) \cos(\omega), \cos(i))'.}{
(sin(i) sin(\omega), -sin(i) cos(\omega), cos(i))'.}

A prograde comet has positive \eqn{\cos(i)}{cos(i)}, negative
\eqn{\cos(i)}{cos(i)} represents a retrograde comet.

\code{class} has the following levels:
\itemize{
 \item \code{COM}: comet orbit not matching any defined orbit class.
 \item \code{CTc}: Chiron-type comet, as defined by Levison and Duncan
 (T_Jupiter > 3; a > a_Jupiter).
 \item \code{ETc}: Encke-type comet, as defined by Levison and Duncan
 (T_Jupiter > 3; a < a_Jupiter).
 \item \code{HTC}: Halley-type comet, classical definition (20y < P < 200y).
 \item \code{HYP}: comets on hyperbolic orbits.
 \item \code{JFc}: Jupiter-family comet, as defined by Levison and Duncan
 (2 < T_Jupiter < 3).
 \item \code{JFC}: Jupiter-family comet, classical definition (P < 20y).
 \item \code{PAR}: comets on parabolic orbits.
}
Hyperbolic and parabolic comets are not periodic; only elliptical comets
are periodic.

The \code{ccf2009} variable gives the observations considered in
Cuesta-Albertos et al. (2009) after fetching in the database in 2007-12-14
for the comets such that
\code{!(class \%in\% c("HYP", "PAR")) & per_y >= 200 & !numbered}. A
periodic comet is \code{numbered} by the IUA only after its second
perihelion passage, and then its \code{id} starts with \code{c}. Due to the
dynamic nature of the data, more comets were added to the database since
2007 and also some observations were updated.

The script performing the data preprocessing is available at
\href{https://github.com/egarpor/sphunif/blob/master/data-raw/comets.R}{
\code{comets.R}}. The data was retrieved on 2020-05-07.
}
\examples{
# Load data
data("comets")

# Add normal vectors
comets$normal <- cbind(sin(comets$i) * sin(comets$om),
                       -sin(comets$i) * cos(comets$om),
                       cos(comets$i))

# Add numbered information
comets$numbered <- substr(comets$id, 1, 1) == "c"

# Tests to be performed
type_tests <- c("PCvM", "PAD", "PRt")

# Excluding the C/1882 R1-X (Great September comet) records with X = B, C, D
comets_ccf2009 <- comets[comets$ccf2009, ][-c(13:15), ]

# Sample size
nrow(comets_ccf2009)

# Tests for the data in Cuesta-Albertos et al. (2009)
tests_ccf2009 <- unif_test(data = comets_ccf2009$normal, type = type_tests,
                           p_value = "asymp")
tests_ccf2009
}
\references{
Cuesta-Albertos, J. A., Cuevas, A., Fraiman, R. (2009) On projection-based
tests for directional and compositional data. \emph{Statistics and
Computing}, 19:367--380. \doi{10.1007/s11222-008-9098-3}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{cir_coord_conv}
\alias{cir_coord_conv}
\alias{Theta_to_X}
\alias{X_to_Theta}
\title{Transforming between polar and Cartesian coordinates}
\usage{
Theta_to_X(Theta)

X_to_Theta(X)
}
\arguments{
\item{Theta}{a \bold{matrix} of size \code{c(n, M)} with \code{M} samples
of size \code{n} of circular data on \eqn{[0, 2\pi)}. Must not contain
\code{NA}'s.}

\item{X}{an \bold{array} of size \code{c(n, 2, M)} containing the Cartesian
coordinates of \code{M} samples of size \code{n} of directions on
\eqn{S^{1}}. Must not contain \code{NA}'s.}
}
\value{
\itemize{
  \item \code{Theta_to_X}: the corresponding \code{X}.
  \item \code{X_to_Theta}: the corresponding \code{Theta}.
}
}
\description{
Transformation between a matrix \code{Theta} containing
\code{M} circular samples of size \code{n} on \eqn{[0, 2\pi)} and an array
\code{X} containing the associated Cartesian coordinates on
\eqn{S^1:=\{{\bf x}\in R^2:||{\bf x}||=1\}}{S^1:=\{x\in R^2:||x||=1\}}.
}
\examples{
# Sample
Theta <- r_unif_cir(n = 10, M = 2)
X <- r_unif_sph(n = 10, p = 2, M = 2)

# Check equality
sum(abs(X - Theta_to_X(X_to_Theta(X))))
sum(abs(Theta - X_to_Theta(Theta_to_X(Theta))))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/sph_distr.R
\name{p_sph_stat_Bingham}
\alias{p_sph_stat_Bingham}
\alias{d_sph_stat_Bingham}
\alias{p_sph_stat_CJ12}
\alias{d_sph_stat_CJ12}
\alias{p_sph_stat_Rayleigh}
\alias{d_sph_stat_Rayleigh}
\alias{p_sph_stat_Rayleigh_HD}
\alias{d_sph_stat_Rayleigh_HD}
\alias{sph_stat_distr}
\alias{p_sph_stat_Ajne}
\alias{d_sph_stat_Ajne}
\alias{p_sph_stat_Bakshaev}
\alias{d_sph_stat_Bakshaev}
\alias{p_sph_stat_Gine_Fn}
\alias{d_sph_stat_Gine_Fn}
\alias{p_sph_stat_Gine_Gn}
\alias{d_sph_stat_Gine_Gn}
\alias{p_sph_stat_PAD}
\alias{d_sph_stat_PAD}
\alias{p_sph_stat_PCvM}
\alias{d_sph_stat_PCvM}
\alias{p_sph_stat_PRt}
\alias{d_sph_stat_PRt}
\alias{p_sph_stat_Riesz}
\alias{d_sph_stat_Riesz}
\title{Asymptotic distributions for spherical uniformity statistics}
\usage{
p_sph_stat_Bingham(x, p)

d_sph_stat_Bingham(x, p)

p_sph_stat_CJ12(x, regime = 1L, beta = 0)

d_sph_stat_CJ12(x, regime = 3L, beta = 0)

p_sph_stat_Rayleigh(x, p)

d_sph_stat_Rayleigh(x, p)

p_sph_stat_Rayleigh_HD(x, p)

d_sph_stat_Rayleigh_HD(x, p)

p_sph_stat_Ajne(x, p, K_max = 1000, thre = 0, ...)

d_sph_stat_Ajne(x, p, K_max = 1000, thre = 0, ...)

p_sph_stat_Bakshaev(x, p, K_max = 1000, thre = 0, ...)

d_sph_stat_Bakshaev(x, p, K_max = 1000, thre = 0, ...)

p_sph_stat_Gine_Fn(x, p, K_max = 1000, thre = 0, ...)

d_sph_stat_Gine_Fn(x, p, K_max = 1000, thre = 0, ...)

p_sph_stat_Gine_Gn(x, p, K_max = 1000, thre = 0, ...)

d_sph_stat_Gine_Gn(x, p, K_max = 1000, thre = 0, ...)

p_sph_stat_PAD(x, p, K_max = 1000, thre = 0, ...)

d_sph_stat_PAD(x, p, K_max = 1000, thre = 0, ...)

p_sph_stat_PCvM(x, p, K_max = 1000, thre = 0, ...)

d_sph_stat_PCvM(x, p, K_max = 1000, thre = 0, ...)

p_sph_stat_PRt(x, p, t = 1/3, K_max = 1000, thre = 0, ...)

d_sph_stat_PRt(x, p, t = 1/3, K_max = 1000, thre = 0, ...)

p_sph_stat_Riesz(x, p, s = 1, K_max = 1000, thre = 0, ...)

d_sph_stat_Riesz(x, p, s = 1, K_max = 1000, thre = 0, ...)
}
\arguments{
\item{x}{a vector of size \code{nx} or a matrix of size \code{c(nx, 1)}.}

\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}

\item{regime}{type of asymptotic regime for the CJ12 test, either \code{1}
(sub-exponential regime), \code{2} (exponential), or \code{3}
(super-exponential; default).}

\item{beta}{\eqn{\beta} parameter in the exponential regime of the CJ12
test, a nonnegative real. Defaults to \code{0}.}

\item{K_max}{integer giving the truncation of the series that compute the
asymptotic p-value of a Sobolev test. Defaults to \code{1e3}.}

\item{thre}{error threshold for the tail probability given by the
the first terms of the truncated series of a Sobolev test. Defaults to
\code{0} (no further truncation).}

\item{...}{further parameters passed to \code{\link{p_Sobolev}} or
\code{\link{d_Sobolev}} (such as \code{x_tail}).}

\item{t}{\eqn{t} parameter for the Rothman and Cressie tests, a real in
\eqn{(0, 1)}. Defaults to \code{1 / 3}.}

\item{s}{\eqn{s} parameter for the \eqn{s}-Riesz test, a real in
\eqn{(0, 2)}. Defaults to \code{1}.}
}
\value{
\itemize{
  \item \code{r_sph_stat_*}: a matrix of size \code{c(n, 1)} containing
  the sample.
  \item \code{p_sph_stat_*}, \code{d_sph_stat_*}: a matrix of size
  \code{c(nx, 1)} with the evaluation of the distribution or density
  functions at \code{x}.
}
}
\description{
Computation of the asymptotic null distributions of
spherical uniformity statistics.
}
\details{
Descriptions and references on most of the asymptotic distributions
are available in García-Portugués and Verdebout (2018).
}
\examples{
# Ajne
curve(d_sph_stat_Ajne(x, p = 3, method = "HBE"), n = 2e2, ylim = c(0, 4))
curve(p_sph_stat_Ajne(x, p = 3, method = "HBE"), n = 2e2, col = 2,
      add = TRUE)

# Bakshaev
curve(d_sph_stat_Bakshaev(x, p = 3, method = "HBE"), to = 5, n = 2e2,
      ylim = c(0, 2))
curve(p_sph_stat_Bakshaev(x, p = 3, method = "HBE"), n = 2e2, col = 2,
      add = TRUE)

# Bingham
curve(d_sph_stat_Bingham(x, p = 3), to = 20, n = 2e2, ylim = c(0, 1))
curve(p_sph_stat_Bingham(x, p = 3), n = 2e2, col = 2, add = TRUE)

# CJ12
curve(d_sph_stat_CJ12(x, regime = 1), from = -10, to = 10, n = 2e2,
      ylim = c(0, 1))
curve(d_sph_stat_CJ12(x, regime = 2, beta = 0.1), n = 2e2, col = 2,
      add = TRUE)
curve(d_sph_stat_CJ12(x, regime = 3), n = 2e2, col = 3, add = TRUE)
curve(p_sph_stat_CJ12(x, regime = 1), n = 2e2, col = 1, add = TRUE)
curve(p_sph_stat_CJ12(x, regime = 2, beta = 0.1), n = 2e2, col = 2,
      add = TRUE)
curve(p_sph_stat_CJ12(x, regime = 3), col = 3, add = TRUE)

# Gine Fn
curve(d_sph_stat_Gine_Fn(x, p = 3, method = "HBE"), to = 2, n = 2e2,
      ylim = c(0, 2))
curve(p_sph_stat_Gine_Fn(x, p = 3, method = "HBE"), n = 2e2, col = 2,
      add = TRUE)

# Gine Gn
curve(d_sph_stat_Gine_Gn(x, p = 3, method = "HBE"), to = 1.5, n = 2e2,
      ylim = c(0, 2.5))
curve(p_sph_stat_Gine_Gn(x, p = 3, method = "HBE"), n = 2e2, col = 2,
      add = TRUE)

# PAD
curve(d_sph_stat_PAD(x, p = 3, method = "HBE"), to = 3, n = 2e2,
      ylim = c(0, 1.5))
curve(p_sph_stat_PAD(x, p = 3, method = "HBE"), n = 2e2, col = 2,
      add = TRUE)

# PCvM
curve(d_sph_stat_PCvM(x, p = 3, method = "HBE"), to = 0.6, n = 2e2,
      ylim = c(0, 7))
curve(p_sph_stat_PCvM(x, p = 3, method = "HBE"), n = 2e2, col = 2,
      add = TRUE)

# PRt
curve(d_sph_stat_PRt(x, p = 3, method = "HBE"), n = 2e2, ylim = c(0, 5))
curve(p_sph_stat_PRt(x, p = 3, method = "HBE"), n = 2e2, col = 2, add = TRUE)

# Rayleigh
curve(d_sph_stat_Rayleigh(x, p = 3), to = 15, n = 2e2, ylim = c(0, 1))
curve(p_sph_stat_Rayleigh(x, p = 3), n = 2e2, col = 2, add = TRUE)

# HD-standardized Rayleigh
curve(d_sph_stat_Rayleigh_HD(x, p = 3), from = -4, to = 4, n = 2e2,
      ylim = c(0, 1))
curve(p_sph_stat_Rayleigh_HD(x, p = 3), n = 2e2, col = 2, add = TRUE)

# Riesz
curve(d_sph_stat_Riesz(x, p = 3, method = "HBE"), n = 2e2, from = 0, to = 5,
      ylim = c(0, 2))
curve(p_sph_stat_Riesz(x, p = 3, method = "HBE"), n = 2e2, col = 2,
      add = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{rhea}
\alias{rhea}
\title{Rhea craters from Hirata (2016)}
\format{
A data frame with 3596 rows and 4 variables:
\describe{
  \item{name}{name of the crater (if named).}
  \item{diameter}{diameter of the crater (in km).}
  \item{theta}{longitude angle \eqn{\theta \in [0, 2\pi)} of the
  crater center.}
  \item{phi}{latitude angle \eqn{\phi \in [-\pi/2, \pi/2]} of the
  crater center.}
}
}
\source{
\url{https://agupubs.onlinelibrary.wiley.com/action/downloadSupplement?doi=10.1002\%2F2015JE004940&file=jgre20485-sup-0002-TableS1.txt}
}
\usage{
rhea
}
\description{
Craters on Rhea from
\href{https://agupubs.onlinelibrary.wiley.com/doi/full/10.1002/2015JE004940}{
Hirata (2016)}.
}
\details{
The \eqn{(\theta, \phi)} angles are such their associated planetocentric
coordinates are:
\deqn{(\cos(\phi) \cos(\theta), \cos(\phi) \sin(\theta), \sin(\phi))',}{
(cos(\phi) cos(\theta), cos(\phi) sin(\theta), sin(\phi))',}
with \eqn{(0, 0, 1)'} denoting the north pole.

The script performing the data preprocessing is available at
\href{https://github.com/egarpor/sphunif/blob/master/data-raw/rhea.R}{
\code{rhea.R}}.
}
\examples{
# Load data
data("rhea")

# Add Cartesian coordinates
rhea$X <- cbind(cos(rhea$theta) * cos(rhea$phi),
                sin(rhea$theta) * cos(rhea$phi),
                sin(rhea$phi))

# Tests
unif_test(data = rhea$X[rhea$diam > 15 & rhea$diam < 20, ],
          type = c("PCvM", "PAD", "PRt"), p_value = "asymp")
}
\references{
Hirata, N. (2016) Differential impact cratering of Saturn's satellites by
heliocentric impactors. \emph{Journal of Geophysical Research: Planets},
121:111--117. \doi{10.1002/2015JE004940}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unif_test.R
\name{unif_test}
\alias{unif_test}
\title{Circular and (hyper)spherical uniformity tests}
\usage{
unif_test(data, type = "all", p_value = "asymp", alpha = c(0.1, 0.05,
  0.01), M = 10000, stats_MC = NULL, crit_val = NULL,
  data_sorted = FALSE, Rayleigh_m = 1, cov_a = 2 * pi, Rothman_t = 1/3,
  Cressie_t = 1/3, Pycke_q = 0.5, Riesz_s = 1, CCF09_dirs = NULL,
  K_CCF09 = 25, CJ12_reg = 3, CJ12_beta = 0, K_max = 10000, ...)
}
\arguments{
\item{data}{sample to perform the test. A matrix of size \code{c(n, p)}
containing a sample of size \code{n} of directions (in Cartesian
coordinates) on \eqn{S^{p-1}}. Alternatively if \code{p = 2}, a matrix of
size \code{c(n, 1)} containing the \code{n} angles on \eqn{[0, 2\pi)} of the
circular sample on \eqn{S^{1}}. Other objects accepted are an array of size
\code{c(n, p, 1)} with directions (in Cartesian coordinates), or a vector of
size \code{n} or an array of size \code{c(n, 1, 1)} with angular data.
Must not contain \code{NA}'s.}

\item{type}{type of test to be applied. A character vector containing any of
the following types of tests, depending on the dimension \eqn{p}:
\itemize{
  \item Circular data: any of the names available at object
  \code{\link{avail_cir_tests}}.
  \item (Hyper)spherical data: any of the names available at object
  \code{\link{avail_sph_tests}}.
}
If \code{type = "all"} (default), then \code{type} is set as
\code{avail_cir_tests} or \code{avail_sph_tests}, depending on the value of
\eqn{p}.}

\item{p_value}{type of \eqn{p}-value computation. Either \code{"MC"} for
employing the approximation by Monte Carlo of the exact null distribution,
\code{"asymp"} (default) for the use of the asymptotic/exact null
distribution (if available), or \code{"crit_val"} for approximation by means
of the table of critical values \code{crit_val}.}

\item{alpha}{vector with significance levels. Defaults to
\code{c(0.10, 0.05, 0.01)}.}

\item{M}{number of Monte Carlo replications for approximating the null
distribution when \code{approx = "MC"}. Also, number of Monte Carlo samples
for approximating the asymptotic distributions based on weighted sums of chi
squared random variables. Defaults to \code{1e4}.}

\item{stats_MC}{a data frame of size \code{c(M, length(type))}, with column
names containing the character vector \code{type}, that results from
extracting \code{$stats_MC} from a call to \code{\link{unif_stat_MC}}. If
provided, the computation of Monte Carlo statistics when \code{approx = "MC"}
is skipped. \code{stats_MC} is checked internally to see if it is sorted.
Internally computed if \code{NULL} (default).}

\item{crit_val}{table with critical values for the tests, to be used if
\code{p_value = "crit_val"}. A data frame, with column names containing the
character vector \code{type} and rows corresponding to the significance
levels \code{alpha}, that results from extracting \code{$crit_val_MC} from
a call to \code{\link{unif_stat_MC}}. Internally computed if
\code{NULL} (default).}

\item{data_sorted}{is the circular data sorted? If \code{TRUE}, certain
statistics are faster to compute. Defaults to \code{FALSE}.}

\item{Rayleigh_m}{integer \eqn{m} for the \eqn{m}-modal Rayleigh test.
Defaults to \code{m = 1} (the standard Rayleigh test).}

\item{cov_a}{\eqn{a_n = a / n} parameter used in the length of the arcs
of the coverage-based tests. Must be positive. Defaults to \code{2 * pi}.}

\item{Rothman_t}{\eqn{t} parameter for the Rothman test, a real in
\eqn{(0, 1)}. Defaults to \code{1 / 3}.}

\item{Cressie_t}{\eqn{t} parameter for the Cressie test, a real in
\eqn{(0, 1)}. Defaults to \code{1 / 3}.}

\item{Pycke_q}{\eqn{q} parameter for the Pycke "\eqn{q}-test", a real in
\eqn{(0, 1)}. Defaults to \code{1 / 2}.}

\item{Riesz_s}{\eqn{s} parameter for the \eqn{s}-Riesz test, a real in
\eqn{(0, 2)}. Defaults to \code{1}.}

\item{CCF09_dirs}{a matrix of size \code{c(n_proj, p)} containing
\code{n_proj} random directions (in Cartesian coordinates) on \eqn{S^{p-1}}
to perform the CCF09 test. If \code{NULL} (default), a sample of size
\code{n_proj = 50} directions is computed internally.}

\item{K_CCF09}{integer giving the truncation of the series present in the
asymptotic distribution of the Kolmogorov-Smirnov statistic. Defaults to
\code{5e2}.}

\item{CJ12_reg}{type of asymptotic regime for CJ12 test, either \code{1}
(sub-exponential regime), \code{2} (exponential), or \code{3}
(super-exponential; default).}

\item{CJ12_beta}{\eqn{\beta} parameter in the exponential regime of CJ12 test,
a positive real.}

\item{K_max}{integer giving the truncation of the series that compute the
asymptotic p-value of a Sobolev test. Defaults to \code{1e4}.}

\item{...}{If \code{p_value = "MC"} or \code{p_value = "crit_val"}, optional
performance parameters to be passed to \code{\link{unif_stat_MC}}:
\code{chunks}, \code{cores}, and \code{seed}.}
}
\value{
If only a \bold{single test} is performed, a list with class
\code{htest} containing the following components:
\itemize{
  \item \code{statistic}: the value of the test statistic.
  \item \code{p.value}: the p-value of the test. If
  \code{p_value = "crit_val"}, an \code{NA}.
  \item \code{alternative}: a character string describing the alternative
  hypothesis.
  \item \code{method}: a character string indicating what type of test was
  performed.
  \item \code{data.name}: a character string giving the name of the data.
  \item \code{reject}: the rejection decision for the levels of significance
  \code{alpha}.
  \item \code{crit_val}: a vector with the critical values for the
  significance levels \code{alpha} used with \code{p_value = "MC"} or
  \code{p_value = "asymp"}.
}
If \bold{several tests} are performed, a \code{type}-named list with
entries for each test given by the above list.
}
\description{
Implementation of several uniformity tests on the (hyper)sphere
\eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{
S^{p-1}:=\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}, with calibration either in
terms of their asymptotic/exact distributions, if available, or Monte Carlo.

\code{unif_test} receives a sample of directions
\eqn{{\bf X}_1,\ldots,{\bf X}_n\in S^{p-1}}{X_1,\ldots,X_n\in S^{p-1}} in
\emph{Cartesian coordinates}, except for the circular case (\eqn{p=2}) in
which the sample can be represented in terms of \emph{angles}
\eqn{\Theta_1,\ldots,\Theta_n\in [0, 2\pi)}.

\code{unif_test} allows to perform several tests within a single call,
facilitating thus the exploration of a dataset by applying several tests.
}
\details{
All the tests reject for large values of the test statistic, so the critical
values for the significance levels \code{alpha} correspond to the
\code{alpha}-upper quantiles of the null distribution of the test statistic.

When \code{p_value = "asymp"}, tests that do not have an implemented or
known asymptotic are omitted, and a warning is generated.

When \code{p_value = "MC"}, it is possible to have a progress bar indicating
the Monte Carlo simulation progress if \code{unif_test} is wrapped with
\code{\link[progressr:with_progress]{progressr::with_progress}} or if
\code{progressr::handlers(global = TRUE)} is invoked (once) by the user.
See the examples below. The progress bar is updated with the number of
finished chunks.

All the statistics are continuous random variables except the
Hodges--Ajne statistic (\code{"Hodges_Ajne"}), the Cressie statistic
(\code{"Cressie"}), and the number of (different) uncovered spacings
(\code{"Num_uncover"}). These three statistics are discrete random variables.

The Monte Carlo calibration for the CCF09 test is made conditionally
on the choice of \cr\code{CCF09_dirs}. That is, all the Monte
Carlo statistics share the same random directions.

Descriptions and references for most of the tests are available
in García-Portugués and Verdebout (2018).
}
\examples{
## Asymptotic distribution

# Circular data
n <- 10
samp_cir <- r_unif_cir(n = n)

# Matrix
unif_test(data = samp_cir, type = "Ajne", p_value = "asymp")

# Vector
unif_test(data = samp_cir[, 1], type = "Ajne", p_value = "asymp")

# Array
unif_test(data = array(samp_cir, dim = c(n, 1, 1)), type = "Ajne",
          p_value = "asymp")
\donttest{
# Several tests
unif_test(data = samp_cir, type = avail_cir_tests, p_value = "asymp")
}
# Spherical data
n <- 10
samp_sph <- r_unif_sph(n = n, p = 3)

# Array
unif_test(data = samp_sph, type = "Bingham", p_value = "asymp")

# Matrix
unif_test(data = samp_sph[, , 1], type = "Bingham", p_value = "asymp")
\donttest{
# Several tests
unif_test(data = samp_sph, type = avail_sph_tests, p_value = "asymp")

## Monte Carlo

# Circular data
unif_test(data = samp_cir, type = "Ajne", p_value = "MC")
unif_test(data = samp_cir, type = avail_cir_tests, p_value = "MC")

# Spherical data
unif_test(data = samp_sph, type = "Bingham", p_value = "MC")
unif_test(data = samp_sph, type = avail_sph_tests, p_value = "MC")

# Caching stats_MC
stats_MC_cir <- unif_stat_MC(n = nrow(samp_cir), p = 2)$stats_MC
stats_MC_sph <- unif_stat_MC(n = nrow(samp_sph), p = 3)$stats_MC
unif_test(data = samp_cir, type = avail_cir_tests,
          p_value = "MC", stats_MC = stats_MC_cir)
unif_test(data = samp_sph, type = avail_sph_tests, p_value = "MC",
          stats_MC = stats_MC_sph)

## Critical values

# Circular data
unif_test(data = samp_cir, type = avail_cir_tests, p_value = "crit_val")

# Spherical data
unif_test(data = samp_sph, type = avail_sph_tests, p_value = "crit_val")

# Caching crit_val
crit_val_cir <- unif_stat_MC(n = n, p = 2)$crit_val_MC
crit_val_sph <- unif_stat_MC(n = n, p = 3)$crit_val_MC
unif_test(data = samp_cir, type = avail_cir_tests,
          p_value = "crit_val", crit_val = crit_val_cir)
unif_test(data = samp_sph, type = avail_sph_tests, p_value = "crit_val",
          crit_val = crit_val_sph)

## Specific arguments

# Rothman
unif_test(data = samp_cir, type = "Rothman", Rothman_t = 0.5)

# CCF09
unif_test(data = samp_sph, type = "CCF09", p_value = "MC",
          CCF09_dirs = samp_sph[1:2, , 1])
unif_test(data = samp_sph, type = "CCF09", p_value = "MC",
          CCF09_dirs = samp_sph[3:4, , 1])

## Using a progress bar when p_value = "MC"

# Define a progress bar
require(progress)
require(progressr)
handlers(handler_progress(
  format = ":spin [:bar] :percent Total: :elapsedfull End \u2248 :eta",
  clear = FALSE))

# Call unif_test() within with_progress()
with_progress(
  unif_test(data = samp_sph, type = avail_sph_tests, p_value = "MC",
            chunks = 10, M = 1e3)
)

# With several cores
with_progress(
  unif_test(data = samp_sph, type = avail_sph_tests, p_value = "MC",
            cores = 2, chunks = 10, M = 1e3)
)

# Instead of using with_progress() each time, it is more practical to run
# handlers(global = TRUE)
# once to activate progress bars in your R session
}
}
\references{
García-Portugués, E. and Verdebout, T. (2018) An overview of uniformity
tests on the hypersphere. \emph{arXiv:1804.00286}.
\url{https://arxiv.org/abs/1804.00286}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/sph_stat.R
\name{sph_stat_Rayleigh}
\alias{sph_stat_Rayleigh}
\alias{sph_stat_Bingham}
\alias{sph_stat_Ajne}
\alias{sph_stat_Gine_Gn}
\alias{sph_stat_Gine_Fn}
\alias{sph_stat_Pycke}
\alias{sph_stat_Bakshaev}
\alias{sph_stat_Riesz}
\alias{sph_stat_PCvM}
\alias{sph_stat_PRt}
\alias{sph_stat_PAD}
\alias{sph_stat_CCF09}
\alias{sph_stat_Rayleigh_HD}
\alias{sph_stat_CJ12}
\alias{sph_stat}
\title{Statistics for testing (hyper)spherical uniformity}
\usage{
sph_stat_Rayleigh(X)

sph_stat_Bingham(X)

sph_stat_Ajne(X, Psi_in_X = FALSE)

sph_stat_Gine_Gn(X, Psi_in_X = FALSE, p = 0L)

sph_stat_Gine_Fn(X, Psi_in_X = FALSE, p = 0L)

sph_stat_Pycke(X, Psi_in_X = FALSE, p = 0L)

sph_stat_Bakshaev(X, Psi_in_X = FALSE, p = 0L)

sph_stat_Riesz(X, Psi_in_X = FALSE, p = 0L, s = 1)

sph_stat_PCvM(X, Psi_in_X = FALSE, p = 0L, N = 160L, L = 1000L)

sph_stat_PRt(X, t = 1/3, Psi_in_X = FALSE, p = 0L, N = 160L, L = 1000L)

sph_stat_PAD(X, Psi_in_X = FALSE, p = 0L, N = 160L, L = 1000L)

sph_stat_CCF09(X, dirs, K_CCF09 = 25L, original = FALSE)

sph_stat_Rayleigh_HD(X)

sph_stat_CJ12(X, regime = 3L, Psi_in_X = FALSE, p = 0L)
}
\arguments{
\item{X}{an \bold{array} of size \code{c(n, p, M)} containing the Cartesian
coordinates of \code{M} samples of size \code{n} of directions on
\eqn{S^{p-1}}. Must not contain \code{NA}'s.}

\item{Psi_in_X}{does \code{X} contain the shortest angles matrix
\eqn{\boldsymbol\Psi}{\Psi} that is obtained with \code{\link{Psi_mat}(X)}?
If \code{FALSE} (default), \eqn{\boldsymbol\Psi}{\Psi} is computed
internally.}

\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}

\item{s}{\eqn{s} parameter for the \eqn{s}-Riesz test, a real in
\eqn{(0, 2)}. Defaults to \code{1}.}

\item{N}{number of points used in the
\link[=Gauss_Legen_nodes]{Gauss-Legendre quadrature}. Defaults to
\code{160}.}

\item{L}{number of discretization points to interpolate angular functions
that require evaluating an integral. Defaults to \code{1e3}.}

\item{t}{\eqn{t} parameter for the Rothman and Cressie tests, a real in
\eqn{(0, 1)}. Defaults to \code{1 / 3}.}

\item{dirs}{a matrix of size \code{c(n_proj, p)} containing \code{n_proj}
random directions (in Cartesian coordinates) on \eqn{S^{p-1}} to perform
the CCF09 test.}

\item{K_CCF09}{integer giving the truncation of the series present in the
asymptotic distribution of the Kolmogorov-Smirnov statistic. Defaults to
\code{5e2}.}

\item{original}{return the CCF09 statistic as originally defined?
If \code{FALSE} (default), a faster and equivalent statistic is computed,
and rejection happens for \emph{large} values of the statistic, which is
consistent with the rest of tests. Otherwise, rejection happens for
\emph{low} values.}

\item{regime}{type of asymptotic regime for the CJ12 test, either \code{1}
(sub-exponential regime), \code{2} (exponential), or \code{3}
(super-exponential; default).}
}
\value{
A matrix of size \code{c(M, 1)} containing the statistics for each
of the \code{M} samples.
}
\description{
Low-level implementation of several statistics for assessing
uniformity on the (hyper)sphere
\eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{S^{p-1}:=
\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}.
}
\details{
Detailed descriptions and references of the statistics are available
in García-Portugués and Verdebout (2018).

The Pycke and CJ12 statistics employ the \emph{scalar products} matrix,
rather than the shortest angles matrix, when \code{Psi_in_X = TRUE}. This
matrix is obtained by setting \code{scalar_prod = TRUE} in
\code{\link{Psi_mat}}.
}
\section{Warning}{

Be careful on avoiding the next bad usages of the functions, which will
produce spurious results:
\itemize{
  \item The directions in \code{X} do \emph{not} have unit norm.
  \item \code{X} does \emph{not} contain \code{Psi_mat(X)} when
  \code{X_in_Theta = TRUE}.
  \item The parameter \code{p} does \emph{not} match with the dimension of
  \eqn{R^p}.
  \item \emph{Not} passing the scalar products matrix to \code{sph_stat_CJ12}
  when \code{Psi_in_X = TRUE}.
  \item The directions in \code{dirs} do \emph{not} have unit norm.
}
}

\examples{
## Sample uniform spherical data

M <- 2
n <- 100
p <- 3
set.seed(123456789)
X <- r_unif_sph(n = n, p = p, M = M)

## Sobolev tests

# Rayleigh
sph_stat_Rayleigh(X)

# Bingham
sph_stat_Bingham(X)

# Ajne
Psi <- Psi_mat(X)
dim(Psi) <- c(dim(Psi), 1)
sph_stat_Ajne(X)
sph_stat_Ajne(Psi, Psi_in_X = TRUE)

# Gine Gn
sph_stat_Gine_Gn(X)
sph_stat_Gine_Gn(Psi, Psi_in_X = TRUE, p = p)

# Gine Fn
sph_stat_Gine_Fn(X)
sph_stat_Gine_Fn(Psi, Psi_in_X = TRUE, p = p)

# Pycke
sph_stat_Pycke(X)
sph_stat_Pycke(Psi, Psi_in_X = TRUE, p = p)

# Bakshaev
sph_stat_Bakshaev(X)
sph_stat_Bakshaev(Psi, Psi_in_X = TRUE, p = p)

# Riesz
sph_stat_Riesz(X, s = 1)
sph_stat_Riesz(Psi, Psi_in_X = TRUE, p = p, s = 1)

# Projected Cramér-von Mises
sph_stat_PCvM(X)
sph_stat_PCvM(Psi, Psi_in_X = TRUE, p = p)

# Projected Rothman
sph_stat_PRt(X)
sph_stat_PRt(Psi, Psi_in_X = TRUE, p = p)

# Projected Anderson-Darling
sph_stat_PAD(X)
sph_stat_PAD(Psi, Psi_in_X = TRUE, p = p)

## Other tests

# CCF09
dirs <- r_unif_sph(n = 3, p = p, M = 1)[, , 1]
sph_stat_CCF09(X, dirs = dirs)

## High-dimensional tests

# Rayleigh HD-Standardized
sph_stat_Rayleigh_HD(X)

# CJ12
sph_stat_CJ12(X, regime = 1)
sph_stat_CJ12(Psi, regime = 1, Psi_in_X = TRUE, p = p)
sph_stat_CJ12(X, regime = 2)
sph_stat_CJ12(Psi, regime = 2, Psi_in_X = TRUE, p = p)
sph_stat_CJ12(X, regime = 3)
sph_stat_CJ12(Psi, regime = 3, Psi_in_X = TRUE, p = p)
}
\references{
García-Portugués, E. and Verdebout, T. (2018) An overview of uniformity
tests on the hypersphere. \emph{arXiv:1804.00286}.
\url{https://arxiv.org/abs/1804.00286}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{planets}
\alias{planets}
\title{Planet orbits}
\format{
A data frame with 9 rows and 3 variables:
\describe{
  \item{planet}{names of the planets and Pluto.}
  \item{i}{inclination; the orbit angle with respect to the ecliptic plane,
  in radians in \eqn{[0, \pi]}.}
  \item{om}{longitude of the ascending node; the angle between the normal
  vector of the orbit and the normal vector of the ecliptic plane, in
  radians in \eqn{[0, 2\pi)}.}
}
}
\source{
Table 2b in \url{https://ssd.jpl.nasa.gov/txt/aprx_pos_planets.pdf}
}
\usage{
planets
}
\description{
Planet orbits data from the
\href{https://ssd.jpl.nasa.gov/?planet_pos}{
JPL Keplerian Elements for Approximate Positions of the Major Planets}.
The normal vector of a planet orbit represents is a vector on \eqn{S^2}.
}
\details{
The normal vector to the ecliptic plane of the planet with inclination
\eqn{i} and longitude of the ascending node \eqn{\omega} is
\deqn{(\sin(i) \sin(\omega), -\sin(i) \cos(\omega), \cos(i))'.}{
(sin(i) sin(\omega), -sin(i) cos(\omega), cos(i))'.}

The script performing the data preprocessing is available at
\href{https://github.com/egarpor/sphunif/blob/master/data-raw/planets.R}{
\code{planets.R}}. The data was retrieved on 2020-05-16.
}
\examples{
# Load data
data("planets")

# Add normal vectors
planets$normal <- cbind(sin(planets$i) * sin(planets$om),
                       -sin(planets$i) * cos(planets$om),
                       cos(planets$i))

# Tests to be performed
type_tests <- c("PCvM", "PAD", "PRt")

# Tests with Pluto
unif_test(data = planets$normal, type = type_tests, p_value = "MC")

# Tests without Pluto
unif_test(data = planets$normal[-9, ], type = type_tests, p_value = "MC")
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{A_theta_x}
\alias{A_theta_x}
\title{Surface area of the intersection of two hyperspherical caps}
\usage{
A_theta_x(theta, x, p, N = 160L, as_matrix = TRUE)
}
\arguments{
\item{theta}{vectors with values in \eqn{[0, \pi]}.}

\item{x}{vector with values in \eqn{[-1, 1]}.}

\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}

\item{N}{number of points used in the
\link[=Gauss_Legen_nodes]{Gauss-Legendre quadrature}. Defaults to
\code{160}.}

\item{as_matrix}{return a matrix with the values of \eqn{A_x(\theta)} on
the grid formed by \code{theta} and \code{x}? If \code{FALSE},
\eqn{A_x(\theta)} is evaluated on \code{theta} and \code{x} if they equal
in size. Defaults to \code{TRUE}.}
}
\value{
A matrix of size \code{c(length(theta), length(x))} containing the
evaluation of \eqn{A_x(\theta)} if \code{as_matrix = TRUE}. Otherwise,
a vector of size \code{c(length(theta)} if \code{theta} and \code{x} equal
in size.
}
\description{
Computation of
\deqn{A_x(\theta_{ij}) := \frac{1}{\omega_p}
\int_{S^{p - 1}} 1_{\{{\bf X}_i'\boldsymbol\gamma \le x,
{\bf X}_j'\boldsymbol\gamma \le x\}}\,\mathrm{d}\boldsymbol\gamma,}{
A_x(\theta_{ij}) := \frac{1}{\omega_p} \int_{S^{p - 1}}
1_{X_i'\gamma \le x, X_j'\gamma \le x} d\gamma,}
where \eqn{\theta_{ij} := \cos^{-1}({\bf X}_i'{\bf X}_j)
\in [0, \pi]}{\theta_{ij} := \cos^{-1}(X_i'X_j) \in [0, \pi]},
\eqn{x \in [-1, 1]}, and \eqn{\omega_{p}} is the surface area of
\eqn{S^{p - 1}}. \eqn{A_x(\theta_{ij})} is the proportion of surface area
of \eqn{S^{p - 1}} covered by the intersection of two hyperspherical caps
centered at \eqn{{\bf X}_i}{X_i} and \eqn{{\bf X}_j}{X_j} and with
common solid angle \eqn{\pi - \cos^{-1}(x)}.
}
\details{
See García-Portugués et al. (2020) for more details about the
\eqn{A_x(\theta)} function.
}
\examples{
# Plot A_x(theta) for several dimensions and x's
A_lines <- function(x, th = seq(0, pi, l = 200)) {

  plot(th, A_theta_x(theta = th, x = x, p = 2), type = "l",
       col = 1, ylim = c(0, 1.25), main = paste("x =", x),
       ylab = expression(A[x](theta)),
       xlab = expression(theta), axes = FALSE)
  axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
       labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
  axis(2); box()
  abline(h = c(0, 1), lty = 2)
  lines(th, A_theta_x(theta = th, x = x, p = 3), col = 2)
  lines(th, A_theta_x(theta = th, x = x, p = 4), col = 3)
  lines(th, A_theta_x(theta = th, x = x, p = 5), col = 4)
  legend("top", lwd = 2, legend = paste("p =", 2:5),
         col = 1:4, cex = 0.75, horiz = TRUE)

}
old_par <- par(mfrow = c(2, 3))
A_lines(x = -0.75)
A_lines(x = -0.25)
A_lines(x = 0)
A_lines(x = 0.25)
A_lines(x = 0.5)
A_lines(x = 0.75)
par(old_par)

# As surface of (theta, x) for several dimensions
A_surf <- function(p, x = seq(-1, 1, l = 201), th = seq(0, pi, l = 201)) {

  col <- c("white", viridisLite::viridis(20))
  breaks <- c(-1, seq(1e-15, 1, l = 21))
  A <- A_theta_x(theta = th, x = x, p = p)
  image(th, x, A, main = paste("p =", p), col = col, breaks = breaks,
        xlab = expression(theta), axes = FALSE)
  axis(1, at = c(0, pi / 4, pi / 2, 3 * pi / 4, pi),
       labels = expression(0, pi / 4, pi / 2, 3 * pi / 4, pi))
  axis(2); box()
  contour(th, x, A, levels = breaks, add = TRUE)

}
old_par <- par(mfrow = c(2, 2))
A_surf(p = 2)
A_surf(p = 3)
A_surf(p = 4)
A_surf(p = 5)
par(old_par)

# No matrix return
th <- seq(0, pi, l = 5)
x <- seq(-1, 1, l = 5)
diag(A_theta_x(theta = th, x = x, p = 2))
A_theta_x(theta = th, x = x, p = 2, as_matrix = FALSE)
}
\references{
García-Portugués, E., Navarro-Esteban, P., Cuesta-Albertos, J. A. (2020)
On a projection-based class of uniformity tests on the hypersphere.
\emph{arXiv:2008.09897}. \url{https://arxiv.org/abs/2008.09897}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unif_test.R
\docType{data}
\name{avail_tests}
\alias{avail_tests}
\alias{avail_cir_tests}
\alias{avail_sph_tests}
\title{Available circular and (hyper)spherical uniformity tests}
\format{
An object of class \code{character} of length 30.

An object of class \code{character} of length 14.
}
\usage{
avail_cir_tests

avail_sph_tests
}
\value{
A character vector whose elements are valid inputs for the
\code{type} argument in \code{\link{unif_test}}, \code{\link{unif_stat}},
\code{\link{unif_stat_distr}}, and \code{\link{unif_stat_MC}}.
\code{avail_cir_tests} provides the available circular tests and
\code{avail_sph_tests} the (hyper)spherical tests.
}
\description{
Listing of the tests implemented in the \code{\link{sphunif}}
package.
}
\examples{
# Circular tests
avail_cir_tests

# Spherical tests
avail_sph_tests
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/distr.R
\name{proj_unif}
\alias{proj_unif}
\alias{d_proj_unif}
\alias{p_proj_unif}
\alias{q_proj_unif}
\alias{r_proj_unif}
\title{Projection of the spherical uniform distribution}
\usage{
d_proj_unif(x, p, log = FALSE)

p_proj_unif(x, p, log = FALSE)

q_proj_unif(u, p)

r_proj_unif(n, p)
}
\arguments{
\item{x}{a vector of size \code{nx} or a matrix of size \code{c(nx, 1)}.}

\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}

\item{log}{compute the logarithm of the density or distribution?}

\item{u}{vector of probabilities.}

\item{n}{sample size employed for computing the statistic.}
}
\value{
A matrix of size \code{c(nx, 1)} with the evaluation of the
density, distribution, or quantile function at \code{x} or \code{u}.
For \code{r_proj_unif}, a random vector of size \code{n}.
}
\description{
Density, distribution, and quantile functions of the
projection of the spherical uniform random variable on an arbitrary
direction, that is, the random variable
\eqn{\boldsymbol{\gamma}'{\bf X}}{\gamma'X}, where \eqn{{\bf X}}{X}
is uniformly distributed on the (hyper)sphere
\eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{S^{p-1}:=
\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}, and
\eqn{\boldsymbol{\gamma}\in S^{p-1}}{\gamma\in S^{p-1}} is an
\emph{arbitrary} projection direction. Note that the distribution is
invariant to the choice of \eqn{\boldsymbol{\gamma}}{\gamma}. Also,
efficient simulation of \eqn{\boldsymbol{\gamma}'{\bf X}}{\gamma'X}.
}
\examples{
# Density function
curve(d_proj_unif(x, p = 2), from = -2, to = 2, n = 2e2, ylim = c(0, 2))
curve(d_proj_unif(x, p = 3), n = 2e2, col = 2, add = TRUE)
curve(d_proj_unif(x, p = 4), n = 2e2, col = 3, add = TRUE)
curve(d_proj_unif(x, p = 5), n = 2e2, col = 4, add = TRUE)
curve(d_proj_unif(x, p = 6), n = 2e2, col = 5, add = TRUE)

# Distribution function
curve(p_proj_unif(x, p = 2), from = -2, to = 2, n = 2e2, ylim = c(0, 1))
curve(p_proj_unif(x, p = 3), n = 2e2, col = 2, add = TRUE)
curve(p_proj_unif(x, p = 4), n = 2e2, col = 3, add = TRUE)
curve(p_proj_unif(x, p = 5), n = 2e2, col = 4, add = TRUE)
curve(p_proj_unif(x, p = 6), n = 2e2, col = 5, add = TRUE)

# Quantile function
curve(q_proj_unif(u = x, p = 2), from = 0, to = 1, n = 2e2, ylim = c(-1, 1))
curve(q_proj_unif(u = x, p = 3), n = 2e2, col = 2, add = TRUE)
curve(q_proj_unif(u = x, p = 4), n = 2e2, col = 3, add = TRUE)
curve(q_proj_unif(u = x, p = 5), n = 2e2, col = 4, add = TRUE)
curve(q_proj_unif(u = x, p = 6), n = 2e2, col = 5, add = TRUE)

# Sampling
hist(r_proj_unif(n = 1e4, p = 4), freq = FALSE, breaks = 50)
curve(d_proj_unif(x, p = 4), n = 2e2, col = 3, add = TRUE)
}
\author{
Eduardo García-Portugués and Paula Navarro-Esteban.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{ecdf_bin}
\alias{ecdf_bin}
\title{Efficient evaluation of the empirical cumulative distribution
function}
\usage{
ecdf_bin(data, sorted_x, data_sorted = FALSE, efic = TRUE, divide_n = TRUE)
}
\arguments{
\item{data}{a vector or column matrix containing the sample.}

\item{sorted_x}{a vector or column matrix with the evaluation points
\bold{sorted increasingly}.}

\item{data_sorted}{is \code{data} is already sorted increasingly?
This avoids sorting the data internally.}

\item{efic}{use the more efficient version of the ecdf evaluation? Set to
\code{FALSE} only for debugging purposes.}

\item{divide_n}{if \code{FALSE}, returns the absolute frequencies instead
of the relative frequencies. Defaults to \code{TRUE}.}
}
\value{
The ecdf evaluated at \code{sorted_x}.
}
\description{
Evaluates the empirical cumulative distribution function
(ecdf) of a sample \code{data} at the evaluation points \code{sorted_x}.
This is done through binary search.
}
\section{Warning}{

Be careful on avoiding the next bad usages of the function, which will
produce spurious results:
\itemize{
  \item \code{sorted_x} is not sorted increasingly.
  \item \code{data} is not sorted increasingly when
  \code{data_sorted = TRUE}-
}
}

\author{
Original code from Douglas Bates'
\url{https://github.com/dmbates/ecdfExample}. Minor adaptations by Eduardo
García-Portugués.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unif_stat_MC.R
\name{unif_stat_MC}
\alias{unif_stat_MC}
\title{Monte Carlo simulation of circular and (hyper)spherical uniformity
statistics}
\usage{
unif_stat_MC(n, type = "all", p, M = 10000, r_H1 = NULL,
  crit_val = NULL, alpha = c(0.1, 0.05, 0.01), return_stats = TRUE,
  stats_sorted = FALSE, chunks = ceiling((n * M)/1e+05), cores = 1,
  seeds = NULL, Rayleigh_m = 1, cov_a = 2 * pi, Rothman_t = 1/3,
  Cressie_t = 1/3, Pycke_q = 0.5, Riesz_s = 1, CCF09_dirs = NULL,
  K_CCF09 = 25, CJ12_reg = 3, ...)
}
\arguments{
\item{n}{sample size.}

\item{type}{type of test to be applied. A character vector containing any of
the following types of tests, depending on the dimension \eqn{p}:
\itemize{
  \item Circular data: any of the names available at object
  \code{\link{avail_cir_tests}}.
  \item (Hyper)spherical data: any of the names available at object
  \code{\link{avail_sph_tests}}.
}
If \code{type = "all"} (default), then \code{type} is set as
\code{avail_cir_tests} or \code{avail_sph_tests}, depending on the value of
\eqn{p}.}

\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}

\item{M}{number of Monte Carlo replications. Defaults to \code{1e4}.}

\item{r_H1}{if provided, the computation of empirical powers is
carried out for the alternative hypothesis sampled with \code{r_H1}.
This must be a function with the same arguments and value as
\code{\link{r_unif_sph}} (see examples). Defaults to \code{NULL}, indicating
that the critical values are estimated from samples of \code{r_unif_sph}.}

\item{crit_val}{if provided, must be the critical values as returned by
\code{$stats_MC} in a call to \code{unif_stat_MC}. They are used for
computing the empirical powers of the tests present in \code{type}.
Defaults to \code{NULL}, which means that no power computation is done.}

\item{alpha}{vector with significance levels. Defaults to
\code{c(0.10, 0.05, 0.01)}.}

\item{return_stats}{return the Monte Carlo statistics? If only the critical
values or powers are desired, \code{FALSE} saves memory in the returned
object. Defaults to \code{TRUE}.}

\item{stats_sorted}{sort the returned Monte Carlo statistics? If
\code{TRUE}, this is useful for evaluating faster the empirical cumulative
distribution function when approximating the distribution in
\code{\link{unif_stat_distr}}. Defaults to \code{FALSE}.}

\item{chunks}{number of chunks to split the \code{M} Monte Carlo
replications. Useful for parallelizing the simulation study in \code{chunks}
tasks containing \code{ceiling(M / chunks)} replications. Useful also for
avoiding memory bottlenecks when \code{M} is large. Defaults to
\cr\code{ceiling((n * M) / 1e5)}.}

\item{cores}{number of cores to perform the simulation. Defaults to \code{1}.}

\item{seeds}{if provided, a vector of size \code{chunks} for fixing the
seeds on each of the simulation chunks (useful for reproducing parallel
simulations). Specifically, for \code{k in 1:chunks}, seeds are
set as \code{set.seed(seeds[k], kind = "Mersenne-Twister")} in each chunk.
Defaults to \code{NULL} (no seed setting is done).}

\item{Rayleigh_m}{integer \eqn{m} for the \eqn{m}-modal Rayleigh test.
Defaults to \code{m = 1} (the standard Rayleigh test).}

\item{cov_a}{\eqn{a_n = a / n} parameter used in the length of the arcs
of the coverage-based tests. Must be positive. Defaults to \code{2 * pi}.}

\item{Rothman_t}{\eqn{t} parameter for the Rothman test, a real in
\eqn{(0, 1)}. Defaults to \code{1 / 3}.}

\item{Cressie_t}{\eqn{t} parameter for the Cressie test, a real in
\eqn{(0, 1)}. Defaults to \code{1 / 3}.}

\item{Pycke_q}{\eqn{q} parameter for the Pycke "\eqn{q}-test", a real in
\eqn{(0, 1)}. Defaults to \code{1 / 2}.}

\item{Riesz_s}{\eqn{s} parameter for the \eqn{s}-Riesz test, a real in
\eqn{(0, 2)}. Defaults to \code{1}.}

\item{CCF09_dirs}{a matrix of size \code{c(n_proj, p)} containing
\code{n_proj} random directions (in Cartesian coordinates) on \eqn{S^{p-1}}
to perform the CCF09 test. If \code{NULL} (default), a sample of size
\code{n_proj = 50} directions is computed internally.}

\item{K_CCF09}{integer giving the truncation of the series present in the
asymptotic distribution of the Kolmogorov-Smirnov statistic. Defaults to
\code{5e2}.}

\item{CJ12_reg}{type of asymptotic regime for CJ12 test, either \code{1}
(sub-exponential regime), \code{2} (exponential), or \code{3}
(super-exponential; default).}

\item{...}{optional arguments to be passed to the \code{r_H1} sampler or to
\code{\link[foreach]{foreach}} (for example, \code{.export} to export global
variables or other functions to the \code{foreach} environment).}
}
\value{
A list with the following entries:
\itemize{
  \item \code{crit_val_MC}: a data frame of size
  \code{c(length(alpha), length(type))}, with column names given by
  \code{type} and rows corresponding to the significance levels \code{alpha},
  that contains the estimated critical values of the tests.
  \item \code{power_MC}: a data frame of size
  \code{c(nrow(crit_val), length(type))}, with column names given by
  \code{type} and rows corresponding to the significance levels of
  \code{crit_val}, that contains the empirical powers of the tests. \code{NA}
  if \code{crit_val = NULL}.
  \item \code{stats_MC}: a data frame of size \code{c(M, length(type))}, with
  column names given by \code{type}, that contains the Monte Carlo
  statistics.
}
}
\description{
Utility for performing Monte Carlo simulation of several
statistics for assessing uniformity on the (hyper)sphere
\eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{
S^{p-1}:=\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}.

\code{unif_stat_MC} provides a convenient wrapper for parallel
evaluation of \code{unif_stat}, the estimation of critical values under the
null distribution, and the computation of empirical powers under the
alternative.
}
\details{
It is possible to have a progress bar if \code{unif_stat_MC} is wrapped with
\code{\link[progressr:with_progress]{progressr::with_progress}} or if
\code{progressr::handlers(global = TRUE)} is invoked (once) by the user.
See the examples below. The progress bar is updated with the number of
finished chunks.

All the tests reject for large values of the test statistic
(\code{max_gap = TRUE} is assumed for the Range test), so the critical
values for the significance levels \code{alpha} correspond to the
\code{alpha}-upper quantiles of the null distribution of the test statistic.

The Monte Carlo simulation for the CCF09 test is made conditionally
on the choice of \code{CCF09_dirs}. That is, all the Monte
Carlo statistics share the same random directions.
}
\examples{
## Critical values

# Single statistic, specific alpha
cir <- unif_stat_MC(n = 10, M = 1e2, type = "Ajne", p = 2, alpha = 0.15)
summary(cir$stats_MC)
cir$crit_val_MC

# All circular statistics
cir <- unif_stat_MC(n = 10, M = 1e2, p = 2)
head(cir$stats_MC)
cir$crit_val_MC

# All spherical statistics
sph <- unif_stat_MC(n = 10, M = 1e2, p = 3)
head(sph$stats_MC)
sph$crit_val_MC

## Using a progress bar

# Define a progress bar
require(progress)
require(progressr)
handlers(handler_progress(
  format = ":spin [:bar] :percent Total: :elapsedfull End \u2248 :eta",
  clear = FALSE))

# Call unif_stat_MC() within with_progress()
with_progress(unif_stat_MC(n = 10, M = 1e2, p = 3, chunks = 10))

# With several cores
with_progress(unif_stat_MC(n = 10, M = 1e2, p = 3, chunks = 10, cores = 2))

# Instead of using with_progress() each time, it is more practical to run
# handlers(global = TRUE)
# once to activate progress bars in your R session

## Power computation

# Single statistic
cir_pow <- unif_stat_MC(n = 10, M = 1e2, type = "Ajne", p = 2,
                        crit_val = cir$crit_val_MC)
cir_pow$crit_val_MC
cir_pow$power_MC

# All circular statistics
cir_pow <- unif_stat_MC(n = 10, M = 1e2, p = 2, crit_val = cir$crit_val_MC)
cir_pow$crit_val_MC
cir_pow$power_MC

# All spherical statistics
sph_pow <- unif_stat_MC(n = 10, M = 1e2, p = 3, crit_val = sph$crit_val_MC)
sph_pow$crit_val_MC
sph_pow$power_MC
\donttest{
## Custom r_H1

# Circular
r_H1 <- function(n, p, M, l = 0.05) {

  stopifnot(p == 2)
  Theta_to_X(matrix(runif(n * M, 0, (2 - l) * pi), n, M))

}
dirs <- r_unif_sph(n = 5, p = 2, M = 1)[, , 1]
cir <- unif_stat_MC(n = 50, M = 1e2, p = 2, CCF09_dirs = dirs)
cir_pow <- unif_stat_MC(n = 50, M = 1e2, p = 2, r_H1 = r_H1, l = 0.10,
                        crit_val = cir$crit_val_MC, CCF09_dirs = dirs)
cir_pow$crit_val_MC
cir_pow$power_MC

# Spherical
r_H1 <- function(n, p, M, l = 0.5) {

  samp <- array(dim = c(n, p, M))
  for (j in 1:M) {

    samp[, , j] <- mvtnorm::rmvnorm(n = n, mean = c(l, rep(0, p - 1)),
                                    sigma = diag(rep(1, p)))
    samp[, , j] <- samp[, , j] / sqrt(rowSums(samp[, , j]^2))

  }
  return(samp)

}
dirs <- r_unif_sph(n = 5, p = 3, M = 1)[, , 1]
sph <- unif_stat_MC(n = 50, M = 1e2, p = 3, CCF09_dirs = dirs)
sph_pow <- unif_stat_MC(n = 50, M = 1e2, p = 3, r_H1 = r_H1, l = 0.5,
                       crit_val = sph$crit_val_MC, CCF09_dirs = dirs)
sph_pow$power_MC

## Pre-built r_H1

# Circular
dirs <- r_unif_sph(n = 5, p = 2, M = 1)[, , 1]
cir_pow <- unif_stat_MC(n = 50, M = 1e2, p = 2, r_H1 = r_alt, alt = "vMF",
                        kappa = 1, crit_val = cir$crit_val_MC,
                        CCF09_dirs = dirs)
cir_pow$power_MC

# Spherical
dirs <- r_unif_sph(n = 5, p = 3, M = 1)[, , 1]
sph_pow <- unif_stat_MC(n = 50, M = 1e2, p = 3, r_H1 = r_alt, alt = "vMF",
                        kappa = 1, crit_val = sph$crit_val_MC,
                        CCF09_dirs = dirs)
sph_pow$power_MC
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{craters}
\alias{craters}
\title{Craters named by the IUA}
\format{
A data frame with 5235 rows and 7 variables:
\describe{
  \item{ID}{database ID.}
  \item{name}{name of the crater.}
  \item{target}{name of the celestial body. A factor with 43 levels,
  such as \code{"Moon"}, \code{"Venus"}, or \code{"Europa"}.}
  \item{target_type}{type of celestial body. A factor with 3 levels:
  \code{"Planet"}, \code{"Moon"}, \code{"Dwarf planet"}, or
  \code{"Asteroid"}.}
  \item{diameter}{diameter of the crater (in km).}
  \item{theta}{longitude angle \eqn{\theta \in [0, 2\pi)} of the
  crater center.}
  \item{phi}{latitude angle \eqn{\phi \in [-\pi/2, \pi/2]} of the
  crater center.}
}
}
\source{
\url{https://planetarynames.wr.usgs.gov/AdvancedSearch}
}
\usage{
craters
}
\description{
\emph{Named} craters of the Solar System by the
\href{https://planetarynames.wr.usgs.gov}{Gazetteer of Planetary
Nomenclature} of the International Astronomical Union (IUA).
}
\details{
"Craters" are understood in the Gazetteer of Planetary Nomenclature as
roughly circular depressions resulting from impact or volcanic activity
(the geological origin is
\href{https://planetarynames.wr.usgs.gov/DescriptorTerms}{unspecified}).

Be aware that the dataset only contains \emph{named} craters by the IUA.
Therefore, there is likely a \bold{high uniform bias} on the distribution
of craters. Presumably the naming process attempts to cover the planet in
a somehow uniform fashion (distant craters are more likely to be named than
neighboring craters). Also, there are substantially more craters in the
listed bodies than those named by the IUA. See \code{\link{venus}} and
\code{\link{rhea}} for more detailed and specific crater datasets.

The \eqn{(\theta, \phi)} angles are such their associated planetocentric
coordinates are:
\deqn{(\cos(\phi) \cos(\theta), \cos(\phi) \sin(\theta), \sin(\phi))',}{
(cos(\phi) cos(\theta), cos(\phi) sin(\theta), sin(\phi))',}
with \eqn{(0, 0, 1)'} denoting the north pole.

The script performing the data preprocessing is available at
\href{https://github.com/egarpor/sphunif/blob/master/data-raw/craters.R}{
\code{craters.R}}. The data was retrieved on 2020-05-31.
}
\examples{
# Load data
data("craters")

# Add Cartesian coordinates
craters$X <- cbind(cos(craters$theta) * cos(craters$phi),
                   sin(craters$theta) * cos(craters$phi),
                   sin(craters$phi))

# Tests to be performed
type_tests <- c("PCvM", "PAD", "PRt")

# Tests for Venus and Rhea
unif_test(data = craters$X[craters$target == "Venus", ], type = type_tests,
          p_value = "asymp")
unif_test(data = craters$X[craters$target == "Rhea", ], type = type_tests,
          p_value = "asymp")
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/int_sph_MC.R
\name{int_sph_MC}
\alias{int_sph_MC}
\title{Monte Carlo integration of functions on the (hyper)sphere}
\usage{
int_sph_MC(f, p, M = 10000, cores = 1, chunks = ceiling(M/1000),
  seeds = NULL, ...)
}
\arguments{
\item{f}{function to be integrated. Its first argument must be the
(hyper)sphere position. Must be vectorized and return a vector of size
\code{nrow(x)} for a matrix input \code{x}. See examples.}

\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}

\item{M}{number of Monte Carlo samples. Defaults to \code{1e4}.}

\item{cores}{number of cores to perform the integration. Defaults to
\code{1}.}

\item{chunks}{number of chunks to split the \code{M} Monte Carlo
samples. Useful for parallelizing the integration in \code{chunks}
tasks containing \code{ceiling(M / chunks)} replications. Useful also for
avoiding memory bottlenecks when \code{M} is large. Defaults to
\cr\code{ceiling(M / 1e3)}.}

\item{seeds}{if provided, a vector of size \code{chunks} for fixing the
seeds on each of the simulation chunks (useful for reproducing parallel
simulations). Specifically, for \code{k in 1:chunks}, seeds are
set as \code{set.seed(seeds[k], kind = "Mersenne-Twister")} in each chunk.
Defaults to \code{NULL} (no seed setting is done).}

\item{...}{optional arguments to be passed to \code{f} or to
\code{\link[foreach]{foreach}} (for example, \code{.export} to export global
variables or other functions to the \code{foreach} environment).}
}
\value{
A scalar with the approximate integral.
}
\description{
Monte Carlo approximation of the integral
\deqn{\int_{S^{p-1}}f(x)\,\mathrm{d}x}{\int_{S^{p-1}} f(x) dx}
of a function \eqn{f:S^{p-1} \rightarrow R} defined on the (hyper)sphere
\eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{
S^{p-1}:=\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}.
}
\details{
It is possible to have a progress bar if \code{int_sph_MC} is wrapped with
\code{\link[progressr:with_progress]{progressr::with_progress}} or if
\code{progressr::handlers(global = TRUE)} is invoked (once) by the user.
See the examples below. The progress bar is updated with the number of
finished chunks.
}
\examples{
## Sequential simulation

# Vectorized functions to be integrated
x1 <- function(x) x[, 1]
quad <- function(x, a = 0) a + rowSums(x^4)

# Approximate \int_{S^{p-1}} x_1 dx = 0
int_sph_MC(f = x1, p = 3, M = 1e4, chunks = 2)

# Approximate \int_{S^{p-1}} (a + \sum_i x_i^4) dx
int_sph_MC(f = quad, p = 2, M = 1e4, a = 0, chunks = 2)

# Compare with Gauss--Legendre integration on S^2
th_k <- Gauss_Legen_nodes(a = 0, b = 2 * pi, N = 40)
w_k <- Gauss_Legen_weights(a = 0, b = 2 * pi, N = 40)
sum(w_k * quad(cbind(cos(th_k), sin(th_k)), a = 1))

## Parallel simulation with a progress bar
\donttest{
# Define a progress bar
require(progress)
require(progressr)
handlers(handler_progress(
  format = ":spin [:bar] :percent Total: :elapsedfull End \u2248 :eta",
  clear = FALSE))

# Call int_sph_MC() within with_progress()
with_progress(int_sph_MC(f = x1, p = 3, cores = 2, M = 1e5, chunks = 100))

# Instead of using with_progress() each time, it is more practical to run
# handlers(global = TRUE)
# once to activate progress bars in your R session
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/cir_distr.R
\name{p_Kolmogorov}
\alias{p_Kolmogorov}
\alias{d_Kolmogorov}
\alias{p_cir_stat_Ajne}
\alias{d_cir_stat_Ajne}
\alias{p_cir_stat_Bingham}
\alias{d_cir_stat_Bingham}
\alias{p_cir_stat_Greenwood}
\alias{d_cir_stat_Greenwood}
\alias{p_cir_stat_Gini}
\alias{d_cir_stat_Gini}
\alias{p_cir_stat_Gini_squared}
\alias{d_cir_stat_Gini_squared}
\alias{p_cir_stat_Hodges_Ajne2}
\alias{p_cir_stat_Hodges_Ajne}
\alias{d_cir_stat_Hodges_Ajne}
\alias{p_cir_stat_Kuiper}
\alias{d_cir_stat_Kuiper}
\alias{p_cir_stat_Log_gaps}
\alias{d_cir_stat_Log_gaps}
\alias{p_cir_stat_Max_uncover}
\alias{d_cir_stat_Max_uncover}
\alias{p_cir_stat_Num_uncover}
\alias{d_cir_stat_Num_uncover}
\alias{p_cir_stat_Pycke}
\alias{d_cir_stat_Pycke}
\alias{p_cir_stat_Vacancy}
\alias{d_cir_stat_Vacancy}
\alias{p_cir_stat_Watson}
\alias{d_cir_stat_Watson}
\alias{p_cir_stat_Watson_1976}
\alias{d_cir_stat_Watson_1976}
\alias{p_cir_stat_Range}
\alias{d_cir_stat_Range}
\alias{p_cir_stat_Rao}
\alias{d_cir_stat_Rao}
\alias{p_cir_stat_Rayleigh}
\alias{d_cir_stat_Rayleigh}
\alias{cir_stat_distr}
\alias{p_cir_stat_Bakshaev}
\alias{d_cir_stat_Bakshaev}
\alias{p_cir_stat_Gine_Fn}
\alias{d_cir_stat_Gine_Fn}
\alias{p_cir_stat_Gine_Gn}
\alias{d_cir_stat_Gine_Gn}
\alias{p_cir_stat_Hermans_Rasson}
\alias{d_cir_stat_Hermans_Rasson}
\alias{p_cir_stat_PAD}
\alias{d_cir_stat_PAD}
\alias{p_cir_stat_PCvM}
\alias{d_cir_stat_PCvM}
\alias{p_cir_stat_PRt}
\alias{d_cir_stat_PRt}
\alias{p_cir_stat_Pycke_q}
\alias{d_cir_stat_Pycke_q}
\alias{p_cir_stat_Rothman}
\alias{d_cir_stat_Rothman}
\alias{p_cir_stat_Riesz}
\alias{d_cir_stat_Riesz}
\title{Asymptotic distributions for circular uniformity statistics}
\usage{
p_Kolmogorov(x, K_Kolmogorov = 25L, alternating = TRUE)

d_Kolmogorov(x, K_Kolmogorov = 25L, alternating = TRUE)

p_cir_stat_Ajne(x, K_Ajne = 15L)

d_cir_stat_Ajne(x, K_Ajne = 15L)

p_cir_stat_Bingham(x)

d_cir_stat_Bingham(x)

p_cir_stat_Greenwood(x)

d_cir_stat_Greenwood(x)

p_cir_stat_Gini(x)

d_cir_stat_Gini(x)

p_cir_stat_Gini_squared(x)

d_cir_stat_Gini_squared(x)

p_cir_stat_Hodges_Ajne2(x, n, asymp_std = FALSE)

p_cir_stat_Hodges_Ajne(x, n, exact = TRUE, asymp_std = FALSE)

d_cir_stat_Hodges_Ajne(x, n, exact = TRUE, asymp_std = FALSE)

p_cir_stat_Kuiper(x, n, K_Kuiper = 12L, second_term = TRUE, Stephens = FALSE)

d_cir_stat_Kuiper(x, n, K_Kuiper = 12L, second_term = TRUE, Stephens = FALSE)

p_cir_stat_Log_gaps(x, abs_val = TRUE)

d_cir_stat_Log_gaps(x, abs_val = TRUE)

p_cir_stat_Max_uncover(x)

d_cir_stat_Max_uncover(x)

p_cir_stat_Num_uncover(x)

d_cir_stat_Num_uncover(x)

p_cir_stat_Pycke(x)

d_cir_stat_Pycke(x)

p_cir_stat_Vacancy(x)

d_cir_stat_Vacancy(x)

p_cir_stat_Watson(x, n = 0L, K_Watson = 25L, Stephens = FALSE)

d_cir_stat_Watson(x, n = 0L, K_Watson = 25L, Stephens = FALSE)

p_cir_stat_Watson_1976(x, K_Watson_1976 = 8L, N = 40L)

d_cir_stat_Watson_1976(x, K_Watson_1976 = 8L)

p_cir_stat_Range(x, n, max_gap = TRUE)

d_cir_stat_Range(x, n, max_gap = TRUE)

p_cir_stat_Rao(x)

d_cir_stat_Rao(x)

p_cir_stat_Rayleigh(x)

d_cir_stat_Rayleigh(x)

p_cir_stat_Bakshaev(x, K_max = 1000, thre = 0, ...)

d_cir_stat_Bakshaev(x, K_max = 1000, thre = 0, ...)

p_cir_stat_Gine_Fn(x, K_max = 1000, thre = 0, ...)

d_cir_stat_Gine_Fn(x, K_max = 1000, thre = 0, ...)

p_cir_stat_Gine_Gn(x, K_max = 1000, thre = 0, ...)

d_cir_stat_Gine_Gn(x, K_max = 1000, thre = 0, ...)

p_cir_stat_Hermans_Rasson(x, K_max = 1000, thre = 0, ...)

d_cir_stat_Hermans_Rasson(x, K_max = 1000, thre = 0, ...)

p_cir_stat_PAD(x, K_max = 1000, thre = 0, ...)

d_cir_stat_PAD(x, K_max = 1000, thre = 0, ...)

p_cir_stat_PCvM(x, K_max = 1000, thre = 0, ...)

d_cir_stat_PCvM(x, K_max = 1000, thre = 0, ...)

p_cir_stat_PRt(x, t = 1/3, K_max = 1000, thre = 0, ...)

d_cir_stat_PRt(x, t = 1/3, K_max = 1000, thre = 0, ...)

p_cir_stat_Pycke_q(x, q = 0.5, K_max = 1000, thre = 0, ...)

d_cir_stat_Pycke_q(x, q = 0.5, K_max = 1000, thre = 0, ...)

p_cir_stat_Rothman(x, t = 1/3, K_max = 1000, thre = 0, ...)

d_cir_stat_Rothman(x, t = 1/3, K_max = 1000, thre = 0, ...)

p_cir_stat_Riesz(x, s = 1, K_max = 1000, thre = 0, ...)

d_cir_stat_Riesz(x, s = 1, K_max = 1000, thre = 0, ...)
}
\arguments{
\item{x}{a vector of size \code{nx} or a matrix of size \code{c(nx, 1)}.}

\item{K_Kolmogorov, K_Kuiper, K_Watson, K_Watson_1976, K_Ajne}{integer giving
the truncation of the series present in the null asymptotic distributions.
For the Kolmogorov-Smirnov-related series defaults to \code{25}; for the
others series defaults to a smaller number.}

\item{alternating}{use the alternating series expansion for the distribution
of the Kolmogorov-Smirnov statistic? Defaults to \code{TRUE}.}

\item{n}{sample size employed for computing the statistic.}

\item{asymp_std}{compute the distribution associated to the normalized
Hodges-Ajne statistic? Defaults to \code{FALSE}.}

\item{exact}{use the exact distribution for the Hodges-Ajne statistic?
Defaults to \code{TRUE}.}

\item{second_term}{use the second-order series expansion for the
distribution of the Kuiper statistic? Defaults to \code{TRUE}.}

\item{Stephens}{compute Stephens (1970) modification so that the null
distribution of the is less dependent on the sample size? The modification
does not alter the test decision.}

\item{abs_val}{compute the distribution associated to the absolute value of
the Darling's log gaps statistic? Defaults to \code{TRUE}.}

\item{N}{number of points used in the
\link[=Gauss_Legen_nodes]{Gauss-Legendre quadrature}. Defaults to \code{40}.}

\item{max_gap}{compute the distribution associated to the maximum gap for
the range statistic? Defaults to \code{TRUE}.}

\item{K_max}{integer giving the truncation of the series that compute the
asymptotic p-value of a Sobolev test. Defaults to \code{1e3}.}

\item{thre}{error threshold for the tail probability given by the
the first terms of the truncated series of a Sobolev test. Defaults to
\code{0} (no further truncation).}

\item{...}{further parameters passed to \code{\link{p_Sobolev}} or
\code{\link{d_Sobolev}} (such as \code{x_tail}).}

\item{t}{\eqn{t} parameter for the Rothman and Cressie tests, a real in
\eqn{(0, 1)}. Defaults to \code{1 / 3}.}

\item{q}{\eqn{q} parameter for the Pycke "\eqn{q}-test", a real in
\eqn{(0, 1)}. Defaults to \code{1 / 2}.}

\item{s}{\eqn{s} parameter for the \eqn{s}-Riesz test, a real in
\eqn{(0, 2)}. Defaults to \code{1}.}
}
\value{
A matrix of size \code{c(nx, 1)} with the evaluation of the
distribution or density function at \code{x}.
}
\description{
Computation of the asymptotic null distributions of circular
uniformity statistics.
}
\details{
Descriptions and references for most of the tests are available
in García-Portugués and Verdebout (2018).
}
\examples{
# Ajne
curve(d_cir_stat_Ajne(x), to = 1.5, n = 2e2, ylim = c(0, 4))
curve(p_cir_stat_Ajne(x), n = 2e2, col = 2, add = TRUE)

# Bakshaev
curve(d_cir_stat_Bakshaev(x, method = "HBE"), to = 6, n = 2e2,
      ylim = c(0, 1))
curve(p_cir_stat_Bakshaev(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)

# Bingham
curve(d_cir_stat_Bingham(x), to = 12, n = 2e2, ylim = c(0, 1))
curve(p_cir_stat_Bingham(x), n = 2e2, col = 2, add = TRUE)
# Greenwood
curve(d_cir_stat_Greenwood(x), from = -6, to = 6, n = 2e2, ylim = c(0, 1))
curve(p_cir_stat_Greenwood(x), n = 2e2, col = 2, add = TRUE)

# Hermans-Rasson
curve(p_cir_stat_Hermans_Rasson(x, method = "HBE"), to = 10, n = 2e2,
      ylim = c(0, 1))
curve(d_cir_stat_Hermans_Rasson(x, method = "HBE"), n = 2e2, add = TRUE,
      col = 2)

# Hodges-Ajne
plot(25:45, d_cir_stat_Hodges_Ajne(cbind(25:45), n = 50), type = "h",
     lwd = 2, ylim = c(0, 1))
lines(25:45, p_cir_stat_Hodges_Ajne(cbind(25:45), n = 50), type = "s",
      col = 2)

# Kolmogorov-Smirnov
curve(d_Kolmogorov(x), to = 3, n = 2e2, ylim = c(0, 2))
curve(p_Kolmogorov(x), n = 2e2, col = 2, add = TRUE)

# Kuiper
curve(d_cir_stat_Kuiper(x, n = 50), to = 3, n = 2e2, ylim = c(0, 2))
curve(p_cir_stat_Kuiper(x, n = 50), n = 2e2, col = 2, add = TRUE)

# Kuiper and Watson with Stephens modification
curve(d_cir_stat_Kuiper(x, n = 8, Stephens = TRUE), to = 2.5, n = 2e2,
      ylim = c(0, 10))
curve(d_cir_stat_Watson(x, n = 8, Stephens = TRUE), n = 2e2, lty = 2,
      add = TRUE)
n <- c(10, 20, 30, 40, 50, 100, 500)
col <- rainbow(length(n))
for (i in seq_along(n)) {
  curve(d_cir_stat_Kuiper(x, n = n[i], Stephens = TRUE), n = 2e2,
        col = col[i], add = TRUE)
  curve(d_cir_stat_Watson(x, n = n[i], Stephens = TRUE), n = 2e2,
        col = col[i], lty = 2, add = TRUE)
}

# Maximum uncovered spacing
curve(d_cir_stat_Max_uncover(x), from = -3, to = 6, n = 2e2, ylim = c(0, 1))
curve(p_cir_stat_Max_uncover(x), n = 2e2, col = 2, add = TRUE)

# Number of uncovered spacing
curve(d_cir_stat_Num_uncover(x), from = -4, to = 4, n = 2e2, ylim = c(0, 1))
curve(p_cir_stat_Num_uncover(x), n = 2e2, col = 2, add = TRUE)

# Log gaps
curve(d_cir_stat_Log_gaps(x), from = -1, to = 4, n = 2e2, ylim = c(0, 1))
curve(p_cir_stat_Log_gaps(x), n = 2e2, col = 2, add = TRUE)

# Gine Fn
curve(d_cir_stat_Gine_Fn(x, method = "HBE"), to = 2.5, n = 2e2,
      ylim = c(0, 2))
curve(p_cir_stat_Gine_Fn(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)

# Gine Gn
curve(d_cir_stat_Gine_Gn(x, method = "HBE"), to = 2.5, n = 2e2,
      ylim = c(0, 2))
curve(p_cir_stat_Gine_Gn(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)

# Gini mean difference
curve(d_cir_stat_Gini(x), from = -4, to = 4, n = 2e2, ylim = c(0, 1))
curve(p_cir_stat_Gini(x), n = 2e2, col = 2, add = TRUE)

# Gini mean squared difference
curve(d_cir_stat_Gini_squared(x), from = -10, to = 10, n = 2e2,
      ylim = c(0, 1))
curve(p_cir_stat_Gini_squared(x), n = 2e2, col = 2, add = TRUE)

# PAD
curve(d_cir_stat_PAD(x, method = "HBE"), to = 3, n = 2e2, ylim = c(0, 1.5))
curve(p_cir_stat_PAD(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)

# PCvM
curve(d_cir_stat_PCvM(x, method = "HBE"), to = 4, n = 2e2, ylim = c(0, 2))
curve(p_cir_stat_PCvM(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)

# PRt
curve(d_cir_stat_PRt(x, method = "HBE"), n = 2e2, ylim = c(0, 5))
curve(p_cir_stat_PRt(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)

# Pycke
curve(d_cir_stat_Pycke(x), from = -5, to = 10, n = 2e2, ylim = c(0, 1))
curve(p_cir_stat_Pycke(x), n = 2e2, col = 2, add = TRUE)

# Pycke q
curve(d_cir_stat_Pycke_q(x, method = "HBE"), to = 15, n = 2e2,
      ylim = c(0, 1))
curve(p_cir_stat_Pycke_q(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)

# Range
curve(d_cir_stat_Range(x, n = 50), to = 2, n = 2e2, ylim = c(0, 4))
curve(p_cir_stat_Range(x, n = 50), n = 2e2, col = 2, add = TRUE)

# Rao
curve(d_cir_stat_Rao(x), from = -6, to = 6, n = 2e2, ylim = c(0, 1))
curve(p_cir_stat_Rao(x), n = 2e2, col = 2, add = TRUE)

# Rayleigh
curve(d_cir_stat_Rayleigh(x), to = 12, n = 2e2, ylim = c(0, 1))
curve(p_cir_stat_Rayleigh(x), n = 2e2, col = 2, add = TRUE)

# Riesz
curve(d_cir_stat_Riesz(x, method = "HBE"), to = 6, n = 2e2,
      ylim = c(0, 1))
curve(p_cir_stat_Riesz(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)

# Rothman
curve(d_cir_stat_Rothman(x, method = "HBE"), n = 2e2, ylim = c(0, 5))
curve(p_cir_stat_Rothman(x, method = "HBE"), n = 2e2, add = TRUE, col = 2)

# Vacancy
curve(d_cir_stat_Vacancy(x), from = -4, to = 4, n = 2e2, ylim = c(0, 1))
curve(p_cir_stat_Vacancy(x), n = 2e2, col = 2, add = TRUE)

# Watson
curve(d_cir_stat_Watson(x), to = 0.5, n = 2e2, ylim = c(0, 15))
curve(p_cir_stat_Watson(x), n = 2e2, col = 2, add = TRUE)

# Watson (1976)
curve(d_cir_stat_Watson_1976(x), to = 1.5, n = 2e2, ylim = c(0, 3))
curve(p_cir_stat_Watson_1976(x), n = 2e2, col = 2, add = TRUE)
}
\references{
García-Portugués, E. and Verdebout, T. (2018) An overview of
uniformity tests on the hypersphere. \emph{arXiv:1804.00286}.
\url{https://arxiv.org/abs/1804.00286}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{r_unif}
\alias{r_unif}
\alias{r_unif_cir}
\alias{r_unif_sph}
\title{Sample uniformly distributed circular and spherical data}
\usage{
r_unif_cir(n, M = 1L, sorted = FALSE)

r_unif_sph(n, p, M = 1L)
}
\arguments{
\item{n}{sample size.}

\item{M}{number of samples of size \code{n}. Defaults to \code{1}.}

\item{sorted}{return each circular sample sorted? Defaults to \code{FALSE}.}

\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}
}
\value{
\itemize{
  \item \code{r_unif_cir}: a \bold{matrix} of size \code{c(n, M)} with
  \code{M} random samples of size \code{n} of uniformly-generated circular
  data on \eqn{[0, 2\pi)}.
  \item \code{r_unif_sph}: an \bold{array} of size \code{c(n, p, M)} with
  \code{M} random samples of size \code{n} of uniformly-generated
  directions on \eqn{S^{p-1}}.
}
}
\description{
Simulation of the uniform distribution on \eqn{[0, 2\pi)} and
\eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{
S^{p-1}:=\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}.
}
\examples{
# A sample on [0, 2*pi)
n <- 5
r_unif_cir(n = n)

# A sample on S^1
p <- 2
samp <- r_unif_sph(n = n, p = p)
samp
rowSums(samp^2)

# A sample on S^2
p <- 3
samp <- r_unif_sph(n = n, p = p)
samp
rowSums(samp^2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/harmonics.R
\name{harmonics}
\alias{harmonics}
\alias{g_i_k}
\title{(Hyper)spherical harmonics}
\usage{
g_i_k(x, i = 1, k = 1, m = NULL, show_m = FALSE)
}
\arguments{
\item{x}{locations in \eqn{S^{p-1}} to evaluate \eqn{g_{i,k}}. Either a
matrix of size \code{c(nx, p)} or a vector of size \code{p}. Normalized
internally if required (with a \code{warning} message).}

\item{i, k}{alternative indexing to refer to the \code{i}-th (hyper)spherical
harmonic of order \code{k}. \code{i} is a positive integer smaller than
\code{\link[=Sobolev]{d_p_k}} and \code{k} is a non-negative integer.}

\item{m}{(hyper)spherical harmonic index, as used in Proposition 2.1. The
index is computed internally from \code{i} and \code{k}. Defaults to
\code{NULL}.}

\item{show_m}{flag to print \code{m} if computed internally when
\code{m = NULL}.}
}
\value{
A vector of size \code{nrow(x)}.
}
\description{
Computation of a certain explicit representation of
(hyper)spherical harmonics on
\eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{
S^{p-1}:=\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}. Details are available in
García-Portugués et al. (2021).
}
\details{
The implementation uses Proposition 2.1 in García-Portugués et al. (2021),
which adapts Theorem 1.5.1 in Dai and Xu (2013) with the correction of
typos in the normalizing constant \eqn{h_\alpha} and in the definition of
the function \eqn{g_\alpha} of the latter theorem.
}
\examples{
n <- 3e3
old_par <- par(mfrow = c(2, 3))
k <- 2
for (i in 1:d_p_k(p = 3, k = k)) {
  X <- r_unif_sph(n = n, p = 3, M = 1)[, , 1]
  col <- rainbow(n)[rank(g_i_k(x = X, k = k, i = i, show_m = TRUE))]
  scatterplot3d::scatterplot3d(X[, 1], X[, 2], X[, 3], color = col,
                               axis = FALSE, pch = 19)
}
for (k in 0:5) {
  X <- r_unif_sph(n = n, p = 3, M = 1)[, , 1]
  col <- rainbow(n)[rank(g_i_k(x = X, k = k, i = 1, show_m = TRUE))]
  scatterplot3d::scatterplot3d(X[, 1], X[, 2], X[, 3], color = col,
                               axis = FALSE, pch = 19)
}
par(old_par)
}
\references{
Dai, F. and Xu, Y. (2013). \emph{Approximation Theory and Harmonic Analysis
on Spheres and Balls}. Springer, New York. \doi{10.1007/978-1-4614-6660-4}

García-Portugués, E., Paindaveine, D., and Verdebout, T. (2021). On the
power of Sobolev tests for isotropy under local rotationally symmetric
alternatives. \emph{arXiv:2108.09874}. \url{https://arxiv.org/abs/2108.09874}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{utils}
\alias{utils}
\alias{n_from_dist_vector}
\alias{t_inv_sqrt_one}
\title{Low-level utilities for \pkg{sphunif}}
\usage{
n_from_dist_vector(n_dist)

t_inv_sqrt_one(t)
}
\arguments{
\item{n_dist}{a positive integer \eqn{(n - 1) * n / 2} for which \eqn{n}
is to be recovered.}

\item{t}{a vector to evaluate \eqn{t / \sqrt{1 - t^2}}.}
}
\description{
Internal and undocumented low-level utilities for
\pkg{sphunif}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{wschisq_utils}
\alias{wschisq_utils}
\alias{r_wschisq_Cpp}
\alias{p_wschisq_MC}
\title{Utilities for weighted sums of non-central chi squared random
variables}
\usage{
r_wschisq_Cpp(n, weights, dfs, ncps)

p_wschisq_MC(x, weights = 0L, dfs = 0L, ncps = 0L, M = 10000L,
  sample = 0L, use_sample = FALSE, x_sorted = FALSE)
}
\arguments{
\item{n}{sample size.}

\item{weights}{vector with the positive weights of the sum. Must have the
same length as \code{dfs}.}

\item{dfs}{vector with the positive degrees of freedom of the chi squared
random variables. Must have the same length as \code{weights}.}

\item{ncps}{non-negative non-centrality parameters. A vector with the same
length as \code{weights}.}

\item{x}{vector of quantiles.}

\item{M}{number of Monte Carlo samples for approximating the distribution.
Defaults to \code{1e4}.}

\item{sample}{if \code{use_sample = TRUE}, the Monte Carlo sample to
approximate the distribution. If not, it is computed internally. Defaults
to \code{1e4}.}

\item{use_sample}{use the already computed \code{sample}? If \code{FALSE}
(default), \code{sample} is computed internally.}
}
\value{
\itemize{
  \item \code{r_wschisq_Cpp}: a matrix of size \code{c(n, 1)}
  containing a random sample.
  \item \code{p_wschisq_MC}: a matrix of size \code{c(nx, 1)}
  with the evaluation of the distribution function at \code{x}.
}
}
\description{
Simulation from a weighted sum of non-central chi squared
random variables and Monte Carlo approximation of its distribution function.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{cir_gaps}
\alias{cir_gaps}
\title{Circular gaps}
\usage{
cir_gaps(Theta, sorted = FALSE)
}
\arguments{
\item{Theta}{a \bold{matrix} of size \code{c(n, M)} with \code{M} samples
of size \code{n} of circular data on \eqn{[0, 2\pi)}. Must not contain
\code{NA}'s.}

\item{sorted}{are the columns of \code{Theta} sorted increasingly? If
\code{TRUE}, performance is improved. If \code{FALSE} (default), each
column of \code{Theta} is sorted internally.}
}
\value{
A matrix of size \code{c(n, M)} containing the \code{n} circular
gaps for each of the \code{M} circular samples.
}
\description{
Computation of the circular gaps of an angular sample
\eqn{\Theta_1,\ldots,\Theta_n} on \eqn{[0, 2\pi)}, defined as
\deqn{\Theta_{(2)} - \Theta_{(1)},\ldots,\Theta_{(n)} - \Theta_{(n - 1)},
2\pi - \Theta_{(n)} - \Theta_{(1)},}
where
\deqn{0 \le \Theta_{(1)} \le \Theta_{(2)} \le \ldots \le
\Theta_{(n)} \le 2\pi.}
}
\section{Warning}{

Be careful on avoiding the next bad usages of \code{cir_gaps}, which will
produce spurious results:
\itemize{
  \item The entries of \code{Theta} are \emph{not} in \eqn{[0, 2\pi)}.
  \item \code{Theta} is \emph{not} sorted increasingly when
  \code{data_sorted = TRUE}.
}
}

\examples{
Theta <- cbind(c(pi, 0, 3 * pi / 2), c(0, 3 * pi / 2, pi), c(5, 3, 1))
cir_gaps(Theta)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sphunif-package.R
\docType{package}
\name{sphunif-package}
\alias{sphunif-package}
\alias{sphunif}
\title{\code{sphunif}: Uniformity Tests on the Circle, Sphere, and
Hypersphere}
\description{
Implementation of uniformity tests on the circle and
(hyper)sphere. The main function of the package is \code{\link{unif_test}},
which conveniently collects more than 30 tests for assessing uniformity on
\eqn{S^{p-1}=\{{\bf x}\in R^p:||{\bf x}||=1\}}{
S^{p-1}=\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}. The test statistics are
implemented in the \code{\link{unif_stat}} function, which allows computing
several statistics to different samples within a single call, thus
facilitating Monte Carlo experiments. Furthermore, the
\code{\link{unif_stat_MC}} function allows parallelizing them in
a simple way. The asymptotic null distributions of the statistics are
available through the function \code{\link{unif_stat_distr}}. The core of
\code{\link{sphunif-package}} is coded in C++ by relying on the
\code{\link[Rcpp]{Rcpp-package}}. The package also provides several
novel datasets and gives the reproducibility for the data application
in García-Portugués, Navarro-Esteban and Cuesta-Albertos (2020)
<arXiv:2008.09897>.
}
\references{
García-Portugués, E. and Verdebout, T. (2018) An overview of uniformity
tests on the hypersphere. \emph{arXiv:1804.00286}.
\url{https://arxiv.org/abs/1804.00286}.

García-Portugués, E., Navarro-Esteban, P., Cuesta-Albertos, J. A. (2020)
On a projection-based class of uniformity tests on the hypersphere.
\emph{arXiv:2008.09897}. \url{https://arxiv.org/abs/2008.09897}

García-Portugués, E., Navarro-Esteban, P., and Cuesta-Albertos, J. A. (2021).
A Cramér–von Mises test of uniformity on the hypersphere. In Balzano, S.,
Porzio, G. C., Salvatore, R., Vistocco, D., and Vichi, M. (Eds.), \emph{
Statistical Learning and Modeling in Data Analysis}, Studies in
Classification, Data Analysis and Knowledge Organization, pp. 107–-116.
Springer, Cham. \doi{10.1007/978-3-030-69944-4_12}.

García-Portugués, E., Paindaveine, D., and Verdebout, T. (2021). On the
power of Sobolev tests for isotropy under local rotationally symmetric
alternatives. \emph{arXiv:2108.09874}. \url{https://arxiv.org/abs/2108.09874}
}
\author{
Eduardo García-Portugués and Thomas Verdebout.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unif-alts.R
\name{locdev}
\alias{locdev}
\alias{f_locdev}
\alias{con_f}
\alias{d_locdev}
\alias{r_locdev}
\alias{cutoff_locdev}
\title{Local projected alternatives to uniformity}
\usage{
f_locdev(z, p, uk)

con_f(f, p, N = 320)

d_locdev(x, mu, f, kappa)

r_locdev(n, mu, f, kappa, F_inv = NULL, ...)

cutoff_locdev(p, K_max = 10000, thre = 0.001, type, Rothman_t = 1/3,
  Pycke_q = 0.5, verbose = FALSE, Gauss = TRUE, N = 320, tol = 1e-06)
}
\arguments{
\item{z}{projected evaluation points for \eqn{f}, a vector with entries on
\eqn{[-1, 1]}.}

\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}

\item{uk}{coefficients \eqn{u_{k, p}} associated to the indexes
\code{1:length(uk)}, a vector.}

\item{f}{angular function defined on \eqn{[-1, 1]}. Must be vectorized.}

\item{N}{number of points used in the \link[=Gauss_Legen_nodes]{
Gauss--Legendre quadrature} for computing the Gegenbauer coefficients.
Defaults to \code{320}.}

\item{x}{locations in \eqn{S^{p-1}} to evaluate the density. Either a
matrix of size \code{c(nx, p)} or a vector of length \code{p}. Normalized
internally if required (with a \code{warning} message).}

\item{mu}{a unit norm vector of size \code{p} giving the axis of rotational
symmetry.}

\item{kappa}{the strength of the local alternative, between \code{0}
and \code{1}.}

\item{n}{sample size, a positive integer.}

\item{F_inv}{quantile function associated to \eqn{f}. Computed by
\code{\link{F_inv_from_f}} if \code{NULL} (default).}

\item{...}{further parameters passed to \code{\link{F_inv_from_f}}.}

\item{K_max}{integer giving the truncation of the series. Defaults to
\code{1e4}.}

\item{thre}{proportion of norm \emph{not} explained by the first terms of the
truncated series. Defaults to \code{1e-3}.}

\item{type}{Sobolev statistic. For \eqn{p = 2}, either \code{"Watson"},
\code{"Rothman"}, \code{"Pycke_q"}, or \code{"Hermans_Rasson"}.
For \eqn{p \ge 2}, \code{"Ajne"}, \code{"Gine_Gn"}, \code{"Gine_Fn"},
\code{"Bakshaev"}, \code{"Riesz"}, \code{"PCvM"}, \code{"PAD"}, or
\code{"PRt"}.}

\item{Rothman_t}{\eqn{t} parameter for the Rothman test, a real in
\eqn{(0, 1)}. Defaults to \code{1 / 3}.}

\item{Pycke_q}{\eqn{q} parameter for the Pycke "\eqn{q}-test", a real in
\eqn{(0, 1)}. Defaults to \code{1 / 2}.}

\item{verbose}{output information about the truncation (\code{TRUE} or
\code{1}) and a diagnostic plot (\code{2})? Defaults to \code{FALSE}.}

\item{Gauss}{use a Gauss--Legendre quadrature rule of \code{N} nodes
in the computation of the Gegenbauer coefficients? Otherwise, call
\code{\link{integrate}}. Defaults to \code{TRUE}.}

\item{tol}{tolerance passed to \code{\link{integrate}}'s \code{rel.tol} and
\code{abs.tol} if \code{Gauss = FALSE}. Defaults to \code{1e-6}.}
}
\value{
\itemize{
  \item \code{f_locdev}: angular function evaluated at \code{x}, a vector.
  \item \code{con_f}: normalizing constant \eqn{c_f} of \eqn{f}, a scalar.
  \item \code{d_locdev}: density function evaluated at \code{x}, a vector.
  \item \code{r_locdev}: a matrix of size \code{c(n, p)} containing a random
  sample from the density \eqn{f_{\kappa, \boldsymbol{\mu}}}{
  f_{\kappa, \mu}}.
  \item \code{cutoff_locdev}: vector of coefficients \eqn{\{u_{k, p}\}}
  automatically truncated according to \code{K_max} and \code{thre}
  (see details).
}
}
\description{
Density and random generation for local projected alternatives
to uniformity with densities
\deqn{f_{\kappa, \boldsymbol{\mu}}({\bf x}): =
\frac{1 - \kappa}{\omega_p} + \kappa f({\bf x}'\boldsymbol{\mu})}{
f_{\kappa, \mu}(x) = (1 - \kappa) / \omega_p + \kappa f(x'\mu)}
where
\deqn{f(z) = \frac{1}{\omega_p}\left\{1 + \sum_{k = 1}^\infty u_{k, p}
C_k^{p / 2 - 1}(z)\right\}}{f(x) = (1 / \omega_p)
\{1 + \sum_{k = 1}^\infty u_{k, p} C_k^(p / 2 - 1)(z)\}}
is the \emph{angular function} controlling the local alternative in a
\link[=Gegenbauer]{Gegenbauer series}, \eqn{0\le \kappa \le 1},
\eqn{\boldsymbol{\mu}}{\mu} is a direction on \eqn{S^{p - 1}}, and
\eqn{\omega_p} is the surface area of \eqn{S^{p - 1}}. The sequence
\eqn{\{u_{k, p}\}} is typically such that
\eqn{u_{k, p} = \left(1 + \frac{2k}{p - 2}\right) b_{k, p}}{
u_{k, p} = (1 + 2k / (p - 2)) b_{k, p}} for the Gegenbauer coefficients
\eqn{\{b_{k, p}\}} of the kernel function of a Sobolev statistic (see the
\link[=Sobolev_coefs]{transformation} between the coefficients \eqn{u_{k, p}}
and \eqn{b_{k, p}}).

Also, automatic truncation of the series \eqn{\sum_{k = 1}^\infty u_{k, p}
C_k^{p / 2 - 1}(z)}{\sum_{k = 1}^\infty u_{k, p} C_k^(p / 2 - 1)(z)}
according to the proportion of \link[=Gegenbauer]{"Gegenbauer norm"}
explained.
}
\details{
See the definitions of local alternatives in Prentice (1978) and in
García-Portugués et al. (2020).

The truncation of \eqn{\sum_{k = 1}^\infty u_{k, p} C_k^{p / 2 - 1}(z)}{
\sum_{k = 1}^\infty u_{k, p} C_k^(p / 2 - 1)(z)} is done to the first
\code{K_max} terms and then up to the index such that the first terms
leave unexplained the proportion \code{thre} of the norm of the whole series.
Setting \code{thre = 0} truncates to \code{K_max} terms exactly. If the
series only contains odd or even non-zero terms, then only \code{K_max / 2}
addends are \emph{effectively} taken into account in the first truncation.
}
\examples{
## Local alternatives diagnostics

loc_alt_diagnostic  <- function(p, type, thre = 1e-3, K_max = 1e3) {

  # Coefficients of the alternative
  uk <- cutoff_locdev(K_max = K_max, p = p, type = type, thre = thre,
                      N = 640)

  old_par <- par(mfrow = c(2, 2))

  # Construction of f
  z <- seq(-1, 1, l = 1e3)
  f <- function(z) f_locdev(z = z, p = p, uk = uk)
  plot(z, f(z), type = "l", xlab = expression(z), ylab = expression(f(z)),
       main = paste0("Local alternative f, ", type, ", p = ", p), log = "y")

  # Projected density on [-1, 1]
  f_proj <- function(z) rotasym::w_p(p = p - 1) * f(z) *
    (1 - z^2)^((p - 3) / 2)
  plot(z, f_proj(z), type = "l", xlab = expression(z),
       ylab = expression(omega[p - 1] * f(z) * (1 - z^2)^{(p - 3) / 2}),
       main = paste0("Projected density, ", type, ", p = ", p), log = "y",
       sub = paste("Integral:", round(con_f(f = f, p = p), 4)))

  # Quantile function for projected density
  mu <- c(rep(0, p - 1), 1)
  F_inv <- F_inv_from_f(f = f, p = p, K = 5e2)
  plot(F_inv, xlab = expression(x), ylab = expression(F^{-1}*(x)),
       main = paste0("Quantile function, ", type, ", p = ", p))

  # Sample from the alternative and plot the projected sample
  n <- 5e4
  samp <- r_locdev(n = n, mu = mu, f = f, kappa = 1, F_inv = F_inv)
  plot(z, f_proj(z), col = 2, type = "l",
       main = paste0("Simulated projected data, ", type, ", p = ", p),
       ylim = c(0, 1.75))
  hist(samp \%*\% mu, freq = FALSE, breaks = seq(-1, 1, l = 50), add = TRUE)

  par(old_par)

}
\donttest{
## Local alternatives for the PCvM test

loc_alt_diagnostic(p = 2, type = "PCvM")
loc_alt_diagnostic(p = 3, type = "PCvM")
loc_alt_diagnostic(p = 4, type = "PCvM")
loc_alt_diagnostic(p = 5, type = "PCvM")
loc_alt_diagnostic(p = 11, type = "PCvM")

## Local alternatives for the PAD test

loc_alt_diagnostic(p = 2, type = "PAD")
loc_alt_diagnostic(p = 3, type = "PAD")
loc_alt_diagnostic(p = 4, type = "PAD")
loc_alt_diagnostic(p = 5, type = "PAD")
loc_alt_diagnostic(p = 11, type = "PAD")

## Local alternatives for the PRt test

loc_alt_diagnostic(p = 2, type = "PRt")
loc_alt_diagnostic(p = 3, type = "PRt")
loc_alt_diagnostic(p = 4, type = "PRt")
loc_alt_diagnostic(p = 5, type = "PRt")
loc_alt_diagnostic(p = 11, type = "PRt")
}
}
\references{
García-Portugués, E., Navarro-Esteban, P., Cuesta-Albertos, J. A. (2020)
On a projection-based class of uniformity tests on the hypersphere.
\emph{arXiv:2008.09897}. \url{https://arxiv.org/abs/2008.09897}

Prentice, M. J. (1978). On invariant tests of uniformity for directions and
orientations. \emph{The Annals of Statistics}, 6(1):169--176.
\doi{10.1214/aos/1176344075}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unif-alts.R
\name{Sobolev_coefs}
\alias{Sobolev_coefs}
\alias{bk_to_vk2}
\alias{bk_to_uk}
\alias{vk2_to_bk}
\alias{vk2_to_uk}
\alias{uk_to_vk2}
\alias{uk_to_bk}
\title{Transformation between different coefficients in Sobolev statistics}
\usage{
bk_to_vk2(bk, p)

bk_to_uk(bk, p, signs = 1)

vk2_to_bk(vk2, p)

vk2_to_uk(vk2, p, signs = 1)

uk_to_vk2(uk, p)

uk_to_bk(uk, p)
}
\arguments{
\item{bk}{coefficients \eqn{b_{k, p}} associated to the indexes
\code{1:length(bk)}, a vector.}

\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}

\item{signs}{signs of the coefficients \eqn{u_{k, p}}, a vector of the
same size as \code{vk2} or \code{bk}, or a scalar. Defaults to \code{1}.}

\item{vk2}{\bold{squared} coefficients \eqn{v_{k, p}^2} associated to the
indexes \code{1:length(vk2)}, a vector.}

\item{uk}{coefficients \eqn{u_{k, p}} associated to the indexes
\code{1:length(uk)}, a vector.}
}
\value{
The corresponding vectors of coefficients \code{vk2}, \code{bk}, or
\code{uk}, depending on the call.
}
\description{
Given a Sobolev statistic
\deqn{S_{n, p} = \sum_{i, j = 1}^n \psi(\cos^{-1}({\bf X}_i'{\bf X}_j)),}{
S_{n, p} = \sum_{i, j = 1}^n \psi(\cos^{-1}(X_i'X_j)),}
for a sample \eqn{{\bf X}_1, \ldots, {\bf X}_n \in S^{p - 1} := \{{\bf x}
\in R^p : ||{\bf x}|| = 1\}}{X_1, \ldots, X_n \in S^{p - 1} :=
\{x \in R^p : ||x|| = 1\}}, \eqn{p\ge 2}, three important sequences
are related to \eqn{S_{n, p}}.
\itemize{
\item \link[=Gegen_coefs]{Gegenbauer coefficients} \eqn{\{b_{k, p}\}} of
\eqn{\psi_p} (see, e.g., the \link[=Pn]{projected-ecdf statistics}), given
by
\deqn{b_{k, p} := \frac{1}{c_{k, p}}\int_0^\pi \psi_p(\theta)
C_k^{p / 2 - 1}(\cos\theta)\,\mathrm{d}\theta.}{
b_{k, p} := \frac{1}{c_{k, p}} \int_0^\pi \psi_p(\theta)
C_k^(p / 2 - 1)(\cos\theta) d\theta.}
\item Weights \eqn{\{v_{k, p}^2\}} of the
\link[=Sobolev]{asymptotic distribution} of the Sobolev statistic,
\eqn{\sum_{k = 1}^\infty v_k^2 \chi^2_{d_{p, k}}}, given by
\deqn{v_{k, p}^2 = \left(1 + \frac{2k}{p - 2}\right)^{-1} b_{k, p},
\quad p \ge 3.}{v_{k, p}^2 = (1 + 2k / (p - 2))^{-1} b_{k, p}, p \ge 3.}
\item Gegenbauer coefficients \eqn{\{u_{k, p}\}} of the
\link[=locdev]{local projected alternative} associated to \eqn{S_{n, p}},
given by
\deqn{u_{k, p} = \left(1 + \frac{2k}{p - 2}\right) v_{k, p},
\quad p \ge 3.}{u_{k, p} = (1 + 2k / (p - 2)) b_{k, p}, p \ge 3.}
}
For \eqn{p = 2}, the factor \eqn{(1 + 2k / (p - 2))} is replaced by \eqn{2}.
}
\details{
See more details in Prentice (1978) and García-Portugués et al. (2020). The
adequate signs of \code{uk} for the \code{"PRt"} \link[=Pn]{Rothman test}
can be retrieved with \code{\link{akx}} and \code{sqr = TRUE}, see the
examples.
}
\examples{
# bk, vk2, and uk for the PCvM test in p = 3
(bk <- Gegen_coefs_Pn(k = 1:5, type = "PCvM", p = 3))
(vk2 <- bk_to_vk2(bk = bk, p = 3))
(uk <- bk_to_uk(bk = bk, p = 3))

# vk2 is the same as
weights_dfs_Sobolev(K_max = 10, thre = 0, p = 3, type = "PCvM")$weights

# bk and uk for the Rothman test in p = 3, with adequate signs
t <- 1 / 3
(bk <- Gegen_coefs_Pn(k = 1:5, type = "PRt", p = 3, Rothman_t = t))
(ak <- akx(x = drop(q_proj_unif(t, p = 3)), p = 3, k = 1:5, sqr = TRUE))
(uk <- bk_to_uk(bk = bk, p = 3, signs = ak))
}
\references{
García-Portugués, E., Navarro-Esteban, P., Cuesta-Albertos, J. A. (2020)
On a projection-based class of uniformity tests on the hypersphere.
\emph{arXiv:2008.09897}. \url{https://arxiv.org/abs/2008.09897}

Prentice, M. J. (1978). On invariant tests of uniformity for directions and
orientations. \emph{The Annals of Statistics}, 6(1):169--176.
\doi{10.1214/aos/1176344075}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{sort_each_col}
\alias{sort_each_col}
\alias{sort_index_each_col}
\title{Sort the columns of a matrix}
\usage{
sort_each_col(A)

sort_index_each_col(A)
}
\arguments{
\item{A}{a matrix of arbitrary dimensions.}
}
\value{
A matrix with the same dimensions as \code{A} such that each
column is sorted increasingly (for \code{sort_each_col}) or contains the
sorting indexes (for \code{sort_index_each_col}).
\code{sort_index_each_col}.
}
\description{
Convenience functions to sort the columns of a matrix in an
increasing way.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{Psi}
\alias{Psi}
\alias{Psi_mat}
\alias{upper_tri_ind}
\title{Shortest angles matrix}
\usage{
Psi_mat(data, ind_tri = 0L, use_ind_tri = FALSE, scalar_prod = FALSE,
  angles_diff = FALSE)

upper_tri_ind(n)
}
\arguments{
\item{data}{an array of size \code{c(n, p, M)} containing the Cartesian
coordinates of \code{M} samples of size \code{n} of directions on
\eqn{S^{p-1}}. Alternatively if \code{p = 2}, an array of size
\code{c(n, 1, M)} containing the angles on \eqn{[0, 2\pi)} of the \code{M}
circular samples of size \code{n} on \eqn{S^{1}}. Must not contain
\code{NA}'s.}

\item{ind_tri}{if \code{use_ind_tri = TRUE}, the vector of 0-based indexes
provided by \code{upper_tri_ind(n)}, which allows to extract the upper
triangular part of the matrix \eqn{\boldsymbol\Psi}{\Psi}. See the examples.}

\item{use_ind_tri}{use the already computed vector index \code{ind_tri}? If
\code{FALSE} (default), \code{ind_tri} is computed internally.}

\item{scalar_prod}{return the scalar products
\eqn{{\bf X}_i'{\bf X}}{X_i'X_j} instead of the shortest angles? Only taken
into account for data in \emph{Cartesian} form. Defaults to
\code{FALSE}.}

\item{angles_diff}{return the (unwrapped) angles difference
\eqn{\Theta_i-\Theta_j} instead of the shortest angles? Only taken into
account for data in \emph{angular} form. Defaults to \code{FALSE}.}

\item{n}{sample size, used to determine the index vector that gives the
upper triangular part of \eqn{\boldsymbol\Psi}{\Psi}.}
}
\value{
\itemize{
  \item \code{Psi_mat}: a matrix of size
  \code{c(n * (n - 1) / 2, M)} containing, for each column, the vector
  half of \eqn{\boldsymbol\Psi}{\Psi} for each of the \code{M} samples.
  \item \code{upper_tri_ind}: a matrix of size \code{n * (n - 1) / 2}
  containing the 0-based linear indexes for extracting the upper triangular
  matrix of a matrix of size \code{c(n, n)}, diagonal excluded, assuming
  column-major order.
}
}
\description{
Efficient computation of the shortest angles matrix
\eqn{\boldsymbol\Psi}{\Psi}, defined as
\deqn{\Psi_{ij}:=\cos^{-1}({\bf X}_i'{\bf X}_j),\quad
i,j=1,\ldots,n,}{\Psi_{ij} = \cos^{-1}(X_i'X_j), i, j = 1, \ldots, n,}
for a sample \eqn{{\bf X}_1,\ldots,{\bf X}_n\in S^{p-1}:=\{{\bf x}\in
R^p:||{\bf x}||=1\}}{X_1, \ldots, X_n \in
S^{p - 1} := \{x \in R^p : ||x|| = 1\}}, \eqn{p\ge 2}.

For a circular sample \eqn{\Theta_1, \ldots, \Theta_n \in [0, 2\pi)},
\eqn{\boldsymbol\Psi}{\Psi} can be expressed as
\deqn{\Psi_{ij}=\pi-|\pi-|\Theta_i-\Theta_j||,\quad
i,j=1,\ldots,n.}{\Psi_{ij}=\pi-|\pi-|\Theta_i-\Theta_j||, i,j=1,\ldots,n.}
}
\section{Warning}{

Be careful on avoiding the next bad usages of \code{Psi_mat}, which will
produce spurious results:
\itemize{
  \item The directions in \code{data} do \emph{not} have unit norm when
  Cartesian coordinates are employed.
  \item The entries of \code{data} are \emph{not} in \eqn{[0, 2\pi)} when
  polar coordinates are employed.
  \item \code{ind_tri} is a vector of size \code{n * (n - 1) / 2} that
  does \emph{not} contain the indexes produced by \code{upper_tri_ind(n)}.
}
}

\examples{
# Shortest angles
n <- 5
X <- r_unif_sph(n = n, p = 2, M = 2)
Theta <- X_to_Theta(X)
dim(Theta) <- c(n, 1, 2)
Psi_mat(X)
Psi_mat(Theta)

# Precompute ind_tri
ind_tri <- upper_tri_ind(n)
Psi_mat(X, ind_tri = ind_tri, use_ind_tri = TRUE)

# Compare with R
A <- acos(tcrossprod(X[, , 1]))
ind <- upper.tri(A)
A[ind]

# Reconstruct matrix
Psi_vec <- Psi_mat(Theta[, , 1, drop = FALSE])
Psi <- matrix(0, nrow = n, ncol = n)
Psi[upper.tri(Psi)] <- Psi_vec
Psi <- Psi + t(Psi)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unif_stat.R
\name{unif_stat}
\alias{unif_stat}
\title{Circular and (hyper)spherical uniformity statistics}
\usage{
unif_stat(data, type = "all", data_sorted = FALSE, Rayleigh_m = 1,
  cov_a = 2 * pi, Rothman_t = 1/3, Cressie_t = 1/3, Pycke_q = 0.5,
  Riesz_s = 1, CCF09_dirs = NULL, K_CCF09 = 25, CJ12_reg = 3)
}
\arguments{
\item{data}{sample to compute the test statistic. An \bold{array} of size
\code{c(n, p, M)} containing \code{M} samples of size \code{n} of directions
(in Cartesian coordinates) on \eqn{S^{p-1}}. Alternatively, a
\bold{matrix} of size \code{c(n, M)} with the angles on \eqn{[0, 2\pi)} of
the \code{M} circular samples of size \code{n} on \eqn{S^{1}}. Other objects
accepted are an array of size \code{c(n, 1, M)} or a vector of size
\code{n} with angular data. Must not contain \code{NA}'s.}

\item{type}{type of test to be applied. A character vector containing any of
the following types of tests, depending on the dimension \eqn{p}:
\itemize{
  \item Circular data: any of the names available at object
  \code{\link{avail_cir_tests}}.
  \item (Hyper)spherical data: any of the names available at object
  \code{\link{avail_sph_tests}}.
}
If \code{type = "all"} (default), then \code{type} is set as
\code{avail_cir_tests} or \code{avail_sph_tests}, depending on the value of
\eqn{p}.}

\item{data_sorted}{is the circular data sorted? If \code{TRUE}, certain
statistics are faster to compute. Defaults to \code{FALSE}.}

\item{Rayleigh_m}{integer \eqn{m} for the \eqn{m}-modal Rayleigh test.
Defaults to \code{m = 1} (the standard Rayleigh test).}

\item{cov_a}{\eqn{a_n = a / n} parameter used in the length of the arcs
of the coverage-based tests. Must be positive. Defaults to \code{2 * pi}.}

\item{Rothman_t}{\eqn{t} parameter for the Rothman test, a real in
\eqn{(0, 1)}. Defaults to \code{1 / 3}.}

\item{Cressie_t}{\eqn{t} parameter for the Cressie test, a real in
\eqn{(0, 1)}. Defaults to \code{1 / 3}.}

\item{Pycke_q}{\eqn{q} parameter for the Pycke "\eqn{q}-test", a real in
\eqn{(0, 1)}. Defaults to \code{1 / 2}.}

\item{Riesz_s}{\eqn{s} parameter for the \eqn{s}-Riesz test, a real in
\eqn{(0, 2)}. Defaults to \code{1}.}

\item{CCF09_dirs}{a matrix of size \code{c(n_proj, p)} containing
\code{n_proj} random directions (in Cartesian coordinates) on \eqn{S^{p-1}}
to perform the CCF09 test. If \code{NULL} (default), a sample of size
\code{n_proj = 50} directions is computed internally.}

\item{K_CCF09}{integer giving the truncation of the series present in the
asymptotic distribution of the Kolmogorov-Smirnov statistic. Defaults to
\code{5e2}.}

\item{CJ12_reg}{type of asymptotic regime for CJ12 test, either \code{1}
(sub-exponential regime), \code{2} (exponential), or \code{3}
(super-exponential; default).}
}
\value{
A data frame of size \code{c(M, length(type))}, with column names
given by \code{type}, that contains the values of the test statistics.
}
\description{
Implementation of several statistics for assessing uniformity
on the (hyper)sphere
\eqn{S^{p-1} := \{{\bf x} \in R^p : ||{\bf x}|| = 1\}}{
S^{p-1} := \{x \in R^p : ||x|| = 1\}}, \eqn{p\ge 2}, for a sample
\eqn{{\bf X}_1,\ldots,{\bf X}_n\in S^{p-1}}{X_1,\ldots,X_n\in S^{p-1}}.

\code{unif_stat} receives a (several) sample(s) of directions in
\emph{Cartesian coordinates}, except for the circular case (\eqn{p=2}) in
which the sample(s) can be \emph{angles}
\eqn{\Theta_1,\ldots,\Theta_n\in [0, 2\pi)}.

\code{unif_stat} allows to compute several statistics to several samples
within a single call, facilitating thus Monte Carlo experiments.
}
\details{
Descriptions and references for most of the statistics are available
in García-Portugués and Verdebout (2018).
}
\examples{
## Circular data

# Sample
n <- 10
M <- 2
Theta <- r_unif_cir(n = n, M = M)

# Matrix
unif_stat(data = Theta, type = "all")

# Array
unif_stat(data = array(Theta, dim = c(n, 1, M)), type = "all")

# Vector
unif_stat(data = Theta[, 1], type = "all")

## Spherical data

# Circular sample in Cartesian coordinates
n <- 10
M <- 2
X <- array(dim = c(n, 2, M))
for (i in 1:M) X[, , i] <- cbind(cos(Theta[, i]), sin(Theta[, i]))

# Array
unif_stat(data = X, type = "all")

# High-dimensional data
X <- r_unif_sph(n = n, p = 3, M = M)
unif_stat(data = X, type = "all")

## Specific arguments

# Rothman
unif_stat(data = Theta, type = "Rothman", Rothman_t = 0.5)

# CCF09
unif_stat(data = X, type = "CCF09", CCF09_dirs = X[, , 1])
unif_stat(data = X, type = "CCF09", CCF09_dirs = X[, , 1], K_CCF09 = 1)

# CJ12
unif_stat(data = X, type = "CJ12", CJ12_reg = 3)
unif_stat(data = X, type = "CJ12", CJ12_reg = 1)
}
\references{
García-Portugués, E. and Verdebout, T. (2018) An overview of uniformity
tests on the hypersphere. \emph{arXiv:1804.00286}.
\url{https://arxiv.org/abs/1804.00286}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{venus}
\alias{venus}
\title{Venus craters}
\format{
A data frame with 967 rows and 4 variables:
\describe{
  \item{name}{name of the crater (if named).}
  \item{diameter}{diameter of the crater (in km).}
  \item{theta}{longitude angle \eqn{\theta \in [0, 2\pi)} of the
  crater center.}
  \item{phi}{latitude angle \eqn{\phi \in [-\pi/2, \pi/2]} of the
  crater center.}
}
}
\source{
\url{https://astrogeology.usgs.gov/search/map/Venus/venuscraters}
}
\usage{
venus
}
\description{
Craters on Venus from the
\href{https://astrogeology.usgs.gov/search/map/Venus/venuscraters}{
USGS Astrogeology Science Center}.
}
\details{
The \eqn{(\theta, \phi)} angles are such their associated planetocentric
coordinates are:
\deqn{(\cos(\phi) \cos(\theta), \cos(\phi) \sin(\theta), \sin(\phi))',}{
(cos(\phi) cos(\theta), cos(\phi) sin(\theta), sin(\phi))',}
with \eqn{(0, 0, 1)'} denoting the north pole.

The script performing the data preprocessing is available at
\href{https://github.com/egarpor/sphunif/blob/master/data-raw/venus.R}{
\code{venus.R}}.
}
\examples{
# Load data
data("venus")

# Add Cartesian coordinates
venus$X <- cbind(cos(venus$theta) * cos(venus$phi),
                 sin(venus$theta) * cos(venus$phi),
                 sin(venus$phi))

# Tests
unif_test(data = venus$X, type = c("PCvM", "PAD", "PRt"), p_value = "asymp")
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{beta_inc}
\alias{beta_inc}
\alias{beta_inc_inv}
\title{The incomplete beta function and its inverse}
\usage{
beta_inc(x, a, b, lower_tail = TRUE, log = FALSE)

beta_inc_inv(u, a, b, lower_tail = TRUE, log = FALSE)
}
\arguments{
\item{x}{a vector of size \code{nx} or a matrix of size \code{c(nx, 1)}.}

\item{a, b}{scalars giving the parameters of the beta function.}

\item{lower_tail}{accumulate the probability from the lower tail? If
\code{FALSE}, the probability is accumulated from the \emph{upper} tail.
Defaults to \code{FALSE}.}

\item{log}{use log-scale? If \code{TRUE}, returns the logarithm of the
incomplete beta function and uses log-scale for \code{u} in \code{beta_inc}.
Defaults to \code{FALSE}.}

\item{u}{a vector of probabilities of size \code{nu} or a matrix of size
\code{c(nu, 1)}.}
}
\value{
\itemize{
  \item \code{beta_inc}: a matrix of size \code{c(nx, 1)} with the
  evaluation of the incomplete beta function at \code{x}.
  \item \code{beta_inc_inv}: a matrix of size \code{c(nu, 1)} with the
  evaluation of the inverse incomplete beta function at \code{u}.
}
}
\description{
Computes the incomplete beta function
\deqn{I_x(a,b):=\int_0^x u^{a-1}(1-u)^{b-1}\,d\mathrm{u},\quad a,b>0}{
I_x(a,b):=\int_0^x u^{a-1}(1-u)^{b-1}du, a,b>0}
and its inverse function.
}
\details{
The functions are mere wrappers to R's internal \code{pbeta} and
\code{qbeta} functions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wschisq.R
\name{wschisq}
\alias{wschisq}
\alias{d_wschisq}
\alias{p_wschisq}
\alias{q_wschisq}
\alias{r_wschisq}
\alias{cutoff_wschisq}
\title{Weighted sums of non-central chi squared random variables}
\usage{
d_wschisq(x, weights, dfs, ncps = 0, method = c("I", "SW", "HBE")[1],
  exact_chisq = TRUE, imhof_epsabs = 1e-06, imhof_epsrel = 1e-06,
  imhof_limit = 10000, grad_method = "simple",
  grad_method.args = list(eps = 1e-07))

p_wschisq(x, weights, dfs, ncps = 0, method = c("I", "SW", "HBE", "MC")[1],
  exact_chisq = TRUE, imhof_epsabs = 1e-06, imhof_epsrel = 1e-06,
  imhof_limit = 10000, M = 10000, MC_sample = NULL)

q_wschisq(u, weights, dfs, ncps = 0, method = c("I", "SW", "HBE", "MC")[1],
  exact_chisq = TRUE, imhof_epsabs = 1e-06, imhof_epsrel = 1e-06,
  imhof_limit = 10000, nlm_gradtol = 1e-06, nlm_iterlim = 1000,
  M = 10000, MC_sample = NULL)

r_wschisq(n, weights, dfs, ncps = 0)

cutoff_wschisq(thre = 1e-04, weights, dfs, ncps = 0, log = FALSE,
  x_tail = NULL)
}
\arguments{
\item{x}{vector of quantiles.}

\item{weights}{vector with the positive weights of the sum. Must have the
same length as \code{dfs}.}

\item{dfs}{vector with the positive degrees of freedom of the chi squared
random variables. Must have the same length as \code{weights}.}

\item{ncps}{non-centrality parameters. Either \code{0} (default) or a
vector with the same length as \code{weights}.}

\item{method}{method for approximating the density, distribution, or
quantile function. Must be \code{"I"} (Imhof), \code{"SW"}
(Satterthwaite--Welch), \code{"HBE"} (Hall--Buckley--Eagleson), or
\code{"MC"} (Monte Carlo; only for distribution or quantile functions).
Defaults to \code{"I"}.}

\item{exact_chisq}{if \code{weights} and \code{dfs} have length one, shall
the \code{\link[stats]{Chisquare}} functions be called? Otherwise, the
approximations are computed for this exact case. Defaults to \code{TRUE}.}

\item{imhof_epsabs, imhof_epsrel, imhof_limit}{precision parameters passed to
\code{\link[CompQuadForm]{imhof}}'s \code{epsabs}, \code{epsrel}, and
\code{limit}, respectively. They default to \code{1e-6}, \code{1e-6},
and \code{1e4}.}

\item{grad_method, grad_method.args}{numerical differentiation parameters
passed to \code{\link[numDeriv]{grad}}'s \code{method} and
\code{method.args}, respectively. They default to \code{"simple"},
and \code{list(eps = 1e-7)} (better precision than \code{imhof_epsabs} to
avoid numerical artifacts).}

\item{M}{number of Monte Carlo samples for approximating the distribution if
\code{method = "MC"}. Defaults to \code{1e4}.}

\item{MC_sample}{if provided, it is employed when \code{method = "MC"}. If
not, it is computed internally.}

\item{u}{vector of probabilities.}

\item{nlm_gradtol, nlm_iterlim}{convergence control parameters passed to
\code{\link[stats]{nlm}}'s \code{gradtol} and \code{iterlim}, respectively.
They default to \code{1e-6} and \code{1e3}.}

\item{n}{sample size.}

\item{thre}{vector with the error thresholds of the tail probability and
mean/variance explained by the first terms of the series. Defaults to
\code{1e-4}. See details.}

\item{log}{are \code{weights} and \code{dfs} given in log-scale? Defaults to
\code{FALSE}.}

\item{x_tail}{scalar evaluation point for determining the upper tail
probability. If \code{NULL}, set to the \code{0.90} quantile of the whole
series, computed by the \code{"HBE"} approximation.}
}
\value{
\itemize{
  \item \code{d_wschisq}: density function evaluated at \code{x}, a vector.
  \item \code{p_wschisq}: distribution function evaluated at \code{x},
  a vector.
  \item \code{q_wschisq}: quantile function evaluated at \code{u}, a vector.
  \item \code{r_wschisq}: a vector of size \code{n} containing a random
  sample.
  \item \code{cutoff_wschisq}: a data frame with the indexes up to which the
  truncated series explains the tail probability with absolute error
  \code{thre}, or the proportion of the mean/variance of the whole series
  that is \emph{not} explained by the truncated series.
}
}
\description{
Approximated density, distribution, and quantile functions for
weighted sums of non-central chi squared random variables:
\deqn{Q_K = \sum_{i = 1}^K w_i \chi^2_{d_i}(\lambda_i),}
where \eqn{w_1, \ldots, w_n} are positive weights, \eqn{d_1, \ldots, d_n}
are positive degrees of freedom, and \eqn{\lambda_1, \ldots, \lambda_n}
are non-negative non-centrality parameters. Also, simulation of \eqn{Q_K}.
}
\details{
Four methods are implemented for approximating the distribution of a
weighted sum of chi squared random variables:
\itemize{
  \item \code{"I"}: Imhof's approximation (Imhof, 1961) for the evaluation
  of the distribution function. If this method is selected, the function
  is simply a wrapper to \code{\link[CompQuadForm]{imhof}} from the
  \code{CompQuadForm} package (Duchesne and Lafaye De Micheaux, 2010).
  \item \code{"SW"}: Satterthwaite--Welch (Satterthwaite, 1946; Welch, 1938)
  approximation, consisting in matching the first \emph{two} moments of
  \eqn{Q_K} with a gamma distribution.
  \item \code{"HBE"}: Hall--Buckley--Eagleson (Hall, 1983; Buckley and
  Eagleson, 1988) approximation, consisting in matching the first
  \emph{three} moments of \eqn{Q_K} with a gamma distribution.
  \item \code{"MC"}: Monte Carlo approximation using the empirical
  cumulative distribution function with \code{M} simulated samples.
}
The Imhof method is exact up to the prescribed numerical accuracy. It is
also the most time-consuming method. The density and quantile functions
for this approximation are obtained by numerical differentiation and
inversion, respectively, of the approximated distribution.

For the methods based on gamma matching, the \code{\link[stats]{GammaDist}}
density, distribution, and quantile functions are invoked. The
Hall--Buckley--Eagleson approximation tends to overperform the
Satterthwaite--Welch approximation.

The Monte Carlo method is relatively inaccurate and slow, but serves as an
unbiased reference of the true distribution function. The inversion of the
empirical cumulative distribution is done by \code{\link[stats]{quantile}}.

An empirical comparison of these and other approximation methods is
given in Bodenham and Adams (2016).

\code{cutoff_wschisq} removes \code{NA}s/\code{NaN}s in \code{weights} or
\code{dfs} with a message. The threshold \code{thre} ensures that the
tail probability of the truncated and whole series differ less than
\code{thre} at \code{x_tail}, or that \code{thre} is the proportion of the
mean/variance of the whole series that is \emph{not} retained. The (upper)
tail probabilities for evaluating truncation are computed using the
Hall--Buckley--Eagleson approximation at \code{x_tail}.
}
\examples{
# Plotting functions for the examples
add_approx_dens <- function(x, dfs, weights, ncps) {

  lines(x, d_wschisq(x, weights = weights, dfs = dfs, ncps = ncps,
                     method = "SW", exact_chisq = FALSE), col = 3)
  lines(x, d_wschisq(x, weights = weights, dfs = dfs, ncps = ncps,
                     method = "HBE", exact_chisq = FALSE), col = 4)
  lines(x, d_wschisq(x, weights = weights, dfs = dfs, ncps = ncps,
                     method = "I", exact_chisq = TRUE), col = 2)
  legend("topright", legend = c("True", "SW", "HBE", "I"), lwd = 2,
         col = c(1, 3:4, 2))

}
add_approx_distr <- function(x, dfs, weights, ncps, ...) {

  lines(x, p_wschisq(x, weights = weights, dfs = dfs, ncps = ncps,
                     method = "SW", exact_chisq = FALSE), col = 3)
  lines(x, p_wschisq(x, weights = weights, dfs = dfs, ncps = ncps,
                     method = "HBE", exact_chisq = FALSE), col = 4)
  lines(x, p_wschisq(x, weights = weights, dfs = dfs, ncps = ncps,
                     method = "MC", exact_chisq = FALSE), col = 5,
                     type = "s")
  lines(x, p_wschisq(x, weights = weights, dfs = dfs, ncps = ncps,
                     method = "I", exact_chisq = TRUE), col = 2)
  legend("bottomright", legend = c("True", "SW", "HBE", "MC", "I"), lwd = 2,
         col = c(1, 3:5, 2))

}
add_approx_quant <- function(u, dfs, weights, ncps, ...) {

  lines(u, q_wschisq(u, weights = weights, dfs = dfs, ncps = ncps,
                     method = "SW", exact_chisq = FALSE), col = 3)
  lines(u, q_wschisq(u, weights = weights, dfs = dfs, ncps = ncps,
                     method = "HBE", exact_chisq = FALSE), col = 4)
  lines(u, q_wschisq(u, weights = weights, dfs = dfs, ncps = ncps,
                     method = "MC", exact_chisq = FALSE), col = 5,
                     type = "s")
  lines(u, q_wschisq(u, weights = weights, dfs = dfs, ncps = ncps,
                     method = "I", exact_chisq = TRUE), col = 2)
  legend("topleft", legend = c("True", "SW", "HBE", "MC", "I"), lwd = 2,
         col = c(1, 3:5, 2))

}

# Validation plots for density, distribution, and quantile functions
u <- seq(0.01, 0.99, l = 100)
old_par <- par(mfrow = c(1, 3))

# Case 1: 1 * ChiSq_3(0) + 1 * ChiSq_3(0) = ChiSq_6(0)
weights <- c(1, 1)
dfs <- c(3, 3)
ncps <- 0
x <- seq(-1, 30, l = 100)
main <- expression(1 * chi[3]^2 * (0) + 1 * chi[3]^2 * (0))
plot(x, dchisq(x, df = 6), type = "l", main = main, ylab = "Density")
add_approx_dens(x = x, weights = weights, dfs = dfs, ncps = ncps)
plot(x, pchisq(x, df = 6), type = "l", main = main, ylab = "Distribution")
add_approx_distr(x = x, weights = weights, dfs = dfs, ncps = ncps)
plot(u, qchisq(u, df = 6), type = "l", main = main, ylab = "Quantile")
add_approx_quant(u = u, weights = weights, dfs = dfs, ncps = ncps)
\donttest{
# Case 2: 2 * ChiSq_3(1) + 1 * ChiSq_6(0.5) + 0.5 * ChiSq_12(0.25)
weights <- c(2, 1, 0.5)
dfs <- c(3, 6, 12)
ncps <- c(1, 0.5, 0.25)
x <- seq(0, 70, l = 100)
main <- expression(2 * chi[3]^2 * (1)+ 1 * chi[6]^2 * (0.5) +
                   0.5 * chi[12]^2 * (0.25))
samp <- r_wschisq(n = 1e4, weights = weights, dfs = dfs, ncps = ncps)
hist(samp, breaks = 50, freq = FALSE, main = main, ylab = "Density",
     xlim = range(x), xlab = "x"); box()
add_approx_dens(x = x, weights = weights, dfs = dfs, ncps = ncps)
plot(x, ecdf(samp)(x), main = main, ylab = "Distribution", type = "s")
add_approx_distr(x = x, weights = weights, dfs = dfs, ncps = ncps)
plot(u, quantile(samp, probs = u), type = "s", main = main,
     ylab = "Quantile")
add_approx_quant(u = u, weights = weights, dfs = dfs, ncps = ncps)

# Case 3: \sum_{k = 1}^K k^(-3) * ChiSq_{5k}(1 / k^2)
K <- 1e2
weights<- 1 / (1:K)^3
dfs <- 5 * 1:K
ncps <- 1 / (1:K)^2
x <- seq(0, 25, l = 100)
main <- substitute(sum(k^(-3) * chi[5 * k]^2 * (1 / k^2), k == 1, K),
                   list(K = K))
samp <- r_wschisq(n = 1e4, weights = weights, dfs = dfs, ncps = ncps)
hist(samp, breaks = 50, freq = FALSE, main = main, ylab = "Density",
     xlim = range(x), xlab = "x"); box()
add_approx_dens(x = x, weights = weights, dfs = dfs, ncps = ncps)
plot(x, ecdf(samp)(x), main = main, ylab = "Distribution", type = "s")
add_approx_distr(x = x, weights = weights, dfs = dfs, ncps = ncps)
plot(u, quantile(samp, probs = u), type = "s", main = main,
     ylab = "Quantile")
add_approx_quant(u = u, weights = weights, dfs = dfs, ncps = ncps)
par(old_par)

# Cutoffs for infinite series of the last example
K <- 1e7
log_weights<- -3 * log(1:K)
log_dfs <- log(5) + log(1:K)
(cutoff <- cutoff_wschisq(thre = 10^(-(1:4)), weights = log_weights,
                          dfs = log_dfs, log = TRUE))

# Approximation
x <- seq(0, 25, l = 100)
l <- length(cutoff$mean)
main <- expression(sum(k^(-3) * chi[5 * k]^2, k == 1, K))
col <- viridisLite::viridis(l)
plot(x, d_wschisq(x, weights = exp(log_weights[1:cutoff$mean[l]]),
                  dfs = exp(log_dfs[1:cutoff$mean[l]])), type = "l",
     ylab = "Density", col = col[l], lwd = 3)
for(i in rev(seq_along(cutoff$mean)[-l])) {
  lines(x, d_wschisq(x, weights = exp(log_weights[1:cutoff$mean[i]]),
                     dfs = exp(log_dfs[1:cutoff$mean[i]])), col = col[i])
}
legend("topright", legend = paste0(rownames(cutoff), " (", cutoff$mean, ")"),
       lwd = 2, col = col)
}
}
\references{
Bodenham, D. A. and Adams, N. M. (2016). A comparison of efficient
approximations for a weighted sum of chi-squared random variables.
\emph{Statistics and Computing}, 26(4):917--928.
\doi{10.1007/s11222-015-9583-4}

Buckley, M. J. and Eagleson, G. K. (1988). An approximation to the
distribution of quadratic forms in normal random variables.
\emph{Australian Journal of Statistics}, 30(1):150--159.
\doi{10.1111/j.1467-842X.1988.tb00471.x}

Duchesne, P. and Lafaye De Micheaux, P. (2010) Computing the distribution
of quadratic forms: Further comparisons between the Liu--Tang--Zhang
approximation and exact methods. \emph{Computational Statistics and Data
Analysis}, 54(4):858--862. \doi{10.1016/j.csda.2009.11.025}

Hall, P. (1983). Chi squared approximations to the distribution of a sum
of independent random variables. \emph{Annals of Probability},
11(4):1028--1036. \doi{10.1214/aop/1176993451}

Imhof, J. P. (1961). Computing the distribution of quadratic forms in normal
variables. \emph{Biometrika}, 48(3/4):419--426.
\doi{10.2307/2332763}

Satterthwaite, F. E. (1946). An approximate distribution of estimates of
variance components. \emph{Biometrics Bulletin}, 2(6):110--114.
\doi{10.2307/3002019}

Welch, B. L. (1938). The significance of the difference between two means
when the population variances are unequal. \emph{Biometrika},
29(3/4):350--362. \doi{10.2307/2332010}
}
\author{
Eduardo García-Portugués and Paula Navarro-Esteban.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Sobolev.R
\name{Sobolev}
\alias{Sobolev}
\alias{d_p_k}
\alias{weights_dfs_Sobolev}
\alias{d_Sobolev}
\alias{p_Sobolev}
\alias{q_Sobolev}
\title{Asymptotic distributions of Sobolev statistics of spherical uniformity}
\usage{
d_p_k(p, k, log = FALSE)

weights_dfs_Sobolev(p, K_max = 1000, thre = 0.001, type, Rothman_t = 1/3,
  Pycke_q = 0.5, Riesz_s = 1, log = FALSE, verbose = TRUE,
  Gauss = TRUE, N = 320, tol = 1e-06, force_positive = TRUE,
  x_tail = NULL)

d_Sobolev(x, p, type, method = c("I", "SW", "HBE")[1], K_max = 1000,
  thre = 0.001, Rothman_t = 1/3, Pycke_q = 0.5, Riesz_s = 1,
  ncps = 0, verbose = TRUE, N = 320, x_tail = NULL, ...)

p_Sobolev(x, p, type, method = c("I", "SW", "HBE", "MC")[1], K_max = 1000,
  thre = 0.001, Rothman_t = 1/3, Pycke_q = 0.5, Riesz_s = 1,
  ncps = 0, verbose = TRUE, N = 320, x_tail = NULL, ...)

q_Sobolev(u, p, type, method = c("I", "SW", "HBE", "MC")[1], K_max = 1000,
  thre = 0.001, Rothman_t = 1/3, Pycke_q = 0.5, Riesz_s = 1,
  ncps = 0, verbose = TRUE, N = 320, x_tail = NULL, ...)
}
\arguments{
\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}

\item{k}{sequence of integer indexes.}

\item{log}{compute the logarithm of \eqn{d_{p,k}}? Defaults to
\code{FALSE}.}

\item{K_max}{integer giving the truncation of the series that compute the
asymptotic p-value of a Sobolev test. Defaults to \code{1e3}.}

\item{thre}{error threshold for the tail probability given by the
the first terms of the truncated series of a Sobolev test. Defaults to
\code{1e-3}.}

\item{type}{Sobolev statistic. For \eqn{p = 2}, either \code{"Watson"},
\code{"Rothman"}, \code{"Pycke_q"}, or \code{"Hermans_Rasson"}.
For \eqn{p \ge 2}, \code{"Ajne"}, \code{"Gine_Gn"}, \code{"Gine_Fn"},
\code{"Bakshaev"}, \code{"Riesz"}, \code{"PCvM"}, \code{"PAD"}, or
\code{"PRt"}.}

\item{Rothman_t}{\eqn{t} parameter for the Rothman test, a real in
\eqn{(0, 1)}. Defaults to \code{1 / 3}.}

\item{Pycke_q}{\eqn{q} parameter for the Pycke "\eqn{q}-test", a real in
\eqn{(0, 1)}. Defaults to \code{1 / 2}.}

\item{Riesz_s}{\eqn{s} parameter for the \eqn{s}-Riesz test, a real in
\eqn{(0, 2)}. Defaults to \code{1}.}

\item{verbose}{output information about the truncation? Defaults to
\code{TRUE}.}

\item{Gauss}{use a Gauss--Legendre quadrature rule of \code{N} nodes
in the computation of the Gegenbauer coefficients? Otherwise, call
\code{\link{integrate}}. Defaults to \code{TRUE}.}

\item{N}{number of points used in the \link[=Gauss_Legen_nodes]{
Gauss--Legendre quadrature} for computing the Gegenbauer coefficients.
Defaults to \code{320}.}

\item{tol}{tolerance passed to \code{\link{integrate}}'s \code{rel.tol} and
\code{abs.tol} if \code{Gauss = FALSE}. Defaults to \code{1e-6}.}

\item{force_positive}{set negative}

\item{x_tail}{scalar evaluation point for determining the upper tail
probability. If \code{NULL}, set to the \code{0.90} quantile of the whole
series, computed by the \code{"HBE"} approximation.}

\item{x}{vector of quantiles.}

\item{method}{method for approximating the density, distribution, or
quantile function. Must be \code{"I"} (Imhof), \code{"SW"}
(Satterthwaite--Welch), \code{"HBE"} (Hall--Buckley--Eagleson), or
\code{"MC"} (Monte Carlo; only for distribution or quantile functions).
Defaults to \code{"I"}.}

\item{ncps}{non-centrality parameters. Either \code{0} (default) or a
vector with the same length as \code{weights}.}

\item{...}{further parameters passed to \code{*_\link{wschisq}}.}

\item{u}{vector of probabilities.}
}
\value{
\itemize{
  \item \code{d_p_k}: a vector of size \code{length(k)} with the
  evaluation of \eqn{d_{p,k}}.
  \item \code{weights_dfs_Sobolev}: a list with entries \code{weights} and
  \code{dfs}, automatically truncated according to \code{K_max} and
  \code{thre} (see details).
  \item \code{d_Sobolev}: density function evaluated at \code{x}, a vector.
  \item \code{p_Sobolev}: distribution function evaluated at \code{x},
  a vector.
  \item \code{q_Sobolev}: quantile function evaluated at \code{u}, a vector.
}
}
\description{
Approximated density, distribution, and quantile functions for
the asymptotic null distributions of Sobolev statistics of uniformity
on \eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{S^{p-1}:=
\{x\in R^p:||x||=1\}}. These asymptotic distributions are infinite
weighted sums of (central) chi squared random variables:
\deqn{\sum_{k = 1}^\infty v_k^2 \chi^2_{d_{p, k}},}
where
\deqn{d_{p, k} := {{p + k - 3}\choose{p - 2}} + {{p + k - 2}\choose{p - 2}}}
is the dimension of the space of eigenfunctions of the Laplacian on
\eqn{S^{p-1}}, \eqn{p\ge 2}, associated to the \eqn{k}-th
eigenvalue, \eqn{k\ge 1}.
}
\details{
The truncation of \eqn{\sum_{k = 1}^\infty v_k^2 \chi^2_{d_{p, k}}} is
done to the first \code{K_max} terms and then up to the index such that
the first terms explain the tail probability at the \code{x_tail} with
an absolute error smaller than \code{thre} (see details in
\code{\link{cutoff_wschisq}}). This automatic truncation takes place when
calling \code{*_Sobolev}. Setting \code{thre = 0} truncates to \code{K_max}
terms exactly. If the series only contains odd or even non-zero terms, then
only \code{K_max / 2} addends are \emph{effectively} taken into account
in the first truncation.
}
\examples{
# Circular-specific statistics
curve(p_Sobolev(x = x, p = 2, type = "Watson", method = "HBE"),
      n = 2e2, ylab = "Distribution", main = "Watson")
curve(p_Sobolev(x = x, p = 2, type = "Rothman", method = "HBE"),
      n = 2e2, ylab = "Distribution", main = "Rothman")
curve(p_Sobolev(x = x, p = 2, type = "Pycke_q", method = "HBE"), to = 10,
      n = 2e2, ylab = "Distribution", main = "Pycke_q")
curve(p_Sobolev(x = x, p = 2, type = "Hermans_Rasson", method = "HBE"),
      to = 10, n = 2e2, ylab = "Distribution", main = "Hermans_Rasson")

# Statistics for arbitrary dimensions
test_statistic <- function(type, to = 1, pmax = 5, M = 1e3, ...) {

  col <- viridisLite::viridis(pmax - 1)
  curve(p_Sobolev(x = x, p = 2, type = type, method = "MC", M = M,
                  ...), to = to, n = 2e2, col = col[pmax - 1],
                  ylab = "Distribution", main = type, ylim = c(0, 1))
  for (p in 3:pmax) {
    curve(p_Sobolev(x = x, p = p, type = type, method = "MC", M = M,
                    ...), add = TRUE, n = 2e2, col = col[pmax - p + 1])
  }
  legend("bottomright", legend = paste("p =", 2:pmax), col = rev(col),
         lwd = 2)

}

# Ajne
test_statistic(type = "Ajne")
\donttest{
# Gine_Gn
test_statistic(type = "Gine_Gn", to = 1.5)

# Gine_Fn
test_statistic(type = "Gine_Fn", to = 2)

# Bakshaev
test_statistic(type = "Bakshaev", to = 3)

# Riesz
test_statistic(type = "Riesz", Riesz_s = 0.5, to = 3)

# PCvM
test_statistic(type = "PCvM", to = 0.6)

# PAD
test_statistic(type = "PAD", to = 3)

# PRt
test_statistic(type = "PRt", Rothman_t = 0.5)

# Quantiles
p <- c(2, 3, 4, 11)
t(sapply(p, function(p) q_Sobolev(u = c(0.10, 0.05, 0.01), p = p,
                                  type = "PCvM")))
t(sapply(p, function(p) q_Sobolev(u = c(0.10, 0.05, 0.01), p = p,
                                  type = "PAD")))
t(sapply(p, function(p) q_Sobolev(u = c(0.10, 0.05, 0.01), p = p,
                                  type = "PRt")))

# Series truncation for thre = 1e-5
sapply(p, function(p) length(weights_dfs_Sobolev(p = p, type = "PCvM")$dfs))
sapply(p, function(p) length(weights_dfs_Sobolev(p = p, type = "PRt")$dfs))
sapply(p, function(p) length(weights_dfs_Sobolev(p = p, type = "PAD")$dfs))
}
}
\author{
Eduardo García-Portugués and Paula Navarro-Esteban.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unif_stat_distr.R
\name{unif_stat_distr}
\alias{unif_stat_distr}
\title{Null distributions for circular and (hyper)spherical uniformity
statistics}
\usage{
unif_stat_distr(x, type, p, n, approx = "asymp", M = 10000,
  stats_MC = NULL, Rothman_t = 1/3, Pycke_q = 0.5, Riesz_s = 1,
  CCF09_dirs = NULL, CJ12_reg = 3, CJ12_beta = 0, Stephens = FALSE,
  K_Kuiper = 25, K_Watson = 25, K_Watson_1976 = 5, K_Ajne = 500,
  K_CCF09 = 25, K_max = 10000, ...)
}
\arguments{
\item{x}{evaluation points for the null distribution(s). Either a vector of
size \code{nx}, if the evaluation points are common for the tests in
\code{type}, or a matrix of size \code{c(nx, length(type))} with columns
containing the evaluation points for each test. Must not contain \code{NA}'s.}

\item{type}{type of test to be applied. A character vector containing any of
the following types of tests, depending on the dimension \eqn{p}:
\itemize{
  \item Circular data: any of the names available at object
  \code{\link{avail_cir_tests}}.
  \item (Hyper)spherical data: any of the names available at object
  \code{\link{avail_sph_tests}}.
}
If \code{type = "all"} (default), then \code{type} is set as
\code{avail_cir_tests} or \code{avail_sph_tests}, depending on the value of
\eqn{p}.}

\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}

\item{n}{sample size employed for computing the statistic.}

\item{approx}{type of approximation to the null distribution, either
\code{"asymp"} (default) for employing the asymptotic null distribution, if
available, or \code{"MC"}, for employing the Monte Carlo approximation of
the exact null distribution.}

\item{M}{number of Monte Carlo replications for approximating the null
distribution when \code{approx = "MC"}. Also, number of Monte Carlo samples
for approximating the asymptotic distributions based on weighted sums of chi
squared random variables. Defaults to \code{1e4}.}

\item{stats_MC}{a data frame of size \code{c(M, length(type))}, with column
names containing the character vector \code{type}, that results from
extracting \code{$stats_MC} from a call to \code{\link{unif_stat_MC}}. If
provided, the computation of Monte Carlo statistics when \code{approx = "MC"}
is skipped. \code{stats_MC} is checked internally to see if it is sorted.
Internally computed if \code{NULL} (default).}

\item{Rothman_t}{\eqn{t} parameter for the Rothman test, a real in
\eqn{(0, 1)}. Defaults to \code{1 / 3}.}

\item{Pycke_q}{\eqn{q} parameter for the Pycke "\eqn{q}-test", a real in
\eqn{(0, 1)}. Defaults to \code{1 / 2}.}

\item{Riesz_s}{\eqn{s} parameter for the \eqn{s}-Riesz test, a real in
\eqn{(0, 2)}. Defaults to \code{1}.}

\item{CCF09_dirs}{a matrix of size \code{c(n_proj, p)} containing
\code{n_proj} random directions (in Cartesian coordinates) on \eqn{S^{p-1}}
to perform the CCF09 test. If \code{NULL} (default), a sample of size
\code{n_proj = 50} directions is computed internally.}

\item{CJ12_reg}{type of asymptotic regime for CJ12 test, either \code{1}
(sub-exponential regime), \code{2} (exponential), or \code{3}
(super-exponential; default).}

\item{CJ12_beta}{\eqn{\beta} parameter in the exponential regime of CJ12 test,
a positive real.}

\item{Stephens}{compute Stephens (1970) modification so that the null
distribution of the is less dependent on the sample size? The modification
does not alter the test decision.}

\item{K_Kuiper, K_Watson, K_Watson_1976, K_Ajne}{integer giving the truncation
of the series present in the null asymptotic distributions. For the
Kolmogorov-Smirnov-related series defaults to \code{25}.}

\item{K_CCF09}{integer giving the truncation of the series present in the
asymptotic distribution of the Kolmogorov-Smirnov statistic. Defaults to
\code{5e2}.}

\item{K_max}{integer giving the truncation of the series that compute the
asymptotic p-value of a Sobolev test. Defaults to \code{1e4}.}

\item{...}{if \code{approx = "MC"}, optional performance parameters to be
passed to \cr\code{\link{unif_stat_MC}}: \code{chunks}, \code{cores},
and \code{seed}.}
}
\value{
A data frame of size \code{c(nx, length(type))}, with column names
given by \code{type}, that contains the values of the null distributions of
the statistics evaluated at \code{x}.
}
\description{
Approximate computation of the null distributions of several
statistics for assessing uniformity on the (hyper)sphere
\eqn{S^{p-1}:=\{{\bf x}\in R^p:||{\bf x}||=1\}}{S^{p-1}:=
\{x\in R^p:||x||=1\}}, \eqn{p\ge 2}. The approximation is done either by
means of the asymptotic distribution or by Monte Carlo.
}
\details{
When \code{approx = "asymp"}, statistics that do not have an implemented or
known asymptotic are omitted, and a warning is generated.

For Sobolev tests, \code{K_max = 1e4} produces probabilities uniformly
accurate with three digits for the \code{"PCvM"}, \code{"PAD"}, and
\code{"PRt"} tests, for dimensions \eqn{p \le 11}. With \code{K_max = 5e4},
these probabilities are uniformly accurate in the fourth digit. With
\code{K_max = 1e3}, only two-digit uniform accuracy is obtained. Uniform
accuracy deteriorates when \eqn{p} increases, e.g., a digit accuracy is lost
when \eqn{p = 51}.

Descriptions and references on most of the asymptotic distributions
are available in García-Portugués and Verdebout (2018).
}
\examples{
## Asymptotic distribution

# Circular statistics
x <- seq(0, 1, l = 5)
unif_stat_distr(x = x, type = "Kuiper", p = 2, n = 10)
unif_stat_distr(x = x, type = c("Ajne", "Kuiper"), p = 2, n = 10)
unif_stat_distr(x = x, type = c("Ajne", "Kuiper"), p = 2, n = 10, K_Ajne = 5)
\donttest{
# All circular statistics
unif_stat_distr(x = x, type = avail_cir_tests, p = 2, n = 10, K_max = 1e3)
}
# Spherical statistics
unif_stat_distr(x = cbind(x, x + 1), type = c("Rayleigh", "Bingham"),
                p = 3, n = 10)
unif_stat_distr(x = cbind(x, x + 1), type = c("Rayleigh", "Bingham"),
                p = 3, n = 10, M = 100)
\donttest{
# All spherical statistics
unif_stat_distr(x = x, type = avail_sph_tests, p = 3, n = 10, K_max = 1e3)

## Monte Carlo distribution

# Circular statistics
x <- seq(0, 5, l = 10)
unif_stat_distr(x = x, type = avail_cir_tests, p = 2, n = 10, approx = "MC")
unif_stat_distr(x = x, type = "Kuiper", p = 2, n = 10, approx = "MC")
unif_stat_distr(x = x, type = c("Ajne", "Kuiper"), p = 2, n = 10,
                approx = "MC")

# Spherical statistics
unif_stat_distr(x = x, type = avail_sph_tests, p = 3, n = 10,
                approx = "MC")
unif_stat_distr(x = cbind(x, x + 1), type = c("Rayleigh", "Bingham"),
                p = 3, n = 10, approx = "MC")
unif_stat_distr(x = cbind(x, x + 1), type = c("Rayleigh", "Bingham"),
                p = 3, n = 10, approx = "MC")

## Specific arguments

# Rothman
unif_stat_distr(x = x, type = "Rothman", p = 2, n = 10, Rothman_t = 0.5,
                approx = "MC")

# CCF09
dirs <- r_unif_sph(n = 5, p = 3, M = 1)[, , 1]
x <- seq(0, 1, l = 10)
unif_stat_distr(x = x, type = "CCF09", p = 3, n = 10, approx = "MC",
                CCF09_dirs = dirs)
unif_stat_distr(x = x, type = "CCF09", p = 3, n = 10, approx = "MC")

# CJ12
unif_stat_distr(x = x, type = "CJ12", p = 3, n = 100, CJ12_reg = 3)
unif_stat_distr(x = x, type = "CJ12", p = 3, n = 100, CJ12_reg = 2,
               CJ12_beta = 0.01)
unif_stat_distr(x = x, type = "CJ12", p = 3, n = 100, CJ12_reg = 1)
}
}
\references{
García-Portugués, E. and Verdebout, T. (2018) An overview of uniformity
tests on the hypersphere. \emph{arXiv:1804.00286}.
\url{https://arxiv.org/abs/1804.00286}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/cir_stat.R
\name{cir_stat_Kuiper}
\alias{cir_stat_Kuiper}
\alias{cir_stat_Watson}
\alias{cir_stat_Watson_1976}
\alias{cir_stat_Range}
\alias{cir_stat_Rao}
\alias{cir_stat_Greenwood}
\alias{cir_stat_Log_gaps}
\alias{cir_stat_Vacancy}
\alias{cir_stat_Max_uncover}
\alias{cir_stat_Num_uncover}
\alias{cir_stat_Gini}
\alias{cir_stat_Gini_squared}
\alias{cir_stat_Ajne}
\alias{cir_stat_Rothman}
\alias{cir_stat_Hodges_Ajne}
\alias{cir_stat_Cressie}
\alias{cir_stat_FG01}
\alias{cir_stat_Rayleigh}
\alias{cir_stat_Bingham}
\alias{cir_stat_Hermans_Rasson}
\alias{cir_stat_Gine_Gn}
\alias{cir_stat_Gine_Fn}
\alias{cir_stat_Pycke}
\alias{cir_stat_Pycke_q}
\alias{cir_stat_Bakshaev}
\alias{cir_stat_Riesz}
\alias{cir_stat_PCvM}
\alias{cir_stat_PRt}
\alias{cir_stat_PAD}
\alias{cir_stat_CCF09}
\alias{cir_stat}
\title{Statistics for testing circular uniformity}
\usage{
cir_stat_Kuiper(Theta, sorted = FALSE, KS = FALSE, Stephens = FALSE)

cir_stat_Watson(Theta, sorted = FALSE, CvM = FALSE, Stephens = FALSE)

cir_stat_Watson_1976(Theta, sorted = FALSE, minus = FALSE)

cir_stat_Range(Theta, sorted = FALSE, gaps_in_Theta = FALSE, max_gap = TRUE)

cir_stat_Rao(Theta, sorted = FALSE, gaps_in_Theta = FALSE)

cir_stat_Greenwood(Theta, sorted = FALSE, gaps_in_Theta = FALSE)

cir_stat_Log_gaps(Theta, sorted = FALSE, gaps_in_Theta = FALSE, abs_val = TRUE)

cir_stat_Vacancy(Theta, a = 2 * pi, sorted = FALSE, gaps_in_Theta = FALSE)

cir_stat_Max_uncover(Theta, a = 2 * pi, sorted = FALSE, gaps_in_Theta = FALSE)

cir_stat_Num_uncover(Theta, a = 2 * pi, sorted = FALSE,
  gaps_in_Theta = FALSE, minus_val = TRUE)

cir_stat_Gini(Theta, sorted = FALSE, gaps_in_Theta = FALSE)

cir_stat_Gini_squared(Theta, sorted = FALSE, gaps_in_Theta = FALSE)

cir_stat_Ajne(Theta, Psi_in_Theta = FALSE)

cir_stat_Rothman(Theta, t = 1/3, Psi_in_Theta = FALSE)

cir_stat_Hodges_Ajne(Theta, asymp_std = FALSE, sorted = FALSE,
  use_Cressie = TRUE)

cir_stat_Cressie(Theta, t = 1/3, sorted = FALSE)

cir_stat_FG01(Theta, sorted = FALSE)

cir_stat_Rayleigh(Theta, m = 1L)

cir_stat_Bingham(Theta)

cir_stat_Hermans_Rasson(Theta, Psi_in_Theta = FALSE)

cir_stat_Gine_Gn(Theta, Psi_in_Theta = FALSE)

cir_stat_Gine_Fn(Theta, Psi_in_Theta = FALSE)

cir_stat_Pycke(Theta, Psi_in_Theta = FALSE)

cir_stat_Pycke_q(Theta, Psi_in_Theta = FALSE, q = 0.5)

cir_stat_Bakshaev(Theta, Psi_in_Theta = FALSE)

cir_stat_Riesz(Theta, Psi_in_Theta = FALSE, s = 1)

cir_stat_PCvM(Theta, Psi_in_Theta = FALSE)

cir_stat_PRt(Theta, t = 1/3, Psi_in_Theta = FALSE)

cir_stat_PAD(Theta, Psi_in_Theta = FALSE, AD = FALSE, sorted = FALSE)

cir_stat_CCF09(Theta, dirs, K_CCF09 = 25L, original = FALSE)
}
\arguments{
\item{Theta}{a \bold{matrix} of size \code{c(n, M)} with \code{M} samples
of size \code{n} of circular data on \eqn{[0, 2\pi)}. Must not contain
\code{NA}'s.}

\item{sorted}{are the columns of \code{Theta} sorted increasingly? If
\code{TRUE}, performance is improved. If \code{FALSE} (default), each
column of \code{Theta} is sorted internally.}

\item{KS}{compute the Kolmogorov-Smirnov statistic (which is
\emph{not} invariant under origin shifts) instead of the Kuiper statistic?
Defaults to \code{FALSE}.}

\item{Stephens}{compute Stephens (1970) modification so that the null
distribution of the is less dependent on the sample size? The modification
does not alter the test decision.}

\item{CvM}{compute the Cramér-von Mises statistic (which is \emph{not}
invariant under origin shifts) instead of the Watson statistic? Defaults to
\code{FALSE}.}

\item{minus}{compute the invariant \eqn{D_n^-} instead of \eqn{D_n^+}?
Defaults to \code{FALSE}.}

\item{gaps_in_Theta}{does \code{Theta} contain the matrix of
\emph{circular gaps} that is obtained with \cr\code{\link{cir_gaps}(Theta)}?
If \code{FALSE} (default), the circular gaps are computed internally.}

\item{max_gap}{compute the maximum gap for the range statistic? If
\code{TRUE} (default), rejection happens for \emph{large} values of
the statistic, which is consistent with the rest of tests. Otherwise,
the minimum gap is computed and rejection happens for \emph{low} values.}

\item{abs_val}{return the absolute value of the Darling's log gaps
statistic? If \code{TRUE} (default), rejection happens for \emph{large}
values of the statistic, which is consistent with the rest of tests.
Otherwise, the signed statistic is computed and rejection happens for
large \emph{absolute} values.}

\item{a}{\eqn{a_n = a / n} parameter used in the length of the arcs
of the coverage-based tests. Must be positive. Defaults to \code{2 * pi}.}

\item{minus_val}{return the negative value of the (standardized) number of
uncovered spacings? If \code{TRUE} (default), rejection happens for
\emph{large} values of the statistic, which is consistent with the rest of
tests. Otherwise, rejection happens for \emph{low} values.}

\item{Psi_in_Theta}{does \code{Theta} contain the shortest angles matrix
\eqn{\boldsymbol\Psi}{\Psi} that is obtained with
\cr\code{\link{Psi_mat}(array(Theta, dim = c(n, 1, M)))}? If \code{FALSE}
(default), \eqn{\boldsymbol\Psi}{\Psi} is computed internally.}

\item{t}{\eqn{t} parameter for the Rothman and Cressie tests, a real in
\eqn{(0, 1)}. Defaults to \code{1 / 3}.}

\item{asymp_std}{normalize the Hodges-Ajne statistic in terms of its
asymptotic distribution? Defaults to \code{FALSE}.}

\item{use_Cressie}{compute the Hodges-Ajne statistic as a particular case
of the Cressie statistic? Defaults to \code{TRUE} as it is more efficient.
If \code{FALSE}, the geometric construction in Ajne (1968) is employed.}

\item{m}{integer \eqn{m} for the \eqn{m}-modal Rayleigh test. Defaults to
\code{m = 1} (the standard Rayleigh test).}

\item{q}{\eqn{q} parameter for the Pycke "\eqn{q}-test", a real in
\eqn{(0, 1)}. Defaults to \code{1 / 2}.}

\item{s}{\eqn{s} parameter for the \eqn{s}-Riesz test, a real in
\eqn{(0, 2)}. Defaults to \code{1}.}

\item{AD}{compute the Anderson-Darling statistic (which is \emph{not}
invariant under origin shifts) instead of the Projected Anderson-Darling
statistic? Defaults to \code{FALSE}.}

\item{dirs}{a matrix of size \code{c(n_proj, 2)} containing \code{n_proj}
random directions (in Cartesian coordinates) on \eqn{S^1} to perform the
CCF09 test.}

\item{K_CCF09}{integer giving the truncation of the series present in the
asymptotic distribution of the Kolmogorov-Smirnov statistic. Defaults to
\code{25}.}

\item{original}{return the CCF09 statistic as originally defined?
If \code{FALSE} (default), a faster and equivalent statistic is computed,
and rejection happens for \emph{large} values of the statistic, which is
consistent with the rest of tests. Otherwise, rejection happens for
\emph{low} values.}
}
\value{
A matrix of size \code{c(M, 1)} containing the statistics for each
of the \code{M} samples.
}
\description{
Low-level implementation of several statistics for assessing
circular uniformity on \eqn{[0, 2\pi)} or, equivalently,
\eqn{S^1:=\{{\bf x}\in R^2:||{\bf x}||=1\}}{S^1:=\{x\in R^2:||x||=1\}}.
}
\details{
Descriptions and references for most of the statistics are available
in García-Portugués and Verdebout (2018).

The statistics \code{cir_stat_PCvM} and \code{cir_stat_PRt} are provided
for the sake of completion, but they equal the more efficiently-implemented
statistics \code{2 * cir_stat_Watson} and \code{cir_stat_Rothman},
respectively.
}
\section{Warning}{

Be careful on avoiding the next bad usages of the functions, which will
produce spurious results:
\itemize{
  \item The entries of \code{Theta} are \emph{not} in \eqn{[0, 2\pi)}.
  \item \code{Theta} does \emph{not} contain the circular gaps when
  \code{gaps_in_Theta = TRUE}.
  \item \code{Theta} is \emph{not} sorted increasingly when
  \code{data_sorted = TRUE}.
  \item \code{Theta} does \emph{not} contain
  \code{\link{Psi_mat}(array(Theta, dim = c(n, 1, M)))} when
  \cr\code{Psi_in_Theta = TRUE}.
  \item The directions in \code{dirs} do \emph{not} have unit norm.
}
}

\examples{
## Sample uniform circular data

M <- 2
n <- 100
set.seed(987202226)
Theta <- r_unif_cir(n = n, M = M)

## Tests based on the empirical cumulative distribution function

# Kuiper
cir_stat_Kuiper(Theta)
cir_stat_Kuiper(Theta, Stephens = TRUE)

# Watson
cir_stat_Watson(Theta)
cir_stat_Watson(Theta, Stephens = TRUE)

# Watson (1976)
cir_stat_Watson_1976(Theta)

## Partition-based tests

# Ajne
Theta_array <- Theta
dim(Theta_array) <- c(nrow(Theta), 1, ncol(Theta))
Psi <- Psi_mat(Theta_array)
cir_stat_Ajne(Theta)
cir_stat_Ajne(Psi, Psi_in_Theta = TRUE)

# Rothman
cir_stat_Rothman(Theta, t = 0.5)
cir_stat_Rothman(Theta)
cir_stat_Rothman(Psi, Psi_in_Theta = TRUE)

# Hodges-Ajne
cir_stat_Hodges_Ajne(Theta)
cir_stat_Hodges_Ajne(Theta, use_Cressie = FALSE)

# Cressie
cir_stat_Cressie(Theta, t = 0.5)
cir_stat_Cressie(Theta)

# FG01
cir_stat_FG01(Theta)

## Spacings-based tests

# Range
cir_stat_Range(Theta)

# Rao
cir_stat_Rao(Theta)

# Greenwood
cir_stat_Greenwood(Theta)

# Log gaps
cir_stat_Log_gaps(Theta)

# Vacancy
cir_stat_Vacancy(Theta)

# Maximum uncovered spacing
cir_stat_Max_uncover(Theta)

# Number of uncovered spacings
cir_stat_Num_uncover(Theta)

# Gini mean difference
cir_stat_Gini(Theta)

# Gini mean squared difference
cir_stat_Gini_squared(Theta)

## Sobolev tests

# Rayleigh
cir_stat_Rayleigh(Theta)
cir_stat_Rayleigh(Theta, m = 2)

# Bingham
cir_stat_Bingham(Theta)

# Hermans-Rasson
cir_stat_Hermans_Rasson(Theta)
cir_stat_Hermans_Rasson(Psi, Psi_in_Theta = TRUE)

# Gine Fn
cir_stat_Gine_Fn(Theta)
cir_stat_Gine_Fn(Psi, Psi_in_Theta = TRUE)

# Gine Gn
cir_stat_Gine_Gn(Theta)
cir_stat_Gine_Gn(Psi, Psi_in_Theta = TRUE)

# Pycke
cir_stat_Pycke(Theta)
cir_stat_Pycke(Psi, Psi_in_Theta = TRUE)

# Pycke q
cir_stat_Pycke_q(Theta)
cir_stat_Pycke_q(Psi, Psi_in_Theta = TRUE)

# Bakshaev
cir_stat_Bakshaev(Theta)
cir_stat_Bakshaev(Psi, Psi_in_Theta = TRUE)

# Riesz
cir_stat_Riesz(Theta, s = 1)
cir_stat_Riesz(Psi, Psi_in_Theta = TRUE, s = 1)
# Projected Cramér-von Mises
cir_stat_PCvM(Theta)
cir_stat_PCvM(Psi, Psi_in_Theta = TRUE)

# Projected Rothman
cir_stat_PRt(Theta, t = 0.5)
cir_stat_PRt(Theta)
cir_stat_PRt(Psi, Psi_in_Theta = TRUE)

# Projected Anderson-Darling
cir_stat_PAD(Theta)
cir_stat_PAD(Psi, Psi_in_Theta = TRUE)

## Other tests

# CCF09
dirs <- r_unif_sph(n = 3, p = 2, M = 1)[, , 1]
cir_stat_CCF09(Theta, dirs = dirs)

## Connection of Kuiper and Watson statistics with KS and CvM, respectively

# Rotate sample for KS and CvM
alpha <- seq(0, 2 * pi, l = 1e4)
KS_alpha <- sapply(alpha, function(a) {
  cir_stat_Kuiper((Theta[, 2, drop = FALSE] + a) \%\% (2 * pi), KS = TRUE)
})
CvM_alpha <- sapply(alpha, function(a) {
  cir_stat_Watson((Theta[, 2, drop = FALSE] + a) \%\% (2 * pi), CvM = TRUE)
})
AD_alpha <- sapply(alpha, function(a) {
  cir_stat_PAD((Theta[, 2, drop = FALSE] + a) \%\% (2 * pi), AD = TRUE)
})

# Kuiper is the maximum rotated KS
plot(alpha, KS_alpha, type = "l")
abline(h = cir_stat_Kuiper(Theta[, 2, drop = FALSE]), col = 2)
points(alpha[which.max(KS_alpha)], max(KS_alpha), col = 2, pch = 16)

# Watson is the minimum rotated CvM
plot(alpha, CvM_alpha, type = "l")
abline(h = cir_stat_Watson(Theta[, 2, drop = FALSE]), col = 2)
points(alpha[which.min(CvM_alpha)], min(CvM_alpha), col = 2, pch = 16)

# Anderson-Darling is the average rotated AD?
plot(alpha, AD_alpha, type = "l")
abline(h = cir_stat_PAD(Theta[, 2, drop = FALSE]), col = 2)
abline(h = mean(AD_alpha), col = 3)
}
\references{
García-Portugués, E. and Verdebout, T. (2018) An overview of uniformity
tests on the hypersphere. \emph{arXiv:1804.00286}.
\url{https://arxiv.org/abs/1804.00286}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/harmonics.R
\name{angles_to_sphere}
\alias{angles_to_sphere}
\alias{sphere_to_angles}
\title{Conversion between angular and Cartesian coordinates of the
(hyper)sphere}
\usage{
angles_to_sphere(theta)

sphere_to_angles(x)
}
\arguments{
\item{theta}{matrix of size \code{c(n, p - 1)} with the angles.}

\item{x}{matrix of size \code{c(n, p)} with the Cartesian coordinates.
Assumed to be of unit norm by rows.}
}
\value{
For \code{angles_to_sphere}, the matrix \code{x}. For
\code{sphere_to_angles}, the matrix \code{theta}.
}
\description{
Transforms the angles \eqn{(\theta_1,\ldots,\theta_{p-1})'} in
\eqn{[0,\pi)^{p-2}\times[-\pi,\pi)} into the Cartesian coordinates
\deqn{(\cos(x_1),\sin(x_1)\cos(x_2),\ldots,
\sin(x_1)\cdots\sin(x_{p-2})\cos(x_{p-1}),
\sin(x_1)\cdots\sin(x_{p-2})\sin(x_{p-1}))'}
of \eqn{S^{p-1}}, and vice versa.
}
\examples{
# Check changes of coordinates
sphere_to_angles(angles_to_sphere(c(pi / 2, 0, pi)))
sphere_to_angles(angles_to_sphere(rbind(c(pi / 2, 0, pi), c(pi, pi / 2, 0))))
angles_to_sphere(sphere_to_angles(c(0, sqrt(0.5), sqrt(0.1), sqrt(0.4))))
angles_to_sphere(sphere_to_angles(
  rbind(c(0, sqrt(0.5), sqrt(0.1), sqrt(0.4)),
        c(0, sqrt(0.5), sqrt(0.5), 0),
        c(0, 1, 0, 0),
        c(0, 0, 0, -1),
        c(0, 0, 1, 0))))

# Circle
sphere_to_angles(angles_to_sphere(0))
sphere_to_angles(angles_to_sphere(cbind(0:3)))
angles_to_sphere(cbind(sphere_to_angles(rbind(c(0, 1), c(1, 0)))))
angles_to_sphere(cbind(sphere_to_angles(rbind(c(0, 1)))))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unif-alts.R
\name{r_alt}
\alias{r_alt}
\title{Sample non-uniformly distributed spherical data}
\usage{
r_alt(n, p, M = 1, alt = "vMF", kappa = 1, nu = 0.5, F_inv = NULL,
  K = 1000, axial_MvMF = TRUE)
}
\arguments{
\item{n}{sample size.}

\item{p}{integer giving the dimension of the ambient space \eqn{R^p} that
contains \eqn{S^{p-1}}.}

\item{M}{number of samples of size \code{n}. Defaults to \code{1}.}

\item{alt}{alternative, must be \code{"vMF"}, \code{"MvMF"},
\code{"ACG"}, \code{"SC"}, \code{"W"}, or \code{"C"}. See details below.}

\item{kappa}{non-negative parameter measuring the strength of the deviation
with respect to uniformity (obtained with \eqn{\kappa = 0}).}

\item{nu}{projection along \eqn{{\bf e}_p}{e_p} controlling the modal
strip of the small circle distribution. Must be in (-1, 1). Defaults to
\code{0.5}.}

\item{F_inv}{quantile function returned by \code{\link{F_inv_from_f}}. Used
for \code{"SC"}, \code{"W"}, and \code{"C"}. Computed by internally if
\code{NULL} (default).}

\item{K}{number of equispaced points on \eqn{[-1, 1]} used for evaluating
\eqn{F^{-1}} and then interpolating. Defaults to \code{1e3}.}

\item{axial_MvMF}{use a mixture of vMF that is axial (i.e., symmetrically
distributed about the origin)? Defaults to \code{TRUE}.}
}
\value{
An \bold{array} of size \code{c(n, p, M)} with \code{M} random
samples of size \code{n} of non-uniformly-generated directions on
\eqn{S^{p-1}}.
}
\description{
Simple simulation of prespecified non-uniform spherical
distributions: von Mises--Fisher (vMF), Mixture of vMF (MvMF),
Angular Central Gaussian (ACG), Small Circle (SC), Watson (W), or
Cauchy-like (C).
}
\details{
The parameter \code{kappa} is used as \eqn{\kappa} in the following
distributions:
\itemize{
  \item \code{"vMF"}: von Mises--Fisher distribution with concentration
  \eqn{\kappa} and directional mean \eqn{{\bf e}_p = (0, 0, \ldots, 1)}{
  e_p = (0, 0, \ldots, 1)}.
  \item \code{"MvMF"}: equally-weighted mixture of \eqn{p} von Mises--Fisher
  distributions with common concentration \eqn{\kappa} and directional means
  \eqn{\pm{\bf e}_1, \ldots, \pm{\bf e}_p}{±e_1, \ldots, ±e_p} if
  \code{axial_MvMF = TRUE}. If \code{axial_MvMF = FALSE}, then only means
  with positive signs are considered.
  \item \code{"ACG"}: Angular Central Gaussian distribution with diagonal
  shape matrix with diagonal given by
  \deqn{(1, \ldots, 1, 1 + \kappa) / (p + \kappa).}
  \item \code{"SC"}: Small Circle distribution with axis mean
  \eqn{{\bf e}_p = (0, 0, \ldots, 1)}{e_p = (0, 0, \ldots, 1)} and
  concentration \eqn{\kappa} about the projection along the mean, \eqn{\nu}.
  \item \code{"W"}: Watson distribution with axis mean
  \eqn{{\bf e}_p = (0, 0, \ldots, 1)}{e_p = (0, 0, \ldots, 1)} and
  concentration \eqn{\kappa}. The Watson distribution is a particular case
  of the Bingham distribution.
  \item \code{"C"}: Cauchy-like distribution with directional mode
  \eqn{{\bf e}_p = (0, 0, \ldots, 1)}{e_p = (0, 0, \ldots, 1)} and
  concentration \eqn{\kappa = \rho / (1 - \rho^2)}. The circular Wrapped
  Cauchy distribution is a particular case of this Cauchy-like distribution.
}

Much faster sampling for \code{"SC"}, \code{"W"}, and \code{"C"} is achieved
providing \code{F_inv}, see examples.
}
\examples{
## Simulation with p = 2

p <- 2
n <- 200
kappa <- 20
nu <- 0.5
rho <- ((2 * kappa + 1) - sqrt(4 * kappa + 1)) / (2 * kappa)
F_inv_SC_2 <- F_inv_from_f(f = function(z) exp(-kappa * (z - nu)^2), p = 2)
F_inv_W_2 <- F_inv_from_f(f = function(z) exp(kappa * z^2), p = 2)
F_inv_C_2 <- F_inv_from_f(f = function(z) (1 - rho^2) /
                            (1 + rho^2 - 2 * rho * z)^(p / 2), p = 2)
x1 <- r_alt(n = n, p = p, alt = "vMF", kappa = kappa)[, , 1]
x2 <- r_alt(n = n, p = p, alt = "MvMF", kappa = kappa)[, , 1]
x3 <- r_alt(n = n, p = p, alt = "ACG", kappa = kappa)[, , 1]
x4 <- r_alt(n = n, p = p, alt = "SC", F_inv = F_inv_SC_2)[, , 1]
x5 <- r_alt(n = n, p = p, alt = "W", F_inv = F_inv_W_2)[, , 1]
x6 <- r_alt(n = n, p = p, alt = "C", F_inv = F_inv_C_2)[, , 1]
r <- runif(n, 0.95, 1.05) # Radius perturbation to improve visualization
plot(r * x1, pch = 16, xlim = c(-1.1, 1.1), ylim = c(-1.1, 1.1), col = 1)
points(r * x2, pch = 16, col = 2)
points(r * x3, pch = 16, col = 3)
points(r * x4, pch = 16, col = 4)
points(r * x5, pch = 16, col = 5)
points(r * x6, pch = 16, col = 6)

## Simulation with p = 3

n <- 200
p <- 3
kappa <- 20
nu <- 0.5
rho <- ((2 * kappa + 1) - sqrt(4 * kappa + 1)) / (2 * kappa)
F_inv_SC_3 <- F_inv_from_f(f = function(z) exp(-kappa * (z - nu)^2), p = 3)
F_inv_W_3 <- F_inv_from_f(f = function(z) exp(kappa * z^2), p = 3)
F_inv_C_3 <- F_inv_from_f(f = function(z) (1 - rho^2) /
                            (1 + rho^2 - 2 * rho * z)^(p / 2), p = 3)
x1 <- r_alt(n = n, p = p, alt = "vMF", kappa = kappa)[, , 1]
x2 <- r_alt(n = n, p = p, alt = "MvMF", kappa = kappa)[, , 1]
x3 <- r_alt(n = n, p = p, alt = "ACG", kappa = kappa)[, , 1]
x4 <- r_alt(n = n, p = p, alt = "SC", F_inv = F_inv_SC_3)[, , 1]
x5 <- r_alt(n = n, p = p, alt = "W", F_inv = F_inv_W_3)[, , 1]
x6 <- r_alt(n = n, p = p, alt = "C", F_inv = F_inv_C_3)[, , 1]
s3d <- scatterplot3d::scatterplot3d(x1, pch = 16, xlim = c(-1.1, 1.1),
                                    ylim = c(-1.1, 1.1), zlim = c(-1.1, 1.1))
s3d$points3d(x2, pch = 16, col = 2)
s3d$points3d(x3, pch = 16, col = 3)
s3d$points3d(x4, pch = 16, col = 4)
s3d$points3d(x5, pch = 16, col = 5)
s3d$points3d(x6, pch = 16, col = 6)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{Gauss_Legen}
\alias{Gauss_Legen}
\alias{Gauss_Legen_nodes}
\alias{Gauss_Legen_weights}
\title{Gauss--Legendre quadrature}
\usage{
Gauss_Legen_nodes(a = -1, b = 1, N = 40L)

Gauss_Legen_weights(a = -1, b = 1, N = 40L)
}
\arguments{
\item{a, b}{scalars giving the interval \eqn{(a, b)}. Defaults to
\eqn{(-1, 1)}.}

\item{N}{number of points used in the Gauss--Legendre quadrature. The
following choices are supported: \code{5}, \code{10}, \code{20},
\code{40}, \code{80}, \code{160}, \code{320}, \code{640}, \code{1280},
\code{2560}, and \code{5120}. Defaults to \code{40}.}
}
\value{
A matrix of size \code{c(N, 1)} with the nodes \eqn{x_k}
(\code{Gauss_Legen_nodes}) or the corresponding weights \eqn{w_k}
(\code{Gauss_Legen_weights}).
}
\description{
Convenience for computing the nodes \eqn{x_k} and weights
\eqn{w_k} of the \emph{Gauss--Legendre} quadrature formula
in \eqn{(a, b)}:
\deqn{\int_a^b f(x) w(x)\,\mathrm{d}x\approx\sum_{k=1}^N w_k f(x_k).}{
\int_a^b f(x) dx\approx\sum_{k=1}^N w_k f(x_k)}.
}
\details{
For \eqn{C^\infty} functions, Gauss--Legendre quadrature
can be very efficient. It is exact for polynomials up to degree
\eqn{2N - 1}.

The nodes and weights up to \eqn{N = 80} were retrieved from
\href{https://dlmf.nist.gov/3.5#v}{NIST} and have \eqn{10^{-21}} precision.
For \eqn{N = 160} onwards, the nodes and weights were computed with the
\code{gauss.quad} function from the \href{https://CRAN.R-project.org/package=statmod}{
\code{statmod}} package (Smyth, 1998), and have \eqn{10^{-15}} precision.
}
\examples{
## Integration of a smooth function in (1, 10)

# Weights and nodes for integrating
x_k <- Gauss_Legen_nodes(a = 1, b = 10, N = 40)
w_k <- Gauss_Legen_weights(a = 1, b = 10, N = 40)

# Check quadrature
f <- function(x) sin(x) * x^2 - log(x + 1)
integrate(f, lower = 1, upper = 10, rel.tol = 1e-12)
sum(w_k * f(x_k))

# Exact for polynomials up to degree 2 * N - 1
f <- function(x) (((x + 0.5) / 1e3)^5 - ((x - 0.5)/ 5)^4 +
  ((x - 0.25) / 10)^2 + 1)^20
sum(w_k * f(x_k))
integrate(f, lower = -1, upper = 1, rel.tol = 1e-12)

## Integration on (0, pi)

# Weights and nodes for integrating
th_k <- Gauss_Legen_nodes(a = 0, b = pi, N = 40)
w_k <- Gauss_Legen_weights(a = 0, b = pi, N = 40)

# Check quadrature
p <- 4
psi <- function(th) -sin(th / 2)
w <- function(th) sin(th)^(p - 2)
integrate(function(th) psi(th) * w(th), lower = 0, upper = pi,
          rel.tol = 1e-12)
sum(w_k * psi(th_k) * w(th_k))

# Integral with Gegenbauer polynomial
k <- 3
C_k <- function(th) drop(Gegen_polyn(theta = th, k = k, p = p))
integrate(function(th) psi(th) * C_k(th) * w(th), lower = 0, upper = pi,
          rel.tol = 1e-12)
th_k <- drop(Gauss_Legen_nodes(a = 0, b = pi, N = 80))
w_k <- drop(Gauss_Legen_weights(a = 0, b = pi, N = 80))
sum(w_k * psi(th_k) * C_k(th_k) * w(th_k))
}
\references{
\emph{NIST Digital Library of Mathematical Functions}. Release
1.0.20 of 2018-09-15. F. W. J. Olver, A. B. Olde Daalhuis, D. W. Lozier,
B. I. Schneider, R. F. Boisvert, C. W. Clark, B. R. Miller,
and B. V. Saunders, eds. \url{https://dlmf.nist.gov/}

Smyth, G. K. (1998). Numerical integration. In: \emph{Encyclopedia of
Biostatistics}, P. Armitage and T. Colton (eds.), Wiley, London,
pp. 3088-3095.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{chisq}
\alias{chisq}
\alias{d_chisq}
\alias{p_chisq}
\title{Density and distribution of a chi squared}
\usage{
d_chisq(x, df, ncp = 0L)

p_chisq(x, df, ncp = 0L)
}
\arguments{
\item{x}{a vector of size \code{nx} or a matrix of size \code{c(nx, 1)}.}

\item{df}{degrees of freedom.}

\item{ncp}{non-centrality parameter.}
}
\value{
A matrix of size \code{c(nx, 1)} with the evaluation of the
density or distribution function at \code{x}.
}
\description{
Computation of the density and distribution functions of a chi
squared.
}
\keyword{internal}
